﻿
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Algorithms;

namespace OpcLabs.BaseLibTest.Algorithms
{
    [TestClass]
    public class EnqueuerTest
    {
        [TestMethod]
        public void NoOverflowTest()
        {
            const int queueCapacity = 10;
            var target = new Enqueuer<object>(() => queueCapacity);

            bool wasOverflow = false;
            target.Overflow += (sender, args) => wasOverflow = true;
            bool wasOverflowClearing = false;
            target.OverflowClearing += (sender, args) => wasOverflowClearing = true;
            bool wasUnderflow = false;
            target.Underflow += (sender, args) => wasUnderflow = true;

            for (int i = 0; i < queueCapacity; i++)
                target.Push(i);

            Assert.IsFalse(wasOverflow);
            Assert.AreEqual(0, target.OverflowCount);
            Assert.IsFalse(wasOverflowClearing);
            Assert.IsFalse(wasUnderflow);
            Assert.AreEqual(0, target.UnderflowCount);
        }

        [TestMethod]
        public void OverflowClearingTest()
        {
            const int queueCapacity = 10;
            var target = new Enqueuer<object>(() => queueCapacity);

            bool wasOverflow = false;
            target.Overflow += (sender, args) => wasOverflow = true;
            bool wasOverflowClearing = false;
            long overflowCountWhenClearing = -1;
            target.OverflowClearing += (sender, args) =>
            {
                wasOverflowClearing = true;
                // ReSharper disable once PossibleNullReferenceException
                overflowCountWhenClearing = ((Enqueuer<object>) sender).OverflowCount;
            };
            bool wasUnderflow = false;
            target.Underflow += (sender, args) => wasUnderflow = true;

            for (int i = 0; i < queueCapacity + 1; i++)
                target.Push(i);

            Assert.IsTrue(wasOverflow);
            Assert.AreEqual(1, target.OverflowCount);
            Assert.IsFalse(wasOverflowClearing);
            Assert.IsFalse(wasUnderflow);
            Assert.AreEqual(0, target.UnderflowCount);

            wasOverflow = false;

            object entry = target.Pull(0);
            Assert.IsNotNull(entry);
            Assert.AreEqual(0, entry);

            Assert.IsFalse(wasOverflow);
            Assert.AreEqual(0, target.OverflowCount);
            Assert.IsTrue(wasOverflowClearing);
            Assert.AreEqual(1, overflowCountWhenClearing);
            Assert.IsFalse(wasUnderflow);
            Assert.AreEqual(0, target.UnderflowCount);
        }

        [TestMethod]
        public void OverflowTest()
        {
            const int queueCapacity = 10;
            var target = new Enqueuer<object>(() => queueCapacity);

            bool wasOverflow = false;
            target.Overflow += (sender, args) => wasOverflow = true;
            bool wasOverflowClearing = false;
            target.OverflowClearing += (sender, args) => wasOverflowClearing = true;
            bool wasUnderflow = false;
            target.Underflow += (sender, args) => wasUnderflow = true;

            for (int i = 0; i < queueCapacity + 1; i++)
                target.Push(i);

            Assert.IsTrue(wasOverflow);
            Assert.AreEqual(1, target.OverflowCount);
            Assert.IsFalse(wasOverflowClearing);
            Assert.IsFalse(wasUnderflow);
            Assert.AreEqual(0, target.UnderflowCount);
        }
    }
}
