﻿
using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    [TestClass]
    public class EasyDAClientTest : EasyDAClientTestBase
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_EmptyString_AllSupportedTechnologies()
        {
            BrowseServersTest_AllSupportedTechnologies("");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_EmptyString_SingleSupportedTechnologies()
        {
            BrowseServersTest_SingleSupportedTechnologies("");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_LocalHost_AllSupportedTechnologies()
        {
            BrowseServersTest_AllSupportedTechnologies("localhost");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_LocalHost_SingleSupportedTechnologies()
        {
            BrowseServersTest_SingleSupportedTechnologies("localhost");
        }

        #endregion



        #region Private Methods

        private void BrowseServersTest_AllSupportedTechnologies([NotNull] string location)
        {
            TestSection.Arrange();
            // IMPROVE: should pre-check the 'location'

            var client = new EasyDAClient();
            OpcTechnologies supportedTechnologies = client.SupportedTechnologies;
            AssertUtilities.WriteLine(supportedTechnologies);

            if (supportedTechnologies == OpcTechnologies.Xml)   // just when XML is the one and only supported
            {
                TestSection.Act();
                try
                {
                    client.BrowseServers(location, supportedTechnologies);
                }
                catch (OpcException)
                {
                    // IMPROVE: test the exception
                    return;
                }
                Assert.Fail("Shouldn't have succeeded");
            }
            else
            {
                TestSection.Act();
                ServerElementCollection serverElements = client.BrowseServers(location, supportedTechnologies);

                TestSection.Assert();
                Assert.IsNotNull(serverElements);
                CollectionAssert.AllItemsAreNotNull(serverElements);
            }
        }

        private void BrowseServersTest_SingleSupportedTechnologies([NotNull] string location)
        {
            TestSection.Arrange();
            // IMPROVE: should pre-check the 'location'
            var client = new EasyDAClient();
            OpcTechnologies supportedTechnologies = client.SupportedTechnologies;

            TestSection.Act();

            var serverElementsDictionary = new Dictionary<OpcTechnologies, ServerElementCollection>();
            for (int technologyBits = 1; technologyBits <= (int)OpcTechnologies.All; technologyBits *= 2)
            {
                var technologies = (OpcTechnologies)technologyBits;
                if ((technologies != OpcTechnologies.Unknown) &&
                    (technologies != OpcTechnologies.Xml) &&    // OPC XML does not support browsing for servers yet
                    ((technologyBits & (int)supportedTechnologies) != 0))
                {
                    ServerElementCollection serverElements = client.BrowseServers(location, technologies);
                    serverElementsDictionary[technologies] = serverElements;
                }
            }

            TestSection.Assert();
            foreach (KeyValuePair<OpcTechnologies, ServerElementCollection> pair in serverElementsDictionary)
            {
                OpcTechnologies technologies = pair.Key;
                ServerElementCollection serverElements = pair.Value;

                AssertUtilities.WriteLine(technologies);
                Assert.IsNotNull(serverElements);
                CollectionAssert.AllItemsAreNotNull(serverElements);
            }
        }

        #endregion
    }
}
