﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;

namespace OpcLabs.EasyOpcLiveTest.Legacy.DataAccess
{
    /// <summary>
    ///This is a test class for EasyDAClientExtensionTest and is intended
    ///to contain all EasyDAClientExtensionTest Unit Tests
    ///</summary>
    [TestClass]
    public class EasyDAClientExtensionTest
    {
        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext { get; set; }

        #region Additional test attributes

        // 
        //You can use the following additional attributes as you write your tests:
        //
        //Use ClassInitialize to run code before running the first test in the class
        //[ClassInitialize()]
        //public static void MyClassInitialize(TestContext testContext)
        //{
        //}
        //
        //Use ClassCleanup to run code after all tests in a class have run
        //[ClassCleanup()]
        //public static void MyClassCleanup()
        //{
        //}
        //
        //Use TestInitialize to run code before running each test
        //[TestInitialize()]
        //public void MyTestInitialize()
        //{
        //}
        //
        //Use TestCleanup to run code after each test has run
        //[TestCleanup()]
        //public void MyTestCleanup()
        //{
        //}
        //

        #endregion

        /// <summary>
        ///A test for ReadItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueArrayOfInt32Test()
        {
            var easyDAClient = new EasyDAClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.ReadValue_ArrayOfI4";
            int[] actual = easyDAClient.ReadItemValueArrayOfInt32(machineName, serverClass, itemId);
            int expectedElement = actual[actual.GetLowerBound(0)];
            for (int i = actual.GetLowerBound(0) + 1; i <= actual.GetUpperBound(0); i++)
                Assert.AreEqual(expectedElement, actual[i]);
        }

        /// <summary>
        ///A test for ReadItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueArrayOfInt32Test1()
        {
            var easyDAClient = new EasyDAClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.ReadValue_ArrayOfI4";
            int[] actual = easyDAClient.ReadItemValueArrayOfInt32(serverDescriptor, itemId);
            int expectedElement = actual[actual.GetLowerBound(0)];
            for (int i = actual.GetLowerBound(0) + 1; i <= actual.GetUpperBound(0); i++ )
                Assert.AreEqual(expectedElement, actual[i]);
        }

        /// <summary>
        ///A test for ReadItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueInt32Test()
        {
            const int value = 892736297;
            var easyDAClient = new EasyDAClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4"; // IMPROVE: Use an item with guaranteed constant value instead
            easyDAClient.WriteItemValue(machineName, serverClass, itemId, value);
            const int expected = value;
            int actual = easyDAClient.ReadItemValueInt32(machineName, serverClass, itemId);
            Assert.AreEqual(expected, actual);
        }

        /// <summary>
        ///A test for ReadItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueInt32Test1()
        {
            const int value = 892736297;
            var easyDAClient = new EasyDAClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_I4"; // IMPROVE: Use an item with guaranteed constant value instead
            easyDAClient.WriteItemValue<int>(serverDescriptor, itemId, value);
            const int expected = value;
            int actual = easyDAClient.ReadItemValueInt32(serverDescriptor, itemId);
            Assert.AreEqual(expected, actual);
        }

        /// <summary>
        ///A test for WriteItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueArrayOfInt32Test()
        {
            var easyDAClient = new EasyDAClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_ArrayOfI4";
            int[] value = { 1, 2, 3 };
            easyDAClient.WriteItemValueArrayOfInt32(machineName, serverClass, itemId, value);
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueArrayOfInt32Test1()
        {
            var easyDAClient = new EasyDAClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_ArrayOfI4";
            int[] value = { 1, 2, 3 };
            easyDAClient.WriteItemValueArrayOfInt32(serverDescriptor, itemId, value);
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueInt32Test()
        {
            var easyDAClient = new EasyDAClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4";
            const int value = 0; // TODO: Initialize to an appropriate value
            easyDAClient.WriteItemValueInt32(machineName, serverClass, itemId, value);
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueInt32Test1()
        {
            var easyDAClient = new EasyDAClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_I4";
            const int value = 0; // TODO: Initialize to an appropriate value
            easyDAClient.WriteItemValueInt32(serverDescriptor, itemId, value);
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for GetAccessRightsPropertyValue
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetAccessRightsPropertyValueTest()
        {
            var easyDAClient = new EasyDAClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4";
            DAAccessRight expected = DAAccessRights.Both;
            DAAccessRight actual = easyDAClient.GetAccessRightsPropertyValue(machineName, serverClass, itemId);
            Assert.AreEqual(expected, actual);
        }
    }
}
