﻿using System;
using System.Globalization;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    public abstract partial class ReadBaseTest
    {
        // ReSharper disable InconsistentNaming
        public abstract class ReadBaseTest_Ctt : ReadBaseTest
        // ReSharper restore InconsistentNaming
        {
            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class ReadBaseTest_Ctt_OpcFoundation_SampleServer_OpcTcp : ReadBaseTest_Ctt
            // ReSharper restore InconsistentNaming
            {
                public ReadBaseTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
                }
            }



            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class ReadBaseTest_Ctt_Softing_DemoServer_OpcTcp : ReadBaseTest_Ctt
            // ReSharper restore InconsistentNaming
            {
                public ReadBaseTest_Ctt_Softing_DemoServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
                }

                [ClassInitialize]
                static public void SocketPrecheck(TestContext context)
                {
                    LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
                }
            }




            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>
            /// Return the minimum value for each data type. To achieve this, configure the server with at least one item of each 
            /// data type and perform a read call for each of those items.
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float
            ///     i. Int16
            ///     j. UInt16
            ///     k. Int32
            ///     l. UInt32
            ///     m. Int64
            ///     n. UInt64
            ///     o. String
            /// </description>
            /// <expectedresult>
            /// Client must display/handle all data types that it supports correctly.
            /// Verify the values shown/logged by the Client match what the Server sent.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015a()
            {
                TestCase_015(typeof(Boolean), false);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015b()
            {
                TestCase_015(typeof(Byte), Byte.MinValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015c()
            {
                TestCase_015(typeof(SByte), SByte.MinValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015d()
            {
                TestCase_015(typeof(Byte[]), new Byte[0]);
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_015e()
            //{
            //    TestCase_015(typeof(Char), Char.MinValue);
            //}

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015f()
            {
                TestCase_015(typeof(DateTime), DateTime.MinValue);
            }

            // CTT injects incorrect value
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015g()
            {
                TestCase_015(typeof(Double), 2.2250738585072E-308d);
            }

            // CTT injects incorrect value
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015h()
            {
                TestCase_015(typeof(Single), 1.175494E-38f);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015i()
            {
                TestCase_015(typeof(Int16), Int16.MinValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015j()
            {
                TestCase_015(typeof(UInt16), UInt16.MinValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015k()
            {
                TestCase_015(typeof(Int32), Int32.MinValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015l()
            {
                TestCase_015(typeof(UInt32), UInt32.MinValue);
            }

            // Not supported by CTT script:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_015m()
            //{
            //    TestCase_015(typeof(Int64), Int64.MinValue);
            //}

            // Not supported by CTT script:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_015n()
            //{
            //    TestCase_015(typeof(UInt64), UInt64.MinValue);
            //}

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_015o()
            {
                TestCase_015(typeof(String), String.Empty);
            }

            /// <description>
            /// Return the maximum value for each data type. To achieve this, configure the server with at least one item of each 
            /// data type and perform a read call for each of those items.
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. Char
            ///     e. DateTime
            ///     f. Double
            ///     g. Float h. Int16
            ///     i. UInt16
            ///     j. Int32
            ///     k. UInt32
            ///     l. Int64
            ///     m. UInt64
            ///</description>
            /// <expectedresult>
            /// Client must display/handle all data types that it supports correctly.
            /// Verify the values shown/logged by the Client match what the Server sent.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016a()
            {
                TestCase_016(typeof(Boolean), true);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016b()
            {
                TestCase_016(typeof(Byte), Byte.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016c()
            {
                TestCase_016(typeof(SByte), SByte.MaxValue);
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_016d()
            //{
            //    TestCase_016(typeof(Char), Char.MaxValue);
            //}

            // CTT injects incorrect value
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_016e()
            //{
            //    TestCase_016(typeof(DateTime), DateTime.MaxValue);
            //}

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016f()
            {
                TestCase_016(typeof(Double), Double.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016g()
            {
                TestCase_016(typeof(Single), Single.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016h()
            {
                TestCase_016(typeof(Int16), Int16.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016i()
            {
                TestCase_016(typeof(UInt16), UInt16.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016j()
            {
                TestCase_016(typeof(Int32), Int32.MaxValue);
            }

            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_016k()
            {
                TestCase_016(typeof(UInt32), UInt32.MaxValue);
            }

            // Not supported by CTT script:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_016l()
            //{
            //    TestCase_016(typeof(Int64), Int64.MaxValue);
            //}

            // Not supported by CTT script:
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_016m()
            //{
            //    TestCase_016(typeof(UInt64), UInt64.MaxValue);
            //}

            /// <description>
            /// Read a single node of each type that is an array. Return the max number of values (array elements) that a Client 
            /// supports. If the Client has no limit then exercise a value of 1000, otherwise use the largest size supported by the Client.
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float i. Guid
            ///     j. Int16
            ///     k. UInt16
            ///     l. Int32
            ///     m. UInt32
            ///     n. Int64
            ///     o. UInt64
            ///     p. String
            ///</description>
            /// <expectedresult>Client is able to consume the maximum number of values for each data-type listed.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018a()
            {
                TestCase_018(typeof(Boolean[]));
            }

            // UA Sample Server incorrectly returns ByteString for UserArray/ByteValue
            //[TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_018b()
            //{
            //    TestCase_018(typeof(Byte[]));
            //}

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018c()
            {
                TestCase_018(typeof(SByte[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018d()
            {
                TestCase_018(typeof(Byte[][]));
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Array")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_018e()
            //{
            //    TestCase_018(typeof(Char[]));
            //}

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018f()
            {
                TestCase_018(typeof(DateTime[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018g()
            {
                TestCase_018(typeof(Double[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018h()
            {
                TestCase_018(typeof(Single[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018i()
            {
                TestCase_018(typeof(Guid[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018j()
            {
                TestCase_018(typeof(Int16[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018k()
            {
                TestCase_018(typeof(UInt16[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018l()
            {
                TestCase_018(typeof(Int32[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018m()
            {
                TestCase_018(typeof(UInt32[]));
            }

            // Not supported by CTT script:
            //[TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_018n()
            //{
            //    TestCase_018(typeof(Int64[]));
            //}

            // Not supported by CTT script:
            //[TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Read")]  // inferred
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_018o()
            //{
            //    TestCase_018(typeof(UInt64[]));
            //}

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_018p()
            {
                TestCase_018(typeof(String[]));
            }

            /// <description>ServiceResult is Good.
            /// Operation level result is a Good code which includes the lower 16-bits set to a `sub-value`.
            /// </description>
            /// <expectedresult>Client detects the code is still good and uses it.</expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_025()
            {
                // ServiceResult is Good.
                //Operation level result is a Good code which includes the lower 16-bits set to a “sub-value”.
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\025.js"))
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        new UAReadArguments(EndpointDescriptor,
                                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));
                    Assert.IsNotNull(attributeData);
                    AssertUtilities.WriteLine(attributeData);
                    AssertUtilities.WriteLine(attributeData.StatusCode.CodeBits);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    Assert.IsNotNull(attributeData.Value);
                    Assert.AreEqual(UACodeBits.GoodClamped, attributeData.StatusCode.CodeBits);
                    StringAssert.Contains(attributeData.StatusCode.ToString(), "GoodClamped");
                }
            }

            /// <description>Randomly inject one of the following ServiceResult codes.
            /// Repeat this test several times each tmie varying the result.
            /// 
            /// 1. ServiceResult = Good 
            /// 2. ServiceResult = Bad_InternalError 
            /// 3. ServiceResult = Bad_InvalidArgument 
            /// 4. ServiceResult = Bad_InvalidState 
            /// 5. ServiceResult = Bad_NothingToDo 
            /// 6. ServiceResult = Bad_OutOfMemory 
            /// 7. ServiceResult = Bad_RequestTooLarge 
            /// 8. ServiceResult = Bad_ResourceUnavailable 
            /// 9. ServiceResult = Bad_Shutdown 
            /// 10. ServiceResult = Bad_Timeout 
            /// 11. ServiceResult = Bad_TooManyOperations 
            /// 12. ServiceResult = Bad_UnexpectedError 
            /// 13. ServiceResult = Bad_UnknownResponse 
            /// 14. ServiceResult = Bad_UserAccessDenied 
            /// 15. ServiceResult = Bad_MaxAgeInvalid 
            /// 16. ServiceResult = Bad_TimestampstoReturnInvalid
            /// 17. ServiceResult = Bad_NodeIdUnknown 
            /// 18. ServiceResult = Bad_NodeIdInvalid 
            /// </description>
            /// <expectedresult>Client logs and/or displays an error that it has received invalid information. 
            /// State the results tested, and your observations for the how the client behaved. 
            /// </expectedresult>
            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_01()
            {
                //TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-01.js", UACodeBits.Good);
                // Another rule kicks in due to the way the CTT script is implemented (see TestCase_Err_002_01 in ValueChangeTest)
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-01.js"))
                {
                    try
                    {
                        MakeClient().Read(new UAReadArguments(EndpointDescriptor,
                                                              UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));
                    }
                    catch (UAException uaException)
                    {
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        AssertServiceException(innerException, UACodeBits.BadUnexpectedError,
                                               "The server returned a list with less");
                        return;
                    }
                    Assert.Fail();
                }
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_02()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-02.js", UACodeBits.BadInternalError);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_03()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-03.js", UACodeBits.BadInvalidArgument);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_04()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-04.js", UACodeBits.BadInvalidState);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_05()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-05.js", UACodeBits.BadNothingToDo);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_06()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-06.js", UACodeBits.BadOutOfMemory);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_07()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-07.js", UACodeBits.BadRequestTooLarge);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_08()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-08.js", UACodeBits.BadResourceUnavailable);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_09()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-09.js", UACodeBits.BadShutdown);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_10()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-10.js", UACodeBits.BadTimeout);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_11()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-11.js", UACodeBits.BadTooManyOperations);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_12()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-12.js", UACodeBits.BadUnexpectedError);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_13()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-13.js", UACodeBits.BadUnknownResponse);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_14()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-14.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_15()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-15.js", UACodeBits.BadMaxAgeInvalid);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_16()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-16.js", UACodeBits.BadTimestampsToReturnInvalid);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_17()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-17.js", UACodeBits.BadNodeIdUnknown);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_001_18()
            {
                TestCase_Err_001(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-001-18.js", UACodeBits.BadNodeIdInvalid);
            }

            /// <description>Inject one of the following Operation Level Results into the first returned item:
            ///
            /// 1. Bad_Timeout
            /// 2. Bad_AttributeIdInvalid 
            /// 3. Bad_DataEncodingInvalid 
            /// 4. Bad_IndexRangeNoData 
            /// 5. Bad_NodeIdUnknown 
            /// 6. Bad_NodeIdInvalid 
            /// 7. Bad_NotReadable 
            /// 8. Bad_UserAccessDenied 
            /// 9. Reserved 
            /// </description>
            /// <expectedresult>
            /// </expectedresult>
            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_01()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-01.js", UACodeBits.BadTimeout);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_02()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-02.js", UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_03()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-03.js", UACodeBits.BadDataEncodingInvalid);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_04()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-04.js", UACodeBits.BadIndexRangeNoData);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_05()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-05.js", UACodeBits.BadNodeIdUnknown);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_06()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-06.js", UACodeBits.BadNodeIdInvalid);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_07()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-07.js", UACodeBits.BadNotReadable);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_08()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-08.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_002_09()
            {
                const int Reserved = 0xC000;    // IMPROVE: have constant for it, in the API?
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-002-09.js", Reserved);
            }

            /// <description>For all results, inject one of the following error Status Code; repeat this test numerous times each time varying the status code:
            ///
            /// 1. Bad_Timeout 
            /// 2. Bad 
            /// 3. Bad_UserAccessDenied 
            /// 4. Good
            /// </description>
            /// <expectedresult>
            /// </expectedresult>
            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_003_01()
            {
                TestCase_Err_003_And_005(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-003-01.js", UACodeBits.BadTimeout);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_003_02()
            {
                TestCase_Err_003_And_005(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-003-02.js", UACodeBits.Bad);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_003_03()
            {
                TestCase_Err_003_And_005(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-003-03.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab, corrected
            [TestMethod]
            public void TestCase_Err_003_04()
            {
                TestCase_Err_003_And_005(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-003-04.js", UACodeBits.Good);
            }

            /// <description>ServiceResult = Good and set error in StatusCode to Bad which includes the lower 16-bits set to
            /// a `sub-value` for the 1st item.
            /// Example: ServiceResult = Good.
            /// StatusCode = Bad_NotConnected for 1st item 
            /// </description>
            /// <expectedresult>Client detects the code is still Bad and does not use it, or uses it in the same context that 
            /// it would for a standard Bad quality code.
            /// Client continues to process remaining items.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004()
            {
                TestCase_Err_002_And_004(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-004.js", UACodeBits.BadNotConnected);
            }

            /// <description>ServiceResult = Good and set error in StatusCode to Bad which includes the lower 16-bits set to a `sub-value` for all items.
            /// Example: ServiceResult = Good.
            /// StatusCode = Bad_DeviceFailure for all items. 
            /// </description>
            /// <expectedresult>Client detects the code is still Bad and does not use it, or uses it in the same context that 
            /// it would for a standard Bad quality code.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005()
            {
                TestCase_Err_003_And_005(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-005.js", UACodeBits.BadDeviceFailure);
            }

            /// <description>ServiceResult = Bad.
            /// StatusCode = GOOD for all items.
            /// </description>
            /// <expectedresult>Client must indicate the Read for all items failed!
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-006.js"))
                {
                    const int count = 3;
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] 
                    {
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Single), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int16), dynamic: true))
                    });
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(count, attributeDataResults.Length);

                    for (int i = 0; i < count; i++)
                    {
                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Assert.IsNotNull(result.Exception);

                        AssertServiceException(result.Exception, UACodeBits.Bad);
                    }
                }
            }

            /// <description>Read 3 or more nodes in a single call.
            /// The CTT returns ServiceResult = Good.
            /// * Results[0 ]= Bad
            /// * Results[1] = Uncertain
            /// * Results[2] = Reserved
            /// </description>
            /// <expectedresult>The affected items must show some indication of the non-good values.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_007()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-007.js"))
                {
                    const int notGoodCount = 3;
                    int[] codeBitsExpected = {UACodeBits.Bad, UACodeBits.Uncertain, 0xC000, UACodeBits.Good};
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] 
                    {
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Single), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int16), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(string), dynamic: true))
                    });
                    Assert.IsNotNull(attributeDataResults);

                    for (int i = 0; i < notGoodCount; i++)
                    {
                        AssertUtilities.WriteLine(i);
                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Assert.IsNull(result.Exception);
                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.IsFalse(attributeData.StatusCode.IsGood);
                        Assert.AreEqual(codeBitsExpected[i], attributeData.StatusCode.CodeBits);
                    }

                    for (int i = notGoodCount; i < attributeDataResults.Length; i++)
                    {
                        AssertUtilities.WriteLine(i);
                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Assert.IsNull(result.Exception);
                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.IsTrue(attributeData.StatusCode.IsGood);
                    }
                }
            }

            /// <description>Return fewer Results than the number of nodes specified in the nodesToRead parameter.
            /// </description>
            /// <expectedresult>Client should log and/or notify the end-user of the error.
            /// The client does not use the data received.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_008()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-008.js"))
                {
                    const int count = 3;
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] 
                    {
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Single), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)),
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int16), dynamic: true))
                    });
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(count, attributeDataResults.Length);

                    for (int i = 0; i < count; i++)
                    {
                        AssertUtilities.WriteLine(i);

                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Exception exception = result.Exception;
                        AssertUtilities.WriteLine(exception);
                        Assert.IsNotNull(exception);
                        Assert.IsTrue(IsClientEngineException(exception, UAClientErrorCodes.FewerReadResultsReturned));
                        Assert.IsNull(result.AttributeData);
                    }
                }
            }

            /// <description>Client issues a Read and requests both timestamps, where MaxAge = 2 seconds.
            /// Return GOOD and set Server timestamp to current time - 2 x MaxAge for all items.
            /// Note: The MaxAge value can be varied for this test.
            /// </description>
            /// <expectedresult>Client reports an error that it has received stale data.
            /// In addition, the Client might choose to:
            /// * Ignore the data
            /// * Display data with a warning
            /// * Accept the data if it is within the acceptable tolerance area
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_009()
            {
                // Holds back the Read Response by 2x MaxAge.
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-009.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true),
                            UAAttributeId.Value,
                            2*1000)
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);
                    Assert.IsTrue(diagnostics.Any(IsStaleTimestampException));
                    Exception warning = diagnostics.First(IsStaleTimestampException);
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            /// <description>Client issues a Read and requests both timestamps, where MaxAge = 2 seconds.
            /// Return GOOD and all set timestamp to current time - 1 day.
            /// </description>
            /// <expectedresult>Client reports an error that it has received stale data.
            /// In addition, the Client might choose to:
            /// * Ignore the data
            /// * Display data with a warning
            /// Accept the data if it is within the acceptable tolerance area
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-010.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true),
                            UAAttributeId.Value,
                            2*1000)
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);
                    Assert.IsTrue(diagnostics.Any(IsStaleTimestampException));
                    Exception warning = diagnostics.First(IsStaleTimestampException);
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            /// <description>Hold back the service response (delay is MaxAge * 2). If MaxAge == 2 then set MaxAge = 2000;
            ///</description>
            /// <expectedresult>Client detects the stale timestamp and indicates a problem with this item by reporting 
            /// Bad_InvalidTimestamp.</expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011()
            {
                // Holds back the Read Response by 2x MaxAge.
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-011.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true),
                            UAAttributeId.Value,
                            2*1000)
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);
                    Assert.IsTrue(diagnostics.Any(IsStaleTimestampException));
                    Exception warning = diagnostics.First(IsStaleTimestampException);
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            /// <description>Return GOOD set timestamp to current time + 25 hours for all items.</description>
            /// <expectedresult>If a client does not display the time then the value should be flagged to indicate an issue. For 
            /// non interactive clients that store or use the value, the value should not be used and an error should be generated. 
            /// A proxy UA server that is a pass-through would be an exception to this.</expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012()
            {
                // Client issues a Read and requests both timestamps.
                // Return GOOD and all set timestamp forward 25-hours.
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-012.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true))
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 2);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);

                    Assert.IsTrue(diagnostics.Any(IsServerTimestampAheadException));
                    Exception serverTimestampAhead = diagnostics.First(IsServerTimestampAheadException);
                    Assert.IsNotNull(serverTimestampAhead);
                    AssertUtilities.WriteLine("Server timestamp ahead warning found:");
                    AssertUtilities.WriteLine(serverTimestampAhead.Message);

                    Assert.IsTrue(diagnostics.Any(IsSourceTimestampAheadException));
                    Exception sourceTimestampAhead = diagnostics.First(IsSourceTimestampAheadException);
                    Assert.IsNotNull(sourceTimestampAhead);
                    AssertUtilities.WriteLine("Source timestamp ahead warning found:");
                    AssertUtilities.WriteLine(sourceTimestampAhead.Message);
                }
            }

            /// <description>Have client specify TimestampsToReturn = BOTH_2 but return only ONE timestamp.</description>
            /// <expectedresult>Client reports the error.
            /// If the Client needs to use the timestamp (for further processing or analysis) then the data is ignored. 
            /// Alternatively, the Client may perform a different action like using a default value (e.g. use `Now`) as based on 
            /// configuration or as documented.</expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_015()
            {
                // Have client specify TimestampsToReturn = BOTH but return
                // only one timestamp.
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-015.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true))
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    AssertUtilities.WriteLine(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    Assert.AreEqual(DateTime.MinValue, attributeData.SourceTimestamp);
                    Assert.AreNotEqual(DateTime.MinValue, attributeData.ServerTimestamp);

                    // IMPROVE: we  have a configurable behavior that would completely refuse the data if timestamp is missing, test it as well

                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);

                    Assert.IsTrue(diagnostics.Any(IsNoSourceTimestampReturnedException));
                    Exception warning = diagnostics.First(IsNoSourceTimestampReturnedException);
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            /// <description>Have client specify TimestampsToReturn = BOTH_2 but return NO timestamps.</description>
            /// <expectedresult>Client reports the error.
            /// If the Client needs to use the timestamp (for further processing or analysis) then the data is ignored. 
            /// Alternatively, the Client may perform a different action like using a default value (e.g. use `Now`) as based on 
            /// configuration or as documented.</expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_016()
            {
                // Have client specify TimestampsToReturn = BOTH_2 but return NO timestamps
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-016.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true))
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    Assert.AreEqual(DateTime.MinValue, attributeData.ServerTimestamp);
                    Assert.AreEqual(DateTime.MinValue, attributeData.SourceTimestamp);

                    // IMPROVE: we  have a configurable behavior that would completely refuse the data if timestamp is missing, test it as well

                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);

                    Assert.IsTrue(diagnostics.Any(IsNoServerTimestampReturnedException));
                    Exception warning1 = diagnostics.First(IsNoServerTimestampReturnedException);
                    Assert.IsNotNull(warning1);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning1.Message);

                    Assert.IsTrue(diagnostics.Any(IsNoSourceTimestampReturnedException));
                    Exception warning2 = diagnostics.First(IsNoSourceTimestampReturnedException);
                    Assert.IsNotNull(warning2);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning2.Message);
                }
            }

            /// <description>Read a node of type Array, the data-type can be anything.
            /// The received values are of a different type.
            /// E.g. read an Int32[] but receive: string[].
            /// </description>
            /// <expectedresult>Client reports an error, and does not process the data.</expectedresult>

            // 20130306 Nathan Pocock <nathan.pocock@opcfoundation.org>: 
            // I have modified the test-case expectations to permit the client to use the data if it has the ability to do so, 
            // otherwise an error must be reported.
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_019_alpha()
            {
                TestCase_Err_019(typeof(Int32[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_019_beta()
            {
                TestCase_Err_019(typeof(String[]));
            }

            /// <description>Read a node of type Array. The values are all empty, e.g. the length of each array is 0 (zero).
            /// </description>
            /// <expectedresult>The client indicates that there are no values returned.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_020()
            {
                // Read a node of type Array.	
                // The values are all empty, e.g. the length of each array is 0 (zero).
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-020.js"))
                {
                    var types = new[]
                    {
                        typeof (Boolean[]), typeof (SByte[]), 
                        //typeof (Byte[]), // CTT does not handle it well
                        typeof (Int16[]), typeof (UInt16[]),
                        typeof (Int32[]),
                        typeof (UInt32[]), typeof (Int64[]), typeof (UInt64[]), typeof (Single[]), typeof (Double[]),
                        typeof (String[]),
                        typeof (DateTime[]), typeof (Guid[]), typeof (Byte[][])
                    };
                    UAReadArguments[] readArgumentArray = types
                        // ReSharper disable AssignNullToNotNullAttribute
                        .Select(type => new UAReadArguments(EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true)))
                        // ReSharper restore AssignNullToNotNullAttribute
                        .ToArray();
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    // ReSharper disable ForCanBeConvertedToForeach
                    for (int i = 0; i < attributeDataResultArray.Length; i++)
                    // ReSharper restore ForCanBeConvertedToForeach
                    {
                        Type type = types[i];
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "{0}: {1}", i, type));
                        UAAttributeDataResult result = attributeDataResultArray[i];
                        Assert.IsNotNull(result);
                        Assert.IsNull(result.Exception);
                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.IsTrue(attributeData.HasGoodStatus);
                        object value = attributeData.Value;
                        Assert.IsNotNull(value);
                        var array = value as Array;
                        Assert.IsNotNull(array);
                        Assert.AreEqual(1, array.Rank);
                        Assert.AreEqual(0, array.GetLength(0));
                        // we think that this is how we indicate that "there are no values returned"
                    }
                }
            }

            /// <description>Read a node of each type that is an array. Return the max number of values (array elements) 
            /// that a Client supports PLUS 1 more. If the Client has no limit, then exercise a value of 1001.
            /// Note: CTT Settings may be needed to specify the Client`s limitation.
            /// </description>
            /// <expectedresult>The Client warns that it received more data than it can consume.
            /// The Client processes the elements that it can support.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_021()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-021.js"))
                {
                    var types = new[]
                    {
                        typeof (Int32[])
                    };
                    UAReadArguments[] readArgumentArray = types
                        // ReSharper disable AssignNullToNotNullAttribute
                        .Select(type => new UAReadArguments(EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true)))
                        // ReSharper restore AssignNullToNotNullAttribute
                        .ToArray();
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    // ReSharper disable ForCanBeConvertedToForeach
                    for (int i = 0; i < attributeDataResultArray.Length; i++)
                    // ReSharper restore ForCanBeConvertedToForeach
                    {
                        Type type = types[i];
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "{0}: {1}", i, type));
                        UAAttributeDataResult result = attributeDataResultArray[i];
                        Assert.IsNotNull(result);
                        Assert.IsNull(result.Exception);
                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.IsTrue(attributeData.HasGoodStatus);
                        object value = attributeData.Value;
                        Assert.IsNotNull(value);
                        var array = value as Array;
                        Assert.IsNotNull(array);
                        Assert.AreEqual(1, array.Rank);
                        Assert.AreEqual(1001, array.Length);
                    }
                }
            }

            /// <description>Request Diagnostics information from server in Request Header but return EMPTY for the result.
            /// </description>
            /// <expectedresult>Client reports that no diagnostics were returned even though they were requested.
            /// </expectedresult>
            [TestCategory("Keyword:Diagnostics")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_022()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-022.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        //new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(float), dynamic: true))
                        new UAReadArguments(EndpointDescriptor, new UANodeId("http://test.org/UA/Data/", "!UnknownIdentifier1")),
                        new UAReadArguments(EndpointDescriptor, new UANodeId("http://test.org/UA/Data/", "!UnknownIdentifier2"))
                    };
                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.DiagnosticsMasks = UADiagnosticsMasks.All;
                    UAAttributeDataResult[] attributeDataResultArray = client.ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsFalse(attributeData.HasGoodStatus);
                    ExceptionCollection diagnostics = result.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);
                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue(diagnostics.Any(exception => IsClientEngineException(exception, UAClientErrorCodes.NoDiagnosticsInfosReturned)));
                    Exception warning = diagnostics.First(exception => IsClientEngineException(exception, UAClientErrorCodes.NoDiagnosticsInfosReturned));
                    // ReSharper restore AssignNullToNotNullAttribute
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            /// <description>Request Diagnostics information from server in Request Header but return: 
            /// 1. More diagnostic entries than nodes requested in NodesToRead
            /// 2. Fewer diagnostic entries than nodes requested in NodesToRead
            /// </description>
            /// <expectedresult>Client reports the mismatched # of entries returned. Client does not process the diagnostic 
            /// info, or if it does then a clear message is present that the information cannot be trusted.
            /// </expectedresult>
            [TestCategory("Keyword:Diagnostics")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_023_01()
            {
                TestCase_Err_023(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-023-01.js");
            }

            [TestCategory("Keyword:Diagnostics")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_023_02()
            {
                Ctt.ClaimScriptBuggyVersion(0xccef93fd);
                TestCase_Err_023(@"Attribute Services\Attribute Client Read Base\Test Cases\Err-023-02.js");
            }

            /// <description>Return float and Double data with NaN values.
            /// </description>
            /// <expectedresult>The client might indicate that it received invalid data, likewise it might also accept the 
            /// NaN values.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_024_alpha()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-024.js"))
                {
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] {
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(float), dynamic: true))});
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(1, attributeDataResults.Length);
                    UAAttributeDataResult attributeDataResult = attributeDataResults[0];
                    Assert.IsNotNull(attributeDataResult);
                    UAAttributeData attributeData = attributeDataResult.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    Assert.IsNotNull(attributeData.Value);

                    ExceptionCollection diagnostics = attributeDataResult.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);

                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue(diagnostics.Any(exception => IsClientEngineException(exception, UAClientErrorCodes.FloatNotANumber)));
                    Exception warning = diagnostics.First(exception => IsClientEngineException(exception, UAClientErrorCodes.FloatNotANumber));
                    // ReSharper restore AssignNullToNotNullAttribute
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            [TestCategory("Keyword:Read")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_024_beta()
            {
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-024.js"))
                {
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] {
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(double), dynamic: true))});
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(1, attributeDataResults.Length);
                    UAAttributeDataResult attributeDataResult = attributeDataResults[0];
                    Assert.IsNotNull(attributeDataResult);
                    UAAttributeData attributeData = attributeDataResult.AttributeData;
                    Assert.IsNotNull(attributeData);
                    Assert.IsTrue(attributeData.HasGoodStatus);
                    Assert.IsNotNull(attributeData.Value);

                    ExceptionCollection diagnostics = attributeDataResult.Diagnostics;
                    Assert.IsNotNull(diagnostics);
                    int count = diagnostics.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 1);
                    CollectionAssert.AllItemsAreNotNull(diagnostics);
                    CollectionAssert.AllItemsAreUnique(diagnostics);

                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue(diagnostics.Any(exception => IsClientEngineException(exception, UAClientErrorCodes.DoubleNotANumber)));
                    Exception warning = diagnostics.First(exception => IsClientEngineException(exception, UAClientErrorCodes.DoubleNotANumber));
                    // ReSharper restore AssignNullToNotNullAttribute
                    Assert.IsNotNull(warning);
                    AssertUtilities.WriteLine("Warning found:");
                    AssertUtilities.WriteLine(warning.Message);
                }
            }

            // 20130306 Nathan Pocock <nathan.pocock@opcfoundation.org>:
            ///// <description>Client issues a Read. UACTT modifies the response so that a difference number of results are 
            ///// returned, e.g. less than requested, or more than requested.</description>
            ///// <expectedresult>Client reports the mismatched # of entries returned. 
            ///// Client does not process the diagnostic info, or if it does then a clear message is present that the data cannot be 
            ///// trusted.
            ///// Client should provide a way for the user to know (if desired) that the Read was successful. 
            ///// Writing an event to a log for every Read is probably not practical since could potentially generate 1000`s of 
            ///// duplicate entries.
            ///// </expectedresult>
            //public void TestCase_Err_025()
            //{
            //    //...
            //}

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_002_And_004([NotNull] string scriptName, int codeBitsExpected)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    const int count = 3;
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] 
                {
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Single), dynamic: true)),
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)),
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int16), dynamic: true))
                });
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(count, attributeDataResults.Length);

                    UAAttributeDataResult result0 = attributeDataResults[0];
                    Assert.IsNotNull(result0);
                    AssertUtilities.WriteLine(result0);
                    Assert.IsNull(result0.Exception);

                    UAAttributeData attributeData0 = result0.AttributeData;
                    Assert.IsNotNull(attributeData0);
                    Assert.AreEqual(codeBitsExpected, attributeData0.StatusCode.CodeBits);
                    //AssertServiceException(result0.Exception, codeBitsExpected);

                    for (int i = 1; i < count; i++)
                    {
                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        Assert.IsNull(result.Exception);

                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.IsTrue(attributeData.StatusCode.IsGood);
                    }
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_003_And_005([NotNull] string scriptName, int codeBitsExpected)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    const int count = 3;
                    UAAttributeDataResult[] attributeDataResults = MakeClient().ReadMultiple(new[] 
                {
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Single), dynamic: true)),
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)),
                    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int16), dynamic: true))
                });
                    Assert.IsNotNull(attributeDataResults);
                    Assert.AreEqual(count, attributeDataResults.Length);

                    for (int i = 0; i < count; i++)
                    {
                        UAAttributeDataResult result = attributeDataResults[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Assert.IsNull(result.Exception);

                        UAAttributeData attributeData = result.AttributeData;
                        Assert.IsNotNull(attributeData);
                        Assert.AreEqual(codeBitsExpected, attributeData.StatusCode.CodeBits);
                        //AssertServiceException(result.Exception, codeBitsExpected);
                    }
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_015([NotNull] Type type, [CanBeNull] object expectedValueNonClsCompliant)
            // ReSharper restore InconsistentNaming
            {
                // Return the minimum value for each data type. To achieve this, configure
                // the server with at least one item of each data type and perform a read
                // call for each of those items.
                // Bool, Byte, SByte, ByteString, Char, DateTime, Double, Float,
                // Int16, UInt16, Int32, UInt32, Int64, UInt64, String
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\015.js"))
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true));
                    //UAAttributeData attributeData = PerformRead(
                    //    MakeClient(),
                    //    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true)));

                    AssertAttributeDataValue(attributeData, Cls.ConvertToClsType(expectedValueNonClsCompliant));
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_016([NotNull] Type type, [CanBeNull] object expectedValueNonClsCompliant)
            // ReSharper restore InconsistentNaming
            {
                // Return the MAXIMUM value for each data type. To achieve this, configure
                // the server with at least one item of each data type and perform a read
                // call for each of those items.
                // Bool, Byte, SByte, ByteString, Char, DateTime, Double, Float,
                // Int16, UInt16, Int32, UInt32, Int64, UInt64, String
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\016.js"))
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true));
                    //UAAttributeData attributeData = PerformRead(
                    //    MakeClient(),
                    //    new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true)));

                    AssertAttributeDataValue(attributeData, Cls.ConvertToClsType(expectedValueNonClsCompliant));
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_001([NotNull] string scriptName, int codeBitsExpected)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    if (!(new UAStatusCode(codeBitsExpected, 0).IsBad))
                    {
                        UAAttributeData attributeData = MakeClient().Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                        Assert.IsNotNull(attributeData);
                        Assert.AreEqual(codeBitsExpected, attributeData.StatusCode.CodeBits);
                        return;
                    }

                    try
                    {
                        MakeClient().Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                    }
                    catch (UAException uaException)
                    {
                        AssertServiceException(uaException.InnerException, codeBitsExpected);
                        return;
                    }
                    Assert.Fail();
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_018([NotNull] Type type)
            // ReSharper restore InconsistentNaming
            {
                // Add following code at the beginning of readResponse018:
                //// Do not intervene if this is just a housekeeping Read
                //var allNamespacesAreZero = true;
                //for (var i = 0; i < request.NodesToRead.length; i++)
                //{
                //    if (request.NodesToRead[i].NodeId.NamespaceIndex != 0)
                //    {
                //        allNamespacesAreZero = false;
                //        break;
                //    }
                //}
                //if (allNamespacesAreZero)
                //    return;

                // Return the MAXIMUM number of elements in an Array for each data type.
                // To achieve this, configure the server with at least one item of each
                // data type and perform a read call for each of those items.
                // Bool, Byte, SByte, ByteString, Char, DateTime, Double, Float,
                // Int16, UInt16, Int32, UInt32, Int64, UInt64, String
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\018.js"))
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true));
                    //UAAttributeData attributeData = PerformRead(
                    //    MakeClient(),
                    //    new UAReadArguments(EndpointDescriptor,
                    //        UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true)));

                    Type expectedType = Cls.GetClsType(type);
                    AssertAttributeDataType(attributeData, expectedType);
                    // IMPROVE: we are not checking the value against what the server has sent

                    object value = attributeData.Value;
                    Assert.IsNotNull(value);
                    var array = value as Array;
                    Assert.IsNotNull(array);
                    Assert.AreEqual(1, array.Rank);
                    Assert.AreEqual(0, array.GetLowerBound(0));
                    Assert.AreEqual(20, array.GetLength(0));
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_019([NotNull] Type type)
            // ReSharper restore InconsistentNaming
            {
                // Read a node of type Array, the data-type can be anything.
                // The received values are of a different type.
                // E.g. read an Int32[] but receive: string[]. 
                using (Ctt.RunScript(CttProjectName, @"Attribute Services\Attribute Client Read Base\Test Cases\Err-019.js"))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(
                            EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray:true))
                    };
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    Assert.IsNotNull(attributeDataResultArray);
                    UAAttributeDataResult result = attributeDataResultArray[0];
                    Assert.IsNotNull(result);
                    Assert.IsNull(result.Exception);
                    UAAttributeData attributeData = result.AttributeData;
                    Assert.IsNotNull(attributeData);
                    object value = attributeData.Value;
                    Assert.IsNotNull(value);
                    Type actualType = value.GetType();
                    Assert.AreNotEqual(type, actualType);
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_023([NotNull] string scriptName)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    UAReadArguments[] readArgumentArray =
                    {
                        new UAReadArguments(EndpointDescriptor, new UANodeId("http://test.org/UA/Data/", "!UnknownIdentifier1")),
                        new UAReadArguments(EndpointDescriptor, new UANodeId("http://test.org/UA/Data/", "!UnknownIdentifier2"))
                    };
                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.DiagnosticsMasks = UADiagnosticsMasks.All;
                    UAAttributeDataResult[] attributeDataResultArray = client.ReadMultiple(readArgumentArray);
                    Assert.IsNotNull(attributeDataResultArray);

                    // ReSharper disable ForCanBeConvertedToForeach
                    for (int i = 0; i < attributeDataResultArray.Length; i++)
                    // ReSharper restore ForCanBeConvertedToForeach
                    {
                        UAAttributeDataResult result = attributeDataResultArray[i];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        AssertUtilities.WriteLine(result.Diagnostics.Count);
                        Assert.IsNull(result.AttributeData);
                        Assert.IsNotNull(result.Exception);

                        // TODO: test more!
                    }
                }
            }

            #endregion
        }
    }
}
