﻿using System;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Parsing;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    public abstract partial class ReadBaseTest
    {
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class ReadBaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : ReadBaseTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public ReadBaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class ReadBaseTest_NoCtt_Softing_DemoServer_Http : ReadBaseTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public ReadBaseTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }
        }
        


        // ReSharper disable InconsistentNaming
        public abstract class ReadBaseTest_NoCtt : ReadBaseTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>
            /// Monitor method call and analyze passed parameters for correctness and adherence to the specifications. 
            /// Reading a single item.
            /// </description>
            /// <expectedresult>
            /// Client correctly receives and processes the results correctly while also checking quality codes.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                    AssertMaxAgeOutcome(recorder, attributeData, typeof(Int32), Int32.MaxValue);
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Read a single valid node, while requesting data with a MaxAge set to 0.
            /// </description>
            /// <expectedresult>
            /// Verify the call is made correctly with the results being processed correctly by the client.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002()
            {
                const double maxAge = 0.0d;
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        new UAReadArguments(EndpointDescriptor,
                                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                            UAAttributeId.Value, UAIndexRangeList.Empty, maxAge));
                    AssertMaxAgeOutcome(recorder, attributeData, typeof(Int32), maxAge);
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If client allows end-user to specify a MaxAge, then end-user specifies a value of say 5 seconds while 
            /// reading one valid node.
            /// </description>
            /// <expectedresult>
            /// Verify client correctly specifies the end-user supplied value to the server into the correct number 
            /// of msec.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003()
            {
                const double maxAge = 5 * 1000;
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        new UAReadArguments(EndpointDescriptor,
                                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                            UAAttributeId.Value, UAIndexRangeList.Empty, maxAge));
                    AssertMaxAgeOutcome(recorder, attributeData, typeof(Int32), maxAge);
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If client allows end-user to specify a MaxAge, attempt to enter MaxAge &lt; 0
            /// </description>
            /// <expectedresult>
            /// Client is expected to send a valid MaxAge.
            /// Note: Client may override the MaxAge with a default value. Likewise, it may simply reject the user-input and not 
            /// issue the Read.
            /// </expectedresult>
            [ExpectedException(typeof(ArgumentOutOfRangeException))]
            [TestCategory("ArgumentChecking")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_004()
            {
                const double maxAge = -5.0d;
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        MakeClient().Read(
                            new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                UAAttributeId.Value, UAIndexRangeList.Empty, maxAge));
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If client allows end-user to specify a MaxAge, attempt to enter MaxAge &gt; Int32 while reading a 
            /// single node value.
            /// </description>
            /// <expectedresult>
            /// Client is expected to send a valid MaxAge.
            /// Note: Client may override the MaxAge with a default value. Likewise, it may simply reject the user-input and not 
            /// issue the Read.
            /// </expectedresult>
            [ExpectedException(typeof(ArgumentOutOfRangeException))]
            [TestCategory("ArgumentChecking")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005()
            {
                const double maxAge = (double)Int32.MaxValue + 5.0d;
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        MakeClient().Read(
                            new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                UAAttributeId.Value, UAIndexRangeList.Empty, maxAge));
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If Client physically able to specify NaN as the value?
            /// </description>
            /// <expectedresult>
            /// Client is expected to send a valid MaxAge.
            /// Note: Client may override the MaxAge with a default value. Likewise, it may simply reject the user-input and not 
            /// issue the Read.
            /// </expectedresult>
            [ExpectedException(typeof(ArgumentOutOfRangeException))]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("ArgumentChecking")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006()
            {
                const double maxAge = Double.NaN;
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        MakeClient().Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                UAAttributeId.Value, UAIndexRangeList.Empty, maxAge));
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>Perform a read of many valid nodes (for example 20+) in a single call, with MaxAge=0.
            /// </description>
            /// <expectedresult>Verify the Client issues one Read call containing all 20 Nodes.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007()
            {
                const double maxAge = 0.0d;
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(
                        UATestUtilities.MakeVolumeNodeDescriptors(NodeCount20Plus, dynamic:true)
                                       .Select(descriptor =>
                                           // ReSharper disable AssignNullToNotNullAttribute
                                               new UAReadArguments(EndpointDescriptor, descriptor, UAAttributeId.Value,
                                                                   UAIndexRangeList.Empty, maxAge))
                        // ReSharper restore AssignNullToNotNullAttribute
                                       .ToArray());

                    Assert.IsNotNull(attributeDataResultArray);
                    int length = attributeDataResultArray.Length;
                    Assert.AreEqual(NodeCount20Plus, length);
                    CollectionAssert.AllItemsAreNotNull(attributeDataResultArray);
                    // ReSharper disable PossibleNullReferenceException
                    Assert.IsTrue(attributeDataResultArray.All(result => result.Exception == null));
                    Assert.IsTrue(
                        attributeDataResultArray.All(
                            result => (result.AttributeData != null) && (result.AttributeData.HasGoodStatus)));
                    // ReSharper restore PossibleNullReferenceException

                    AssertReadRecordMaxAges(recorder, NodeCount20Plus, maxAge);
                    // IMPROVE: verify that the nodes in the read record are the correct ones
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>If client allows end-user to specify a MaxAge, then end-user specifies a value of say 5 seconds while 
            /// reading many valid nodes (20+)
            /// </description>
            /// <expectedresult>Verify client correctly specifies the end-user supplied value to the server into the correct 
            /// duration and that one Read call is made.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008()
            {
                const double maxAge = 5 * 1000;
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(
                        UATestUtilities.MakeVolumeNodeDescriptors(NodeCount20Plus, dynamic: true)
                                       .Select(descriptor =>
                                           // ReSharper disable AssignNullToNotNullAttribute
                                               new UAReadArguments(EndpointDescriptor, descriptor, UAAttributeId.Value,
                                                                   UAIndexRangeList.Empty, maxAge))
                        // ReSharper restore AssignNullToNotNullAttribute
                                       .ToArray());

                    Assert.IsNotNull(attributeDataResultArray);
                    int length = attributeDataResultArray.Length;
                    Assert.AreEqual(NodeCount20Plus, length);
                    CollectionAssert.AllItemsAreNotNull(attributeDataResultArray);
                    // ReSharper disable PossibleNullReferenceException
                    Assert.IsTrue(attributeDataResultArray.All(result => result.Exception == null));
                    Assert.IsTrue(
                        attributeDataResultArray.All(
                            result => (result.AttributeData != null) && (result.AttributeData.HasGoodStatus)));
                    // ReSharper restore PossibleNullReferenceException

                    AssertReadRecordMaxAges(recorder, NodeCount20Plus, maxAge);
                    // IMPROVE: verify that the nodes in the read record are the correct ones
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>If client allows end-user to specify a MaxAge, attempt to enter MaxAge > Int32 while reading many 
            /// valid nodes (20+)
            /// </description>
            /// <expectedresult>Client is expected to send a valid MaxAge, and that one Read call is made.
            /// </expectedresult>
            [ExpectedException(typeof(ArgumentOutOfRangeException))]
            [TestCategory("ArgumentChecking")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009()
            {
                // We require that no call is made and an exception is thrown - that should be an allowed expected result too.
                const double maxAge = (double)Int32.MaxValue + 5.0d;
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        MakeClient().ReadMultiple(
                            UATestUtilities.MakeVolumeNodeDescriptors(NodeCount20Plus, dynamic: true)
                                           .Select(descriptor =>
                                               // ReSharper disable AssignNullToNotNullAttribute
                                                   new UAReadArguments(EndpointDescriptor, descriptor, UAAttributeId.Value,
                                                                       UAIndexRangeList.Empty, maxAge))
                            // ReSharper restore AssignNullToNotNullAttribute
                                           .ToArray());
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>If client allows end-user to specify a MaxAge, then end-user specifies a “fractional” value of say 
            /// 5.75 msecs while reading many valid nodes (20+)
            /// </description>
            /// <expectedresult>Verify client correctly specifies the end-user supplied value to the server into the correct 
            /// duration and that one Read call is made.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010()
            {
                const double maxAge = 5.75d;
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(
                        UATestUtilities.MakeVolumeNodeDescriptors(NodeCount20Plus, dynamic: true)
                                       .Select(descriptor =>
                                           // ReSharper disable AssignNullToNotNullAttribute
                                               new UAReadArguments(EndpointDescriptor, descriptor, UAAttributeId.Value,
                                                                   UAIndexRangeList.Empty, maxAge))
                        // ReSharper restore AssignNullToNotNullAttribute
                                       .ToArray());

                    AssertAttributeDataResultArray(attributeDataResultArray, NodeCount20Plus, mustBeGoodStatus: true);
                    AssertReadRecordMaxAges(recorder, NodeCount20Plus, maxAge);
                    // IMPROVE: verify that the nodes in the read record are the correct ones
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If Client allows the end-user to choose the TimestampsToReturn, then read 1 Node and test with each possible value 
            /// i.e.:
            ///     a. SOURCE_0
            ///     b. SERVER_1
            ///     c. BOTH_2
            ///     d. NEITHER_3
            /// </description>
            /// <expectedresult>Check that client sends the correct (as picked from a Client user interface) TimestampsToReturn. 
            /// For the settings made, verify that the timestamp received by the client and shown to the user matches what was 
            /// passed up from the server – either the server value or the source.
            /// If the client has the ability to display BOTH timestamps then both values are verified and compared against what 
            /// was sent by the Server.
            /// </expectedresult>
            // IMPROVE: when made possible by our API, implement tests for 11a, b, d.
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011c()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeData attributeData = MakeClient().Read(
                        new UAReadArguments(EndpointDescriptor,
                                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                    AssertTimestampsOutcome(recorder, new[] { attributeData }, typeof(Int32));
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If Client allows the end-user to choose the TimestampsToReturn, then while reading many valid nodes (20+) set 
            /// TimestampsToReturn =
            ///     a. SOURCE_0
            ///     b. SERVER_1
            ///     c. BOTH_2
            ///     d. NEITHER_3
            /// </description>
            /// <expectedresult>Check that client sends the correct (as picked from a Client user interface) TimestampsToReturn. 
            /// For the settings made, verify that for all Nodes the timestamps received by the client and shown to the user 
            /// matches what was passed up from the server – either the server value or the source.
            /// If the client has the ability to display BOTH timestamps then both values are verified and compared against what 
            /// was sent by the Server, for all Nodes.
            /// </expectedresult>
            // IMPROVE: when made possible by our API, implement tests for 12a, b, d.
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012c()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(
                        UATestUtilities.MakeVolumeNodeDescriptors(NodeCount20Plus, dynamic: true)
                                       .Select(descriptor =>
                                           // ReSharper disable AssignNullToNotNullAttribute
                                               new UAReadArguments(EndpointDescriptor, descriptor))
                        // ReSharper restore AssignNullToNotNullAttribute
                                       .ToArray());

                    AssertAttributeDataResultArray(attributeDataResultArray, NodeCount20Plus, mustBeGoodStatus: true);
                    // ReSharper disable PossibleNullReferenceException
                    AssertTimestampsOutcome(recorder, attributeDataResultArray.Select(result => result.AttributeData).ToArray(),
                                            typeof(Int64));
                    // ReSharper restore PossibleNullReferenceException
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// If the Client permits the end-user to manually enter a NodeId (versus being required to pick a Node from a 
            /// Browser dialog) then the end-user specifies invalid NodeIds, as in:
            ///      An empty string
            ///      Invalid string representation of a NodeId.
            ///      If NodeId is GUID, specify a non GUID value and/or invalid GUID syntax and/or incorrect GUID (id not known).
            ///      If NodeId is numeric, specify a negative number.
            /// </description>
            /// <expectedresult>Ideally, the client will detect any anomalies in the NodeId syntax and will report an error.
            /// The Client should be prepared to handle and report any errors returned by the Server.
            /// </expectedresult>
            [ExpectedException(typeof(ArgumentException))]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("ArgumentChecking")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_alpha()
            {
                EasyUAClient client = MakeClient();
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        client.Read(new UAReadArguments(EndpointDescriptor, new UANodeId("")));
                    }
                    catch (Exception exception)
                    {
                        AssertUtilities.WriteLine($"Exception in the Read statement: {exception}");
                        throw;
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            [ExpectedException(typeof(UANodeIdFormatException))]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("ArgumentChecking")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_beta()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    EasyUAClient client = MakeClient();
                    try
                    {
                        var nodeDescriptor = new UANodeDescriptor("ns=INVALID_STRING_REPRESENTATION_OF_A_NODEID");
                        client.Read(new UAReadArguments(EndpointDescriptor, nodeDescriptor));
                    }
                    catch (Exception exception)
                    {
                        AssertUtilities.WriteLine($"Exception in the Read statement: {exception}");
                        throw;
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            [ExpectedException(typeof(FormatException))]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("ArgumentChecking")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_gamma()
            {
                // IMPROVE: do not understand what is meant by "a non GUID value and/or ... incorrect GUID (id not known)", no test
                using (var recorder = new UAOperationRecorder())
                {
                    EasyUAClient client = MakeClient();
                    try
                    {
                        client.Read(new UAReadArguments(EndpointDescriptor,
                                                 new UANodeId("namespace3", Guid.Parse("INVALID_GUID_SYNTAX"))));
                    }
                    catch (Exception exception)
                    {
                        AssertUtilities.WriteLine($"Exception in the Read statement: {exception}");
                        throw;
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            [ExpectedException(typeof(OverflowException))]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("ArgumentChecking")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_delta()
            {
                // IMPROVE: do not understand what is meant by "a non GUID value and/or ... incorrect GUID (id not known)", no test
                using (var recorder = new UAOperationRecorder())
                {
                    EasyUAClient client = MakeClient();
                    try
                    {
                        client.Read(new UAReadArguments(EndpointDescriptor, new UANodeId("namespace4", -5)));
                    }
                    catch (Exception exception)
                    {
                        AssertUtilities.WriteLine($"Exception in Read: {exception}");
                        throw;
                    }
                    finally
                    {
                        Assert.AreEqual(0, recorder.ReadRecordList.Count);
                    }
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Read the value of every attribute for any given node.
            /// Note: Ideally the Read will be a single call containing all requests. If not possible, then several Read calls 
            /// would be acceptable.
            /// </description>
            /// <expectedresult>
            /// Verify the Client does read ALL attributes of the selected node.
            /// Verify the values displayed/logged by the Client match what the Server sent.
            /// </expectedresult>
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    Array attributeIds = typeof(UAAttributeId).GetEnumValues();
                    Debug.Assert(attributeIds != null);
                    int count = attributeIds.Length - 1;
                    var readArgumentArray = new UAReadArguments[count];
                    int j = 0;
                    for (int i = 0; i < attributeIds.Length; i++)
                    {
                        // ReSharper disable PossibleNullReferenceException
                        var attributeId = (UAAttributeId)attributeIds.GetValue(i);
                        // ReSharper restore PossibleNullReferenceException
                        if (attributeId == UAAttributeId.None)
                            continue;
                        readArgumentArray[j] = new UAReadArguments(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                            attributeId);
                        j++;
                    }

                    UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                    AssertAttributeDataResultArray(attributeDataResultArray, count, mustBeGoodStatus: false);
                    for (int i = 0; i < count; i++)
                    {
                        // ReSharper disable once PossibleNullReferenceException
                        UAAttributeData attributeData = attributeDataResultArray[i].AttributeData;
                        Assert.IsNotNull(attributeData);
                        AssertUtilities.WriteLine(
                            // ReSharper disable once PossibleNullReferenceException
                            $"attributeId: {readArgumentArray[i].AttributeId}; attributeData: {attributeData}");
                    }

                    // ReSharper disable PossibleNullReferenceException
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.NodeId)].AttributeData, typeof(UANodeId));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.NodeClass)].AttributeData, typeof(Int32));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.BrowseName)].AttributeData, typeof(UAQualifiedName));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.WriteMask)].AttributeData, typeof(Int64));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.UserWriteMask)].AttributeData, typeof(Int64));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.Value)].AttributeData, typeof(Int32));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.DataType)].AttributeData, typeof(UANodeId));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.ValueRank)].AttributeData, typeof(Int32));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.AccessLevel)].AttributeData, typeof(Byte));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.UserAccessLevel)].AttributeData, typeof(Byte));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.MinimumSamplingInterval)].AttributeData, typeof(Double));
                    AssertAttributeDataType(attributeDataResultArray[Array.FindIndex(readArgumentArray, argument =>
                        argument.AttributeId == UAAttributeId.Historizing)].AttributeData, typeof(Boolean));
                    // ReSharper restore PossibleNullReferenceException

                    int readCount = recorder.ReadRecordList.Count;
                    Assert.IsTrue(readCount < count);
                }
            }
#endif

            /// <description>
            /// Return a value (Value attribute) of each possible array data type supported by the client. To achieve this 
            /// configure the server with at least one item of each data type and perform a read call for each of those items.
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float i. Guid
            ///     j. Int16
            ///     k. UInt16
            ///     l. Int32
            ///     m. UInt32
            ///     n. Int64
            ///     o. UInt64
            ///     p. String
            ///</description>
            /// <expectedresult>
            /// Client must display/handle all data types that it supports correctly.
            /// Verify the values shown/logged by the Client match what the Server sent.
            /// </expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017a()
            {
                TestCase_017(typeof(Boolean[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017b()
            {
                TestCase_017(typeof(Byte[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017c()
            {
                TestCase_017(typeof(SByte[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017d()
            {
                TestCase_017(typeof(Byte[][]));
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Array")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_017e()
            //{
            //    TestCase_017(typeof(Char[]));
            //}

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017f()
            {
                TestCase_017(typeof(DateTime[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017g()
            {
                TestCase_017(typeof(Double[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017h()
            {
                TestCase_017(typeof(Single[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017i()
            {
                TestCase_017(typeof(Guid[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017j()
            {
                TestCase_017(typeof(Int16[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017k()
            {
                TestCase_017(typeof(UInt16[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017l()
            {
                TestCase_017(typeof(Int32[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017m()
            {
                TestCase_017(typeof(UInt32[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017n()
            {
                TestCase_017(typeof(Int64[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017o()
            {
                TestCase_017(typeof(UInt64[]));
            }

            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017p()
            {
                TestCase_017(typeof(String[]));
            }

            /// <description>
            /// Read multiple nodes where each is an array of each data-type supported by the Client:
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float i. Guid
            ///     j. Int16
            ///     k. UInt16
            ///     l. Int32
            ///     m. UInt32
            ///     n. Int64
            ///     o. UInt64
            ///     p. String
            /// Note: This test may include reading all data-types in one call, or multiple data-types in one call. But all 
            /// supported types must be covered.
            ///</description>
            /// <expectedresult>Values displayed/reported in client match those sent by the server.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019()
            {
                Type[] typeArray = GetTestTypeArray();
                int count = typeArray.Length;
                UAReadArguments[] readArgumentArray = typeArray.Select(type =>
                    new UAReadArguments(EndpointDescriptor,
                        // ReSharper disable AssignNullToNotNullAttribute
                                        UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true))).ToArray();
                // ReSharper restore AssignNullToNotNullAttribute

                UAAttributeDataResult[] attributeDataResultArray = MakeClient().ReadMultiple(readArgumentArray);

                for (int i = 0; i < count; i++)
                {
                    Type type = typeArray[i];
                    Debug.Assert(type != null);
                    AssertUtilities.WriteLine(type);

                    UAAttributeDataResult attributeDataResult = attributeDataResultArray[i];
                    Assert.IsNotNull(attributeDataResult);
                    AssertUtilities.WriteLine(attributeDataResult);
                    Assert.IsNull(attributeDataResult.Exception);
                    UAAttributeData attributeData = attributeDataResult.AttributeData;
                    Assert.IsNotNull(attributeData);
                    
                    Type expectedType = Cls.GetClsType(type);
                    AssertAttributeDataType(attributeData, expectedType);
                    // IMPROVE: we are not checking the value against what the server has sent

                    object value = attributeData.Value;
                    Assert.IsNotNull(value);
                    var array = value as Array;
                    Assert.IsNotNull(array);
                    Assert.AreEqual(1, array.Rank);
                    Assert.AreEqual(0, array.GetLowerBound(0));
                }
            }

            /// <description>
            /// Read a node specifying a single element ONLY within the array, e.g. IndexRange = `4`. Do this for each supported 
            /// data-type: 
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float i. Guid
            ///     j. Int16
            ///     k. UInt16
            ///     l. Int32
            ///     m. UInt32
            ///     n. Int64
            ///     o. UInt64
            ///     p. String
            ///</description>
            /// <expectedresult>Client correctly requests the element within the array. Verify the value shown/logged by the Client 
            /// matches the value sent by the Server.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_020()
            {
                TestCase_IndexRangeList(new UAIndexRangeList(new[] { new UAIndexRange(1) }));   // IMPROVE: use 4
            }

            /// <description>
            /// Read a range of values within an array, e.g. IndexRange = `3:5`. Do this for each supported data-type: 
            ///     a. Bool
            ///     b. Byte
            ///     c. SByte
            ///     d. ByteString
            ///     e. Char
            ///     f. DateTime
            ///     g. Double
            ///     h. Float i. Guid
            ///     j. Int16
            ///     k. UInt16
            ///     l. Int32
            ///     m. UInt32
            ///     n. Int64
            ///     o. UInt64
            ///     p. String
            ///</description>
            /// <expectedresult>Client correctly requests the elements within the array. Verify the values shown/logged by the 
            /// Client match what the Server sent.</expectedresult>
            [TestCategory("Keyword:Array")]
            [TestCategory("Keyword:Read")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_021()
            {
                TestCase_IndexRangeList(new UAIndexRangeList(new[] { new UAIndexRange(1, 2) }));    // IMPROVE: use 3:5
            }

            // ReSharper restore InconsistentNaming
#endregion



#region Private Methods

            // ReSharper disable InconsistentNaming
            private void TestCase_017([NotNull] Type type)
            // ReSharper restore InconsistentNaming
            {
                UAAttributeData attributeData = MakeClient().Read(
                    EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true));
                //UAAttributeData attributeData = PerformRead(
                //    MakeClient(),
                //    new UAReadArguments(EndpointDescriptor,
                //        UATestUtilities.MakeNodeDescriptor(type, dynamic: true, preferArray: true)));

                Type expectedType = Cls.GetClsType(type);
                AssertAttributeDataType(attributeData, expectedType);
                // IMPROVE: we are not checking the value against what the server has sent
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_IndexRangeList([NotNull] UAIndexRangeList indexRangeList)
            // ReSharper restore InconsistentNaming
            {
                // The OPC Foundation Sample Server and Softing Demo Server, don't always provide the required indices.
                // We therefore need to write first.

                EasyUAClient client = MakeClient();
                var typeArray = GetTestTypeArray();
                int count = typeArray.Length;

                UAWriteValueArguments[] writeValueArgumentArray = typeArray.Select(type =>
                    new UAWriteValueArguments(EndpointDescriptor,
                        // ReSharper disable AssignNullToNotNullAttribute
                        UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray: true),
                        // ReSharper disable PossibleNullReferenceException
                        Array.CreateInstance(type.GetElementType(), indexRangeList[0].Maximum + 5), indexRangeList, typeof(object)))
                    // ReSharper restore AssignNullToNotNullAttribute
                    .ToArray();

                client.WriteMultipleValues(writeValueArgumentArray);

                UAReadArguments[] readArgumentArray = typeArray.Select(type =>
                    new UAReadArguments(EndpointDescriptor,
                        // ReSharper disable AssignNullToNotNullAttribute
                        UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray: true),
                        UAAttributeId.Value, indexRangeList))
                    // ReSharper restore AssignNullToNotNullAttribute
                    .ToArray();

                UAAttributeDataResult[] attributeDataResultArray = client.ReadMultiple(readArgumentArray);

                for (int i = 0; i < count; i++)
                {
                    UAAttributeDataResult attributeDataResult = attributeDataResultArray[i];
                    Assert.IsNotNull(attributeDataResult);
                    Assert.IsNull(attributeDataResult.Exception);
                    UAAttributeData attributeData = attributeDataResult.AttributeData;
                    Assert.IsNotNull(attributeData);

                    Type type = typeArray[i];
                    Debug.Assert(type != null);

                    Type expectedType = Cls.GetClsType(type);
                    AssertAttributeDataType(attributeData, expectedType);
                    // IMPROVE: we are not checking whether we are sending the correct server to the server
                    // IMPROVE: we are not checking the value against what the server has sent

                    object value = attributeData.Value;
                    Assert.IsNotNull(value);
                    var array = value as Array;
                    Assert.IsNotNull(array);
                    Assert.AreEqual(indexRangeList.Count, array.Rank);
                    for (int j = 0; j < array.Rank; j++)
                    {
                        Assert.AreEqual(0, array.GetLowerBound(j));
                        UAIndexRange indexRange = indexRangeList[j];
                        Debug.Assert(indexRange != null);
                        AssertUtilities.WriteLine(indexRange);
                        Assert.AreEqual(indexRange.Span - 1, array.GetUpperBound(j));
                    }
                }
            }

#endregion
        }
    }
}
