﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    public abstract partial class WriteBaseTest
    {
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class WriteBaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : WriteBaseTest_NoCtt
            // ReSharper restore InconsistentNaming
        {
            public WriteBaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class WriteBaseTest_NoCtt_Softing_DemoServer_Http : WriteBaseTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            #region Public Constructors

            public WriteBaseTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            #endregion



            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }



            #region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>
            /// Write to numerous Nodes, i.e. 10+ and demonstrate that this occurs in a single call.
            /// </description>
            /// <expectedresult>
            /// Client able to package multiple writes into a single call.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    var random = new Random();
                    UAWriteValueArguments[] argumentsArray =
                        UATestUtilities.MakeVolumeNodeDescriptors(NodeCount10Plus, dynamic: false)
                        .Select(descriptor =>
                            // ReSharper disable AssignNullToNotNullAttribute
                           new UAWriteValueArguments(EndpointDescriptor, descriptor, random.Next()))
                        // ReSharper restore AssignNullToNotNullAttribute
                        .ToArray();

                    UAWriteResult[] resultArray = MakeClient().WriteMultipleValues(argumentsArray);
                    Assert.IsNotNull(resultArray);
                    List<Exception> exceptions =
                        // ReSharper disable PossibleNullReferenceException
                        resultArray.Where(result => result.Exception != null)
                                   .Select(result => result.Exception)
                                   .ToList();
                    // ReSharper restore PossibleNullReferenceException
                    exceptions.ForEach(AssertUtilities.WriteLine);
                    // ReSharper disable PossibleNullReferenceException
                    Assert.IsTrue(resultArray.All(result => result.Exception == null));
                    // ReSharper restore PossibleNullReferenceException

                    int writeCount = recorder.WriteRecordList.Count;
                    Assert.AreEqual(1, writeCount);
                }
            }
#endif

            // ReSharper restore InconsistentNaming
#endregion
        }



        // ReSharper disable InconsistentNaming
        public abstract class WriteBaseTest_NoCtt : WriteBaseTest
        // ReSharper restore InconsistentNaming
        {
#region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>
            /// Monitor method call and analyze passed parameters for correctness and adherence to the specifications.
            /// </description>
            /// <expectedresult>
            /// Write successful.
            /// Client checks the results of the write call.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    MakeClient().WriteValue(
                        EndpointDescriptor, 
                        UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false),
                        new Random().Next());
                    int writeCount = recorder.WriteRecordList.Count;
                    Assert.AreEqual(1, writeCount);
                }
            }
#endif

// TestCase_001: We do not test "Client checks the results of the write call.", as there are other test cases for it.

#if (NETFRAMEWORK)
            /// <description>
            /// Client can write multiple values to the same node in a single call, where the node is not an array type.
            /// </description>
            /// <expectedresult>
            /// A single call is performed that will contain all write operations.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_004()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    var writeValueArgumentsArray = new[]
                        {
                            new UAWriteValueArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false), 1),
                            new UAWriteValueArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false), 2),
                            new UAWriteValueArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false), 3)
                        };
                    UAWriteResult[] resultArray = MakeClient().WriteMultipleValues(writeValueArgumentsArray);
                    Assert.IsNotNull(resultArray);
                    List<Exception> exceptions =
                        // ReSharper disable PossibleNullReferenceException
                        resultArray.Where(result => result.Exception != null)
                                   .Select(result => result.Exception)
                                   .ToList();
                    // ReSharper restore PossibleNullReferenceException
                    exceptions.ForEach(AssertUtilities.WriteLine);
                    // ReSharper disable PossibleNullReferenceException
                    Assert.IsTrue(resultArray.All(result => result.Exception == null));
                    // ReSharper restore PossibleNullReferenceException

                    int writeCount = recorder.WriteRecordList.Count;
                    Assert.AreEqual(1, writeCount);
                    WriteRecord writeRecord = recorder.WriteRecordList[0];
                    Debug.Assert(writeRecord != null);
                    WriteValueCollection sdkWriteValueCollection = writeRecord.NodesToWrite;
                    Assert.IsNotNull(sdkWriteValueCollection);
                    Assert.AreEqual(3, sdkWriteValueCollection.Count);
                }
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Demonstrate the ability to write to any attribute, and handle the call correctly:
            /// a. NodeId
            /// b. NodeClass
            /// c. BrowseName
            /// d. DisplayName
            /// e. Description
            /// f. WriteMask
            /// g. UserWriteMask
            /// h. IsAbstract
            /// i. Symmetric
            /// j. InverseName
            /// k. ContainsNo Loops l. EventNotifier
            /// m. DataType
            /// n. ValueRank
            /// o. Array Dimentions
            /// p. AccessLevel
            /// q. UserAccess Level
            /// r. Minimum Sampling Interval
            /// s. Historizing
            /// t. Executable
            /// u. UserExecutable
            /// </description>
            /// <expectedresult>
            /// Some attributes can be expected to fail. However, the call must be monitored for conformance and the results 
            /// to be handled by the client correctly and with sufficient information to be made available to the end-user in 
            /// the event that a write might and in some cases will fail.
            /// A Client may prevent writes to certain attributes and this will be acceptable if consistent with documentation.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005a()
            {
                TestCase_005(UAAttributeId.NodeId, new UANodeId(UAUtilities.OpcUANamespaceUriString, "identifier"), new NodeId("identifier", 0), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005b()
            {
                TestCase_005(UAAttributeId.NodeClass, UANodeClass.ObjectType, (Int32)NodeClass.ObjectType, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005c()
            {
                TestCase_005(UAAttributeId.BrowseName, new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "name"), new QualifiedName("name", 0), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005d()
            {
                TestCase_005(UAAttributeId.DisplayName, "aString3", new LocalizedText("aString3"), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005e()
            {
                TestCase_005(UAAttributeId.Description, "aString2", new LocalizedText("aString2"), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005f()
            {
                TestCase_005(UAAttributeId.WriteMask, 0x55555554, 0x55555554u, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005g()
            {
                TestCase_005(UAAttributeId.UserWriteMask, 0x55555554, 0x55555554u, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005h()
            {
                TestCase_005(UAAttributeId.IsAbstract, true, true, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005i()
            {
                TestCase_005(UAAttributeId.Symmetric, false, false, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005j()
            {
                TestCase_005(UAAttributeId.InverseName, "aString1", new LocalizedText("aString1"), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005k()
            {
                TestCase_005(UAAttributeId.ContainsNoLoops, true, true, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005l()
            {
                TestCase_005(UAAttributeId.EventNotifier, 252, (Byte)252, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005m()
            {
                TestCase_005(UAAttributeId.DataType, new UANodeId(UAUtilities.OpcUANamespaceUriString, "identifier"), new NodeId("identifier", 0), ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005n()
            {
                TestCase_005(UAAttributeId.ValueRank, 16, 16, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005o()
            {
                TestCase_005(UAAttributeId.ArrayDimensions, new[] { 10, 20, 30 }, new UInt32[] { 10, 20, 30 }, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005p()
            {
                TestCase_005(UAAttributeId.AccessLevel, 253, (Byte)253, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005q()
            {
                TestCase_005(UAAttributeId.UserAccessLevel, 254, (Byte)254, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005r()
            {
                const float value = 1000.37f;
                TestCase_005(UAAttributeId.MinimumSamplingInterval, value, (Double)value, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005s()
            {
                TestCase_005(UAAttributeId.Historizing, true, true, ex => 
                    (ex is UAServiceException) && 
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadNotWritable);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005t()
            {
                TestCase_005(UAAttributeId.Executable, false, false, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005u()
            {
                TestCase_005(UAAttributeId.UserExecutable, true, true, ex =>
                    (ex is UAServiceException) &&
                    ((UAServiceException)ex).ServiceResult == UACodeBits.BadAttributeIdInvalid);
            }
#endif

            /// <description>
            /// Write to the Value of one item of each supported data type.
            /// a. Bool
            /// b. Byte
            /// c. SByte
            /// d. ByteString 
            /// e. Char
            /// f. DateTime
            /// g. Double
            /// h. Float i. Guid
            /// j. Int16
            /// k. UInt16
            /// l. Int32
            /// m. UInt32
            /// n. Int64
            /// o. UInt64
            /// p. String
            /// </description>
            /// <expectedresult>
            /// Write successful.
            /// Client checks the results of the write call.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006a()
            {
                TestCase_006(typeof(Boolean), false);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006b()
            {
                TestCase_006(typeof(Byte), 200);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006c()
            {
                TestCase_006(typeof(SByte), -100);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006d()
            {
                TestCase_006(typeof(Byte[]), new [] { 150, 151, 152, 153, 154 });
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Write")]
            ////[TestCategory("Keyword:Array")]   // deemed improper
            ////[TestCategory("Keyword:Diagnostics")]   // deemed improper
            ////[TestCategory("Keyword:Inferred")]   // deemed improper
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_006e()
            //{
            //    TestCase_006(typeof(Char));
            //}

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006f()
            {
                TestCase_006(typeof(DateTime), new DateTime(1967, 8, 27, 23, 30, 0));
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006g()
            {
                TestCase_006(typeof(Double), -5.13e100d);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006h()
            {
                TestCase_006(typeof(Single), -4.24e20f);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006i()
            {
                TestCase_006(typeof(Guid), Guid.NewGuid());
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006j()
            {
                TestCase_006(typeof(Int16), -32000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006k()
            {
                TestCase_006(typeof(UInt16), 64000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006l()
            {
                TestCase_006(typeof(Int32), -32000*1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006m()
            {
                TestCase_006(typeof(UInt32), 64000*1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006n()
            {
                TestCase_006(typeof(Int64), -32000*1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006o()
            {
                TestCase_006(typeof(UInt64), 64000*1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006p()
            {
                TestCase_006(typeof(String), "Jak chudak do jeste vetsi nouze prisel.");
            }

            // 20130425 NathanP:
            // This intent here is to have multiple write transactions within one call. Now, those items might be for example: 
            // a.	1xInt16 + 1xInt32 + 1xFloat etc.
            // b.	10xInt16 + 1xFloat + 100xString etc.
            // IMPROVE: make the test more complex, according to the above explanation
            /// <description>
            /// In a single call, Write to multiple items of each supported data-type::
            /// a. Bool
            /// b. Byte
            /// c. SByte
            /// d. ByteString 
            /// e. Char
            /// f. DateTime
            /// g. Double
            /// h. Float i. Guid
            /// j. Int16
            /// k. UInt16
            /// l. Int32
            /// m. UInt32
            /// n. Int64
            /// o. UInt64
            /// p. String
            /// </description>
            /// <expectedresult>
            /// Write successful.
            /// Client checks the results of the write call.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007a()
            {
                TestCase_007(typeof(Boolean), false);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007b()
            {
                TestCase_007(typeof(Byte), 200);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007c()
            {
                TestCase_007(typeof(SByte), -100);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007d()
            {
                TestCase_007(typeof(Byte[]), new[] { 150, 151, 152, 153, 154 });
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Write")]
            ////[TestCategory("Keyword:Array")]   // deemed improper
            ////[TestCategory("Keyword:Diagnostics")]   // deemed improper
            ////[TestCategory("Keyword:Inferred")]   // deemed improper
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_007e()
            //{
            //    TestCase_007(typeof(Char));
            //}

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007f()
            {
                TestCase_007(typeof(DateTime), new DateTime(1967, 8, 27, 23, 30, 0));
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007g()
            {
                TestCase_007(typeof(Double), -5.13e100d);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007h()
            {
                TestCase_007(typeof(Single), -4.24e20f);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007i()
            {
                TestCase_007(typeof(Guid), Guid.NewGuid());
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007j()
            {
                TestCase_007(typeof(Int16), -32000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007k()
            {
                TestCase_007(typeof(UInt16), 64000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007l()
            {
                TestCase_007(typeof(Int32), -32000 * 1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007m()
            {
                TestCase_007(typeof(UInt32), 64000 * 1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007n()
            {
                TestCase_007(typeof(Int64), -32000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007o()
            {
                TestCase_007(typeof(UInt64), 64000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007p()
            {
                TestCase_007(typeof(String), "Jak chudak do jeste vetsi nouze prisel.");
            }

            /// <description>
            /// Check the client can write values that are within the acceptable range of each supported data type: 
            /// a. Bool
            /// b. Byte
            /// c. SByte
            /// d. Char
            /// e. DateTime
            /// f. Double
            /// g. Float h. Int16
            /// i. UInt16
            /// j. Int32
            /// k. UInt32
            /// l. Int64
            /// m. UInt64
            /// </description>
            /// <expectedresult>
            /// Client should not send values that are in excess of what is supported by the data-type defined for a 
            /// particular node, e.g. Client does not write a Uint16 value (e.g. 33000) to an Byte value.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008a_alpha()
            {
                TestCase_008_alpha(typeof(Boolean), false);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008b_alpha()
            {
                TestCase_008_alpha(typeof(Byte), 200);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008c_alpha()
            {
                TestCase_008_alpha(typeof(SByte), -100);
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Write")]
            ////[TestCategory("Keyword:Array")]   // deemed improper
            ////[TestCategory("Keyword:Diagnostics")]   // deemed improper
            ////[TestCategory("Keyword:Inferred")]   // deemed improper
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_008d_alpha()
            //{
            //    TestCase_008(typeof(Char));
            //}

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008e_alpha()
            {
                TestCase_008_alpha(typeof(DateTime), new DateTime(1967, 8, 27, 23, 30, 0));
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008f_alpha()
            {
                TestCase_008_alpha(typeof(Double), -5.13e100d);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008g_alpha()
            {
                TestCase_008_alpha(typeof(Single), -4.24e20f);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008h_alpha()
            {
                TestCase_008_alpha(typeof(Int16), -32000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008i_alpha()
            {
                TestCase_008_alpha(typeof(UInt16), 64000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008j_alpha()
            {
                TestCase_008_alpha(typeof(Int32), -32000 * 1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008k_alpha()
            {
                TestCase_008_alpha(typeof(UInt32), 64000 * 1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008l_alpha()
            {
                TestCase_008_alpha(typeof(Int64), -32000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008m_alpha()
            {
                TestCase_008_alpha(typeof(UInt64), 64000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008a_beta()
            {
                TestCase_008_beta(typeof(Boolean), "yes");
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008b_beta()
            {
                TestCase_008_beta(typeof(Byte), 33000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008c_beta()
            {
                TestCase_008_beta(typeof(SByte), 128);
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Write")]
            ////[TestCategory("Keyword:Array")]   // deemed improper
            ////[TestCategory("Keyword:Diagnostics")]   // deemed improper
            ////[TestCategory("Keyword:Inferred")]   // deemed improper
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_008d_beta()
            //{
            //    TestCase_008(typeof(Char));
            //}

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008e_beta()
            {
                TestCase_008_beta(typeof(DateTime), DateTime.MaxValue.ToOADate() + 1);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008f_beta()
            {
                TestCase_008_beta(typeof(Double), "xyz");
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008g_beta()
            {
                TestCase_008_beta(typeof(Single), -4.24e208d);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008h_beta()
            {
                TestCase_008_beta(typeof(Int16), -33000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008i_beta()
            {
                TestCase_008_beta(typeof(UInt16), -1000);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008j_beta()
            {
                TestCase_008_beta(typeof(Int32), -32000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008k_beta()
            {
                TestCase_008_beta(typeof(UInt32), 64000 * 1000000000L);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008l_beta()
            {
                TestCase_008_beta(typeof(Int64), 5e208d);
            }

            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_008m_beta()
            {
                TestCase_008_beta(typeof(UInt64), 6e208d);
            }

            /// <description>
            /// IndexRange=``. Write to an item of each supported data-type (single-dimension array): 
            /// a. Bool
            /// b. Byte
            /// c. SByte
            /// d. ByteString 
            /// e. Char
            /// f. DateTime
            /// g. Double
            /// h. Float i. Guid
            /// j. Int16
            /// k. UInt16
            /// l. Int32
            /// m. UInt32
            /// n. Int64
            /// o. UInt64
            /// p. String
            /// </description>
            /// <expectedresult>
            /// Write successful.
            /// Client checks the results of the write call.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")] 
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009a()
            {
                TestCase_009(typeof(Boolean[]), new [] {false, true, false, true, false});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009b()
            {
                TestCase_009(typeof(Byte[]), new [] {200, 201, 202, 203, 203});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009c()
            {
                TestCase_009(typeof(SByte[]), new [] {-100, 101, -102, 103, 104});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009d()
            {
                TestCase_009(typeof(Byte[][]), new []
                    {
                        new [] {150, 151, 152, 153, 154},
                        new [] {160, 161, 162, 163, 164}, 
                        new [] {170, 171, 172, 173, 174}, 
                        new [] {180, 181, 182, 183, 184}, 
                        new [] {190, 191, 192, 193, 194} 
                    });
            }

            // There is no such type in UA:
            //[TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   
            ////[TestCategory("Keyword:Diagnostics")]   // deemed improper
            ////[TestCategory("Keyword:Inferred")]   // deemed improper
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]
            //[TestMethod]
            //public void TestCase_009e()
            //{
            //    TestCase_009(typeof(Char[]));
            //}

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009f()
            {
                TestCase_009(typeof(DateTime[]), new []
                    {
                        new DateTime(1967, 8, 27, 23, 30, 0),
                        new DateTime(1941, 3, 11),
                        new DateTime(1940, 12, 17),
                        new DateTime(2001, 9, 27),
                        new DateTime(2005, 12, 9)
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009g()
            {
                TestCase_009(typeof(Double[]), new [] {-5.13e100d, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009h()
            {
                TestCase_009(typeof(Single[]), new [] {-4.24e20f, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009i()
            {
                TestCase_009(typeof(Guid[]), new [] { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009j()
            {
                TestCase_009(typeof(Int16[]), new [] {-32000, 32001, -32002, 32003, -32004});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009k()
            {
                TestCase_009(typeof(UInt16[]), new []{64000, 64001, 64002, 64003, 64004});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009l()
            {
                TestCase_009(typeof(Int32[]), new [] {-32000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009m()
            {
                TestCase_009(typeof(UInt32[]), new [] {64000 * 1000, 1, 2, 3, 4});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009n()
            {
                TestCase_009(typeof(Int64[]), new [] {-32000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009o()
            {
                TestCase_009(typeof(UInt64[]), new [] {64000 * 1000000000L, 1, 2, 3, 4});
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_009p()
            {
                TestCase_009(typeof(String[]), new [] 
                {
                    "Jak chudak do jeste vetsi nouze prisel.",
                    "Dlouhy, Siroky a Kratkozraky.",
                    "Kasparkuv hrobecek.",
                    "O trinacti tchynich.",
                    "-.-.-.-.-.-.-.-.-.-." });
            }

#if (NETFRAMEWORK)
            /// <description>
            /// IndexRange=``. Write to an item of each data-type (2-dimension array): 
            /// Bool, Byte, SByte, ByteString, DateTime, Double, Float, Guid, Int16, UInt16, Int32, UInt32, Int64, UInt64, 
            /// String. 
            /// </description>
            /// <expectedresult>
            /// Client able to write a value of each data-type that it supports.
            /// Verify that values written to server match those sent by the client.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011a()
            {
                TestCase_011(typeof (Boolean[,]), new[,]
                {
                    { false, true, false, true, false },
                    { true, false, true, false, true}
                });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011b()
            {
                TestCase_011(typeof(Byte[,]), new[,]
                    {
                        { 200, 201, 202, 203, 204 },
                        { 210, 211, 212, 213, 214 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011c()
            {
                TestCase_011(typeof(SByte[,]), new[,]
                    {
                        { -100, 101, -102, 103, 104 },
                        { -110, 111, -112, 113, 114 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011d()
            {
                TestCase_011(typeof(Byte[,][]), new[,]
                    {
                        {
                            new [] {150, 151, 152, 153, 154},
                            new [] {160, 161, 162, 163, 164}, 
                            new [] {170, 171, 172, 173, 174}, 
                            new [] {180, 181, 182, 183, 184}, 
                            new [] {190, 191, 192, 193, 194} 
                        },
                        {
                            new [] {50, 51, 52, 53, 54},
                            new [] {60, 61, 62, 63, 64}, 
                            new [] {70, 71, 72, 73, 74}, 
                            new [] {80, 81, 82, 83, 84}, 
                            new [] {90, 91, 92, 93, 94} 
                        }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011f()
            {
                TestCase_011(typeof(DateTime[,]), new[,]
                    {
                        {
                            new DateTime(1967, 8, 27, 23, 30, 0),
                            new DateTime(1941, 3, 11),
                            new DateTime(1940, 12, 17),
                            new DateTime(2001, 9, 27),
                            new DateTime(2005, 12, 9)
                        },
                        {
                            new DateTime(1968, 8, 27, 23, 30, 0),
                            new DateTime(1942, 3, 11),
                            new DateTime(1941, 12, 17),
                            new DateTime(2002, 9, 27),
                            new DateTime(2006, 12, 9)
                        }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011g()
            {
                TestCase_011(typeof(Double[,]), new[,]
                    {
                        { -5.13e100d, 1, 2, 3, 4 },
                        { -15.13e100d, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011h()
            {
                TestCase_011(typeof(Single[,]), new[,]
                    {
                        { -4.24e20f, 1, 2, 3, 4 },
                        { -14.24e20f, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011i()
            {
                TestCase_011(typeof(Guid[,]), new[,]
                    {
                        { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() },
                        { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011j()
            {
                TestCase_011(typeof(Int16[,]), new[,]
                    {
                        { -32000, 32001, -32002, 32003, -32004 },
                        { -32010, 32011, -32012, 32013, -32014 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011k()
            {
                TestCase_011(typeof(UInt16[,]), new[,]
                    {
                        { 64000, 64001, 64002, 64003, 64004 },
                        { 64010, 64011, 64012, 64013, 64014 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011l()
            {
                TestCase_011(typeof(Int32[,]), new[,]
                    {
                        { -32000 * 1000, 1, 2, 3, 4 },
                        { -32010 * 1000, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011m()
            {
                TestCase_011(typeof(UInt32[,]), new[,]
                    {
                        { 64000 * 1000, 1, 2, 3, 4 },
                        { 64010 * 1000, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011n()
            {
                TestCase_011(typeof(Int64[,]), new[,]
                    {
                        { -32000 * 1000000000L, 1, 2, 3, 4 },
                        { -32010 * 1000000000L, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011o()
            {
                TestCase_011(typeof(UInt64[,]), new[,]
                    {
                        { 64000 * 1000000000L, 1, 2, 3, 4 },
                        { 64010 * 1000000000L, 11, 12, 13, 14 }
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011p()
            {
                TestCase_011(typeof(String[,]), new[,]
                    {
                        {
                            "Jak chudak do jeste vetsi nouze prisel.",
                            "Dlouhy, Siroky a Kratkozraky.",
                            "Kasparkuv hrobecek.",
                            "O trinacti tchynich.",
                            "-.-.-.-.-.-.-.-.-.-." 
                        },
                        {
                            "1-Jak chudak do jeste vetsi nouze prisel.",
                            "1-Dlouhy, Siroky a Kratkozraky.",
                            "1-Kasparkuv hrobecek.",
                            "1-O trinacti tchynich.",
                            "1--.-.-.-.-.-.-.-.-.-." 
                        }
                    });
            }
#endif

            // 20130425 NathanP:
            // This intent here is to have multiple write transactions within one call. Now, those items might be for example: 
            // a.	1xInt16 + 1xInt32 + 1xFloat etc.
            // b.	10xInt16 + 1xFloat + 100xString etc.
            // IMPROVE: make the test more complex, according to the above explanation
            /// <description>
            /// IndexRange=``. In a single call, Write to multiple nodes of each supported data-type (single-dimension array): 
            /// Bool, Byte, SByte, ByteString, DateTime, Double, Float, Guid, Int16, UInt16, Int32, UInt32, Int64, UInt64, String. 
            /// </description>
            /// <expectedresult>
            /// Client able to write a value of each data-type that it supports.
            /// Verify that values written to server match those sent by the client.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010a()
            {
                TestCase_010(typeof(Boolean[]), new[] { false, true, false, true, false });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010b()
            {
                TestCase_010(typeof(Byte[]), new[] { 200, 201, 202, 203, 203 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010c()
            {
                TestCase_010(typeof(SByte[]), new[] { -100, 101, -102, 103, 104 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010d()
            {
                TestCase_010(typeof(Byte[][]), new[]
                    {
                        new [] {150, 151, 152, 153, 154},
                        new [] {160, 161, 162, 163, 164}, 
                        new [] {170, 171, 172, 173, 174}, 
                        new [] {180, 181, 182, 183, 184}, 
                        new [] {190, 191, 192, 193, 194} 
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010f()
            {
                TestCase_010(typeof(DateTime[]), new[]
                    {
                        new DateTime(1967, 8, 27, 23, 30, 0),
                        new DateTime(1941, 3, 11),
                        new DateTime(1940, 12, 17),
                        new DateTime(2001, 9, 27),
                        new DateTime(2005, 12, 9)
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010g()
            {
                TestCase_010(typeof(Double[]), new[] { -5.13e100d, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010h()
            {
                TestCase_010(typeof(Single[]), new[] { -4.24e20f, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010i()
            {
                TestCase_010(typeof(Guid[]), new[] { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010j()
            {
                TestCase_010(typeof(Int16[]), new[] { -32000, 32001, -32002, 32003, -32004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010k()
            {
                TestCase_010(typeof(UInt16[]), new[] { 64000, 64001, 64002, 64003, 64004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010l()
            {
                TestCase_010(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010m()
            {
                TestCase_010(typeof(UInt32[]), new[] { 64000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010n()
            {
                TestCase_010(typeof(Int64[]), new[] { -32000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010o()
            {
                TestCase_010(typeof(UInt64[]), new[] { 64000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_010p()
            {
                TestCase_010(typeof(String[]), new[] 
                {
                    "Jak chudak do jeste vetsi nouze prisel.",
                    "Dlouhy, Siroky a Kratkozraky.",
                    "Kasparkuv hrobecek.",
                    "O trinacti tchynich.",
                    "-.-.-.-.-.-.-.-.-.-." });
            }

            /// <description>
            /// Write an entire array of each data-type and increase the size of the array (single-dimension array): 
            /// Bool, Byte, SByte, ByteString, DateTime, Double, Float, Guid, Int16, UInt16, Int32, UInt32, Int64, UInt64, String.
            /// Note: Test requires use of arrays whose sizes are dynamic.
            /// </description>
            /// <expectedresult>
            /// Client able to write a value of each data-type that it supports. Verify that values written to server match 
            /// those sent by the client.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012a()
            {
                TestCase_012(typeof(Boolean[]), new[] { false, true, false, true, false });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012b()
            {
                TestCase_012(typeof(Byte[]), new[] { 200, 201, 202, 203, 203 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012c()
            {
                TestCase_012(typeof(SByte[]), new[] { -100, 101, -102, 103, 104 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012d()
            {
                TestCase_012(typeof(Byte[][]), new[]
                    {
                        new [] {150, 151, 152, 153, 154},
                        new [] {160, 161, 162, 163, 164}, 
                        new [] {170, 171, 172, 173, 174}, 
                        new [] {180, 181, 182, 183, 184}, 
                        new [] {190, 191, 192, 193, 194} 
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012f()
            {
                TestCase_012(typeof(DateTime[]), new[]
                    {
                        new DateTime(1967, 8, 27, 23, 30, 0),
                        new DateTime(1941, 3, 11),
                        new DateTime(1940, 12, 17),
                        new DateTime(2001, 9, 27),
                        new DateTime(2005, 12, 9)
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012g()
            {
                TestCase_012(typeof(Double[]), new[] { -5.13e100d, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012h()
            {
                TestCase_012(typeof(Single[]), new[] { -4.24e20f, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012i()
            {
                TestCase_012(typeof(Guid[]), new[] { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012j()
            {
                TestCase_012(typeof(Int16[]), new[] { -32000, 32001, -32002, 32003, -32004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012k()
            {
                TestCase_012(typeof(UInt16[]), new[] { 64000, 64001, 64002, 64003, 64004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012l()
            {
                TestCase_012(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012m()
            {
                TestCase_012(typeof(UInt32[]), new[] { 64000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012n()
            {
                TestCase_012(typeof(Int64[]), new[] { -32000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012o()
            {
                TestCase_012(typeof(UInt64[]), new[] { 64000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012p()
            {
                TestCase_012(typeof(String[]), new[] 
                {
                    "Jak chudak do jeste vetsi nouze prisel.",
                    "Dlouhy, Siroky a Kratkozraky.",
                    "Kasparkuv hrobecek.",
                    "O trinacti tchynich.",
                    "-.-.-.-.-.-.-.-.-.-." });
            }

            /// <description>
            /// Write an entire array of each data-type and decrease the size of the array (single-dimension array): 
            /// Bool, Byte, SByte, ByteString, DateTime, Double, Float, Guid, Int16, UInt16, Int32, UInt32, Int64, UInt64, String.
            /// Note: Test requires use of arrays whose sizes are dynamic.
            /// </description>
            /// <expectedresult>
            /// Client able to write a value of each data-type that it supports. Verify that values written to server match 
            /// those sent by the client.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014a()
            {
                TestCase_014(typeof(Boolean[]), new[] { false, true, false, true, false });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014b()
            {
                TestCase_014(typeof(Byte[]), new[] { 200, 201, 202, 203, 203 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014c()
            {
                TestCase_014(typeof(SByte[]), new[] { -100, 101, -102, 103, 104 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014d()
            {
                TestCase_014(typeof(Byte[][]), new[]
                    {
                        new [] {150, 151, 152, 153, 154},
                        new [] {160, 161, 162, 163, 164}, 
                        new [] {170, 171, 172, 173, 174}, 
                        new [] {180, 181, 182, 183, 184}, 
                        new [] {190, 191, 192, 193, 194} 
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014f()
            {
                TestCase_014(typeof(DateTime[]), new[]
                    {
                        new DateTime(1967, 8, 27, 23, 30, 0),
                        new DateTime(1941, 3, 11),
                        new DateTime(1940, 12, 17),
                        new DateTime(2001, 9, 27),
                        new DateTime(2005, 12, 9)
                    });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014g()
            {
                TestCase_014(typeof(Double[]), new[] { -5.13e100d, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014h()
            {
                TestCase_014(typeof(Single[]), new[] { -4.24e20f, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014i()
            {
                TestCase_014(typeof(Guid[]), new[] { Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid() });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014j()
            {
                TestCase_014(typeof(Int16[]), new[] { -32000, 32001, -32002, 32003, -32004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014k()
            {
                TestCase_014(typeof(UInt16[]), new[] { 64000, 64001, 64002, 64003, 64004 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014l()
            {
                TestCase_014(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014m()
            {
                TestCase_014(typeof(UInt32[]), new[] { 64000 * 1000, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014n()
            {
                TestCase_014(typeof(Int64[]), new[] { -32000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014o()
            {
                TestCase_014(typeof(UInt64[]), new[] { 64000 * 1000000000L, 1, 2, 3, 4 });
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014p()
            {
                TestCase_014(typeof(String[]), new[] 
                {
                    "Jak chudak do jeste vetsi nouze prisel.",
                    "Dlouhy, Siroky a Kratkozraky.",
                    "Kasparkuv hrobecek.",
                    "O trinacti tchynich.",
                    "-.-.-.-.-.-.-.-.-.-." });
            }

#if (NETFRAMEWORK)
            /// <description>
            /// Using IndexRange, write to just one element within a single-dimension array, e.g.: 
            /// * 1st item only
            /// * 2nd item only
            /// * 4th item only
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_016_alpha()
            {
                TestCase_016(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 }, 0);
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_016_beta()
            {
                TestCase_016(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 }, 1);
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_016_gamma()
            {
                TestCase_016(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4 }, 3);
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Using IndexRange, write to every other (odd numbered) element within a single-dimension array.
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017()
            {
                TestCase_017(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4, 23, 34, 45, 56, 67 });
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Using IndexRange, write to a range of elements, i.e. positions 3 through 6 of a 10-element single-dimension 
            /// array.
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_018()
            {
                TestCase_018(typeof(Int32[]), new[] { -32000 * 1000, 1, 2, 3, 4, 23, 34, 45, 56, 67 }, 3, 6);
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Using a 2-dimensional array, write values using IndexRange=`2:3,2:3`.
            /// Repeat the test each time varying the IndexRange.
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019()
            {
                TestCase_019(typeof(Int32[,]), typeof(Int32[,]),
                    new[,]
                        {
                            { 1, 2, 3, 4, 5},
                            { 11, 12, 13, 14, 15},
                            { 21, 22, 23, 24, 25},
                            { 31, 32, 33, 34, 35},
                            { 41, 42, 43, 44, 45}
                        },
                    new[,]
                        {
                            { 101, 102 },
                            { 111, 112 }
                        });
            }
            // IMPROVE: do the varying of IndexRange
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Specify an IndexRange of `2:3` on a String.
            /// Repeat the test using a ByteString.
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_020_alpha()
            {
                TestCase_020(typeof(string), "abcdefgh", "AB");
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_020_beta()
            {
                TestCase_020(typeof(Byte[]), new Byte[] {11, 22, 33, 44, 55, 66, 77, 88}, new Byte[] { 1, 2 });
            }
#endif

#if (NETFRAMEWORK)
            /// <description>
            /// Specify an IndexRange of `2:3,2:3` on a String array.
            /// Repeat the same test on a ByteString array.
            /// </description>
            /// <expectedresult>
            /// Verify the correct value was written to the specific element of the array only.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_021_alpha()
            {
                TestCase_021(typeof(string[]), typeof(string),
                    new[] { "0abcdefgh", "1abcdefgh", "2abcdefgh", "3abcdefgh", "4abcdefgh" } ,
                    "AB");
            }

            [TestCategory("Keyword:Write")]
            [TestCategory("Keyword:Array")]
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_021_beta()
            {
                TestCase_021(typeof(Byte[][]), typeof(Byte[]),
                    new []
                        {
                            new Byte[] { 1, 2, 3, 4, 5, 6, 7, 8 },
                            new Byte[] { 11, 12, 13, 14, 15, 16, 17, 18 },
                            new Byte[] { 21, 22, 23, 24, 25, 26, 27, 28 },
                            new Byte[] { 31, 32, 33, 34, 35, 36, 37, 38 },
                            new Byte[] { 41, 42, 43, 44, 45, 46, 47, 48 }
                        }, 
                    new Byte[] { 101, 102 });
            }
#endif

#if (NETFRAMEWORK)
            // IMPROVE: we currently cannot specify a different locale
            /// <description>
            /// Write to a Node Attribute of type LocalizedText, and attempt to overwrite the text and also use a different 
            /// LocaleId.
            /// </description>
            /// <expectedresult>
            /// Confirm that the server received the text in the different locale from the client.
            /// </expectedresult>
            [TestCategory("Keyword:Write")]
            //[TestCategory("Keyword:Array")]   // deemed improper
            //[TestCategory("Keyword:Diagnostics")]   // deemed improper
            //[TestCategory("Keyword:Inferred")]   // deemed improper
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_022_alpha()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    EasyUAClient client = MakeClient();
                    UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                        typeof(int), dynamic: false, preferArray: true);

                    var arguments = new UAWriteValueArguments(
                        EndpointDescriptor,
                        nodeDescriptor,
                        "abcdefgh",
                        UAAttributeId.Description);

                    try
                    {
                        client.WriteValue(arguments);
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);

                        // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                        if (serviceException.ServiceResult != UACodeBits.BadNotWritable)
                            throw;
                    }

                    Assert.AreEqual(1, recorder.WriteRecordList.Count);
                    WriteRecord sdkWriteRecord = recorder.WriteRecordList[0];
                    Debug.Assert(sdkWriteRecord != null);

                    WriteValueCollection sdkWriteValueCollection = sdkWriteRecord.NodesToWrite;
                    Assert.IsNotNull(sdkWriteValueCollection);
                    Assert.AreEqual(1, sdkWriteValueCollection.Count);
                    WriteValue sdkWriteValue = sdkWriteValueCollection[0];
                    Assert.IsNotNull(sdkWriteValue);
                    Assert.AreEqual((uint)UAAttributeId.Description, sdkWriteValue.AttributeId);
                    DataValue sdkDataValue = sdkWriteValue.Value;
                    Assert.IsNotNull(sdkDataValue);
                    
                    Assert.AreEqual(new LocalizedText("abcdefgh"), sdkDataValue.Value);
                }
                // IMPROVE: Verify that values written to server match those sent by the client.
            }
#endif

            // ReSharper restore InconsistentNaming
#endregion
        }
    }
}
