﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    [TestClass]
    public partial class WriteBaseTest : EndpointTestBase
    {
        #region Private Constants

        private const int NodeCount10Plus = 20; // the number used when "10+" is mentioned in the spec

        #endregion



        #region Static Private Methods

        static private void AssertWriteIndexRange(int minimum, int maximum, [NotNull] WriteRecord sdkWriteRecord)
        {
            string expected = String.Format(CultureInfo.InvariantCulture, "{0}:{1}", minimum, maximum);
            AssertWriteIndexRange(expected, sdkWriteRecord);
        }

        static private void AssertWriteIndexRange(string expectedIndexRange, [NotNull] WriteRecord sdkWriteRecord)
        {
            WriteValueCollection sdkWriteValueCollection = sdkWriteRecord.NodesToWrite;
            Assert.IsNotNull(sdkWriteValueCollection);
            Assert.AreEqual(1, sdkWriteValueCollection.Count);
            WriteValue sdkWriteValue = sdkWriteValueCollection[0];
            Assert.IsNotNull(sdkWriteValue);
            string indexRange = sdkWriteValue.IndexRange;
            Assert.IsNotNull(indexRange);
            AssertUtilities.WriteLine(indexRange);
            Assert.AreEqual(expectedIndexRange, indexRange);
        }

        static private void AssertWriteSingleIndex(int index, [NotNull] WriteRecord sdkWriteRecord)
        {
            WriteValueCollection sdkWriteValueCollection = sdkWriteRecord.NodesToWrite;
            Assert.IsNotNull(sdkWriteValueCollection);
            Assert.AreEqual(1, sdkWriteValueCollection.Count);
            WriteValue sdkWriteValue = sdkWriteValueCollection[0];
            Assert.IsNotNull(sdkWriteValue);
            string indexRange = sdkWriteValue.IndexRange;
            Assert.IsNotNull(indexRange);
            AssertUtilities.WriteLine(indexRange);
            Assert.AreEqual(index.ToString(CultureInfo.InvariantCulture), indexRange);
        }

        //static private bool IsUAServiceException([NotNull] Exception exception, int codeBitsExpected)
        //{
        //    var serviceException = exception as UAServiceException;
        //    if (serviceException == null)
        //        return false;
        //    return serviceException.ServiceResult == codeBitsExpected;
        //}

#endregion



#region Test Methods
// ReSharper disable InconsistentNaming

// ReSharper restore InconsistentNaming
#endregion



#region Private Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_005(
        // ReSharper restore InconsistentNaming
            UAAttributeId attributeId, [CanBeNull] object value, [CanBeNull] object expectedSdkValue, 
            [NotNull] Predicate<Exception> expectedExceptionPredicate)
        {
            using (var recorder = new UAOperationRecorder())
            {
                AssertUtilities.WriteLine(attributeId);
                AssertUtilities.WriteLine((uint)attributeId);
                EasyUAClient client = MakeClient();
                Exception exception = null;
                try
                {
                    var arguments = new UAWriteValueArguments(
                        EndpointDescriptor,
                        UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: false), value, attributeId);
                    Debug.Assert(arguments.AttributeId == attributeId);
                    AssertUtilities.WriteLine(arguments.AttributeType);
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    exception = uaException.InnerException;
                    Assert.IsNotNull(exception);
                    AssertUtilities.WriteLine(exception.GetType());
                }
                int writeCount = recorder.WriteRecordList.Count;
                Assert.AreEqual(1, writeCount);
                WriteRecord writeRecord = recorder.WriteRecordList[0];
                Debug.Assert(writeRecord != null);
                WriteValueCollection sdkWriteValueCollection = writeRecord.NodesToWrite;
                Assert.IsNotNull(sdkWriteValueCollection);
                Assert.AreEqual(1, sdkWriteValueCollection.Count);
                WriteValue sdkWriteValue = sdkWriteValueCollection[0];
                Assert.IsNotNull(sdkWriteValue);
                Assert.AreEqual((uint)attributeId, sdkWriteValue.AttributeId);
                DataValue sdkDataValue = sdkWriteValue.Value;
                Assert.IsNotNull(sdkDataValue);
                object sdkValue = sdkDataValue.Value;
                Type expectedType = expectedSdkValue?.GetType();
                Type sdkValueType = sdkValue?.GetType();
                if ((expectedType != null) && (sdkValueType != null) && expectedType.IsArray)
                {
                    Assert.IsTrue(sdkValueType.IsArray);
                    Assert.AreEqual(expectedType, sdkValueType);

                    var expectedArray = (Array)expectedSdkValue;
                    var actualArray = (Array)sdkValue;

                    Assert.IsTrue(
                        ((IStructuralEquatable)expectedArray).Equals(actualArray, EqualityComparer<Object>.Default));
                }
                else
                    Assert.AreEqual(expectedSdkValue, sdkValue);

                Assert.IsTrue(expectedExceptionPredicate(exception));
            }
        }
#endif

        // ReSharper disable InconsistentNaming
        private void TestCase_006([NotNull] Type type, [CanBeNull] object value)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false/*, preferArray:false*/), 
                value,
                type);
            client.WriteValue(arguments);
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_007([NotNull] Type type, [CanBeNull] object value)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments1 = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false/*, preferArray:false, shared:false*/),
                value,
                type);
            var arguments2 = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray:false, shared: true),
                value,
                type);
            UAWriteResult[] resultArray = client.WriteMultipleValues(new [] {arguments1, arguments2});
            Assert.IsNotNull(resultArray);
            // ReSharper disable PossibleNullReferenceException
            Assert.IsTrue(resultArray.All(result => result.Exception == null));
            // ReSharper restore PossibleNullReferenceException
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_008_alpha([NotNull] Type type, [CanBeNull] object value)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            client.WriteValue(EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false/*, preferArray:false*/),
                value,
                type);
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_008_beta([NotNull] Type type, [CanBeNull] object value)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            try
            {
                client.WriteValue(EndpointDescriptor,
                    UATestUtilities.MakeNodeDescriptor(type, dynamic: false/*, preferArray:false*/),
                    value,
                    type);
            }
            catch (UAException uaException)
            {
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                Type innerExceptionType = innerException.GetType();
                AssertUtilities.WriteLine(innerExceptionType);
                Assert.IsTrue(
                    (innerExceptionType == typeof (OverflowException)) ||
                    (innerExceptionType == typeof (FormatException)) ||
                    (innerExceptionType == typeof(InvalidCastException)));
                return;
            }
            Assert.Fail();
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_009([NotNull] Type type, [CanBeNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray:true),
                array,
                type);
            client.WriteValue(arguments);
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_010([NotNull] Type type, [CanBeNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments1 = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false/*, preferArray:false, shared:false*/),
                array,
                type);
            var arguments2 = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray: false, shared: true),
                array,
                type);
            UAWriteResult[] resultArray = client.WriteMultipleValues(new[] { arguments1, arguments2 });
            Assert.IsNotNull(resultArray);
            // ReSharper disable PossibleNullReferenceException
            Assert.IsTrue(resultArray.All(result => result.Exception == null));
            // ReSharper restore PossibleNullReferenceException

            // IMPROVE: Verify that values written to server match those sent by the client.
        }

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_011([NotNull] Type type, [CanBeNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                Type nodeType = type.GetElementType();
                    // this is not the right type, but the servers we are using do not support 2-dimensional arrays, so we must pick some node at least
                Debug.Assert(nodeType != null);
                var arguments = new UAWriteValueArguments(
                    EndpointDescriptor,
                    UATestUtilities.MakeNodeDescriptor(nodeType, dynamic: false, preferArray: true),
                    array,
                    type);

                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                    // we know it's not gonna work - just need to test what we are sending out
                {
                    AssertServiceException(uaException.InnerException, UACodeBits.BadTypeMismatch);

                    Assert.AreEqual(1, recorder.WriteRecordList.Count);
                    WriteRecord sdkWriteRecord = recorder.WriteRecordList[0];
                    Debug.Assert(sdkWriteRecord != null);

                    WriteValueCollection sdkNodesToWrite = sdkWriteRecord.NodesToWrite;
                    Assert.IsNotNull(sdkNodesToWrite);
                    Assert.AreEqual(1, sdkNodesToWrite.Count);
                    WriteValue sdkWriteValue = sdkNodesToWrite[0];
                    Assert.IsNotNull(sdkWriteValue);
                    DataValue sdkDataValue = sdkWriteValue.Value;
                    Assert.IsNotNull(sdkDataValue);
                    object value = sdkDataValue.Value;
                    Assert.IsNotNull(value);
                    var sdkMatrix = value as Matrix;
                    Assert.IsNotNull(sdkMatrix);
                    int[] dimensions = sdkMatrix.Dimensions;
                    Assert.IsNotNull(dimensions);
                    Assert.AreEqual(2, dimensions.Length);
                    // IMPROVE: test more

                    return;
                }
                Assert.Fail();
            }
        }
#endif

        // ReSharper disable InconsistentNaming
        private void TestCase_012([NotNull] Type type, [NotNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray: true),
                array,
                type);
            client.WriteValue(arguments);

            Type arrayType = array.GetType();
            AssertUtilities.WriteLine(arrayType);
            Type elementType = arrayType.GetElementType();
            Debug.Assert(elementType != null);
            AssertUtilities.WriteLine(elementType);
            int length = array.Length;
            AssertUtilities.WriteLine(length);
            Array newArray = Array.CreateInstance(elementType, length + 1);
            AssertUtilities.WriteLine(newArray.GetType());
            array.CopyTo(newArray, 0);
            newArray.SetValue(array.GetValue(length - 1), length);
            AssertUtilities.WriteLine(newArray.Length);

            arguments.Value = newArray;

            client.WriteValue(arguments);

            // IMPROVE: Verify that values written to server match those sent by the client.
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_014([NotNull] Type type, [NotNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            EasyUAClient client = MakeClient();
            var arguments = new UAWriteValueArguments(
                EndpointDescriptor,
                UATestUtilities.MakeNodeDescriptor(type, dynamic: false, preferArray: true),
                array,
                type);
            client.WriteValue(arguments);

            Type arrayType = array.GetType();
            AssertUtilities.WriteLine(arrayType);
            Type elementType = arrayType.GetElementType();
            Debug.Assert(elementType != null);
            AssertUtilities.WriteLine(elementType);
            int length = array.Length;
            AssertUtilities.WriteLine(length);
            Array newArray = Array.CreateInstance(elementType, length - 1);
            AssertUtilities.WriteLine(newArray.GetType());
            Array.Copy(array, 0, newArray, 0, length - 1);
            AssertUtilities.WriteLine(newArray.Length);

            arguments.Value = newArray;

            client.WriteValue(arguments);

            // IMPROVE: Verify that values written to server match those sent by the client.
        }

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_016([NotNull] Type type, [NotNull] Array array, int index)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    type, dynamic: false, preferArray: true);

                var arguments1 = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    array,
                    type);
                client.WriteValue(arguments1);

                object value = array.GetValue(index ^ 1);
                AssertUtilities.WriteLine(value);

                var arguments2 = new UAWriteValueArguments(
                    EndpointDescriptor, nodeDescriptor, new[] {value}, new UAIndexRangeList {index}, type);

                try
                {
                    client.WriteValue(arguments2);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if (serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid)
                        throw;
                }

                Assert.AreEqual(2, recorder.WriteRecordList.Count);
                WriteRecord sdkWriteRecord = recorder.WriteRecordList[1];
                Debug.Assert(sdkWriteRecord != null);
                AssertWriteSingleIndex(index, sdkWriteRecord);
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_017([NotNull] Type type, [NotNull] Array array)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    type, dynamic: false, preferArray: true);

                var arguments1 = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    array,
                    type);
                client.WriteValue(arguments1);

                int count = 0;
                for (int index = 1; index < array.Length; index += 2)
                {
                    object value = array.GetValue(index);
                    AssertUtilities.WriteLine(value);

                    var arguments2 = new UAWriteValueArguments(
                        EndpointDescriptor, nodeDescriptor, new[] { value }, new UAIndexRangeList { index }, type);

                    try
                    {
                        client.WriteValue(arguments2);
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);

                        // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                        if (serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid)
                            throw;
                    }

                    count++;
                }

                Assert.AreEqual(1 + count, recorder.WriteRecordList.Count);
                for (int i = 1; i < recorder.WriteRecordList.Count; i++)
                {
                    AssertUtilities.WriteLine(i);
                    WriteRecord sdkWriteRecord = recorder.WriteRecordList[i];
                    Debug.Assert(sdkWriteRecord != null);
                    int index = (i - 1)*2 + 1;
                    AssertWriteSingleIndex(index, sdkWriteRecord);
                }
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_018([NotNull] Type type, [NotNull] Array array, int minimum, int maximum)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    type, dynamic: false, preferArray: true);

                var arguments = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    array,
                    type);
                client.WriteValue(arguments);

                Type arrayType = array.GetType();
                AssertUtilities.WriteLine(arrayType);
                Type elementType = arrayType.GetElementType();
                Debug.Assert(elementType != null);
                AssertUtilities.WriteLine(elementType);
                int length = array.Length;
                AssertUtilities.WriteLine(length);
                int newLength = maximum - minimum + 1;
                AssertUtilities.WriteLine(newLength);
                Array newArray = Array.CreateInstance(elementType, newLength);
                AssertUtilities.WriteLine(newArray.GetType());
                Array.Copy(array, minimum ^ 1, newArray, 0, newLength);
                AssertUtilities.WriteLine(newArray.Length);

                arguments.Value = newArray;
                arguments.IndexRangeList = UAIndexRangeList.OneDimension(minimum, maximum);

                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if (serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid)
                        throw;
                }

                Assert.AreEqual(2, recorder.WriteRecordList.Count);
                WriteRecord sdkWriteRecord = recorder.WriteRecordList[1];
                Debug.Assert(sdkWriteRecord != null);
                AssertWriteIndexRange(minimum, maximum, sdkWriteRecord);
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_019([NotNull] Type fullType, [NotNull] Type partialType, object fullValue, object partialValue)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                Type nodeType = fullType.GetElementType();
                Debug.Assert(nodeType != null);
                // this is not the right type, but the servers we are using do not support 2-dimensional arrays, so we must pick some node at least
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    nodeType, dynamic: false/*, preferArray: false*/);

                var arguments = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    fullValue,
                    fullType);
                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if ((serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid) &&
                        (serviceException.ServiceResult != UACodeBits.BadTypeMismatch))
                        throw;
                }

                arguments.Value = partialValue;
                arguments.IndexRangeList = new UAIndexRangeList { new UAIndexRange(2, 3), new UAIndexRange(2, 3) };
                arguments.ValueType = partialType;

                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if ((serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid) &&
                        (serviceException.ServiceResult != UACodeBits.BadTypeMismatch))
                        throw;
                }

                Assert.AreEqual(2, recorder.WriteRecordList.Count);
                WriteRecord sdkWriteRecord = recorder.WriteRecordList[1];
                Debug.Assert(sdkWriteRecord != null);
                AssertWriteIndexRange("2:3,2:3", sdkWriteRecord);
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_020([NotNull] Type type, object fullValue, object partialValue)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    type, dynamic: false/*, preferArray: false*/);

                var arguments = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    fullValue,
                    type);
                client.WriteValue(arguments);

                arguments.Value = partialValue;
                arguments.IndexRangeList = UAIndexRangeList.OneDimension(2, 3);
                arguments.ValueType = type;

                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if ((serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid) &&
                        (serviceException.ServiceResult != UACodeBits.BadTypeMismatch))
                        throw;
                }

                Assert.AreEqual(2, recorder.WriteRecordList.Count);
                WriteRecord sdkWriteRecord = recorder.WriteRecordList[1];
                Debug.Assert(sdkWriteRecord != null);
                AssertWriteIndexRange(2, 3, sdkWriteRecord);
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_021([NotNull] Type fullType, [NotNull] Type partialType, object fullValue, object partialValue)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                UANodeDescriptor nodeDescriptor = UATestUtilities.MakeNodeDescriptor(
                    fullType, dynamic: false/*, preferArray: false*/);

                var arguments = new UAWriteValueArguments(
                    EndpointDescriptor,
                    nodeDescriptor,
                    fullValue,
                    fullType);
                client.WriteValue(arguments);

                arguments.Value = partialValue;
                arguments.IndexRangeList = new UAIndexRangeList { new UAIndexRange(2, 3), new UAIndexRange(2, 3) };
                arguments.ValueType = partialType;

                try
                {
                    client.WriteValue(arguments);
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                    if ((serviceException.ServiceResult != UACodeBits.BadIndexRangeInvalid) &&
                        (serviceException.ServiceResult != UACodeBits.BadTypeMismatch))
                        throw;
                }

                Assert.AreEqual(2, recorder.WriteRecordList.Count);
                WriteRecord sdkWriteRecord = recorder.WriteRecordList[1];
                Debug.Assert(sdkWriteRecord != null);
                AssertWriteIndexRange("2:3,2:3", sdkWriteRecord);
            }
            // IMPROVE: Verify that values written to server match those sent by the client.
        }
#endif

#endregion
    }
}
