﻿using System;
using System.Diagnostics;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    public abstract partial class WriteQualityAndTimestampTest
    {
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class WriteQualityAndTimestampTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : WriteQualityAndTimestampTest_NoCtt
            // ReSharper restore InconsistentNaming
        {
            public WriteQualityAndTimestampTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class WriteQualityAndTimestampTest_NoCtt_Softing_DemoServer_Http : WriteQualityAndTimestampTest_NoCtt
            // ReSharper restore InconsistentNaming
        {
            public WriteQualityAndTimestampTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }
        }



        // ReSharper disable InconsistentNaming
        public abstract class WriteQualityAndTimestampTest_NoCtt : WriteQualityAndTimestampTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>
            /// Write to an item, changing its Value, Quality ServerTimestamp and SourceTimestamp.
            /// </description>
            /// <expectedresult>
            /// All units are to be verifiably written.
            /// </expectedresult>
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001()
            {
                var attributeData = new UAAttributeData(
                    new Random().Next(), 
                    UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning),
                    new DateTime(1967, 8, 27),
                    new DateTime(2001, 9, 27, 23, 45, 0));
                TestCase_001_to_004(attributeData);
            }

            /// <description>
            /// Write to an item, changing its Value, Quality and ServerTimestamp only.
            /// </description>
            /// <expectedresult>
            /// All units are to be verifiably written.
            /// </expectedresult>
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002()
            {
                var attributeData = new UAAttributeData(
                    new Random().Next(),
                    UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning))
                    {
                        ServerTimestamp = new DateTime(2001, 9, 27, 23, 45, 0, 0, DateTimeKind.Utc)
                    };
                TestCase_001_to_004(attributeData);
            }

            /// <description>
            /// Write to an item, changing its Value and Quality only to something like:
            /// a. Good
            /// b. Good_Clamped 
            /// c. Good_Overload
            /// d. Bad
            /// e. Bad_NotConnected
            /// </description>
            /// <expectedresult>
            /// For each quality code that a Client supports writing, must be tested and verified. If server does not allow 
            /// a quality code to be written, client must handle response from server and indicate this clearly.
            /// </expectedresult>
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003a()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                    {
                        StatusCode = new UAStatusCode(UACodeBits.Good, 0)
                    };
                TestCase_001_to_004(attributeData);
            }

            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003b()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                {
                    StatusCode = new UAStatusCode(UACodeBits.GoodClamped, 0)
                };
                TestCase_001_to_004(attributeData);
            }
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003c()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                {
                    StatusCode = new UAStatusCode(UACodeBits.GoodOverload, 0)
                };
                TestCase_001_to_004(attributeData);
            }

            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003d()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                {
                    StatusCode = new UAStatusCode(UACodeBits.Bad, 0)
                };
                TestCase_001_to_004(attributeData);
            }
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003e()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                {
                    StatusCode = new UAStatusCode(UACodeBits.BadNotConnected, 0)
                };
                TestCase_001_to_004(attributeData);
            }

            /// <description>
            /// Client to write a Vendor specific quality code that uses the lower-level bits (sub-status code).
            /// </description>
            /// <expectedresult>
            /// Client able to write a custom quality code.
            /// Client prepared to have the request rejected if the Server does not belong to the same Vendor or Server does 
            /// not allow the quality code to be written.
            /// </expectedresult>
            [TestCategory("Keyword:Write")] // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_004()
            {
                var attributeData = new UAAttributeData(new Random().Next())
                {
                    StatusCode = new UAStatusCode(UACodeBits.Uncertain, 5555)
                };
                TestCase_001_to_004(attributeData);
            }
#endif

            // ReSharper restore InconsistentNaming
#endregion


#region Private Methods

#if (NETFRAMEWORK)
            // ReSharper disable InconsistentNaming
            private void TestCase_001_to_004([NotNull] UAAttributeData attributeData)
            // ReSharper restore InconsistentNaming
            {
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        MakeClient().Write(
                            EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false),
                            attributeData);
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);

                        // Some servers simply don't support this; but we just need to check what we have sent in, primarily.
                        if (serviceException.ServiceResult != UACodeBits.BadWriteNotSupported)
                            throw;
                    }
                    int writeCount = recorder.WriteRecordList.Count;
                    Assert.AreEqual(1, writeCount);

                    WriteRecord sdkWriteRecord = recorder.WriteRecordList[0];
                    Debug.Assert(sdkWriteRecord != null);

                    WriteValueCollection sdkWriteValueCollection = sdkWriteRecord.NodesToWrite;
                    Assert.IsNotNull(sdkWriteValueCollection);
                    Assert.AreEqual(1, sdkWriteValueCollection.Count);
                    WriteValue sdkWriteValue = sdkWriteValueCollection[0];
                    Assert.IsNotNull(sdkWriteValue);

                    DataValue sdkDataValue = sdkWriteValue.Value;
                    Assert.IsNotNull(sdkDataValue);
                    Assert.AreEqual(attributeData.Value, sdkDataValue.Value);
                    Assert.AreEqual(attributeData.StatusCode.ToInt64(), (Int64)sdkDataValue.StatusCode);
                    Assert.AreEqual(attributeData.ServerTimestamp, sdkDataValue.ServerTimestamp);
                    Assert.AreEqual(attributeData.SourceTimestamp, sdkDataValue.SourceTimestamp);
                }
            }
#endif

#endregion
        }
    }
}
