﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Reflection;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.ComponentModel.Linking;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.DataAccess;
using OpcLabs.EasyOpc.UA.DataAccess.Extensions;
using OpcLabs.EasyOpc.UA.InformationModel;
using OpcLabs.EasyOpc.UA.Navigation;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DataAccess
{
    public partial class BasicTest
    {
        // ReSharper disable InconsistentNaming
        public class BasicTest_Ctt : BasicTest
        // ReSharper restore InconsistentNaming
        {
            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class BasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp : BasicTest_Ctt
            // ReSharper restore InconsistentNaming
            {
                public BasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
                }
            }



            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class BasicTest_Ctt_Softing_DemoServer_OpcTcp : BasicTest_Ctt
            // ReSharper restore InconsistentNaming
            {
                public BasicTest_Ctt_Softing_DemoServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
                }

                [ClassInitialize]
                static public void SocketPrecheck(TestContext context)
                {
                    LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
                }


                #region Test Methods
                // ReSharper disable InconsistentNaming

                /// <description>Inject the following on the DataItem Type:
                /// 1.When reading the Definition property value (when read) it is a very long string and NOT a formula.
                /// 2.When reading the ValuePrecision property value contains a negative number.
                /// 3.On a Float/Double node, when reading the ValuePrecision value returns NaN. 
                /// </description>
                /// <expectedresult>
                /// a.Client may be able to parse the string and determine the validity of the value. But it is expected that most 
                /// Clients will simply show the value only. Therefore, the Client is able to display the value correctly which 
                /// may include some form of truncating very large strings to fit any screen-space or database field length etc.
                /// b.The client reports a warning that the Server has returned a value indicating the receipt of invalid value.
                /// The Client can ignore the ValuePrecision.
                /// c.The client reports a warning that the Server has returned a value indicating the receipt of invalid value.
                /// The Client can ignore the ValuePrecision.
                /// </expectedresult>
                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_001_01()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-001-01.js"))
                    {
                        const string dataItemPath = "[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16";
                        UABrowsePath browsePath = UABrowsePath.Parse(dataItemPath, "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        // We must put something (non-null) into the definition initially, otherwise the CTT script
                        // won't be able to recognize it as a string value to be modified.
                        MakeClient().WriteValue(
                            EndpointDescriptor, 
                            UABrowsePath.Parse(dataItemPath + ".[Definition]", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                            "anInitialDefinitionString");

                        // Read the value of DataItemType using the model client
                        UAModelClient modelClient = MakeModelClient();

                        var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                        modelClient.ReadNode(dataItemNode);
                        Assert.IsNull(dataItemNode.Exception);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                        Assert.AreEqual(0, dataItemNode.ErrorCode);
                        Assert.AreEqual("", dataItemNode.ErrorMessage);
                        Assert.IsTrue(dataItemNode.StatusCode.IsGood);

                        string definition = dataItemNode.Definition;
                        Assert.IsNotNull(definition);
                        int length = definition.Length;
                        AssertUtilities.WriteLine(length);
                        AssertUtilities.WriteLine(definition);
                        Assert.IsTrue(length > 1000);
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_001_02()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-001-02.js"))
                    {
                        const string dataItemPath = "[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16";
                        UABrowsePath browsePath = UABrowsePath.Parse(dataItemPath, "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        // Read the value of DataItemType using the model client
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                        modelClient.ReadNode(dataItemNode);
                        Assert.IsNull(dataItemNode.Exception);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                        Assert.AreEqual(0, dataItemNode.ErrorCode);
                        Assert.AreEqual("", dataItemNode.ErrorMessage);
                        Assert.IsTrue(dataItemNode.StatusCode.IsGood);

                        double? valuePrecision = dataItemNode.ValuePrecision;
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "ValuePrecision: {0}", valuePrecision));
                        Assert.IsFalse(valuePrecision.HasValue);

                        Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                        UAModelAccessFailureEventArgs modelAccessFailureEventArgs = modelAccessFailureEventArgsList[0];
                        Assert.IsNotNull(modelAccessFailureEventArgs);
                        AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                        Exception exception = modelAccessFailureEventArgs.Exception;
                        Assert.IsNotNull(exception);
                        var targetAccessException = exception as TargetAccessException;
                        Assert.IsNotNull(targetAccessException);
                        AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                        Assert.IsNotNull(abstractLinkingTarget);
                        var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                        Assert.IsNotNull(linkingTarget);
                        Assert.AreEqual("ValuePrecision", linkingTarget.MemberInfo.Name);

                        Exception innerException1 = targetAccessException.InnerException;
                        Assert.IsNotNull(innerException1);
                        var targetInvocationException = innerException1 as TargetInvocationException;
                        Assert.IsNotNull(targetInvocationException);

                        Exception innerException2 = targetInvocationException.InnerException;
                        Assert.IsNotNull(innerException2);
                        StringAssert.Contains(innerException2.Message,
                            "The ValuePrecision property of OPC-UA DataItem node cannot contain a negative number");
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_001_03()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-001-03.js"))
                    {
                        const string dataItemPath = "[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16";
                        UABrowsePath browsePath = UABrowsePath.Parse(dataItemPath, "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        // Read the value of DataItemType using the model client
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                        modelClient.ReadNode(dataItemNode);
                        Assert.IsNull(dataItemNode.Exception);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                        Assert.AreEqual(0, dataItemNode.ErrorCode);
                        Assert.AreEqual("", dataItemNode.ErrorMessage);
                        Assert.IsTrue(dataItemNode.StatusCode.IsGood);

                        double? valuePrecision = dataItemNode.ValuePrecision;
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "ValuePrecision: {0}", valuePrecision));
                        Assert.IsFalse(valuePrecision.HasValue);

                        Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                        UAModelAccessFailureEventArgs modelAccessFailureEventArgs = modelAccessFailureEventArgsList[0];
                        Assert.IsNotNull(modelAccessFailureEventArgs);
                        AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                        Exception exception = modelAccessFailureEventArgs.Exception;
                        Assert.IsNotNull(exception);
                        var targetAccessException = exception as TargetAccessException;
                        Assert.IsNotNull(targetAccessException);
                        AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                        Assert.IsNotNull(abstractLinkingTarget);
                        var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                        Assert.IsNotNull(linkingTarget);
                        Assert.AreEqual("ValuePrecision", linkingTarget.MemberInfo.Name);

                        Exception innerException1 = targetAccessException.InnerException;
                        Assert.IsNotNull(innerException1);
                        var targetInvocationException = innerException1 as TargetInvocationException;
                        Assert.IsNotNull(targetInvocationException);

                        Exception innerException2 = targetInvocationException.InnerException;
                        Assert.IsNotNull(innerException2);
                        StringAssert.Contains(innerException2.Message,
                            "The ValuePrecision property of OPC-UA DataItem node cannot contain a NaN value (not a number)");
                    }
                }

                /// <description>Inject the following error codes on an Analog Item Type
                /// 1.When reading the InstrumentRange is returned with NaN values for:
                /// • High value only
                /// • Low value only
                /// • High and Low values
                /// 2.When reading the InstrumentRange is returned with matching values, e.g. 0.
                /// 3.When reading the EngineeringUnits the DisplayName is null/empty.
                /// 4.When reading the EngineeringUnits the Description is null/empty.
                /// 5.When reading the EngineeringUnits the namespcaeURi is null/empty.
                /// 6.When reading the EURange a NaN is returned for:
                /// • High value only
                /// • Low value only
                /// • High and Low values
                /// 7.When reading the EURange the High and Low values are switched.
                /// 8.When reading the EURange the High and Low values are the same 
                /// </description>
                /// <expectedresult>
                /// a.Client reports an error to the user. The client is unable to use the InstrumentRange values since they 
                /// are invalid/incomplete.
                /// The Client can ignore the InstrumentRange.
                /// b.Client reports an error to the user. The client is unable to use the InstrumentRange values since they 
                /// are invalid/incomplete.
                /// The Client can ignore the InstrumentRange
                /// c.The Client should be able to still display the EngineeringUnits based on the DisplayName and Description 
                /// values (within the EUInformation structure).
                /// If the Client provides specialized behavior for well-defined EngineeringUnit definitions then the Client 
                /// could optionally ignore the value and report a reason to the end-user.
                /// d.The Client assumes that no EngineeringUnits are defined for the selected Node.
                /// e.The Client simply uses an empty-string for the description.
                /// f.In every case the EURange is either incomplete or invalid.
                /// The Client should report a warning to the end-user that the Server is returning incomplete/invalid values.
                /// The Client should not use the EURange.
                /// g.The Client may report a warning to the end-user.
                /// h.Client reports an error to the user. The client is unable to use the EURange values since they are 
                /// invalid.
                /// The Client can ignore the EURange.
                /// </expectedresult>
                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_01()
                {
                    Ctt.ClaimScriptBuggyVersion(0x92f2f46e);
                    Ctt.ClaimScriptBuggyVersion(0x43c6526a);
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-01.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);

                        {
                            modelClient.ReadNode(analogItemNode,
                                                 // ReSharper disable AssignNullToNotNullAttribute
                                                 enumerable => !enumerable.Contains(UAModellingTags.MandatoryProperty));
                            // ReSharper restore AssignNullToNotNullAttribute
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            UARange instrumentRange = analogItemNode.InstrumentRange;
                            Assert.IsNull(instrumentRange);

                            Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[0];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("InstrumentRange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            StringAssert.Contains(innerException2.Message,
                                                  "The InstrumentRange.High property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)");
                        }

                        {
                            modelClient.ReadNode(analogItemNode,
                                // ReSharper disable AssignNullToNotNullAttribute
                                                 enumerable => !enumerable.Contains(UAModellingTags.MandatoryProperty));
                            // ReSharper restore AssignNullToNotNullAttribute
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            UARange instrumentRange = analogItemNode.InstrumentRange;
                            Assert.IsNull(instrumentRange);

                            Assert.AreEqual(2, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[1];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("InstrumentRange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            StringAssert.Contains(innerException2.Message,
                                                  "The InstrumentRange.Low property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)");
                        }

                        {
                            modelClient.ReadNode(analogItemNode,
                                // ReSharper disable AssignNullToNotNullAttribute
                                                 enumerable => !enumerable.Contains(UAModellingTags.MandatoryProperty));
                            // ReSharper restore AssignNullToNotNullAttribute
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            UARange instrumentRange = analogItemNode.InstrumentRange;
                            Assert.IsNull(instrumentRange);

                            Assert.AreEqual(3, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[2];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("InstrumentRange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            Assert.IsTrue(
                                innerException2.Message.Contains(
                                    "The InstrumentRange.High property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)") ||
                                innerException2.Message.Contains(
                                    "The InstrumentRange.Low property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)"));
                        }
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_02()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-02.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                        modelClient.ReadNode(analogItemNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                        Assert.AreEqual(0, analogItemNode.ErrorCode);
                        Assert.AreEqual("", analogItemNode.ErrorMessage);
                        Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                        UARange instrumentRange = analogItemNode.InstrumentRange;
                        Assert.IsNull(instrumentRange);

                        Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                        UAModelAccessFailureEventArgs modelAccessFailureEventArgs = modelAccessFailureEventArgsList[0];
                        Assert.IsNotNull(modelAccessFailureEventArgs);
                        AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                        Exception exception = modelAccessFailureEventArgs.Exception;
                        Assert.IsNotNull(exception);
                        var targetAccessException = exception as TargetAccessException;
                        Assert.IsNotNull(targetAccessException);
                        AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                        Assert.IsNotNull(abstractLinkingTarget);
                        var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                        Assert.IsNotNull(linkingTarget);
                        Assert.AreEqual("InstrumentRange", linkingTarget.MemberInfo.Name);

                        Exception innerException1 = targetAccessException.InnerException;
                        Assert.IsNotNull(innerException1);
                        var targetInvocationException = innerException1 as TargetInvocationException;
                        Assert.IsNotNull(targetInvocationException);

                        Exception innerException2 = targetInvocationException.InnerException;
                        Assert.IsNotNull(innerException2);
                        StringAssert.Contains(innerException2.Message,
                            "The InstrumentRange property of OPC-UA AnalogItem node cannot contain an empty range (a range with the same High and Low values)");
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_03()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-03.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                        modelClient.ReadNode(analogItemNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                        Assert.AreEqual(0, analogItemNode.ErrorCode);
                        Assert.AreEqual("", analogItemNode.ErrorMessage);
                        Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                        UAEUInformation engineeringUnits = analogItemNode.EngineeringUnits;
                        Assert.IsNotNull(engineeringUnits);
                        Assert.AreEqual("", engineeringUnits.DisplayName);
                        string description = engineeringUnits.Description;
                        Assert.AreNotEqual("", description);
                        StringAssert.Contains(description, engineeringUnits.ToString());
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_04()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-04.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                        modelClient.ReadNode(analogItemNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                        Assert.AreEqual(0, analogItemNode.ErrorCode);
                        Assert.AreEqual("", analogItemNode.ErrorMessage);
                        Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                        UAEUInformation engineeringUnits = analogItemNode.EngineeringUnits;
                        Assert.IsNotNull(engineeringUnits);
                        string description = engineeringUnits.Description;
                        Assert.AreEqual("", description);
                        Assert.AreEqual("", engineeringUnits.ToString());
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_05()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-05.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                        modelClient.ReadNode(analogItemNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                        Assert.AreEqual(0, analogItemNode.ErrorCode);
                        Assert.AreEqual("", analogItemNode.ErrorMessage);
                        Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                        UAEUInformation engineeringUnits = analogItemNode.EngineeringUnits;
                        Assert.IsNotNull(engineeringUnits);
                        string namespaceUriString = engineeringUnits.NamespaceUriString;
                        Assert.AreEqual("", namespaceUriString);
                        Assert.AreEqual("", engineeringUnits.ToString());
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_06()
                {
                    Ctt.ClaimScriptBuggyVersion(0x57c633c8);
                    Ctt.ClaimScriptBuggyVersion(0x4711fd56);
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-06.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);

                        {
                            modelClient.ReadNode(analogItemNode, UAModellingTags.MandatoryProperty);
                            // ReSharper restore AssignNullToNotNullAttribute
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            Assert.IsTrue(analogItemNode.IsDefaultEURange);

                            Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[0];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("EURange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            StringAssert.Contains(innerException2.Message,
                                                  "The EURange.High property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)");
                        }

                        {
                            modelClient.ReadNode(analogItemNode, UAModellingTags.MandatoryProperty);
                            // ReSharper restore AssignNullToNotNullAttribute
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            Assert.IsTrue(analogItemNode.IsDefaultEURange);

                            Assert.AreEqual(2, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[1];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("EURange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            StringAssert.Contains(innerException2.Message,
                                                  "The EURange.Low property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)");
                        }

                        {
                            modelClient.ReadNode(analogItemNode, UAModellingTags.MandatoryProperty);
                            AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}",
                                                          analogItemNode));

                            Assert.IsTrue(analogItemNode.IsDefaultEURange);

                            Assert.AreEqual(3, modelAccessFailureEventArgsList.Count);
                            UAModelAccessFailureEventArgs modelAccessFailureEventArgs =
                                modelAccessFailureEventArgsList[2];
                            Assert.IsNotNull(modelAccessFailureEventArgs);
                            AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                            Exception exception = modelAccessFailureEventArgs.Exception;
                            Assert.IsNotNull(exception);
                            var targetAccessException = exception as TargetAccessException;
                            Assert.IsNotNull(targetAccessException);
                            AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                            Assert.IsNotNull(abstractLinkingTarget);
                            var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                            Assert.IsNotNull(linkingTarget);
                            Assert.AreEqual("EURange", linkingTarget.MemberInfo.Name);

                            Exception innerException1 = targetAccessException.InnerException;
                            Assert.IsNotNull(innerException1);
                            var targetInvocationException = innerException1 as TargetInvocationException;
                            Assert.IsNotNull(targetInvocationException);

                            Exception innerException2 = targetInvocationException.InnerException;
                            Assert.IsNotNull(innerException2);
                            Assert.IsTrue(
                                innerException2.Message.Contains(
                                    "The EURange.High property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)") ||
                                innerException2.Message.Contains(
                                    "The EURange.Low property of OPC-UA AnalogItem node cannot contain a NaN value (not a number)"));
                        }
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")] // inferred
                [TestMethod]
                public void TestCase_Err_002_07()
                {
                    using (
                        Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-07.js"))
                    {
                        EasyUAClient client = MakeClient();

                        UANodeDescriptor nodeDescriptor = UABrowsePath.Parse(
                            "[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float.[EURange]",
                            "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        ValueResult[] resultArray = client.ReadMultipleValues(new[]
                            {
                                new UAReadArguments(EndpointDescriptor, nodeDescriptor)
                            });
                        Assert.IsNotNull(resultArray);
                        Assert.AreEqual(1, resultArray.Length);

                        ValueResult result = resultArray[0];
                        Assert.IsNotNull(result);
                        AssertUtilities.WriteLine(result);
                        Assert.IsNull(result.Exception);
                        object value = result.Value;
                        Assert.IsNotNull(value);
                        AssertUtilities.WriteLine(value);
                        var range = value as UARange;
                        Assert.IsNotNull(range);
                        Assert.IsTrue(range.IsReversed);

                        ExceptionCollection diagnostics = result.Diagnostics;
                        Assert.IsNotNull(diagnostics);
                        Assert.IsTrue(diagnostics.Count >= 1);
                        // ReSharper disable AssignNullToNotNullAttribute
                        Assert.IsTrue(diagnostics.Any(exception => IsClientEngineException(exception, UAClientErrorCodes.RangeLimitsReversed)));
                        Exception warning = diagnostics.First(exception => IsClientEngineException(exception, UAClientErrorCodes.RangeLimitsReversed));
                        // ReSharper restore AssignNullToNotNullAttribute
                        Assert.IsNotNull(warning);
                        AssertUtilities.WriteLine("Warning found:");
                        AssertUtilities.WriteLine(warning.Message);
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_002_08()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-002-08.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();
                        var modelAccessFailureEventArgsList = new List<UAModelAccessFailureEventArgs>();
                        modelClient.ModelAccessFailure += (sender, args) => modelAccessFailureEventArgsList.Add(args);

                        UANodeDescriptor nodeDescriptor =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                        modelClient.ReadNode(analogItemNode, UAModellingTags.MandatoryProperty);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                        Assert.AreEqual(0, analogItemNode.ErrorCode);
                        Assert.AreEqual("", analogItemNode.ErrorMessage);
                        Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                        Assert.IsTrue(analogItemNode.IsDefaultEURange);

                        Assert.AreEqual(1, modelAccessFailureEventArgsList.Count);
                        UAModelAccessFailureEventArgs modelAccessFailureEventArgs = modelAccessFailureEventArgsList[0];
                        Assert.IsNotNull(modelAccessFailureEventArgs);
                        AssertUtilities.WriteLine(modelAccessFailureEventArgs);

                        Exception exception = modelAccessFailureEventArgs.Exception;
                        Assert.IsNotNull(exception);
                        var targetAccessException = exception as TargetAccessException;
                        Assert.IsNotNull(targetAccessException);
                        AbstractLinkingTarget abstractLinkingTarget = targetAccessException.LinkingTarget;
                        Assert.IsNotNull(abstractLinkingTarget);
                        var linkingTarget = abstractLinkingTarget as ObjectMemberLinkingTarget;
                        Assert.IsNotNull(linkingTarget);
                        Assert.AreEqual("EURange", linkingTarget.MemberInfo.Name);

                        Exception innerException1 = targetAccessException.InnerException;
                        Assert.IsNotNull(innerException1);
                        var targetInvocationException = innerException1 as TargetInvocationException;
                        Assert.IsNotNull(targetInvocationException);

                        Exception innerException2 = targetInvocationException.InnerException;
                        Assert.IsNotNull(innerException2);
                        StringAssert.Contains(innerException2.Message,
                            "The EURange property of OPC-UA AnalogItem node cannot contain an empty range (a range with the same High and Low values)");
                    }
                }

                /// <description>Inject the following error codes on a TwoStateDiscreteType
                ///
                /// 1.When reading a node of this Type, the text returned for the TrueState is empty, and the value returned 
                /// is True.
                /// 2.When reading a node of this Type, the text returned for the FalseState is empty, and the value returned 
                /// is False.
                /// </description>
                /// <expectedresult>
                /// a.Client may simply choose to display/use empty text, or override and use a text-equivalent to True, i.e. 
                /// “True”.
                /// b.Client may simply choose to display/use empty text, or override and use a text-equivalent to False, i.e. 
                /// “False”. 
                /// </expectedresult>
                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_003_01_alpha()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-003-01.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath = UABrowsePath.Parse(
                            "[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", 
                            "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, browsePath);
                        twoStateDiscreteNode.SubstituteEmptyStates = false;

                        twoStateDiscreteNode.Value = true;
                        modelClient.WriteNode(twoStateDiscreteNode, UAModellingTags.VariableValue);

                        modelClient.ReadNode(twoStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                        Assert.IsNull(twoStateDiscreteNode.Exception);
                        Assert.AreEqual(0, twoStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", twoStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(twoStateDiscreteNode.StatusCode.IsGood);

                        Assert.IsTrue(twoStateDiscreteNode.Value);
                        Assert.AreEqual("", twoStateDiscreteNode.CurrentState);
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_003_01_beta()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-003-01.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath = UABrowsePath.Parse(
                            "[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1",
                            "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, browsePath);

                        twoStateDiscreteNode.Value = true;
                        modelClient.WriteNode(twoStateDiscreteNode, UAModellingTags.VariableValue);

                        modelClient.ReadNode(twoStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                        Assert.IsNull(twoStateDiscreteNode.Exception);
                        Assert.AreEqual(0, twoStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", twoStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(twoStateDiscreteNode.StatusCode.IsGood);

                        Assert.IsTrue(twoStateDiscreteNode.Value);
                        Assert.AreEqual("True", twoStateDiscreteNode.CurrentState);
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_003_02_alpha()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-003-02.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath = UABrowsePath.Parse(
                            "[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1",
                            "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, browsePath);
                        twoStateDiscreteNode.SubstituteEmptyStates = false;

                        twoStateDiscreteNode.Value = false;
                        modelClient.WriteNode(twoStateDiscreteNode, UAModellingTags.VariableValue);

                        modelClient.ReadNode(twoStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                        Assert.IsNull(twoStateDiscreteNode.Exception);
                        Assert.AreEqual(0, twoStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", twoStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(twoStateDiscreteNode.StatusCode.IsGood);

                        Assert.IsFalse(twoStateDiscreteNode.Value);
                        Assert.AreEqual("", twoStateDiscreteNode.CurrentState);
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_003_02_beta()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-003-02.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath = UABrowsePath.Parse(
                            "[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1",
                            "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                        var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, browsePath);

                        twoStateDiscreteNode.Value = false;
                        modelClient.WriteNode(twoStateDiscreteNode, UAModellingTags.VariableValue);

                        modelClient.ReadNode(twoStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                        Assert.IsNull(twoStateDiscreteNode.Exception);
                        Assert.AreEqual(0, twoStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", twoStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(twoStateDiscreteNode.StatusCode.IsGood);

                        Assert.IsFalse(twoStateDiscreteNode.Value);
                        Assert.AreEqual("False", twoStateDiscreteNode.CurrentState);
                    }
                }

                /// <description>Inject the following error codes on a MultiState Discrete Type
                /// 1.When reading a node of this Type the value returned contains a value that exceeds the size of the enumerated list, e.g.: 
                /// • Value returned = 100
                /// • Enumerated list length = 5.
                /// 2.When reading a node of this Type the value returned contains a value that is within the index range but the value within the enumerated list is empty. 
                /// </description>
                /// <expectedresult>
                /// Client reports the error/warning to the end-user.Client uses the raw value and ignores the EnumStrings.
                /// </expectedresult>
                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_004_01()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-004-01.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/MultiState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        var multiStateDiscreteNode = modelClient.GetMultiStateDiscreteNode<Int64, UInt32>(EndpointDescriptor, browsePath);
                        modelClient.ReadNode(multiStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "MultiStateDiscreteNode: {0}", multiStateDiscreteNode));

                        Assert.IsNull(multiStateDiscreteNode.Exception);
                        Assert.AreEqual(0, multiStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", multiStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(multiStateDiscreteNode.StatusCode.IsGood);

                        Int64 value = multiStateDiscreteNode.Value;
                        AssertUtilities.WriteLine(value);
                        
                        string displayValue = multiStateDiscreteNode.DisplayValue();
                        AssertUtilities.WriteLine(displayValue);
                        StringAssert.Contains(displayValue, "invalid!");
                        StringAssert.Contains(displayValue, value.ToString(CultureInfo.CurrentCulture));
                        string currentState = multiStateDiscreteNode.CurrentState;
                        AssertUtilities.WriteLine(currentState);
                        StringAssert.Contains(currentState, value.ToString(CultureInfo.CurrentCulture));
                    }
                }

                [TestCategory("Live")]
                [TestCategory("OpcCompliance")]
                [TestCategory("Type:CTT")]  // inferred
                [TestMethod]
                public void TestCase_Err_004_02()
                {
                    using (Ctt.RunScript(CttProjectName, @"Data Access\Data Access Client Basic\Test Cases\Err-004-02.js"))
                    {
                        UAModelClient modelClient = MakeModelClient();

                        UABrowsePath browsePath =
                            UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/MultiState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                        var multiStateDiscreteNode = modelClient.GetMultiStateDiscreteNode<Int64, UInt32>(EndpointDescriptor, browsePath);
                        modelClient.ReadNode(multiStateDiscreteNode);
                        AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "MultiStateDiscreteNode: {0}", multiStateDiscreteNode));

                        Assert.IsNull(multiStateDiscreteNode.Exception);
                        Assert.AreEqual(0, multiStateDiscreteNode.ErrorCode);
                        Assert.AreEqual("", multiStateDiscreteNode.ErrorMessage);
                        Assert.IsTrue(multiStateDiscreteNode.StatusCode.IsGood);

                        Int64 value = multiStateDiscreteNode.Value;
                        AssertUtilities.WriteLine(value);

                        string displayValue = multiStateDiscreteNode.DisplayValue();
                        AssertUtilities.WriteLine(displayValue);
                        StringAssert.Contains(displayValue, "invalid!");
                        StringAssert.Contains(displayValue, value.ToString(CultureInfo.CurrentCulture));
                        string currentState = multiStateDiscreteNode.CurrentState;
                        AssertUtilities.WriteLine(currentState);
                        StringAssert.Contains(currentState, value.ToString(CultureInfo.CurrentCulture));
                    }
                }

                // ReSharper restore InconsistentNaming
                #endregion
            }



            #region Test Methods
            // ReSharper disable InconsistentNaming

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            #endregion
        }
    }
}
