﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.QualityTools.Testing.Fakes;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua.Fakes;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.DataAccess;
using OpcLabs.EasyOpc.UA.DataAccess.Extensions;
using OpcLabs.EasyOpc.UA.InformationModel;
using OpcLabs.EasyOpc.UA.Navigation;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;
#if (NETFRAMEWORK)
#endif
#if (NETFRAMEWORK)
#endif

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DataAccess
{
    public partial class BasicTest
    {
        // ReSharper disable InconsistentNaming
        public class BasicTest_NoCtt : BasicTest
        // ReSharper restore InconsistentNaming
        {
#region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>Client able to Browse and display the Nodes of DataItemType and also query the References and the 
            /// [optional] Properties.
            /// </description> 
            /// <expectedresult>Information is processed accordingly.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001_alpha()
            {
                // Browsing for the DataItemType nodes themselves
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Float"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Int16"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001_beta()
            {
                // Browsing into the DataItemType nodes
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Definition"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "ValuePrecision"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001_gamma_Traditional()
            {
                // Read the value of DataItemType using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(Int16), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001_gamma_ModelClient()
            {
                // Read the value of DataItemType using the model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableValue);
                Assert.IsNull(dataItemNode.Exception);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                Assert.AreEqual(0, dataItemNode.ErrorCode);
                Assert.AreEqual("", dataItemNode.ErrorMessage);
                Assert.IsTrue(dataItemNode.StatusCode.IsGood);

                Int16 value = dataItemNode.Value;
                AssertUtilities.WriteLine(value);
            }

            /// <description>Client able to read/display/use the optional Definition property
            /// </description> 
            /// <expectedresult>If the property is not available then the Client can choose to report the missing property, or 
            /// simply ignore and move on.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002_alpha_Traditional()
            {
                // Read just the Definition property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "Definition");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                if (value == null)
                    return;

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(String), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002_alpha_ModelClient()
            {
                // Obtain the Definition property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableProperty);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                string definition = dataItemNode.Definition;
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Definition: {0}", definition));
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002_beta_Traditional()
            {
                // Error - Read just the Definition property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "Definition");
                try
                {
                    MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                catch (UAException uaException)
                {
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadNoMatch, serviceException.ServiceResult.StatusCode.CodeBits);

                    return;
                }
                Assert.Fail();
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002_beta_ModelClient()
            {
                // Error - Obtain the Definition property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableProperty);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                string definition = dataItemNode.Definition;
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Definition: {0}", definition));
            }

            /// <description>The Definition property value is an empty string.
            /// </description> 
            /// <expectedresult>This is valid.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003_Traditional()
            {
                // using traditional means
                EasyUAClient client = MakeClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "Definition");
                client.WriteValue(EndpointDescriptor, browsePath, "");
                object value = client.ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Assert.AreEqual("", value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003_ModelClient()
            {
                // using model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                dataItemNode.Definition = "";
                modelClient.WriteNode(dataItemNode, UAModellingTags.VariableProperty);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableProperty);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));
                string definition = dataItemNode.Definition;
                Assert.IsNotNull(definition);
                Assert.AreEqual("", definition);
            }

            /// <description>Write a value to the Definition property 
            /// </description> 
            /// <expectedresult>Client should be prepared for the call to succeed or fail. If the call fails then the Client 
            /// reports the error.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_004_alpha_Traditional()
            {
                // using traditional means
                var random = new Random();
                EasyUAClient client = MakeClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "Definition");
                string testDefinition = String.Format(CultureInfo.InvariantCulture, "TestDefinition({0})", random.Next());
                AssertUtilities.WriteLine(testDefinition);
                client.WriteValue(EndpointDescriptor, browsePath, testDefinition);
                object value = client.ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);
                Assert.AreEqual(testDefinition, value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_004_alpha_ModelClient()
            {
                // using the model client
                var random = new Random();
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                string testDefinition = String.Format(CultureInfo.InvariantCulture, "TestDefinition({0})", random.Next());
                AssertUtilities.WriteLine(testDefinition);
                var dataItemNode = modelClient.GetDataItemNode<float>(EndpointDescriptor, browsePath);
                dataItemNode.Definition = testDefinition;
                modelClient.WriteNode(dataItemNode, UAModellingTags.VariableProperty);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableProperty);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));
                string definition = dataItemNode.Definition;
                Assert.IsNotNull(definition);
                Assert.AreEqual(testDefinition, definition);
            }

            // IMPROVE: TestCase_004_beta_XXXX, test the failure cases

            /// <description>Client able to read/display/use the optional ValuePrecision property - if available.
            /// </description> 
            /// <expectedresult>If the property is not available then the Client can choose to report the missing property, or 
            /// simply ignore and move on.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005_alpha_Traditional()
            {
                // Read just the ValuePrecision property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "ValuePrecision");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(Double), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005_alpha_ModelClient()
            {
                // Obtain the ValuePrecision property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(dataItemNode);
                Assert.IsNull(dataItemNode.Exception);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                double? valuePrecision = dataItemNode.ValuePrecision;
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "ValuePrecision: {0}", valuePrecision));
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005_beta_Traditional()
            {
                // Error - Read just the ValuePrecision property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "ValuePrecision");
                try
                {
                    MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                catch (UAException uaException)
                {
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadNoMatch, serviceException.ServiceResult.StatusCode.CodeBits);

                    return;
                }
                Assert.Fail();
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_005_beta_ModelClient()
            {
                // Error - Obtain the ValuePrecision property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(dataItemNode);
                Assert.IsNull(dataItemNode.Exception);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));

                double? valuePrecision = dataItemNode.ValuePrecision;
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "ValuePrecision: {0}", valuePrecision));
            }

            /// <description>Write a value to the ValuePrecision property - if available.
            /// </description> 
            /// <expectedresult>Client should be prepared for the call to succeed or fail. If the call fails then the Client 
            /// reports the error.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006_alpha_Traditional()
            {
                // using traditional means
                var random = new Random();
                EasyUAClient client = MakeClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "ValuePrecision");
                double testValuePrecision = random.Next(10000) / 100.0;
                AssertUtilities.WriteLine(testValuePrecision);
                client.WriteValue(EndpointDescriptor, browsePath, testValuePrecision);
                object value = client.ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);
                Assert.AreEqual(testValuePrecision, value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_006_alpha_ModelClient()
            {
                // using the model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                double testValuePrecision = new Random().NextDouble();
                AssertUtilities.WriteLine(testValuePrecision);
                var dataItemNode = modelClient.GetDataItemNode<float>(EndpointDescriptor, browsePath);
                dataItemNode.ValuePrecision = testValuePrecision;
                modelClient.WriteNode(dataItemNode, UAModellingTags.VariableProperty);
                modelClient.ReadNode(dataItemNode, UAModellingTags.VariableProperty);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "DataItemNode: {0}", dataItemNode));
                Assert.IsTrue(dataItemNode.ValuePrecision.HasValue);
                double valuePrecision = dataItemNode.ValuePrecision.Value;
                Assert.AreEqual(testValuePrecision, valuePrecision);
            }

            // IMPROVE: TestCase_006_beta_XXXX, test the failure cases

            /// <description>Write a value to any Node (of this type).
            /// </description> 
            /// <expectedresult>Client correctly formats the value (according to the data-type of the Node) and issues the write.
            /// Client should be prepared for the call to succeed or fail and to report any errors to the end-user.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007_alpha_Traditional()
            {
                EasyUAClient client = MakeClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var testValue = (Int16)new Random().Next(Int16.MinValue, Int16.MaxValue + 1);
                AssertUtilities.WriteLine(testValue);
                client.WriteValue(EndpointDescriptor, browsePath, testValue);

                object value = client.ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Assert.AreEqual(testValue, value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007_alpha_ModelClient_1()
            {
                // the right way of doing it
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);

                var testValue = (Int16)new Random().Next(Int16.MinValue, Int16.MaxValue + 1);
                AssertUtilities.WriteLine(testValue);
                dataItemNode.Value = testValue;
                modelClient.WriteNode(dataItemNode, UAModellingTags.VariableValue);

                Assert.IsNull(dataItemNode.Exception);
                Assert.AreEqual(0, dataItemNode.ErrorCode);
                Assert.AreEqual("", dataItemNode.ErrorMessage);
                Assert.IsTrue(dataItemNode.StatusCode.IsGood);

                Int16 value = dataItemNode.Value;
                Assert.AreEqual(testValue, value);
            }

            [ExpectedException(typeof(InvalidCastException))]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007_alpha_ModelClient_2()
            {
                // try to fool it by using UABaseVariable with a type of 'object'
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                modelClient.GetNode<UABaseVariable<object, object>>(EndpointDescriptor, browsePath);
            }

            [ExpectedException(typeof(InvalidCastException))]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_007_alpha_ModelClient_3()
            {
                // try to fool it by using UADataItemNode with a wider type
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                modelClient.GetDataItemNode<Int32>(EndpointDescriptor, browsePath);
            }

            // IMPROVE: have TestCase_007_beta_XXXX, test the failure cases (errors from the server side)

            /// <description>Client able to add and remove MonitoredItems of this type to/from a Subscription.
            /// </description> 
            /// <expectedresult>There`s no reason for Nodes of this type to behave any differently to other Nodes. Therefore 
            /// these calls are expected to succeed unless the Server refuses based on its level of support and/or access rights.
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("TimeDomain")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011_alpha_Traditional()
            {
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                    MakeClient(),
                    new EasyUAMonitoredItemArguments(null, EndpointDescriptor, browsePath, 100),
                    MillisecondsToCollectSingleNotification);

                // ReSharper disable PossibleMultipleEnumeration
                int count = eventArgsEnumerable.Count();
                AssertUtilities.WriteLine($"count: {count}");
                Assert.IsTrue(count >= 1);
                AssertNoFailures(eventArgsEnumerable);
                // ReSharper restore PossibleMultipleEnumeration
            }

            [TestCategory("Live")]
            [TestCategory("TimeDomain")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_011_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);
                int count = 0;
                dataItemNode.AttributeDataChanged += (sender, args) => { count++; };
                modelClient.SubscribeNode(true, dataItemNode, UAModellingTags.VariableValue);
                try
                {
                    AssertUtilities.WaitingPhase(MillisecondsToCollectSingleNotification);
                    AssertUtilities.WriteLine($"count: {count}");
                    Assert.IsTrue(count >= 1);
                    Assert.IsNull(dataItemNode.Exception);
                    AssertUtilities.WriteLine(dataItemNode.Value);
                }
                finally
                {
                    modelClient.SubscribeNode(false, dataItemNode, UAModellingTags.VariableValue);
                    // IMPROVE: use dynamic item, and verify that no more changes are coming
                }
            }

            // IMPROVE: have TestCase_011_beta_XXXX, for the described server failure response

#if (NETFRAMEWORK)
            /// <description>Client able to add and remove MonitoredItems of this type to/from a Subscription while also 
            /// specifying a DeadbandAbsolute filter of 10.
            /// </description> 
            /// <expectedresult>There`s no reason for Nodes of this type to behave any differently to other Nodes. Therefore 
            /// these calls are expected to succeed unless the Server refuses based on its level of support and/or access rights.
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("TimeDomain")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012_alpha_Traditional()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    const double deadbandValue = 10;
                    UABrowsePath browsePath = UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16",
                                                                 "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                    EasyUAClient client = MakeClient();
                    IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        client,
                        new EasyUAMonitoredItemArguments(null, EndpointDescriptor, browsePath,
                            new UAMonitoringParameters(100, new UADataChangeFilter(UADeadbandType.Absolute, deadbandValue))),
                        MillisecondsToCollectSingleNotification);

                    AssertAbsoluteDeadband(recorder, deadbandValue);

                    // ReSharper disable PossibleMultipleEnumeration
                    int count = eventArgsEnumerable.Count();
                    AssertUtilities.WriteLine($"count: {count}");
                    Assert.IsTrue(count >= 1);
                    AssertNoFailures(eventArgsEnumerable);
                    // ReSharper restore PossibleMultipleEnumeration
                }
            }
#endif

#if (NETFRAMEWORK)
            [TestCategory("Live")]
            [TestCategory("TimeDomain")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_012_alpha_ModelClient()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    const double deadbandValue = 10;
                    var modelClient = new UAModelClient { Client = { Isolated = true } };
                    modelClient.ResolveNodeParameters += (sender, args) =>
                    {
                        // ReSharper disable PossibleNullReferenceException
                        args.MonitoringParameters.DataChangeFilter = new UADataChangeFilter(UADeadbandType.Absolute, deadbandValue);
                        // ReSharper restore PossibleNullReferenceException
                    };

                    UABrowsePath browsePath =
                        UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DataItem/Int16",
                                           "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                    var dataItemNode = modelClient.GetDataItemNode<Int16>(EndpointDescriptor, browsePath);

                    int count = 0;
                    dataItemNode.AttributeDataChanged += (sender, args) => { count++; };
                    modelClient.SubscribeNode(true, dataItemNode, UAModellingTags.VariableValue);
                    try
                    {
                        AssertUtilities.WaitingPhase(MillisecondsToCollectSingleNotification);

                        AssertAbsoluteDeadband(recorder, deadbandValue);

                        AssertUtilities.WriteLine($"count: {count}");
                        Assert.IsTrue(count >= 1);
                        Assert.IsNull(dataItemNode.Exception);
                        AssertUtilities.WriteLine(dataItemNode.Value);
                    }
                    finally
                    {
                        modelClient.SubscribeNode(false, dataItemNode, UAModellingTags.VariableValue);
                        // IMPROVE: use dynamic item, and verify that no more changes are coming
                    }
                }
            }
#endif

            // IMPROVE: have TestCase_012_beta_XXXX, for the described server failure response

            /// <description>Client able to Browse and display the Nodes of AnalogItemType and also query the References and the 
            /// [optional and mandatory] Properties.
            /// </description> 
            /// <expectedresult>Information is processed accordingly.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_alpha()
            {
                // Browsing for the AnalogItemType nodes themselves
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Float"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Int16"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_beta()
            {
                // Browsing into the AnalogItemType nodes
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                //Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "Definition"));
                //Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "ValuePrecision"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "InstrumentRange"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "EURange"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "EngineeringUnits"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_gamma_Traditional()
            {
                // Read the value of AnalogItemType using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(Int16), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_013_gamma_ModelClient()
            {
                // Read the value of AnalogItemType using the model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<Int16>(EndpointDescriptor, browsePath);
                modelClient.ReadNode(analogItemNode, UAModellingTags.VariableValue);
                Assert.IsNull(analogItemNode.Exception);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.AreEqual(0, analogItemNode.ErrorCode);
                Assert.AreEqual("", analogItemNode.ErrorMessage);
                Assert.IsTrue(analogItemNode.StatusCode.IsGood);

                Int16 value = analogItemNode.Value;
                AssertUtilities.WriteLine(value);
            }

            /// <description>Write a value to any Node (of this type) for each data-type, i.e. Int, Float, DateTime etc.
            /// </description> 
            /// <expectedresult>Client correctly formats the value (according to the data-type of the Node) and issues the write.
            /// Client should be prepared for the call to succeed or fail and to report any errors to the end-user.
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_SByte()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/SByte", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(SByte),
                    (SByte)11, (SByte)12);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Byte()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Byte", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Byte),
                    (Byte)21, (Byte)22);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Int16()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Int16),
                    (Int16)31, (Int16)32);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_UInt16()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(UInt16),
                    (UInt16)41, (UInt16)42);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Int32()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Int32),
                    // ReSharper disable RedundantCast
                    (Int32)51, (Int32)52);
                // ReSharper restore RedundantCast
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_UInt32()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(UInt32),
                    (UInt32)61, (UInt32)62);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Int64()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int64", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Int64),
                    (Int64)51, (Int64)52);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_UInt64()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt64", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(UInt64),
                    (UInt64)61, (UInt64)62);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Double()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Double", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Double),
                    // ReSharper disable RedundantCast
                    (Double)1.1, (Double)1.2);
                // ReSharper restore RedundantCast
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_Float()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Single),
                    (Single)2.1, (Single)2.2);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfInt16()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Int16[]),
                    new[] { (Int16)31, (Int16)33, (Int16)35, (Int16)37 }, new[] { (Int16)32, (Int16)34, (Int16)36, (Int16)38 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfInt32()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Int32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Int32[]),
                    // ReSharper disable RedundantCast
                    new[] { (Int32)41, (Int32)43, (Int32)45, (Int32)47 }, new[] { (Int32)42, (Int32)44, (Int32)46, (Int32)48 });
                // ReSharper restore RedundantCast
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfUInt16()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/UInt16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(UInt16[]),
                    new[] { (UInt16)31, (UInt16)33, (UInt16)35, (UInt16)37 }, new[] { (UInt16)32, (UInt16)34, (UInt16)36, (UInt16)38 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfUInt32()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/UInt32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(UInt32[]),
                    new[] { (UInt32)41, (UInt32)43, (UInt32)45, (UInt32)47 }, new[] { (UInt32)42, (UInt32)44, (UInt32)46, (UInt32)48 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfFloat()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Single[]),
                    new[] { (Single)1.1, (Single)1.3, (Single)1.5, (Single)1.7 }, new[] { (Single)1.2, (Single)1.4, (Single)1.6, (Single)1.8 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_Traditional_ArrayOfDouble()
            {
                TestCase_014_alpha_Traditional(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Double", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    typeof(Double[]),
                    // ReSharper disable RedundantCast
                    new[] { (Double)2.1, (Double)2.3, (Double)2.5, (Double)2.7 }, new[] { (Double)2.2, (Double)2.4, (Double)2.6, (Double)2.8 });
                // ReSharper restore RedundantCast
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_SByte()
            {
                TestCase_014_alpha_ModelClient<Int16, SByte>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/SByte", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    11, 12);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Byte()
            {
                TestCase_014_alpha_ModelClient<Byte>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Byte", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    21, 22);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Int16()
            {
                TestCase_014_alpha_ModelClient<Int16>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    31, 32);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_UInt16()
            {
                TestCase_014_alpha_ModelClient<Int32, UInt16>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    41, 42);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Int32()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Int32>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    51, 52);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_UInt32()
            {
                TestCase_014_alpha_ModelClient<Int64, UInt32>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    61, 62);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Int64()
            {
                TestCase_014_alpha_ModelClient<Int64>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Int64", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    51, 52);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_UInt64()
            {
                TestCase_014_alpha_ModelClient<Decimal, UInt64>(
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/UInt64", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    61, 62);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Double()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Double>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Double", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    1.1, 1.2);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_Float()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Single>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    2.1f, 2.2f);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfInt16()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Int16[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Int16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new Int16[] { 31, 33, 35, 37 }, new Int16[] { 32, 34, 36, 38 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfInt32()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Int32[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Int32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new[] { 41, 43, 45, 47 }, new[] { 42, 44, 46, 48 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfUInt16()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Int32[], UInt16[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/UInt16", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new[] { 31, 33, 35, 37 }, new[] { 32, 34, 36, 38 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfUInt32()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Int64[], UInt32[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/UInt32", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new Int64[] { 41, 43, 45, 47 }, new Int64[] { 42, 44, 46, 48 });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfFloat()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Single[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new[] { 1.1f, 1.3f, 1.5f, 1.7f }, new[] { 1.2f, 1.4f, 1.6f, 1.8f });
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_014_alpha_ModelClient_ArrayOfDouble()
            {
                // ReSharper disable RedundantTypeArgumentsOfMethod
                TestCase_014_alpha_ModelClient<Double[]>(
                    // ReSharper restore RedundantTypeArgumentsOfMethod
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType Arrays/Double", "http://opcfoundation.org/Quickstarts/ReferenceApplications"),
                    new[] { 2.1, 2.3, 2.5, 2.7 }, new[] { 2.2, 2.4, 2.6, 2.8 });
            }

            // IMPROVE: test more data types - currently we are doing only those supported by the Softing UA Demo Server
            // IMPROVE: have TestCase_014_beta_XXXX, for the described server failure response

            /// <description>Client able to read/display/use the optional InstrumentRange property - if available. This 
            /// means that the Client is able to read/display/use both the min and max values of the range.
            /// </description> 
            /// <expectedresult>If the property is not available then the Client can choose to report the missing property, 
            /// or simply ignore and move on.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_015_alpha_Traditional()
            {
                // Read just the InstrumentRange property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "InstrumentRange");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                var instrumentRange = value as UARange;
                Assert.IsNotNull(instrumentRange);
                Assert.AreEqual(0.0, instrumentRange.Low);
                Assert.AreEqual(120.0, instrumentRange.High);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_015_alpha_ModelClient()
            {
                // Obtain the InstrumentRange property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange instrumentRange = analogItemNode.InstrumentRange;
                Assert.IsNotNull(instrumentRange);
                Assert.AreEqual(0.0, instrumentRange.Low);
                Assert.AreEqual(120.0, instrumentRange.High);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_015_beta_Traditional()
            {
                // Error - Read just the InstrumentRange property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "InstrumentRange");
                try
                {
                    MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                catch (UAException uaException)
                {
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadNoMatch, serviceException.ServiceResult.StatusCode.CodeBits);

                    return;
                }
                Assert.Fail();
            }

            [ExpectedException(typeof(InvalidCastException))]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_015_beta_ModelClient()
            {
                // Error - Obtain the InstrumentRange property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
            }

#if (NETFRAMEWORK)
            /// <description>The InstrumentRange returned has the High and Low values reversed, where:
            /// * High=10; Low=100.
            /// * High=-10; Low=-20;
            /// </description> 
            /// <expectedresult>The Client might not care if the values are transposed and that would be fine. If the Client 
            /// uses the InstrumentRange to determine valid values to write to the Node then it should warn the end-user that 
            /// the values received are back-to-front.
            /// Client should be prepared for the call to succeed or fail and to report any errors to the end-user.
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestCategory("Unavailable")]
            [TestMethod]
            public void TestCase_016_alpha_Traditional()
            {
                TestCase_016_Traditional(100, 10);
            }
#endif

#if (NETFRAMEWORK)
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_016_beta_Traditional()
            {
                TestCase_016_Traditional(-10, -20); // modified with respect to description that was incorrect
            }
#endif

#if (NETFRAMEWORK)
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestCategory("Unavailable")]
            [TestMethod]
            public void TestCase_016_alpha_ModelClient()
            {
                TestCase_016_ModelClient(100, 10);
            }
#endif

#if (NETFRAMEWORK)
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_016_beta_ModelClient()
            {
                TestCase_016_ModelClient(-10, -20); // modified with respect to description that was incorrect
            }
#endif

            // IMPROVE: have TestCase_016_gamma_XXXX, for the described server failure response

            /// <description>Client able to read/display/use the EURange property.
            /// </description> 
            /// <expectedresult>If the property is not available then the Client can choose to report the missing property, 
            /// or simply ignore and move on.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017_alpha_Traditional()
            {
                // Read just the EURange property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "EURange");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                var euRange = value as UARange;
                Assert.IsNotNull(euRange);
                Assert.AreEqual(0.0, euRange.Low);
                Assert.AreEqual(100.0, euRange.High);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017_alpha_ModelClient()
            {
                // Obtain the EURange property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange euRange = analogItemNode.EURange;
                Assert.IsNotNull(euRange);
                Assert.AreEqual(0.0, euRange.Low);
                Assert.AreEqual(100.0, euRange.High);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017_beta_Traditional()
            {
                // Error - Read just the EURange property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "EURange");
                try
                {
                    MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                catch (UAException uaException)
                {
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadNoMatch, serviceException.ServiceResult.StatusCode.CodeBits);

                    return;
                }
                Assert.Fail();
            }

            [ExpectedException(typeof(InvalidCastException))]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_017_beta_ModelClient()
            {
                // Error - Obtain the EURange property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
            }

            /// <description>Client able to read/display/use the optional EngineeringUnits property - if available. This 
            /// means that the Client is able to read/display/use the DisplayName and/or Description.
            /// </description> 
            /// <expectedresult>If the property is not available then the Client can choose to report the missing property, 
            /// or simply ignore and move on.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_018_alpha_Traditional()
            {
                // Read just the EngineeringUnits property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "EngineeringUnits");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                var engineeringUnits = value as UAEUInformation;
                Assert.IsNotNull(engineeringUnits);
                AssertUtilities.WriteLine(engineeringUnits);
                Assert.AreEqual("liters", engineeringUnits.DisplayName);
                Assert.AreEqual("liters", engineeringUnits.Description);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_018_alpha_ModelClient()
            {
                // Obtain the EngineeringUnits property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UAEUInformation engineeringUnits = analogItemNode.EngineeringUnits;
                Assert.IsNotNull(engineeringUnits);
                AssertUtilities.WriteLine(engineeringUnits);
                Assert.AreEqual("liters", engineeringUnits.DisplayName);
                Assert.AreEqual("liters", engineeringUnits.Description);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_018_beta_Traditional()
            {
                // Error - Read just the EngineeringUnits property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "EngineeringUnits");
                try
                {
                    MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                catch (UAException uaException)
                {
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadNoMatch, serviceException.ServiceResult.StatusCode.CodeBits);

                    return;
                }
                Assert.Fail();
            }

            [ExpectedException(typeof(InvalidCastException))]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_018_beta_ModelClient()
            {
                // Error - Obtain the EngineeringUnits property value using the model client
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/All Profiles/Scalar/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
            }

            /// <description>Client able to Browse and display the Nodes of DiscreteItemType derived types, and also query the 
            /// [mandatory] Properties.
            /// </description> 
            /// <expectedresult>Information is processed accordingly.</expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_alpha()
            {
                // Browsing for the nodes themselves
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "MultiState1"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "TwoState1"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_beta_MultiStateDiscreteType()
            {
                // Browsing into the MultiStateDiscreteType nodes
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/MultiState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "EnumStrings"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_beta_TwoStateDiscreteType()
            {
                // Browsing into the TwoStateDiscreteType nodes
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                UANodeElementCollection nodeElementCollection = MakeClient().BrowseDataVariables(EndpointDescriptor, browsePath);
                Assert.IsNotNull(nodeElementCollection);

                // ReSharper disable PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "FalseState"));
                Assert.IsTrue(nodeElementCollection.Any(element => element.BrowseName.Name == "TrueState"));
                // ReSharper restore PossibleNullReferenceException
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_gamma_Traditional_MultiStateDiscreteType()
            {
                // Read the value of MultiStateDiscreteType using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/MultiState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(Int64), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_gamma_TraditionalTwoStateDiscreteType()
            {
                // Read the value of TwoStateDiscreteType using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                object value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                Assert.IsNotNull(value);

                Type type = value.GetType();
                AssertUtilities.WriteLine(type);
                AssertUtilities.WriteLine(value);
                Assert.AreEqual(typeof(bool), type);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_gamma_ModelClient_MultiStateDiscreteType()
            {
                // Read the value of MultiStateDiscreteType using the model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/MultiState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var multiStateDiscreteNode = modelClient.GetMultiStateDiscreteNode<Int64, UInt32>(EndpointDescriptor, browsePath);
                modelClient.ReadNode(multiStateDiscreteNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "MultiStateDiscreteNode: {0}", multiStateDiscreteNode));

                Assert.IsNull(multiStateDiscreteNode.Exception);
                Assert.AreEqual(0, multiStateDiscreteNode.ErrorCode);
                Assert.AreEqual("", multiStateDiscreteNode.ErrorMessage);
                Assert.IsTrue(multiStateDiscreteNode.StatusCode.IsGood);

                Int64 value = multiStateDiscreteNode.Value;
                AssertUtilities.WriteLine(value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_019_gamma_ModelClient_TwoStateDiscreteType()
            {
                // Read the value of TwoStateDiscreteType using the model client
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, browsePath);
                modelClient.ReadNode(twoStateDiscreteNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                Assert.IsNull(twoStateDiscreteNode.Exception);
                Assert.AreEqual(0, twoStateDiscreteNode.ErrorCode);
                Assert.AreEqual("", twoStateDiscreteNode.ErrorMessage);
                Assert.IsTrue(twoStateDiscreteNode.StatusCode.IsGood);

                bool value = twoStateDiscreteNode.Value;
                AssertUtilities.WriteLine(value);
            }

            /// <description>Using an AnalogItemType node with an InstrumentRange property
            /// 1. Write a value that is within the defined range. 
            /// 2. Write a value that exceeds the lower value of the range.
            /// 3. Write a value that exceeds the upper value of the range.
            /// </description> 
            /// <expectedresult>Step Number Action Expected Result(s) 
            /// 1  
            /// Client correctly formats the value (according to the data-type of the Node) and issues the write. Client should 
            /// be prepared for the call to succeed or fail and to report any errors to the end-user.  
            /// 2  
            /// Client correctly formats the value (according to the data-type of the Node) and issues the write. Client should 
            /// be prepared for the call to succeed or fail and to report any errors to the end-user.  
            /// 3  
            /// Client correctly formats the value (according to the data-type of the Node) and issues the write. Client should 
            /// be prepared for the call to succeed or fail and to report any errors to the end-user.  
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_020_01_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange instrumentRange = analogItemNode.InstrumentRange;
                Assert.IsNotNull(instrumentRange);

                // The server actually checks the values against EURange, so we need to make sure we are inside EURange as well.
                UARange range = null;
                try
                {
                    range = UARange.Intersect(instrumentRange, analogItemNode.EURange);
                }
                catch (ArithmeticException)
                {
                    Assert.Inconclusive();  // we cannot determine the right test value in this case
                }
                AssertUtilities.WriteLine(range);
                var testValue = (float)(range.LowBound + range.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNull(exception);

                object value = MakeClient().ReadValue(EndpointDescriptor, nodeDescriptor);
                Assert.AreEqual(testValue, value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_020_02_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange instrumentRange = analogItemNode.InstrumentRange;
                Assert.IsNotNull(instrumentRange);

                // The server actually checks the values against EURange, so we need to make sure we are outside EURange as well.
                var testValue = (float)(Math.Min(instrumentRange.LowBound, analogItemNode.EURange.LowBound) -
                    instrumentRange.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNotNull(exception);

                var serviceException = exception as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadOutOfRange, serviceException.ServiceResult.StatusCode.CodeBits);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_020_03_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange instrumentRange = analogItemNode.InstrumentRange;
                Assert.IsNotNull(instrumentRange);

                // The server actually checks the values against EURange, so we need to make sure we are outside EURange as well.
                var testValue = (float)(Math.Max(instrumentRange.HighBound, analogItemNode.EURange.HighBound) +
                    instrumentRange.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNotNull(exception);

                var serviceException = exception as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadOutOfRange, serviceException.ServiceResult.StatusCode.CodeBits);
            }

            // IMPROVE: have TestCase_020_XX_beta, for the described server failure response

            /// <description>Using any AnalogItemType node:
            /// 1. Write a value that is within the defined EURange property range. 
            /// 2. Write a value that exceeds the lower value of the EURange property range. 
            /// 3. Write a value that exceeds the upper value of the EURange property range. 
            /// </description> 
            /// <expectedresult>Client correctly formats the value (according to the data-type of the Node) and issues the write.
            /// Client should be prepared for the call to succeed or fail and to report any errors to the end-user. 
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_021_01_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange euRange = analogItemNode.EURange;
                Assert.IsNotNull(euRange);

                var testValue = (float)(euRange.LowBound + euRange.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNull(exception);

                object value = MakeClient().ReadValue(EndpointDescriptor, nodeDescriptor);
                Assert.AreEqual(testValue, value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_021_02_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange euRange = analogItemNode.EURange;
                Assert.IsNotNull(euRange);

                var testValue = (float)(euRange.LowBound - euRange.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNotNull(exception);

                var serviceException = exception as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadOutOfRange, serviceException.ServiceResult.StatusCode.CodeBits);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_021_03_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var analogItemNode = modelClient.GetAnalogItemNode<float>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(analogItemNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "AnalogItemNode: {0}", analogItemNode));

                Assert.IsNull(analogItemNode.Exception);
                UARange euRange = analogItemNode.EURange;
                Assert.IsNotNull(euRange);

                var testValue = (float)(euRange.HighBound + euRange.Size * new Random().NextDouble());
                AssertUtilities.WriteLine(testValue);
                analogItemNode.Value = testValue;
                modelClient.WriteNode(analogItemNode, UAModellingTags.VariableValue);
                Exception exception = analogItemNode.Exception;
                AssertUtilities.WriteLine(exception);
                Assert.IsNotNull(exception);

                var serviceException = exception as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadOutOfRange, serviceException.ServiceResult.StatusCode.CodeBits);
            }

            // IMPROVE: have TestCase_021_XX_beta, for the described server failure response

            /// <description>TwoStateDiscreteType. 
            /// 1. Client able to read the value of a Node of this type and display/use it in raw format (i.e. True/False) 
            /// 2. Write a value (True/False) to a node of this type.
            /// 3. Subscribe to a node of this type and prove that dataChange notifications can be received and processed. 
            /// Unsubscribe the node. 
            /// 4. Client able read the value of a Node of this type and to display use the LocalizedText value that represents 
            /// the discrete value. This test will be verified when the value is True and then again when it is False. 
            /// 5. Client tries to write a new string value to the TrueState and/or FalseState property. 
            /// </description> 
            /// <expectedresult>Step Number Action Expected Result(s) 
            /// 1  
            /// Information is used accordingly by the Client.  
            /// 2  
            /// The request is issued correctly by the client. Any errors returned from the Server are reported to the end-user.  
            /// 3  
            /// This is standard behavior and therefore should be correctly processed by the Client. Any errors returned from 
            /// the Server are reported to the end-user.  
            /// 4  
            /// The Client displays/uses the LocalizedText value if supported, or uses the raw value (true/false) instead.  
            /// 5  
            /// Client should be prepared for the call to succeed or fail. Any errors are to be reported to the end-user.  
            /// </expectedresult>
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_022_01_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(twoStateDiscreteNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));

                Assert.IsNull(twoStateDiscreteNode.Exception);
                bool value = twoStateDiscreteNode.Value;
                AssertUtilities.WriteLine(value);
            }

            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_022_02_alpha_ModelClient()
            {
                UAModelClient modelClient = MakeModelClient();
                UANodeDescriptor nodeDescriptor =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");

                var twoStateDiscreteNode = modelClient.GetNode<UATwoStateDiscreteNode>(EndpointDescriptor, nodeDescriptor);
                modelClient.ReadNode(twoStateDiscreteNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));
                Assert.IsNull(twoStateDiscreteNode.Exception);

                bool testValue = !twoStateDiscreteNode.Value;
                twoStateDiscreteNode.Value = testValue;
                modelClient.WriteNode(twoStateDiscreteNode, UAModellingTags.VariableValue);
                Assert.IsNull(twoStateDiscreteNode.Exception);

                twoStateDiscreteNode.Value = !testValue;
                modelClient.ReadNode(twoStateDiscreteNode);
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "TwoStateDiscreteNode: {0}", twoStateDiscreteNode));
                Assert.IsNull(twoStateDiscreteNode.Exception);
                Assert.AreEqual(testValue, twoStateDiscreteNode.Value);
            }

            //private class CollectingTwoStateDiscreteNode : UATwoStateDiscreteNode
            //{
            //    public override bool Value
            //    {
            //        get
            //        {
            //            return base.Value;
            //        }
            //        set
            //        {
            //            base.Value = value;
            //        }
            //    }
            //}

            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_022_03_alpha_ModelClient()
            //{
            //    UAModelClient modelClient = MakeModelClient();
            //    UANodeDescriptor nodeDescriptor =
            //        UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/DiscreteType/TwoState1", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
            // TODO: !


            //}

            // IMPROVE: have TestCase_022_XX_beta, for the described server failure response

            // ReSharper restore InconsistentNaming
#endregion



#region Private Methods

            // ReSharper disable InconsistentNaming
            private void TestCase_014_alpha_ModelClient<TValue>([NotNull] UANodeDescriptor nodeDescriptor,
                // ReSharper restore InconsistentNaming
                TValue testValue1, TValue testValue2)
            {
                TestCase_014_alpha_ModelClient<TValue, TValue>(nodeDescriptor, testValue1, testValue2);
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_014_alpha_ModelClient<TValue, TAttribute>([NotNull] UANodeDescriptor nodeDescriptor,
                // ReSharper restore InconsistentNaming
                TValue testValue1, TValue testValue2)
            {
                UAModelClient modelClient = MakeModelClient();
                var node = modelClient.GetAnalogItemNode<TValue, TAttribute>(EndpointDescriptor, nodeDescriptor);

                node.Value = testValue1;
                modelClient.WriteNode(node, UAModellingTags.VariableValue);
                AssertUtilities.WriteLine(node.ErrorMessage);
                Assert.IsNull(node.Exception);
                node.Value = testValue2;
                modelClient.ReadNode(node, UAModellingTags.VariableValue);
                AssertUtilities.WriteLine(node.ErrorMessage);
                Assert.IsNull(node.Exception);
                TValue value1 = node.Value;
                // ReSharper disable CompareNonConstrainedGenericWithNull
                if (typeof(TValue).IsArray && (testValue1 != null))
                    // ReSharper restore CompareNonConstrainedGenericWithNull
                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue((testValue1 as IEnumerable).Cast<object>().Select(
                        Cls.ConvertToClsType).SequenceEqual((value1 as IEnumerable).Cast<object>()));
                // ReSharper restore AssignNullToNotNullAttribute
                else
                    Assert.AreEqual(Cls.ConvertToClsType(testValue1), value1);

                node.Value = testValue2;
                modelClient.WriteNode(node, UAModellingTags.VariableValue);
                AssertUtilities.WriteLine(node.ErrorMessage);
                Assert.IsNull(node.Exception);
                node.Value = testValue1;
                modelClient.ReadNode(node, UAModellingTags.VariableValue);
                AssertUtilities.WriteLine(node.ErrorMessage);
                Assert.IsNull(node.Exception);
                TValue value2 = node.Value;
                // ReSharper disable CompareNonConstrainedGenericWithNull
                if (typeof(TValue).IsArray && (testValue2 != null))
                    // ReSharper restore CompareNonConstrainedGenericWithNull
                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue((testValue2 as IEnumerable).Cast<object>().Select(
                        Cls.ConvertToClsType).SequenceEqual((value2 as IEnumerable).Cast<object>()));
                // ReSharper restore AssignNullToNotNullAttribute
                else
                    Assert.AreEqual(Cls.ConvertToClsType(testValue2), value2);
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_014_alpha_Traditional([NotNull] UANodeDescriptor nodeDescriptor, [NotNull] Type type,
                // ReSharper restore InconsistentNaming
                object testValue1, object testValue2)
            {
                EasyUAClient client = MakeClient();

                client.WriteValue(EndpointDescriptor, nodeDescriptor, testValue1, type);
                object value1 = client.ReadValue(EndpointDescriptor, nodeDescriptor);
                if (type.IsArray && (testValue1 != null))
                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue((testValue1 as IEnumerable).Cast<object>().Select(
                        Cls.ConvertToClsType).SequenceEqual((value1 as IEnumerable).Cast<object>()));
                // ReSharper restore AssignNullToNotNullAttribute
                else
                    Assert.AreEqual(Cls.ConvertToClsType(testValue1), value1);

                client.WriteValue(EndpointDescriptor, nodeDescriptor, testValue2, type);
                object value2 = client.ReadValue(EndpointDescriptor, nodeDescriptor);
                if (type.IsArray && (testValue2 != null))
                    // ReSharper disable AssignNullToNotNullAttribute
                    Assert.IsTrue((testValue2 as IEnumerable).Cast<object>().Select(
                        Cls.ConvertToClsType).SequenceEqual((value2 as IEnumerable).Cast<object>()));
                // ReSharper restore AssignNullToNotNullAttribute
                else
                    Assert.AreEqual(Cls.ConvertToClsType(testValue2), value2);
            }

#if (NETFRAMEWORK)
            // ReSharper disable InconsistentNaming
            private void TestCase_016_Traditional(double low, double high)
            // ReSharper restore InconsistentNaming
            {
                // Read just the InstrumentRange property value using traditional means
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                browsePath += new UAQualifiedName(UAUtilities.OpcUANamespaceUriString, "InstrumentRange");
                object value;
                using (ShimsContext.Create())
                {
                    // ReSharper disable PossibleNullReferenceException
                    ShimRange.AllInstances.LowSetDouble = (range, d) => ShimsContext.ExecuteWithoutShims(() => range.Low = low);
                    ShimRange.AllInstances.HighSetDouble = (range, d) => ShimsContext.ExecuteWithoutShims(() => range.High = high);
                    // ReSharper restore PossibleNullReferenceException
                    value = MakeClient().ReadValue(EndpointDescriptor, browsePath);
                }
                Assert.IsNotNull(value);

                var instrumentRange = value as UARange;
                Assert.IsNotNull(instrumentRange);
                AssertUtilities.WriteLine(instrumentRange);

                Assert.AreEqual(low, instrumentRange.Low);
                Assert.AreEqual(high, instrumentRange.High);
                // The IsReversed property - this is our "warning"
                Assert.IsTrue(instrumentRange.IsReversed);
                StringAssert.Contains(instrumentRange.ToString(), "reversed!");
                // The LowBound and HighBound properties
                Assert.AreEqual(high, instrumentRange.LowBound);
                Assert.AreEqual(low, instrumentRange.HighBound);
                // The InRange method
                Assert.IsTrue(instrumentRange.InRange((low + high) / 2.0));
                Assert.IsFalse(instrumentRange.InRange(high - 10.0));
                Assert.IsFalse(instrumentRange.InRange(low + 10.0));
            }
#endif

#if (NETFRAMEWORK)
            // ReSharper disable InconsistentNaming
            private void TestCase_016_ModelClient(double low, double high)
            // ReSharper restore InconsistentNaming
            {
                UAModelClient modelClient = MakeModelClient();
                UABrowsePath browsePath =
                    UABrowsePath.Parse("[ObjectsFolder]/CTT/Static/DA Profile/AnalogType/Float", "http://opcfoundation.org/Quickstarts/ReferenceApplications");
                var node = modelClient.GetAnalogItemNode<float, float>(EndpointDescriptor, browsePath);
                using (ShimsContext.Create())
                {
                    // ReSharper disable PossibleNullReferenceException
                    ShimRange.AllInstances.LowSetDouble = (range, d) => ShimsContext.ExecuteWithoutShims(() => range.Low = low);
                    ShimRange.AllInstances.HighSetDouble = (range, d) => ShimsContext.ExecuteWithoutShims(() => range.High = high);
                    // ReSharper restore PossibleNullReferenceException
                    modelClient.ReadNode(node, UAModellingTags.VariableProperty);
                }

                var instrumentRange = node.InstrumentRange;
                Assert.IsNotNull(instrumentRange);
                AssertUtilities.WriteLine(instrumentRange);

                Assert.AreEqual(low, instrumentRange.Low);
                Assert.AreEqual(high, instrumentRange.High);
                // The IsReversed property - this is our "warning"
                Assert.IsTrue(instrumentRange.IsReversed);
                StringAssert.Contains(instrumentRange.ToString(), "reversed!");
                // The LowBound and HighBound properties
                Assert.AreEqual(high, instrumentRange.LowBound);
                Assert.AreEqual(low, instrumentRange.HighBound);
                // The InRange method
                Assert.IsTrue(instrumentRange.InRange((low + high) / 2.0));
                Assert.IsFalse(instrumentRange.InRange(high - 10.0));
                Assert.IsFalse(instrumentRange.InRange(low + 10.0));
            }
#endif

#endregion
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BasicTest_NoCtt_Softing_DemoServer_Http : BasicTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public BasicTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BasicTest_NoCtt_Softing_DemoServer_OpcTcp : BasicTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public BasicTest_NoCtt_Softing_DemoServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }
    }
}
