﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DataAccess
{
    public partial class DeadbandTest
    {
        // ReSharper disable InconsistentNaming
        public class DeadbandTest_NoCtt : DeadbandTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>Monitor an Integer node while specifying a DeadbandPercent value of 10.
            /// </description> 
            /// <expectedresult>Client correctly requests the Deadband value.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]    // because we need to wait for the notification a little
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001()
            {
                const double deadbandValue = 10.0;
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(int), dynamic: true);
                var dataChangeFilter = new UADataChangeFilter(UADeadbandType.Percent, deadbandValue);
                monitoredItemArguments.MonitoringParameters.DataChangeFilter = dataChangeFilter;

                TestCase_001_And_002(monitoredItemArguments);
            }

            /// <description>Monitor a Float node while specifying a DeadbandPercent value of 0.5.
            /// </description> 
            /// <expectedresult>Client correctly requests the Deadband value.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]    // because we need to wait for the notification a little
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002()
            {
                const double deadbandValue = 0.5;
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(float), dynamic: true);
                //var monitoredItemArguments = 
                //    new EasyUAMonitoredItemArguments(null, EndpointDescriptor,
                //        new UANodeDescriptor(UABrowsePath.Parse("[ObjectsFolder]/Boilers/Boiler #1/DrumX001/LIX001/Output", "http://opcfoundation.org/UA/Boiler/")), 
                //        100);
                var dataChangeFilter = new UADataChangeFilter(UADeadbandType.Percent, deadbandValue);
                monitoredItemArguments.MonitoringParameters.DataChangeFilter = dataChangeFilter;

                TestCase_001_And_002(monitoredItemArguments);
            }
#endif

            /// <description>Specify a non-Analog Node (any Node that does not have EURange defined) and try to monitor it while 
            /// specifying a DeadbandPercent value of 10.
            /// </description> 
            /// <expectedresult>Ideally the Client should not permit the request from the end-user.
            /// Client reports the error returned from the Server.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_003()
            {
                const double deadbandValue = 10.0;
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(bool), dynamic: true);
                var dataChangeFilter = new UADataChangeFilter(UADeadbandType.Percent, deadbandValue);
                monitoredItemArguments.MonitoringParameters.DataChangeFilter = dataChangeFilter;

                using (EasyUAClient client = MakeClient())
                {
                    IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        client, monitoredItemArguments, MillisecondsToCollectMultipleNotifications);

                    // ReSharper disable PossibleMultipleEnumeration
                    int count = eventArgsEnumerable.Count();
                    Assert.AreEqual(1, count);
                    EasyUAMonitoredItemChangedEventArgs eventArgs = eventArgsEnumerable.First();
                    Debug.Assert(eventArgs != null);
                    // ReSharper restore PossibleMultipleEnumeration
                    Exception exception = eventArgs.Exception;
                    Assert.IsNotNull(exception);
                    Debug.WriteLine(exception.ToString());
                    var serviceException = exception as UAServiceException;
                    Assert.IsNotNull(serviceException);
                    Assert.AreEqual(UACodeBits.BadFilterNotAllowed, serviceException.ServiceResult.StatusCode.CodeBits);
                }
            }

            // The CTT script for the test cases below exists, but doesn't appear to work, which is in line with the test being mark "Unavailable"
            ///// <description>Error injection for Bad Service Results
            ///// 1.ServiceResult = Bad_FilterNotAllowed.
            ///// 2.ServiceResult = Bad_MonitoredItemFilterInvalid
            ///// 3.ServiceResult = Bad_MonitoredItemFilterUnsupported 
            ///// </description> 
            ///// <expectedresult>Client reports the error.</expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            //[TestCategory("Keyword:Unavailable")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]    // because we need to wait for the notification a little
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_001_01()
            //{
            //    // ServiceResult = “Bad_FilterNotAllowed”
            //    TestCase_Err_001(@"Data Access\Data Access Client Deadband\Test Cases\Err-001-01.js", 
            //        UACodeBits.BadFilterNotAllowed);
            //}

            //[TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            //[TestCategory("Keyword:Unavailable")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]    // because we need to wait for the notification a little
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_001_02()
            //{
            //    // ServiceResult = “Bad_MonitoredItemFilterInvalid"
            //    TestCase_Err_001(@"Data Access\Data Access Client Deadband\Test Cases\Err-001-02.js", 
            //        UACodeBits.BadMonitoredItemFilterInvalid);
            //}

            //[TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            //[TestCategory("Keyword:Unavailable")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]    // because we need to wait for the notification a little
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_001_03()
            //{
            //    // ServiceResult = “Bad_MonitoredItemFilterUnsupported"
            //    TestCase_Err_001(@"Data Access\Data Access Client Deadband\Test Cases\Err-001-03.js", 
            //        UACodeBits.BadMonitoredItemFilterUnsupported);
            //}

            // ReSharper restore InconsistentNaming
#endregion
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class DeadbandTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : DeadbandTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public DeadbandTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class DeadbandTest_NoCtt_Softing_DemoServer_Http : DeadbandTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public DeadbandTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }
        }
    }
}
