﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DataAccess
{
    public partial class DeadbandTest : EndpointTestBase
    {
        #region Private Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_001_And_002([NotNull] EasyUAMonitoredItemArguments monitoredItemArguments)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            using (EasyUAClient client = MakeClient())
            {
                client.SubscribeMonitoredItem(monitoredItemArguments);
                AssertUtilities.WaitingPhase(2*1000);   // allow notification to come

                Assert.AreEqual(1, recorder.CreateMonitoredItemsRecordList.Count - UATestUtilities.BuiltInCreateMonitoredItemsCount);
                CreateMonitoredItemsRecord createMonitoredItemsRecord = recorder.CreateMonitoredItemsRecordList[UATestUtilities.BuiltInCreateMonitoredItemsCount];
                Debug.Assert(createMonitoredItemsRecord != null);
                MonitoredItemCreateRequestCollection monitoredItemCreateRequestCollection =
                    createMonitoredItemsRecord.ItemsToCreate;
                Assert.IsNotNull(monitoredItemCreateRequestCollection);
                Assert.AreEqual(1, monitoredItemCreateRequestCollection.Count);
                MonitoredItemCreateRequest monitoredItemCreateRequest = monitoredItemCreateRequestCollection[0];
                Assert.IsNotNull(monitoredItemCreateRequest);
                MonitoringParameters monitoringParameters = monitoredItemCreateRequest.RequestedParameters;
                Assert.IsNotNull(monitoringParameters);
                ExtensionObject filter = monitoringParameters.Filter;
                Assert.IsNotNull(filter);
                object body = filter.Body;
                Assert.IsNotNull(body);
                var sdkDataChangeFilter = body as DataChangeFilter;
                Assert.IsNotNull(sdkDataChangeFilter);

                UADataChangeFilter dataChangeFilter = monitoredItemArguments.MonitoringParameters.DataChangeFilter;
                Debug.Assert(dataChangeFilter != null);
                Assert.AreEqual((uint)dataChangeFilter.DeadbandType, sdkDataChangeFilter.DeadbandType);
                Assert.AreEqual(dataChangeFilter.DeadbandValue, sdkDataChangeFilter.DeadbandValue);
            }
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_001([NotNull] string scriptName, Int32 expectedCodeBits)
        // ReSharper restore InconsistentNaming
        {
            const double deadbandValue = 10.0;
            using (Ctt.RunScript(CttProjectName, scriptName))
            using (EasyUAClient client = MakeClient())
            {
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(float), dynamic: true);
                var dataChangeFilter = new UADataChangeFilter(UADeadbandType.Percent, deadbandValue);
                monitoredItemArguments.MonitoringParameters.DataChangeFilter = dataChangeFilter;

                IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs
                    (
                        client, monitoredItemArguments, MillisecondsToCollectMultipleNotifications);

                // ReSharper disable PossibleMultipleEnumeration
                int count = eventArgsEnumerable.Count();
                Assert.AreEqual(1, count);
                EasyUAMonitoredItemChangedEventArgs eventArgs = eventArgsEnumerable.First();
                Debug.Assert(eventArgs != null);
                // ReSharper restore PossibleMultipleEnumeration
                Exception exception = eventArgs.Exception;
                Assert.IsNotNull(exception);
                Debug.WriteLine(exception.ToString());
                var serviceException = exception as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(expectedCodeBits, serviceException.ServiceResult.StatusCode.CodeBits);
            }
        }
#endif

#endregion
    }
}
