﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DiscoveryServices
{
    public partial class FindServersBasicTest
    {
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class FindServersBasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp : FindServersBasicTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public FindServersBasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }
        }



        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class FindServersBasicTest_Ctt_Softing_DemoServer_OpcTcp : FindServersBasicTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public FindServersBasicTest_Ctt_Softing_DemoServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }



        // ReSharper disable InconsistentNaming
        public class FindServersBasicTest_Ctt : FindServersBasicTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>ServiceResult = Good, Servers[] are returned as an empty list.
            /// </description>
            /// <expectedresult>
            /// Client handles the situation, and it may choose to log and/or displays a warning to the end-user but this is 
            /// not an expectation. It could simply show an empty screen!
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005()
            {
                using (Ctt.RunScript(CttProjectName, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-005.js"))
                {
                    EasyUAClient client = MakeClient();
                    UADiscoveryElementCollection discoveryElementCollection =
                        client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                    Assert.IsNotNull(discoveryElementCollection);
                    Assert.AreEqual(0, discoveryElementCollection.Count);
                }
            }

            /// <description>
            /// ServiceResult = Good, but duplicate the contents of the Servers collection so that it is returned with 
            /// duplicate entries, i.e. size = size * 2.
            /// </description>
            /// <expectedresult>
            /// Client may choose to indicate to the end-user that duplicate entries were received.
            /// An ideal Client would remove any duplicate entries, but this is not an expectation.
            /// If the Client simply displayed what it received then this would be acceptable providing that the ServiceResult=Good.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006()
            {
                UADiscoveryElementCollection cttDiscoveryElementCollection;
                using (Ctt.RunScript(CttProjectName, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-006.js"))
                {
                    EasyUAClient cttClient = MakeClient();
                    cttDiscoveryElementCollection = cttClient.FindLocalApplications(
                        new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                    Assert.IsNotNull(cttDiscoveryElementCollection);
                    Assert.AreNotEqual(0, cttDiscoveryElementCollection.Count);
                }

                EasyUAClient plainClient = MakeClient();
                UADiscoveryElementCollection plainDiscoveryElementCollection = plainClient.FindLocalApplications(
                    new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(plainDiscoveryElementCollection);
                Assert.AreNotEqual(0, plainDiscoveryElementCollection.Count);

                CollectionAssert.AreEquivalent(plainDiscoveryElementCollection, cttDiscoveryElementCollection);
            }

            /// <description>ServiceResult = Good.
            /// The Servers[] list is huge (100+ elements). The data within the results can be randomly generated to `appear` real.
            /// </description>
            /// <expectedresult>
            /// Client does not crash due to the large quantity of data. The client processes it in a reasonable manner (if 
            /// the list is displayed - it is scrollable or paged, but is available in some manner that can be accessed)
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_007()
            {
                using (Ctt.RunScript(CttProjectName, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-007.js"))
                {
                    EasyUAClient client = MakeClient();
                    UADiscoveryElementCollection discoveryElementCollection =
                        client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                    Assert.IsNotNull(discoveryElementCollection);
                    int count = discoveryElementCollection.Count;
                    AssertUtilities.WriteLine(count);
                    Assert.IsTrue(count >= 50); // Client and DiscoveryServer apps are not eligible
                }
            }

            /// <description>Client receives a mix-match of ApplicationUri`s in the Servers parameter: 
            /// Servers [0] URI contains a host name.
            /// Servers [1] URI as valid IPv4.
            /// Servers [2] URI as valid IPv6.
            /// Servers [3] URI is empty.
            /// </description>
            /// <expectedresult>
            /// Client accepts the first 3 parameters but generates an error/log of the 4th.
            /// IPv6 addresses are different to IPv4 but should be validated nonetheless even if IPv6 is not 
            /// supported/available on the Client`s computer.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_008()
            {
                TestCase_Err_008(false);
            }

            /// <description>Servers[0] ApplicationURI contains invalid IP, e.g. 192.168.0.300 or 256.100.100.100 etc.
            /// </description>
            /// <expectedresult>
            /// Client rejects the IP address and logs and/or displays a message to the end-user.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_009()
            {
                TestCase_Err_009(false);
            }

            /// <description>ServiceResult = Good, the first Servers [] item`s ProductUri parameter is empty.
            /// </description>
            /// <expectedresult>
            /// ProductUri is a required field, so client should log and/or display a warning to the end-user and/or visually 
            /// identifies the first Server differently such to indicate that it is in an error state, perhaps a different 
            /// color or is disabled (not clickable) etc.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010()
            {
                TestCase_Err_010(false);
            }

            /// <description>ServiceResult = Good, the first Servers [] item`s ApplicationName parameter is empty.
            /// </description>
            /// <expectedresult>
            /// Client should log and/or display a warning to the end-user and/or display a warning to the end-user and/or 
            /// visually identifies the first Server differently such to indicate that it is in an error state, perhaps 
            /// a different color or is disabled (not clickable) etc.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011()
            {
                TestCase_Err_011(false);
            }

            /// <description>ServiceResult = Good, the first Servers[] ApplicationType is SERVER_0 and DiscoveryUrls[] is 
            /// empty.
            /// </description>
            /// <expectedresult>
            /// Client reports no availability of the UA Server since no discovery Endpoints are available and/or visually 
            /// identifies the first Server differently such to indicate that it is in an error state, perhaps a different 
            /// color or is disabled (not clickable) etc.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_013()
            {
                TestCase_Err_013_And_016(false, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-013.js");
            }

            /// <description>ServiceResult = Good, the first Servers [] item`s ApplicationType field is set to CLIENT_1.
            /// </description>
            /// <expectedresult>
            /// The Client should log and/or display a message to the end-user that the list of OPC UA Servers returned 
            /// contains a UA Client.
            /// The record might not be used/displayed within a browser etc. but if it is then it should be displayed 
            /// differently.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_014()
            {
                TestCase_Err_014(false);
            }

            /// <description>ServiceResult = Good, the first Servers[] ApplicationType is CLIENTANDSERVER_1 and 
            /// DiscoveryUrls[] is empty.
            /// </description>
            /// <expectedresult>
            /// Client reports no availability of the UA Server since no discovery Endpoints are available and/or visually 
            /// identifies the first Server differently such to indicate that it is in an error state, perhaps a different 
            /// color or is disabled (not clickable) etc.
            /// </expectedresult>
            [TestCategory("Keyword:FindServers")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_016()
            {
                TestCase_Err_013_And_016(false, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-016.js");
            }

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            #endregion
        }
    }
}
