﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DiscoveryServices
{
    public partial class FindServersBasicTest : EndpointTestBase
    {
        #region Protected Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_008(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName,
                                    @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-008.js"))
            {
                EasyUAClient client = MakeClient();
                UADiscoveryElementCollection discoveryElementCollection =
                    client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId ==
                                                                                  UAEngineEventIds
                                                                                      .ApplicationDescriptionApplicationUriInvalid));
                // ReSharper restore PossibleNullReferenceException

                if (inverted)
                {
                    Assert.IsNull(logEntryEventArgs);
                    return;
                }

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

                int count = discoveryElementCollection.Count;
                AssertUtilities.WriteLine(count);
                foreach (UADiscoveryElement discoveryElement in discoveryElementCollection)
                    AssertUtilities.WriteLine(discoveryElement);

                switch (Ctt.ScriptCrc)
                {
                    case 0x03321b2d:
                        // do nothing 
                        Assert.Inconclusive();
                        break;
                    case 0xda299f6b:
                    case 0x031f38f4:
                        Assert.AreEqual(1, count);

                        // ReSharper disable PossibleNullReferenceException
                        Assert.IsTrue(discoveryElementCollection.Any(element => element.ApplicationUriString.Contains("localhost")));
                        // ReSharper restore PossibleNullReferenceException
                        Assert.Inconclusive();
                        break;
                    default:
                        Assert.AreEqual(3, count);

                        // ReSharper disable PossibleNullReferenceException
                        Assert.IsTrue(discoveryElementCollection.Any(element => element.ApplicationUriString.Contains("localhost")));
                        Assert.IsTrue(discoveryElementCollection.Any(element => element.ApplicationUriString.Contains("127.0.0.1")));
                        Assert.IsTrue(discoveryElementCollection.Any(element => element.ApplicationUriString.Contains("[::1]")));
                        break;
                    // ReSharper restore PossibleNullReferenceException
                }
            }
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_009(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            UADiscoveryElementCollection discoveryElementCollection;
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-009.js"))
            {
                EasyUAClient client = MakeClient();
                discoveryElementCollection =
                    client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .ApplicationDescriptionApplicationUriIPAddressInvalid));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }
            
            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            int cttCount = discoveryElementCollection.Count;
            AssertUtilities.WriteLine(cttCount);

            // ReSharper disable PossibleNullReferenceException
            Assert.IsFalse(
                discoveryElementCollection.Any(element => element.ApplicationUriString.Contains("192.168.0.300")));
            // ReSharper restore PossibleNullReferenceException

            EasyUAClient plainClient = MakeClient();
            UADiscoveryElementCollection plainDiscoveryElementCollection =
                plainClient.FindLocalApplications(new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
            Assert.IsNotNull(plainDiscoveryElementCollection);
            int plainCount = plainDiscoveryElementCollection.Count;
            AssertUtilities.WriteLine(plainCount);

            Assert.IsTrue(cttCount < plainCount);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_010(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            UADiscoveryElementCollection discoveryElementCollection;
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName,
                                    @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-010.js"))
            {
                EasyUAClient client = MakeClient();
                discoveryElementCollection = client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ApplicationDescriptionProductUriNullOrEmpty));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            int cttCount = discoveryElementCollection.Count;
            AssertUtilities.WriteLine(cttCount);

            // ReSharper disable PossibleNullReferenceException
            Assert.IsFalse(discoveryElementCollection.Any(element => string.IsNullOrEmpty(element.ProductUriString)));
            // ReSharper restore PossibleNullReferenceException

            EasyUAClient plainClient = MakeClient();
            UADiscoveryElementCollection plainDiscoveryElementCollection =
                plainClient.FindLocalApplications(new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
            Assert.IsNotNull(plainDiscoveryElementCollection);
            int plainCount = plainDiscoveryElementCollection.Count;
            AssertUtilities.WriteLine(plainCount);

            Assert.IsTrue(cttCount < plainCount);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_011(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            UADiscoveryElementCollection discoveryElementCollection;
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName,
                                    @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-011.js"))
            {
                EasyUAClient client = MakeClient();
                discoveryElementCollection = client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ApplicationDescriptionApplicationNameNullOrEmpty));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            int cttCount = discoveryElementCollection.Count;
            AssertUtilities.WriteLine(cttCount);

            // ReSharper disable PossibleNullReferenceException
            Assert.IsTrue(discoveryElementCollection.Any(element => string.IsNullOrEmpty(element.ApplicationName)));
            // ReSharper restore PossibleNullReferenceException

            EasyUAClient plainClient = MakeClient();
            UADiscoveryElementCollection plainDiscoveryElementCollection =
                plainClient.FindLocalApplications(new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
            Assert.IsNotNull(plainDiscoveryElementCollection);
            int plainCount = plainDiscoveryElementCollection.Count;
            AssertUtilities.WriteLine(plainCount);

            Assert.AreEqual(plainCount, cttCount);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_014(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            UADiscoveryElementCollection discoveryElementCollection;
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName,
                                    @"Discovery Services\Discovery Client Find Servers Basic\Test Cases\Err-014.js"))
            {
                EasyUAClient client = MakeClient();
                discoveryElementCollection = client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ApplicationDescriptionApplicationTypeClient));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            int cttCount = discoveryElementCollection.Count;
            AssertUtilities.WriteLine(cttCount);

            EasyUAClient plainClient = MakeClient();
            UADiscoveryElementCollection plainDiscoveryElementCollection =
                plainClient.FindLocalApplications(new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
            Assert.IsNotNull(plainDiscoveryElementCollection);
            int plainCount = plainDiscoveryElementCollection.Count;
            AssertUtilities.WriteLine(plainCount);

            Assert.IsTrue(cttCount < plainCount);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_013_And_016(
            // ReSharper restore InconsistentNaming
            bool inverted,
            [NotNull] string scriptName)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            UADiscoveryElementCollection discoveryElementCollection;
            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                discoveryElementCollection = client.FindLocalApplications(new[] { EndpointDescriptor.UrlString }, UAApplicationTypes.Server);
                Assert.IsNotNull(discoveryElementCollection);
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == UAEngineEventIds.ApplicationDescriptionDiscoveryUrlsEmpty));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            int cttCount = discoveryElementCollection.Count;
            AssertUtilities.WriteLine(cttCount);

            EasyUAClient plainClient = MakeClient();
            UADiscoveryElementCollection plainDiscoveryElementCollection = plainClient.FindLocalApplications(
                new[] { EndpointDescriptorDirect.UrlString }, UAApplicationTypes.Server);
            Assert.IsNotNull(plainDiscoveryElementCollection);
            int plainCount = plainDiscoveryElementCollection.Count;
            AssertUtilities.WriteLine(plainCount);

            Assert.IsTrue(cttCount < plainCount);
        }
#endif

#endregion
    }
}
