﻿using System;
using System.Globalization;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DiscoveryServices
{
    public partial class GetEndpointsBasicTest
    {
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class GetEndpointsBasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp : GetEndpointsBasicTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public GetEndpointsBasicTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
            }
        }



        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class GetEndpointsBasicTest_Ctt_Softing_DemoServer_OpcTcp : GetEndpointsBasicTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public GetEndpointsBasicTest_Ctt_Softing_DemoServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }



        // ReSharper disable InconsistentNaming
        public class GetEndpointsBasicTest_Ctt : GetEndpointsBasicTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
            // ReSharper disable InconsistentNaming

            // The behavior of the client may depend on whether the endpoint description modified by the script is the one 
            // the client wants to use (depending on its other characteristics), or not, and the other of endpoint may not be 
            // stable. We are therefore trying to modify other endpoints as well, and test the outcome based on knowledge 
            // about the structure of server endpoint descriptions.

            /// <description>Randomly inject one of the following status codes into the Service Result.
            /// Repeat this test numerous times each time varying the code:
            /// 
            /// 1. Bad_NothingToDo
            /// 2. Bad_InternalError
            /// 3. Bad_OutOfMemory
            /// 4. Bad_RequestTooLarge
            /// 5. Bad_ResponseTooLarge
            /// 6. Bad_UnexpectedError
            /// 7. Bad_UserAccessDenied
            /// 8. Bad_Timeout 
            /// </description>
            /// <expectedresult>
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_01()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-01.js",
                                 UACodeBits.BadNothingToDo);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_02()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-02.js",
                                 UACodeBits.BadInternalError);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_03()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-03.js",
                                 UACodeBits.BadOutOfMemory);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_04()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-04.js",
                                 UACodeBits.BadRequestTooLarge);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_05()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-05.js",
                                 UACodeBits.BadResponseTooLarge);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_06()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-06.js",
                                 UACodeBits.BadUnexpectedError);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_07()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-07.js",
                                 UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // incorrectly marked as Lab in the list
            [TestMethod]
            public void TestCase_Err_001_08()
            {
                TestCase_Err_001(@"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-001-08.js",
                                 UACodeBits.BadTimeout);
            }

            /// <description>ServiceResult = Good.
            /// Endpoints are empty.
            /// </description>
            /// <expectedresult>Client reports that nothing was received.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_002()
            {
                using (Ctt.RunScript(CttProjectName, 
                    @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-002.js"))
                {
                    try
                    {
                        EasyUAClient client = MakeClient();
                        client.Read(EndpointDescriptor,
                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
                    }
                    catch (UAException uaException)
                    {
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        var clientEngineException = innerException as UAClientEngineException;
                        Assert.IsNotNull(clientEngineException);
                        Assert.AreEqual(UAClientErrorCodes.CannotPreselectBestEndpoint, clientEngineException.ErrorCode);
                        string message = clientEngineException.Message;
                        StringAssert.Contains(message, "0 endpoints");
                        return;
                    }
                    Assert.Fail();
                }
            }

            /// <description>ServiceResult = Good, the first EndPoints[] endpointUrl is empty.
            /// </description>
            /// <expectedresult>
            /// Client may choose to filter the endpoint from a display and/or report to the end-user that an invalid endpoint 
            /// was received.</expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecurityNone_False()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecurityNone, Sign.Negative);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecuritySign_False()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecuritySign, Sign.Negative);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecuritySignAndEncrypt_False()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecuritySign, Sign.Negative);
            }

            // MessageSecurityPreference should have no influence on the outcome

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecurityNone_True()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecurityNone, Sign.Positive);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecuritySign_True()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecuritySign, Sign.Positive);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_SecuritySignAndEncrypt_True()
            {
                TestCase_Err_005(UAMessageSecurityModes.SecuritySign, Sign.Positive);
            }

            /// <description>ServiceResult = Good, EndPoints [0] endPointUrl uses different IP/Hostname than defined in 
            /// ServerCertificate.
            /// </description>
            /// <expectedresult>[Spoofing scenario] Client must verify that endpointUrl matches the HostNames specified in 
            /// the Certificate.
            /// Client may choose to filter the endpoint from a display and/or report to the end-user that an invalid endpoint 
            /// was received citing the hostname specifically.
            /// For interactive Clients they may choose to prompt the end-user to override and accept the use of the endpoint 
            /// in order to connect to the UA Server.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_alpha()
            {
                TestCase_Err_006_alpha(inverted:false);
            }

            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_beta()
            {
                using (Ctt.RunScript(CttProjectName, 
                    @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-006.js"))
                {
                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.AllowEndpointDomainPrompt = false;
                    client.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.EnforceSameSite = false;
                    client.IsolatedParameters.SessionParameters.RequireServerEndpointsMatchDiscovery = false;
                    client.IsolatedParameters.SessionParameters.UpdateEndpointBeforeConnect = false;

                    // Make sure we select the endpoint that is affected by the CTT script.
                    client.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.MessageSecurityPreference =
                        Sign.Positive;
                    try
                    {
                        client.Read(EndpointDescriptor,
                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);

                        Assert.AreEqual(UACodeBits.BadCertificateHostNameInvalid, 
                            serviceException.ServiceResult.StatusCode.CodeBits);
                        return;
                    }
                    Assert.Fail();
                }
            }

            /// <description>ServiceResult = Good. Endpoints[0].ApplicationUri is empty.
            /// </description>
            /// <expectedresult>
            /// Client identifies the incomplete endpoint description.
            /// Client discards the endpoint.
            /// Client may report a message that an invalid record was received.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_007()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-007.js",
                    UAEngineEventIds.EndpointDescriptionApplicationUriNullOrEmpty);
            }

            /// <description>ServiceResult = Good. Endpoints[0].DiscoveryUrls is empty.
            /// </description>
            /// <expectedresult>
            /// Client identifies the incomplete endpoint description.
            /// Client discards the endpoint.
            /// Client may report a message that an invalid record was received.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_008()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-008.js", 
                    UAEngineEventIds.EndpointDescriptionDiscoveryUrlsEmpty);
            }

            /// <description>ServiceResult = Good, EndPoints[0] server is null.
            /// </description>
            /// <expectedresult>
            /// Client identifies the incomplete endpoint description.
            /// Client discards the endpoint.
            /// Client may report a message that an invalid record was received.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_009()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-009.js",
                    UAEngineEventIds.EndpointDescriptionServerNullOrEmpty);
            }

            /// <description>
            /// ServiceResult = Good. Endpoints[0].Server.ApplicationType is CLIENT_1, gatewayServerUri is not empty.
            /// </description>
            /// <expectedresult>
            /// Client ignores or marks the record differently so as to not be used.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-010.js",
                    UAEngineEventIds.EndpointDescriptionApplicationTypeClient);
            }

            /// <description>
            /// ServiceResult = Good. Endpoints[0].Server.ApplicationType is CLIENT_1, discoveryProfileUri is not empty.
            /// </description>
            /// <expectedresult>
            /// Client ignores or marks the record differently so as to not be used.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-011.js",
                    UAEngineEventIds.EndpointDescriptionApplicationTypeClient);
            }

            /// <description>
            /// ServiceResult = Good. Endpoints[0].Server.ApplicationType is CLIENT_1 discoveryUrls is not empty.
            /// </description>
            /// <expectedresult>
            /// Client ignores or marks the record differently so as to not be used.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012()
            {
                Test_Err_003_To_013(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-012.js",
                    UAEngineEventIds.EndpointDescriptionApplicationTypeClient);
            }

            /// <description>
            /// ServiceResult = Good, endpoints[0].SecurityMode is INVALID_0.
            /// </description>
            /// <expectedresult>
            /// Client is able to identify the invalid security mode and does not use the endpoint for a connection.
            /// A Client with a GUI may filter the record or display it in an error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_021()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-021.js",
                    UAEngineEventIds.EndpointDescriptionSecurityModeInvalid);
            }

            /// <description>
            /// ServiceResult = Good, one of the endpointDescriptions contains an invalid SecurityPolicyUri (empty).
            /// </description>
            /// <expectedresult>
            /// Client is able to identify the invalid security mode and does not use the endpoint for a connection.
            /// A Client with a GUI may filter the record or display it in an error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_022()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-022.js",
                    UAEngineEventIds.EndpointDescriptionSecurityPolicyUriEmpty);
            }

            /// <description>
            /// ServiceResult = Good.
            /// EndpointDescriptions[0].SecurityPolicyUri contains a string that is not formatted as a Uri.
            /// </description>
            /// <expectedresult>
            /// Client is able to identify the invalid security mode and does not use the endpoint for a connection.
            /// A Client with a GUI may filter the record or display it in an error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_023()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-023.js",
                    UAEngineEventIds.EndpointDescriptionSecurityPolicyUriInvalid);
            }

            /// <description>
            /// ServiceResult = Good. Response includes at least one endpoint where userIdentityTokens.tokenType is 
            /// ISSUEDTOKEN_3 but userIdentityTokens.issuedtokenType is empty.
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays an error to the end-user about misbehaving server and does not use the endpoint 
            /// for a connection. The endpoint is not displayed or is displayed while being identified in error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_026()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-026.js",
                    UAEngineEventIds.EndpointDescriptionIssuedTokenTypeEmpty);
            }

            /// <description>
            /// ServiceResult = Good.
            /// endpointDescriptions[0].transportProfileUri is empty.
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays an error to the end-user about misbehaving server and does not use the endpoint 
            /// for a connection. The endpoint is not displayed or is displayed while being identified in error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_028()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-028.js",
                    UAEngineEventIds.EndpointDescriptionTransportProfileUriEmpty);
            }

            /// <description>
            /// ServiceResult = Good.
            /// endpointDescriptions[0].transportProfileUri is not a valid Uri.
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays an error to the end-user about misbehaving server and does not use the endpoint 
            /// for a connection. The endpoint is not displayed or is displayed while being identified in error state.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_029()
            {
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-029.js",
                    UAEngineEventIds.EndpointDescriptionTransportProfileUriInvalid);
            }

            /// <description>
            /// ServiceResult = Good.
            /// endpointDescriptions[0].securityLevel is 0.
            /// </description>
            /// <expectedresult>
            /// If client displays the information in a graphical environment, then it should also warn an end-user about 
            /// the risks of using an insecure remote connection. This warning can occur on attempting to connect or as part 
            /// of what is being displayed. 
            /// If the client does not display information, then it should log or generate a warning about an insecure 
            /// connection.
            /// </expectedresult>
            [TestCategory("Keyword:GetEndpoints")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_032()
            {
                // IMPROVE: it may be an accident that this test passes. It is different from the others.
                Test_Err_021_To_031(false, @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-032.js",
                    UAEngineEventIds.SelectedEndpointWithZeroSecurityLevel);
            }

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            // Note that only endpoints with the same scheme should be counted into the total
            private int CountOfServerEndpointDescriptions(UAMessageSecurityModes messageSecurityModes)
            {
                int count = 0;
                if ((messageSecurityModes & UAMessageSecurityModes.SecurityNone) != 0) count += 1;  // 2 total
                if ((messageSecurityModes & UAMessageSecurityModes.SecuritySign) != 0) count += 1;  // 2 total
                if ((messageSecurityModes & UAMessageSecurityModes.SecuritySignAndEncrypt) != 0) count += 0;    // 2 total
                return count;
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_005(UAMessageSecurityModes messageSecurityModes, Sign messageSecurityPreference)
            // ReSharper restore InconsistentNaming
            {
                const int countOfEndpointDescriptions = 6;
                const int countOfModifiedEndpointDescriptions = 1;
                int countOfAllowedEndpointDescriptions = CountOfServerEndpointDescriptions(messageSecurityModes);

                int countOfEndpointSelectionFailures = 0;
                int countOfCttUnsupported = 0;
                const string scriptNameFormatIndex1AndMore =
                    @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\OpcLabs-Err-005-Index-{0:D2}.js";
                for (int index = 0; index < countOfEndpointDescriptions; index++)
                {
                    AssertUtilities.WriteSeparatorLine();
                    AssertUtilities.WriteLine(index);

                    string scriptName = (index == 0)
                        ? @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-005.js"
                        : String.Format(CultureInfo.InvariantCulture, scriptNameFormatIndex1AndMore, index);
                    using (Ctt.RunScript(CttProjectName, scriptName))
                    {
                        EasyUAClient client = MakeClient();
                        client.IsolatedParameters.SessionParameters.RequireServerEndpointsMatchDiscovery = false;
                        client.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.AllowedMessageSecurityModes =
                            messageSecurityModes;
                        client.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.MessageSecurityPreference = messageSecurityPreference;

                        try
                        {
                            client.Read(EndpointDescriptor,
                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));

                            //if ((messageSecurityModes & UAMessageSecurityModes.SecurityNone) == 0)
                            //    Assert.Fail("CTT cannot be used with message security, exception expected but not thrown.");
                        }
                        catch (UAException uaException)
                        {
                            Exception innerException = uaException.InnerException;
                            Assert.IsNotNull(innerException);
                            if (innerException is UAServiceException serviceException)
                            {
                                int codeBits = serviceException.ServiceResult;
                                Assert.IsTrue(
                                    (codeBits == UACodeBits.BadSecureChannelClosed) ||
                                    (codeBits == UACodeBits.BadSecurityPolicyRejected));
                                AssertUtilities.WriteLine("CTT unsupported.");
                                countOfCttUnsupported++;
                                continue;
                            }
                            if (innerException is UAClientEngineException clientEngineException)
                            {
                                Assert.AreEqual(UAClientErrorCodes.CannotPreselectBestEndpoint, clientEngineException.ErrorCode);
                                AssertUtilities.WriteLine("Endpoint selection failure.");
                                countOfEndpointSelectionFailures++;
                                continue;
                            }
                            Assert.Fail();
                        }
                    }
                }

                if (countOfAllowedEndpointDescriptions > countOfModifiedEndpointDescriptions)
                    // We have not modified enough endpoint descriptors to cause any failure
                    Assert.AreEqual(0, countOfEndpointSelectionFailures);
                else
                    Assert.AreNotEqual(0, countOfEndpointSelectionFailures);    // IMPROVE: calculate and expect the precise count

                if ((messageSecurityModes & UAMessageSecurityModes.SecurityNone) != 0)
                    Assert.AreEqual(0, countOfCttUnsupported);
            }

            #endregion
        }
    }
}
