﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.DiscoveryServices
{
    public partial class GetEndpointsBasicTest : EndpointTestBase
    {
        #region Private Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void Test_Err_003_To_013(
        // ReSharper restore InconsistentNaming
            bool inverted, [NotNull] string scriptName, int eventIdExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                // must not throw
                client.Read(EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                // ReSharper disable PossibleNullReferenceException
                (args.EventId == eventIdExpected));
                // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void Test_Err_021_To_031(
        // ReSharper restore InconsistentNaming
            bool inverted, [NotNull] string scriptName, int eventIdExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.RequireServerEndpointsMatchDiscovery = false;
                // must not throw
                client.Read(EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                // ReSharper disable PossibleNullReferenceException
                (args.EventId == eventIdExpected));
                // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_001(
        // ReSharper restore InconsistentNaming
            [NotNull] string scriptName, int codeBitsExpected)
        {
            using (Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                try
                {
                    client.Read(EndpointDescriptor,
                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine(uaException);
                    Exception innerException = uaException.InnerException;
                    Assert.IsNotNull(innerException);
                    var serviceException = innerException as UAServiceException;
                    Assert.IsNotNull(serviceException);

                    Assert.AreEqual(codeBitsExpected, serviceException.ServiceResult.StatusCode.CodeBits);
                    return;
                }
                Assert.Fail();
            }
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_Err_006_alpha(
        // ReSharper restore InconsistentNaming
            bool inverted)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, 
                @"Discovery Services\Discovery Client Get Endpoints Basic\Test Cases\Err-006.js"))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.EndpointDescriptionChecks.RequireHostMatchCertificate = true;
                client.IsolatedParameters.SessionParameters.RequireServerEndpointsMatchDiscovery = false;    // prevent error
                // This will select and work with another endpoint, one that is not affected by the CTT script.
                // must not throw
                client.Read(EndpointDescriptor,
                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: false));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                // ReSharper disable PossibleNullReferenceException
                (args.EventId == UAEngineEventIds.EndpointDescriptionEndpointUrlHostNotInCertificate));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            StringAssert.Contains(logEntryEventArgs.Message, "\"otherxyz\"");
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }
#endif

#endregion
    }
}
