﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.MonitoredItemServices
{
    public partial class ModifyTest
    {
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]   // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ModifyTest_Ctt_OpcFoundation_SampleServer_OpcTcp : ModifyTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public ModifyTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
            }
        }



        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]   // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ModifyTest_Ctt_Softing_DemoServer_OpcTcp : ModifyTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public ModifyTest_Ctt_Softing_DemoServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }



        // ReSharper disable InconsistentNaming
        public class ModifyTest_Ctt : ModifyTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>
            /// Inject the following testcases
            /// 1. serviceResult is Good, but the revisedQueueSize != requested queueSize.
            /// 2. serviceResult is Good, but the revisedQueueSize is decades larger than the requested queueSize.
            /// 3. serviceResult is Good, but the revisedQueueSize is 0. 
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays an information message to the end-user.
            /// </expectedresult>
            [TestCategory("Keyword:ModifyMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_01()
            {
                TestCase_Err_005(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-005-01.js",
                                 UAEngineEventIds.RevisedQueueSizeDifferent, LogEntryType.Information);
            }

            [TestCategory("Keyword:ModifyMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_02()
            {
                TestCase_Err_005(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-005-02.js",
                                 UAEngineEventIds.RevisedQueueSizeAboveToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:ModifyMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_03()
            {
                TestCase_Err_005(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-005-03.js",
                                 UAEngineEventIds.RevisedQueueSizeInvalid, LogEntryType.Warning);
            }

            /// <description>
            /// Inject the following test cases
            /// 1. ServiceResult = Bad_NothingToDo
            /// 2. ServiceResult = Bad_TooManyOperations
            /// 3. ServiceResult = Bad_SubscriptionIdInvalid
            /// 4. ServiceResult = Good, but the first result is set to Bad_MonitoredItemIdInvalid.
            /// 5. ServiceResult = Good, and all results are set to Bad_MonitoredItemIdInvalid. 
            /// </description>
            /// <expectedresult>
            /// a,b,c,e: Client reports the error to the end-user and does no further processing of the results.
            /// d: Client reports each result in error. Remaining items are still processed. 
            /// </expectedresult>
            [TestCategory("Keyword:DeleteMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_01()
            {
                TestCase_Err_012(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-012-01.js",
                                 UAEngineEventIds.SubscriptionPerformChangesFailure, LogEntryType.Warning);
            }

            [TestCategory("Keyword:DeleteMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_02()
            {
                TestCase_Err_012(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-012-02.js",
                                 UAEngineEventIds.SubscriptionPerformChangesFailure, LogEntryType.Warning);
            }

            [TestCategory("Keyword:DeleteMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_03()
            {
                TestCase_Err_012(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-012-03.js",
                                 UAEngineEventIds.SubscriptionPerformChangesFailure, LogEntryType.Warning);
            }

            [TestCategory("Keyword:DeleteMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_04()
            {
                TestCase_Err_012(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-012-04.js",
                                 UAEngineEventIds.MonitoredItemDeleteError, LogEntryType.Information);
            }

            [TestCategory("Keyword:DeleteMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_05()
            {
                TestCase_Err_012(false, @"Monitored Item Services\Monitor Client Modify\Test Cases\Err-012-05.js",
                                 UAEngineEventIds.MonitoredItemDeleteError, LogEntryType.Information);
            }

            // ReSharper restore InconsistentNaming
            #endregion
        }



        #region Private Methods

        #endregion
    }
}
