﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.MonitoredItemServices
{
    public partial class ModifyTest : EndpointTestBase
    {
        #region Protected Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_005(
            // ReSharper restore InconsistentNaming
            bool inverted,
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                monitoredItemArguments.MonitoringParameters.QueueSize = 20;
                monitoredItemArguments.SubscriptionParameters.PublishingInterval = 50;  // set explicitly (do not use zero)
                var modifiedMonitoringParameters =
                    new UAMonitoringParameters(monitoredItemArguments.MonitoringParameters);
                modifiedMonitoringParameters.SamplingInterval += 500;
                EasyUAClient client = MakeClient();
                CollectMonitoredItemChangedEventArgs(client,
                    monitoredItemArguments,
                    handles =>
                    {
                        Debug.Assert(handles != null);
                        AssertUtilities.WaitingPhase(MillisecondsToCollectSingleNotification);
                        client.ChangeMonitoredItemSubscription(handles[0], modifiedMonitoringParameters);
                        AssertUtilities.WaitingPhase(MillisecondsToCollectSingleNotification);
                    }
                );
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == eventIdExpected));
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            // ReSharper restore PossibleNullReferenceException
            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_012(
            // ReSharper restore InconsistentNaming
            bool inverted,
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();

                EasyUAMonitoredItemArguments monitoredItemArguments1 = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                // If we had just one monitored item, the DeleteMonitoredItems service may not be called, because of 
                // optimization. We therefore neeed at least one more monitored item on the same subscription.
                EasyUAMonitoredItemArguments monitoredItemArguments2 = MakeMonitoredItemArguments(typeof(Int16), dynamic: false);
                int handle2 = client.SubscribeMonitoredItem(monitoredItemArguments2);
                try
                {
                    CollectMonitoredItemChangedEventArgs(
                        client,
                        monitoredItemArguments1,
                        MillisecondsToCollectSingleNotification);
                }
                finally
                {
                    client.UnsubscribeMonitoredItem(handle2);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
        }
#endif

#endregion
    }
}
