﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.MonitoredItemServices
{
    public partial class ValueChangeTest
    {
        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_Ctt : ValueChangeTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>
            /// Randomly inject one of the following ServiceResult codes. Repeat this test several times each time varying the 
            /// result
            ///     1. ServiceResult = Bad_TooManySubscriptions
            ///     2. ServiceResult = Bad_NothingToDo
            ///     3. ServiceResult = Bad_TimestampsToReturn Invalid
            ///     4. ServiceResult = Bad_SubscriptionIdInvalid 
            ///</description>
            /// <expectedresult>Client logs and/or displays an error to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_001_01()
            {
                // Injects error "Bad_TooManySubscriptions" into the Service Result when calling CreateMonitoredItemsHelper. 
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-001-01.js"))
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Single), dynamic: true),
                        MillisecondsToCollectSingleNotification);

                    EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs.Exception, UACodeBits.BadTooManySubscriptions);
                }
            }

            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_001_02()
            {
                // Injects error "Bad_NothingToDo" into the Service Result when calling CreateMonitoredItemsHelper.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-001-02.js"))
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Single), dynamic: true),
                        MillisecondsToCollectSingleNotification);

                    EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs.Exception, UACodeBits.BadNothingToDo);
                }
            }

            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_001_03()
            {
                //  Injects error "Bad_TimestampsToReturnInvalid" into the Service Result when calling CreateMonitoredItemsHelper.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-001-03.js"))
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Single), dynamic: true),
                        MillisecondsToCollectSingleNotification);

                    EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs.Exception, UACodeBits.BadTimestampsToReturnInvalid);
                }
            }

            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_001_04()
            {
                // Injects error "Bad_SubscriptionIdInvalid" into the Service Result when calling CreateMonitoredItemsHelper.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-001-04.js"))
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Single), dynamic: true),
                        MillisecondsToCollectSingleNotification);

                    EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs.Exception, UACodeBits.BadSubscriptionIdInvalid);
                }
            }

            /// <description>
            /// Inject the following test cases
            ///     1. serviceResult is Good, but the results are empty.
            ///     2. serviceResult is Good, but there are fewer results than requests, i.e. remove 1 result item.
            ///     3. serviceResult is Good, but there are more results than requests, i.e. duplicate the first result and add to 
            ///        the end of the results.
            /// </description>
            /// <expectedresult>
            /// a. Client logs and/or displays an error to the end-user. 
            /// b. Client processes the results that it does receive. Client reports an error about the missing item by perhaps 
            ///    showing it as disabled or revises its status/quality as Bad.
            /// c. Client processes the results that it receives. Client reports the problem by perhaps showing them as disabled 
            ///    or revising their status/quality as Bad. 
            /// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_002_01()
            {
                // Clears the results parameter.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-002-01.js"))
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Single), dynamic: true),
                        MillisecondsToCollectSingleNotification);

                    EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs.Exception, UACodeBits.BadUnexpectedError,
                        "The server returned a list with less elements");
                }
            }

            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_002_02()
            {
                // Removes one result item from the collection.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-002-02.js"))
                {
                    var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
                        new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
                    int monitoredItemCount = monitoredItemArgumentsArray.Length;
                    IEnumerable<EasyUADataChangeNotificationEventArgs>[] eventArgsEnumerableArray =
                        CollectMultipleDataChangeNotificationEventArgs(
                            MakeClient(),
                            monitoredItemArgumentsArray,
                            MillisecondsToCollectMultipleNotifications);
                    Assert.AreEqual(monitoredItemCount, eventArgsEnumerableArray.Length);

                    // ReSharper disable AssignNullToNotNullAttribute
                    int countOfNone = eventArgsEnumerableArray.Count(eventArgsEnumerable => !eventArgsEnumerable.Any());
                    int countOfSingle = eventArgsEnumerableArray.Count(eventArgsEnumerable => eventArgsEnumerable.Count() == 1);
                    int countOfMultiple = eventArgsEnumerableArray.Count(eventArgsEnumerable => eventArgsEnumerable.Count() > 1);
                    // ReSharper restore AssignNullToNotNullAttribute
                    AssertUtilities.WriteLine($"countOfNone: {countOfNone}");
                    AssertUtilities.WriteLine($"countOfSingle: {countOfSingle}");
                    AssertUtilities.WriteLine($"countOfMultiple: {countOfMultiple}");

                    Assert.AreEqual(0, countOfNone);
                    Assert.AreEqual(1, countOfSingle);
                    Assert.AreEqual(monitoredItemCount - 1, countOfMultiple);

                    foreach (IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable in eventArgsEnumerableArray)
                    {
                        // ReSharper disable PossibleMultipleEnumeration
                        Debug.Assert(eventArgsEnumerable != null);
                        if (eventArgsEnumerable.Count() == 1)
                        {
                            EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                            AssertServiceException(eventArgs.Exception, UACodeBits.BadUnexpectedError,
                                "The server returned less results");
                        }
                        else
                            AssertNoFailures(eventArgsEnumerable);
                        // ReSharper restore PossibleMultipleEnumeration
                    }
                }
            }

            [TestCategory("Keyword:CreateMonitoredItems")]  // inferred
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]  // inferred
            [TestMethod]
            public void TestCase_Err_002_03()
            {
                // Adds one result item to the collection.
                using (Ctt.RunScript(CttProjectName, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-002-03.js"))
                {
                    var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
                        new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
                    int monitoredItemCount = monitoredItemArgumentsArray.Length;
                    IEnumerable<EasyUADataChangeNotificationEventArgs>[] eventArgsEnumerableArray =
                        CollectMultipleDataChangeNotificationEventArgs(
                            MakeClient(),
                            monitoredItemArgumentsArray,
                            MillisecondsToCollectSingleNotification);
                    Assert.AreEqual(monitoredItemCount, eventArgsEnumerableArray.Length);

                    // ReSharper disable AssignNullToNotNullAttribute
                    int countOfNone = eventArgsEnumerableArray.Count(eventArgsEnumerable => !eventArgsEnumerable.Any());
                    int countOfSingle = eventArgsEnumerableArray.Count(eventArgsEnumerable => eventArgsEnumerable.Count() == 1);
                    int countOfMultiple = eventArgsEnumerableArray.Count(eventArgsEnumerable => eventArgsEnumerable.Count() > 1);
                    // ReSharper restore AssignNullToNotNullAttribute
                    AssertUtilities.WriteLine($"countOfNone: {countOfNone}");
                    AssertUtilities.WriteLine($"countOfSingle: {countOfSingle}");
                    AssertUtilities.WriteLine($"countOfMultiple: {countOfMultiple}");

                    Assert.AreEqual(0, countOfNone);
                    Assert.AreEqual(monitoredItemCount, countOfSingle);
                    Assert.AreEqual(0, countOfMultiple);

                    foreach (IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable in eventArgsEnumerableArray)
                    {
                        Debug.Assert(eventArgsEnumerable != null);
                        EasyUADataChangeNotificationEventArgs eventArgs = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                        AssertServiceException(eventArgs.Exception, UACodeBits.BadUnexpectedError,
                            "The server returned a list with more elements");
                    }
                }
            }

            /// <description>
            /// ServiceResult = Good, but first item result is set to one of the operation level error codes as specified in 
            /// Part 4, section 5.12.2.3:
            ///     1. Bad_MonitoringModeInvalid
            ///     2. Bad_NodeIdInvalid
            ///     3. Bad_NodeIdUnknown
            ///     4. Bad_AttributeIdInvalid
            ///     5. Bad_IndexRangeInvalid
            ///     6. Bad_DataEncodingInvalid
            ///     7. Bad_DataEncodingUnsupported
            ///     8. Bad_UserAccessDenied
            ///     9. Bad_MonitoredItemFilterInvalid
            ///     10. Bad_MonitoredItemFilterUnsupported
            ///     11. Bad_FilterNotAllowed
            /// </description>
            /// <expectedresult>Client logs and/or displays an error to the end-user.
            /// The client indicates the problematic nodes, perhaps shows them as disabled or revises their status/quality as Bad.
            /// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_01()
            {
                // Injects error "Bad_MonitoredItemIdInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-01.js", UACodeBits.BadMonitoredItemIdInvalid);   // test script uses an incorrect code
                //TestCase_Err_003_(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-008a", UACodeBits.BadMonitoringModeInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_02()
            {
                // Injects error "Bad_NodeIdInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-02.js", UACodeBits.BadNodeIdInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_03()
            {
                // Injects error "Bad_NodeIdUnknown" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-03.js", UACodeBits.BadNodeIdUnknown);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_04()
            {
                // Injects error "Bad_AttributeIdInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-04.js", UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_05()
            {
                // Injects error "Bad_IndexRangeInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-05.js", UACodeBits.BadIndexRangeInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_06()
            {
                // Injects error "Bad_DataEncodingInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-06.js", UACodeBits.BadDataEncodingInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_07()
            {
                // Injects error "Bad_DataEncodingUnsupported" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-07.js", UACodeBits.BadDataEncodingUnsupported);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_08()
            {
                // Injects error "Bad_UserAccessDenied" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-08.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_09()
            {
                // Injects error "BadMonitoredItemFilterInvalid" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-09.js", UACodeBits.BadMonitoredItemFilterInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_10()
            {
                //  Injects error "BadMonitoredItemFilterUnsupported" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-10.js", UACodeBits.BadMonitoredItemFilterUnsupported);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_11()
            {
                // Injects error "Bad_FilterNotAllowed" into the first operation-level Result
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_003(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-003-11.js", UACodeBits.BadFilterNotAllowed);
            }

            /// <description>
            /// ServiceResult = Good, but the results for all items are set to one of the operation level error codes:
            ///     1. Bad_MonitoringModeInvalid
            ///     2. Bad_NodeIdInvalid
            ///     3. Bad_NodeIdUnknown
            ///     4. Bad_AttributeIdInvalid
            ///     5. Bad_IndexRangeInvalid
            ///     6. Bad_DataEncodingInvalid
            ///     7. Bad_DataEncodingUnsupported
            ///     8. Bad_UserAccessDenied
            ///     9. Bad_MonitoredItemFilterInvalid
            ///     10. Bad_MonitoredItemFilterUnsupported
            ///     11. Bad_FilterNotAllowed
            /// </description>
            /// <expectedresult>Client logs and/or displays an error to the end-user.
            /// The client indicates the problematic nodes, perhaps shows them as disabled or revises their status/quality as Bad.
            /// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_01()
            {
                // Injects error "Bad_MonitoringModeInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-01.js", UACodeBits.BadMonitoringModeInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_02()
            {
                // Injects error "Bad_NodeIdInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-02.js", UACodeBits.BadNodeIdInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_03()
            {
                // Injects error "Bad_NodeIdUnknown" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-03.js", UACodeBits.BadNodeIdUnknown);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_04()
            {
                // Injects error "Bad_AttributeIdInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-04.js", UACodeBits.BadAttributeIdInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_05()
            {
                // Injects error "Bad_IndexRangeInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-05.js", UACodeBits.BadIndexRangeInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_06()
            {
                // Injects error "Bad_DataEncodingInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-06.js", UACodeBits.BadDataEncodingInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_07()
            {
                // Injects error "Bad_DataEncodingUnsupported" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-07.js", UACodeBits.BadDataEncodingUnsupported);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_08()
            {
                // Injects error "Bad_UserAccessDenied" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-08.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_09()
            {
                // Injects error "BadMonitoredItemFilterInvalid" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-09.js", UACodeBits.BadMonitoredItemFilterInvalid);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_10()
            {
                //  Injects error "BadMonitoredItemFilterUnsupported" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-10.js", UACodeBits.BadMonitoredItemFilterUnsupported);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_11()
            {
                // Injects error "Bad_FilterNotAllowed" into the all operation-level Results
                // when calling CREATEMONITOREDITEMS.
                TestCase_Err_004(@"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-004-11.js", UACodeBits.BadFilterNotAllowed);
            }

            /// <description>
            /// Inject the following error codes:
            /// 1. serviceResult is Good, but the revisedSamplingInterval != requested samplingInterval, e.g. Max Double.
            /// 2. serviceResult is Good, but the revisedSamplingInterval is decades larger than the requested samplingInterval.
            /// 3. serviceResult is Good, but the revisedSamplingInterval is decades smaller than the requested samplingInterval.
            /// 4. serviceResult is Good, but the revisedSamplingInterval is a Negative number.
            /// 5. serviceResult is Good, but the revisedSamplingInterval is NaN. 
            /// </description>
            /// <expectedresult>
            /// a. Client may log and/or display an information message to the end-user. 
            /// b. Client may log and/or display an information message to the end-user. 
            /// c. Client logs and/or displays a warning message to the end-user. 
            /// d. Client logs and/or displays an error message to the end-user. e. Client logs and/or displays an error
            /// message to the end-user. 
            /// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_01()
            {
                // IMPROVE: add a test for RevisedSamplingIntervalDifferent (it's not this one...)
                TestCase_Err_006(false, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-01.js",
                                 UAEngineEventIds.RevisedSamplingIntervalAboveToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_02()
            {
                TestCase_Err_006(false, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-02.js",
                                 UAEngineEventIds.RevisedSamplingIntervalAboveToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_03()
            {
                TestCase_Err_006(false, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-03.js",
                                 UAEngineEventIds.RevisedSamplingIntervalBelowToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_04()
            {
                TestCase_Err_006(false, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-04.js",
                                 UAEngineEventIds.RevisedSamplingIntervalInvalid, LogEntryType.Warning);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006_05()
            {
                TestCase_Err_006(false, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-05.js",
                                 UAEngineEventIds.RevisedSamplingIntervalInvalid, LogEntryType.Warning);
            }

            // ReSharper restore InconsistentNaming
            #endregion
        }



        #region Private Methods

        [NotNull]
        private EasyUADataChangeNotificationEventArgs AssertAndGetSingleEventArgs(
            [NotNull] IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable)
        {
            // ReSharper disable PossibleMultipleEnumeration
            int count = eventArgsEnumerable.Count();
            AssertUtilities.WriteLine($"eventArgList.Count: {count}");
            //Assert.IsTrue(count >= 1);    // see swallowNotificationsInFailure
            Assert.AreEqual(1, count);
            EasyUADataChangeNotificationEventArgs eventArgs = eventArgsEnumerable.First();
            Assert.IsNotNull(eventArgs);
            return eventArgs;
            // ReSharper restore PossibleMultipleEnumeration
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_Err_003([NotNull] string scriptName, int codeBitsExpected)
        // ReSharper restore InconsistentNaming
        {
            using (Ctt.RunScript(CttProjectName, scriptName))
            {
                var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
                    new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
                int monitoredItemCount = monitoredItemArgumentsArray.Length;
                IEnumerable<EasyUADataChangeNotificationEventArgs>[] eventArgsEnumerableArray =
                    CollectMultipleDataChangeNotificationEventArgs(
                        MakeClient(),
                        monitoredItemArgumentsArray,
                        MillisecondsToCollectMultipleNotifications);
                Assert.AreEqual(monitoredItemCount, eventArgsEnumerableArray.Length);

                // The first monitored item
                IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable0 = eventArgsEnumerableArray[0];
                Debug.Assert(eventArgsEnumerable0 != null);
                EasyUADataChangeNotificationEventArgs eventArgs0 = AssertAndGetSingleEventArgs(eventArgsEnumerable0);
                AssertServiceException(eventArgs0.Exception, codeBitsExpected);

                // The second and subsequent monitored items
                for (int i = 1; i < monitoredItemCount; i++)
                {
                    // ReSharper disable PossibleMultipleEnumeration
                    IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable = eventArgsEnumerableArray[i];
                    Debug.Assert(eventArgsEnumerable != null);
                    Assert.IsTrue(eventArgsEnumerable.Count() > 1);
                    AssertNoFailures(eventArgsEnumerable);
                    // ReSharper restore PossibleMultipleEnumeration
                }
            }
        }

        // ReSharper disable InconsistentNaming
        private void TestCase_Err_004([NotNull] string scriptName, int codeBitsExpected)
        // ReSharper restore InconsistentNaming
        {
            using (Ctt.RunScript(CttProjectName, scriptName))
            {
                var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
                    new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
                int monitoredItemCount = monitoredItemArgumentsArray.Length;
                IEnumerable<EasyUADataChangeNotificationEventArgs>[] eventArgsEnumerableArray =
                    CollectMultipleDataChangeNotificationEventArgs(
                        MakeClient(),
                        monitoredItemArgumentsArray,
                        MillisecondsToCollectSingleNotification);
                Assert.AreEqual(monitoredItemCount, eventArgsEnumerableArray.Length);

                for (int i = 0; i < monitoredItemCount; i++)
                {
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable = eventArgsEnumerableArray[0];
                    Debug.Assert(eventArgsEnumerable != null);
                    EasyUADataChangeNotificationEventArgs eventArgs0 = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    AssertServiceException(eventArgs0.Exception, codeBitsExpected);
                }
            }
        }

        #endregion
    }
}
