﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.MonitoredItemServices
{
    public partial class ValueChangeTest
    {
        [TestClass] // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : ValueChangeTest_NoCtt
            // ReSharper restore InconsistentNaming
        {
            public ValueChangeTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }
        }



        [TestClass] // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_NoCtt_Softing_DemoServer_OpcTcp : ValueChangeTest_NoCtt
            // ReSharper restore InconsistentNaming
        {
            public ValueChangeTest_NoCtt_Softing_DemoServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }



        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_NoCtt : ValueChangeTest
            // ReSharper restore InconsistentNaming
        {
            #region Test Methods

// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
/// <description>Able to add one MonitoredItem of any supported data-type, to the correct/specified subscription.
/// </description> <expectedresult>Analyze the parameters for compliance and validity.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_001()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable =
 CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Int32), dynamic: true),
                        MillisecondsToCollectMultipleNotifications);

                    // ReSharper disable PossibleMultipleEnumeration
                    int count = eventArgsEnumerable.Count();
                    AssertUtilities.WriteLine($"count: {count}");
                    Assert.IsTrue(count > 1);
                    AssertNoFailures(eventArgsEnumerable);
                    Assert.AreEqual(1, recorder.CreateMonitoredItemsRecordList.Count - UATestUtilities.BuiltInCreateMonitoredItemsCount);
                    // ReSharper restore PossibleMultipleEnumeration
                }
            }
#endif

#if (NETFRAMEWORK)
/// <description>Add multiple MonitoredItems, varying the data-types of each node. Vary the ReadValueId (Node/attribute 
/// combination) when adding multiple monitored items to a subscription. Also include the EventNotifier attribute of 
/// nodes which generate events (if client supports events).
/// </description>
/// <expectedresult>Analyze the parameters for compliance and validity.
/// A single call that specifies all nodes is desired.</expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_002()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
                        new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
                    int monitoredItemCount = monitoredItemArgumentsArray.Length;

                    IEnumerable<EasyUADataChangeNotificationEventArgs>[] eventArgsEnumerableArray =
                        CollectMultipleDataChangeNotificationEventArgs(
                            MakeClient(),
                            monitoredItemArgumentsArray,
                            MillisecondsToCollectMultipleNotifications);
                    Assert.AreEqual(monitoredItemCount, eventArgsEnumerableArray.Length);

                    foreach (
                        IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable in eventArgsEnumerableArray)
                    {
                        // ReSharper disable PossibleMultipleEnumeration
                        Debug.Assert(eventArgsEnumerable != null);
                        int eventArgsCount = eventArgsEnumerable.Count();
                        AssertUtilities.WriteLine($"eventArgsCount: {eventArgsCount}");
                        Assert.IsTrue(eventArgsCount > 1);
                        AssertNoFailures(eventArgsEnumerable);
                        // ReSharper restore PossibleMultipleEnumeration
                    }

                    Assert.AreEqual(1, recorder.CreateMonitoredItemsRecordList.Count - UATestUtilities.BuiltInCreateMonitoredItemsCount);
                }
            }
#endif

            ///// <description>The Client defines 2 subscriptions whose active state is Enabled.
            ///// The Client is to add 1 or more MonitoredItems to both subscriptions, with each node while varying the data-type.
            ///// </description>
            ///// <expectedresult>The client will make at least 2 calls to CreateMonitoredItems. Ideally 1 call per subscription 
            ///// (specifying all MonitoredItems to add to the subscription).
            ///// Client can show/report that it is receiving data from both subscriptions.
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_003()
            //{
            //    Assert.Inconclusive("Test not implemented!");
            //}

            ///// <description>
            ///// The Client defines 2 subscriptions where the 1st PublishingMode=Enabled, the 2nd PublishingMode is disabled.
            ///// The Client is to add 1 or more MonitoredItems to the 2nd subscription..
            ///// </description>
            ///// <expectedresult>
            ///// The client will make 1 call to CreateMonitoredItems, passing all of the MonitoredItems in it.
            ///// The Client allows items to be added to the subscription although it is not Publishing.
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_004()
            //{
            //    Assert.Inconclusive("Test not implemented!");
            //}

            ///// <description>Add 4(or more) MonitoredItems varying the timestampsToReturn parameter.</description>
            ///// <expectedresult>
            ///// The client will make 4 calls to CreateMonitoredItem each time specifying the same subscriptionId.
            ///// The client can show/report the timestamp data per MonitoredItem as applicable by the request.
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_005()
            //{
            //    Assert.Inconclusive("Test not implemented!");
            //}

            ///// <description>
            ///// Add a large number of MonitoredItems to an existing (PublishingMode = enabled) subscription. Vary the “large” 
            ///// number, for example:
            /////     a. 20 items
            /////     b. 50 items
            /////     c. 100 items
            /////     d. 500 items
            /////     e. 1000 items.
            ///// </description>
            ///// <expectedresult>
            ///// The client should optimize the number of nodes it specifies in the request. The client may attempt 1 call, or 
            ///// break-up the larger requests into multiple smaller requests. In any case, the client must be prepared to receive 
            ///// an error indicating too many operations which if received must be handled gracefully.
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_006()
            //{
            //    Assert.Inconclusive("Test not implemented!");
            //}

            ///// <description>
            ///// Add multiple MonitoredItems each time varying the data-type of the specified Node, as well as varying the 
            ///// MonitoringMode such that:
            /////     a. itemsToCreate[0] = Sampling
            /////     b. itemsToCreate[1] = Reporting
            /////     c. itemsToCreate[2] = Disabled
            ///// </description>
            ///// <expectedresult>
            ///// The client should optimize the number of nodes it specifies in the request. 1 request is possible.
            ///// The client will show that it is receiving a value for the 2nd node specified (Reporting).
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateMonitoredItems")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("TimeDomain")]
            //[TestCategory("Type:Lab")]  // inferred
            //[TestMethod]
            //public void TestCase_007()
            //{
            //    Assert.Inconclusive("Test not implemented!");
            //}

            /// <description>
            /// Client adds 1 MonitoredItem specifying an attribute of Value (13)
            /// </description>
            /// <expectedresult>
            /// The attribute is correctly specified and the client can demonstrate receiving a value for this MonitoredItem.
            /// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")] // inferred
            [TestMethod]
            public void TestCase_008()
            {
                IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable =
                    CollectMonitoredItemChangedEventArgs(
                        MakeClient(),
                        MakeMonitoredItemArguments(typeof(Int32), dynamic: true),
                        MillisecondsToCollectMultipleNotifications);

                // ReSharper disable PossibleMultipleEnumeration
                int count = eventArgsEnumerable.Count();
                AssertUtilities.WriteLine($"count: {count}");
                foreach (EasyUADataChangeNotificationEventArgs eventArgs in eventArgsEnumerable)
                {
                    Debug.Assert(eventArgs != null);
                    // ReSharper disable once AssignNullToNotNullAttribute
                    AssertUtilities.WriteLine(eventArgs.ToString());
                }

                Assert.IsTrue(count > 1);
                AssertNoFailures(eventArgsEnumerable);
                // ReSharper restore PossibleMultipleEnumeration
            }

#if (NETFRAMEWORK)
/// <description>
/// Client adds 1 MonitoredItem specifying an attribute of DisplayName (4)
/// </description>
/// <expectedresult>
/// The attribute is correctly specified and the client can demonstrate receiving a value for this MonitoredItem.
/// </expectedresult>
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:Lab")]  // inferred
            [TestMethod]
            public void TestCase_009()
            {
                using (var recorder = new UAOperationRecorder())
                {
                    EasyUAMonitoredItemArguments dynamicMonitoredItemArguments =
                        MakeMonitoredItemArguments(typeof(Int32), dynamic: true);
                    dynamicMonitoredItemArguments.AttributeId = UAAttributeId.DisplayName;
                    IEnumerable<EasyUADataChangeNotificationEventArgs> eventArgsEnumerable =
 CollectMonitoredItemChangedEventArgs
                        (
                            MakeClient(),
                            dynamicMonitoredItemArguments,
                            MillisecondsToCollectSingleNotification);

                    // ReSharper disable PossibleMultipleEnumeration
                    int count = eventArgsEnumerable.Count();
                    AssertUtilities.WriteLine($"count: {count}");
                    foreach (EasyUADataChangeNotificationEventArgs eventArgs in eventArgsEnumerable)
                    {
                        Debug.Assert(eventArgs != null);
                        // ReSharper disable once AssignNullToNotNullAttribute
                        AssertUtilities.WriteLine(eventArgs.ToString());
                    }
                    Assert.AreEqual(1, count);
                    EasyUADataChangeNotificationEventArgs eventArgs0 = AssertAndGetSingleEventArgs(eventArgsEnumerable);
                    Assert.IsNull(eventArgs0.Exception);
                    UAAttributeData attributeData = eventArgs0.AttributeData;
                    Assert.IsNotNull(attributeData);
                    object value = attributeData.Value;
                    Assert.IsNotNull(value);
                    Assert.AreEqual("Int32Value", value);
                    Assert.AreEqual(1, recorder.CreateMonitoredItemsRecordList.Count - UATestUtilities.BuiltInCreateMonitoredItemsCount);
                    CreateMonitoredItemsRecord createMonitoredItemsRecord =
 recorder.CreateMonitoredItemsRecordList[UATestUtilities.BuiltInCreateMonitoredItemsCount];
                    Debug.Assert(createMonitoredItemsRecord != null);
                    MonitoredItemCreateRequestCollection monitoredItemCreateRequestCollection =
                        createMonitoredItemsRecord.ItemsToCreate;
                    Assert.IsNotNull(monitoredItemCreateRequestCollection);
                    Assert.AreEqual(1, monitoredItemCreateRequestCollection.Count);
                    MonitoredItemCreateRequest monitoredItemCreateRequest = monitoredItemCreateRequestCollection[0];
                    Assert.IsNotNull(monitoredItemCreateRequest);
                    ReadValueId itemToMonitor = monitoredItemCreateRequest.ItemToMonitor;
                    Assert.IsNotNull(itemToMonitor);
                    Assert.AreEqual( /*Opc.Ua.*/Attributes.DisplayName, itemToMonitor.AttributeId);

                    // ReSharper restore PossibleMultipleEnumeration
                }
            }
#endif

#if (NETFRAMEWORK)
            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_Err_006_01_Inverted()
            {
                // IMPROVE: add a test for RevisedSamplingIntervalDifferent (it's not this one...)
                TestCase_Err_006(true, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-01.js",
                                 UAEngineEventIds.RevisedSamplingIntervalAboveToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_Err_006_02_Inverted()
            {
                TestCase_Err_006(true, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-02.js",
                                 UAEngineEventIds.RevisedSamplingIntervalAboveToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_Err_006_03_Inverted()
            {
                TestCase_Err_006(true, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-03.js",
                                 UAEngineEventIds.RevisedSamplingIntervalBelowToleranceRange, LogEntryType.Information);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_Err_006_04_Inverted()
            {
                TestCase_Err_006(true, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-04.js",
                                 UAEngineEventIds.RevisedSamplingIntervalInvalid, LogEntryType.Warning);
            }

            [TestCategory("Keyword:CreateMonitoredItems")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestMethod]
            public void TestCase_Err_006_05_Inverted()
            {
                TestCase_Err_006(true, @"Monitored Item Services\Monitor Client Value Change\Test Cases\Err-006-05.js",
                                 UAEngineEventIds.RevisedSamplingIntervalInvalid, LogEntryType.Warning);
            }
#endif

            // ReSharper restore InconsistentNaming

            #endregion
        }



#if (NETFRAMEWORK)
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]   // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_Ctt_OpcFoundation_SampleServer_OpcTcp : ValueChangeTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public ValueChangeTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
            }
        }
#endif



#if (NETFRAMEWORK)
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]   // currently CTT does not inject the changes as it should
        // ReSharper disable InconsistentNaming
        public class ValueChangeTest_Ctt_Softing_DemoServer_OpcTcp : ValueChangeTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public ValueChangeTest_Ctt_Softing_DemoServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }
#endif
    }
}
