﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System.Collections.Generic;
using JetBrains.Annotations;
using Opc.Ua;
using Opc.Ua.Client;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    class SdkOperationRecord : OperationRecord
    {
        // Inputs
        public RequestHeader RequestHeader;

        // Outputs
        public ResponseHeader Result { get; set; }
        public DiagnosticInfoCollection DiagnosticInfos { get; set; }
    }

    class SdkSessionOperationRecord : OperationRecord
    {
        public SdkSessionOperationRecord([NotNull] Session session)
        {
            Session = session;
        }

        [NotNull]
        public Session Session { get; }
    }

    class SdkSessionClientOperationRecord : SdkOperationRecord
    {
        public SdkSessionClientOperationRecord([NotNull] SessionClient sessionClient)
        {
            SessionClient = sessionClient;
        }

        [NotNull]
        public SessionClient SessionClient { get; }
    }

    class CloseSessionRecord : SdkSessionClientOperationRecord
    {
        public CloseSessionRecord([NotNull] SessionClient sessionClient)
            : base(sessionClient)
        {
        }

        // Inputs
        public bool DeleteSubscriptions;
    }

    //public class ActivateSessionRecord : SdkSessionClientOperationRecord
    //{
    //    // Inputs
    //    public RequestHeader RequestHeader;
    //    public SignatureData ClientSignature;
    //    public SignedSoftwareCertificateCollection ClientSoftwareCertificates;
    //    public StringCollection LocaleIds;
    //    public ExtensionObject UserIdentityToken;
    //    public SignatureData UserTokenSignature;

    //    // Outputs
    //    public byte[] ServerNonce;
    //    public StatusCodeCollection Results;
    //    public DiagnosticInfoCollection DiagnosticInfos;
    //}

    class CreateMonitoredItemsRecord : SdkSessionClientOperationRecord
    {
        public CreateMonitoredItemsRecord([NotNull] SessionClient sessionClient)
            : base(sessionClient)
        {
        }

        // Inputs
        public uint SubscriptionId;
        public TimestampsToReturn TimestampsToReturn;
        public MonitoredItemCreateRequestCollection ItemsToCreate;

        // Outputs
        public MonitoredItemCreateResultCollection Results { get; set; }
    }

    class SessionOpenRecord : SdkSessionOperationRecord
    {
        public SessionOpenRecord([NotNull] Session session)
            : base(session)
        {
        }

        // Inputs
        public string SessionName;
        public uint SessionTimeout;
        public IUserIdentity Identity;
        public IList<string> PreferredLocales;
    }

    class ReadRecord : SdkSessionClientOperationRecord
    {
        public ReadRecord([NotNull] SessionClient sessionClient)
            : base(sessionClient)
        {
        }

        // Inputs
        public double MaxAge;
        public TimestampsToReturn TimestampsToReturn;
        public ReadValueIdCollection NodesToRead;

        // Outputs
        public DataValueCollection Results;
    }

    class WriteRecord : SdkSessionClientOperationRecord
    {
        public WriteRecord([NotNull] SessionClient sessionClient)
            : base(sessionClient)
        {
        }

        // Inputs
        public WriteValueCollection NodesToWrite;

        // Outputs
        public StatusCodeCollection Results;
    }
}
