﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SessionServices
{
    public partial class BaseTest
    {
        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BaseTest_Ctt_OpcFoundation_SampleServer_OpcTcp : BaseTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public BaseTest_Ctt_OpcFoundation_SampleServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
            }
        }



        [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BaseTest_Ctt_Softing_DemoServer_OpcTcp : BaseTest_Ctt
        // ReSharper restore InconsistentNaming
        {
            public BaseTest_Ctt_Softing_DemoServer_OpcTcp()
            {
                CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
            }
        }



        // ReSharper disable InconsistentNaming
        public class BaseTest_Ctt : BaseTest
        // ReSharper restore InconsistentNaming
        {
            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>if a client supports a configurable requestedSessionTimeout then Client specifies a really short 
            /// requestedSessionTimeout, e.g. 1-second.
            /// CTT adjusts the return rate to 60-seconds.
            /// Test script looks to see if the requestedSessionTimeout &lt;= 10 secs.
            /// </description>
            /// <expectedresult>Client is prepared to receive a revised session timeout, and then logs and/or reports an 
            /// information message to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_005()
            {
                TestCase_005(false);
            }

            /// <description>If a client supports a configurable requestedsessiontimeout then Client specifies a really long 
            /// requestedSessionTimeout, e.g. 1 week.
            /// CTT adjusts the return rate to 60-seconds.
            /// Test script looks to see if the requestedSessionTimeout is &gt;= 60 secs.
            /// </description>
            /// <expectedresult>Client is prepared to receive a revised session timeout, and then logs and/or reports an 
            /// information message to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_006()
            {
                TestCase_006(false);
            }

            /// <description>Randomly inject one of the following status codes into the Service Result.
            /// Repeat this test multiple times, each time varying the code:
            ///
            /// 1. Bad_SecureChannelIdInvalid
            /// 2. Bad_NonceInvalid
            /// 3. Bad_SecurityChecksFailed
            /// 4. Bad_CertificateTimeInvalid
            /// 5. Bad_CertificateIssuerTimeInvalid
            /// 6. Bad_CertificateHostNameInvalid
            /// 7. Bad_CertificateUriInvalid
            /// 8. Bad_CertificateUseNotAllowed
            /// 9. Bad_CertificateIssuerUseNotAllowed
            /// 10. Bad_CertificateUntrusted
            /// 11. Bad_CertificateRevocationUnknown
            /// 12. Bad_CertificateIssuerRevocationUnknown
            /// 13. Bad_CertificateRevoked
            /// 14. Bad_CertificateIssuerRevoked
            /// 15. Bad_TooManySessions.
            /// 16. Bad_ServerUriInvalid 
            /// </description>
            /// <expectedresult>Client does not crash, but instead reports an error to the end-user.
            /// The Client terminates/cancels the session request.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_01()
            {
                // Injects service result = "Bad_SecureChannelIdInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-01.js", UACodeBits.BadSecureChannelIdInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_02()
            {
                // Injects service result = "Bad_NonceInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-02.js", UACodeBits.BadNonceInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_03()
            {
                //  Injects service result = "Bad_SecurityChecksFailed".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-03.js", UACodeBits.BadSecurityChecksFailed);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_04()
            {
                // Injects service result = "Bad_CertificateTimeInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-04.js", UACodeBits.BadCertificateTimeInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_05()
            {
                // Injects service result = "Bad_CertificateIssuerTimeInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-05.js", UACodeBits.BadCertificateIssuerTimeInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_06()
            {
                // Injects service result = "Bad_CertificateHostnameInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-06.js", UACodeBits.BadCertificateHostNameInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_07()
            {
                // Injects service result = "Bad_CertificateUriInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-07.js", UACodeBits.BadCertificateUriInvalid);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_08()
            {
                // Injects service result = "Bad_CertificateUseNotAllowed".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-08.js", UACodeBits.BadCertificateUseNotAllowed);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_09()
            {
                // Injects service result = "Bad_CertificateIssuerUseNotAllowed".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-09.js", UACodeBits.BadCertificateIssuerUseNotAllowed);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_10()
            {
                // Injects service result = "Bad_CertificateUntrusted".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-10.js", UACodeBits.BadCertificateUntrusted);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_11()
            {
                // Injects service result = "Bad_CertificateRevocationUnknown".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-11.js", UACodeBits.BadCertificateRevocationUnknown);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_12()
            {
                // Injects service result = "Bad_CertificateIssuerRevocationUnknown".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-12.js", UACodeBits.BadCertificateIssuerRevocationUnknown);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_13()
            {
                // Injects service result = "Bad_CertificateRevoked".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-13.js", UACodeBits.BadCertificateRevoked);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_14()
            {
                // Injects service result = "Bad_CertificateIssuerRevoked".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-14.js", UACodeBits.BadCertificateIssuerRevoked);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_15()
            {
                //  Injects service result = "Bad_TooManySessions".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-15.js", UACodeBits.BadTooManySessions);
            }

            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_001_16()
            {
                // Injects service result = "Bad_ServerUriInvalid".
                TestCase_Err_001(@"Session Services\Session Client Base\Test Cases\Err-001-16.js", UACodeBits.BadServerUriInvalid);
            }

            /// <description>sessionId returned is the same as another already-open session.
            ///</description>
            /// <expectedresult>Client closes this Session, leaving the existing session alone.
            /// Client then logs and/or displays an error to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_002()
            {
                TestCase_Err_002(false);
            }

            /// <description>Client successfully establishes a first session. Client attempts to establish another session but 
            /// receives a sessionId = null.
            ///</description>
            /// <expectedresult>Client closes this Session, leaving the existing session alone.
            /// Client then reports an error to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003()
            {
                TestCase_Err_003(false);
            }

            /// <description>Client opens 2 (or more) sessions with the UA Server. The first is established with success. 
            /// The second responds with the same authenticationToken as used by the first session.
            ///</description>
            /// <expectedresult>Client reports an error and does not connect to the UA Server.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_006()
            {
                TestCase_Err_006(false);
            }

            /// <description>revisedSessionTimeout is decades long, e.g. requestedSessionTimeout * 10.
            ///</description>
            /// <expectedresult>Client logs this information for diagnostic purposes only.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_007()
            {
                TestCase_Err_007(false);
            }

            // cannot currently test this due to CTT script problem; and, the test case seemed to disappear
            ///// <description>Server returns a revisedSessionTimeout that is unrealistically too short, i.e. 1-second.
            /////</description>
            ///// <expectedresult>Client logs and/or displays a warning message to the end-user.
            ///// Verify Client does not connect to the Server without end-user confirmation due to the potential for network 
            ///// saturation and overloading of the Server.</expectedresult>
            //[TestCategory("Keyword:CreateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_008()
            //{
            //    // Revising the session timeout to an insane low level, 1-sec.
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-008.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //        EasyUAClient client = MakeClient();
            //        // Do any operation that requires a UA session, in order to cause the session be created.
            //        client.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //        // ReSharper disable PossibleNullReferenceException
            //        LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //            (args.EventId == UAEngineEventIds.RevisedSessionTimeoutUnusuallyShort));
            //        // ReSharper restore PossibleNullReferenceException
            //        Assert.IsTrue(logEntryEventArgs != null);
            //        AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //        Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            //    }
            //}

            /// <description>Server returns a revisedSessionTimeout that is unrealistically too long, i.e. 1-week.
            ///</description>
            /// <expectedresult>Client logs and/or displays a warning to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_009()
            {
                TestCase_Err_009(false);
            }

            // cannot currently test this due to CTT script problem; and, the test case seemed to disappear
            ///// <description>Server returns an unrealistically short revisedSessionTimeout of 100ms (or less).
            /////</description>
            ///// <expectedresult>Client reports a warning to the end-user.
            ///// Client may also elect to prompt the user to disconnect from the UA Server, or to simply disconnect from the 
            ///// UA Server.</expectedresult>
            //[TestCategory("Keyword:CreateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_010()
            //{
            //    //  Revising the session timeout to be 100msec.
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-010.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //        EasyUAClient client = MakeClient();
            //        try
            //        {
            //            // Do any operation that requires a UA session, in order to cause the session be created.
            //            client.Read(new UAReadArguments(EndpointDescriptor,
            //                                            UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            //        }
            //        catch (UAException uaException)
            //        {
            //            AssertUtilities.WriteLine(uaException.Message);
            //            Exception innerException = uaException.InnerException;
            //            Assert.IsNotNull(innerException);
            //            AssertUtilities.WriteLine(innerException.Message);
            //            var clientEngineException = innerException as UAClientEngineException;
            //            Assert.IsNotNull(clientEngineException);
            //            Assert.AreEqual(UAErrorCodes.RevisedSessionTimeoutBelowMinimum, clientEngineException.ErrorCode);

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //            // ReSharper disable PossibleNullReferenceException
            //            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //                (args.EventId == UAEngineEventIds.RevisedSessionTimeoutBelowMinimum));
            //            // ReSharper restore PossibleNullReferenceException
            //            Assert.IsTrue(logEntryEventArgs != null);
            //            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            //            return;
            //        }
            //        Assert.Fail();
            //    }
            //}

            /// <description>serverNonce returned is less than 32 bytes.
            ///</description>
            /// <expectedresult>Client closes the session and then logs and/or displays an error to the end-user.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011()
            {
                TestCase_Err_011(false);
            }

            // cannot currently test this due to CTT script problem; and, the test case seemed to disappear
            ///// <description>Client successfully establishes a session with the UA Server. Client attempts to establish a second 
            ///// (concurrent) session with the same UA server, but the serverNonce returned is the same one received in the first 
            ///// session response.
            ///// </description>
            ///// <expectedresult>Client closes the session and then logs and/or displays an error to the end-user.
            ///// The original session remains open/alive.</expectedresult>
            //[TestCategory("Keyword:CreateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_012()
            //{
            //    // ServerNonce is re-used
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-012.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //        EasyUAClient client1 = MakeClient();
            //        // Do any operation that requires a UA session, in order to cause the session be created.
            //        client1.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

            //        try
            //        {
            //            EasyUAClient client2 = MakeClient();
            //            // Do any operation that requires a UA session, in order to cause the session be created.
            //            client2.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            //        }
            //        catch (UAException uaException)
            //        {
            //            AssertUtilities.WriteLine(uaException.Message);
            //            Exception innerException = uaException.InnerException;
            //            Assert.IsNotNull(innerException);
            //            AssertUtilities.WriteLine(innerException.Message);
            //            var clientEngineException = innerException as UAClientEngineException;
            //            Assert.IsNotNull(clientEngineException);
            //            Assert.AreEqual(UAErrorCodes.ServerNonceDuplicate, clientEngineException.ErrorCode);

            //            // IMPROVE: verify that the session has been truly closed

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //            // ReSharper disable PossibleNullReferenceException
            //            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //                (args.EventId == UAEngineEventIds.ServerNonceDuplicate));
            //            // ReSharper restore PossibleNullReferenceException
            //            Assert.IsTrue(logEntryEventArgs != null);
            //            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            //            // The call below uses the original session and must not throw.
            //            client1.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

            //            return;
            //        }
            //        Assert.Fail();
            //    }
            //}

            /// <description>The EndpointDescriptions returned from the Discovery endpoint are different to those in the 
            /// CreateSession response.
            /// </description>
            /// <expectedresult>Client disconnects and reports the error.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_020()
            {
                TestCase_Err_020(false);
            }

            /// <description>serverEndpoints is empty.
            /// </description>
            /// <expectedresult>Session is dropped. Endpoints must match what was received from GetEndpoints.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_021()
            {
                TestCase_Err_021(false);
            }

            /// <description>ServiceResult = Good, the list of serverEndpoints[] is reduced by 1 endpoint to force a difference 
            /// with the Discovery services.
            /// </description>
            /// <expectedresult>Client closes the session.
            /// Endpoints must match what was received from GetEndpoints.
            /// Client logs and/or displays an error to the end user.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_022()
            {
                TestCase_Err_022(false);
            }

            //From: Nathan Pocock [mailto:nathan.pocock@opcfoundation.org] 
            //Sent: Wednesday, June 05, 2013 8:23 PM
            //To: Zbynek Zahradnik; deepthi.teegala@opcfoundation.org
            //Subject: RE: Question to /Session Services/Session Client Base/Test Cases/Err-023

            //Hello Zbynek,

            //This is one of those test-cases that you should ignore. There are no/zero UA Servers (or Clients) that have Software Certificates yet. The reason? Because it took ages for the UAWG to define them, and we currently do not have any implementation to create them. 

            //In UA Part 4 5.6.3.2 (ActivateSession parameters) the productUri parameter contains this description “The productUri contained in the SoftwareCertificates shall match the productUri in the ApplicationDescription passed by the Client in the CreateSession requests. Certificates without matching productUri should be ignored. ” so I think you’re right that this test-case needs revision anyway. This is one of our original test-cases from 2009, back when we had limited understanding of UA. Interestingly enough, there’s no commentary about whether the client should conduct the same validation as the server does; I think it should from a security standpoint. 

            //I propose that we disable this test-case until we have SoftwareCertificate capabilities.

            //Thanks,


            //Nathan.
            //From: Zbynek Zahradnik [mailto:zbynekz@opclabs.com] 
            //Sent: Wednesday, June 5, 2013 10:26 AM
            //To: deepthi.teegala@opcfoundation.org; nathan.pocock@opcfoundation.org
            //Subject: Question to /Session Services/Session Client Base/Test Cases/Err-023
            //Deepthi, Nathan,
            //I have a question to this test.
            //Err-023	CTT 	CreateSession	productUris contained in the returned SoftwareCertificates do not match the 
            //productUris in the EndpointDescription used by the Client to connect to the Server.	Client closes session 
            //calling CloseSession.
            //Client logs and/or display an error to the end user.
            //First, what does it mean?
            //a)	For every productUri in SoftwareCertificates, such productUri must be in EndpointDescriptions?
            //OR
            //b)	For every producturi in EndpointDescriptions, such productUri must be in SoftwareCertificates?
            //Can you point me to place in the specification that describes this?
            //Currently, we are doing closely to b), and we are always logging the problem, but we only reject the connection 
            //if a parameter EasyUAClient.SessionParameters.RequireMatchingServerSoftwareCertificates is set to ‘true’ (which 
            //is not by default). The reason for this is interoperability, as in reality I found that too many servers return 
            //no software certificates at all (can you tell me a single one that does? – all servers I am currently testing 
            //with, return zero software certificates). Technically, I can change the default to requiring this easily, but 
            //I struggle with doing it a lot. In the client business, our aim is to be able to connect whenever possible, and 
            //turning down servers may mean that customers will simply go shop somewhere else, to a client that is NOT even 
            //certified, to get the results they want, and we lose the business. 
            //BUT, if a) is the right option, the current test script makes no sense unless used with a server that returns 
            //at least one software certificate. Where can I get one?
            //Also, in  UA 1.02, it actually says the SoftwareCertificates during connection time are deprecated and should 
            //be empty. So, that further disqualifies b) as an interoperable option. And, it kind of makes a) unnecessary 
            // as well.
            //OR, is there a third option? E.g. do as b), but only check it SoftwareCertificates contains AT LEAST ONE 
            //certificate?
            //AND, the test description says “the productUris in the EndpointDescription used by the Client to connect to 
            //the Server”, but that’s not true either – the productUris are not “used” like this – instead, the server 
            //returns the EndpointDescriptions in CreateSession.
            //All in all, I think that the whole test needs much more thought, or clarification.
            /// <description>productUris contained in the returned SoftwareCertificates do not match the productUris in the 
            /// EndpointDescription used by the Client to connect to the Server.
            /// </description>
            /// <expectedresult>Client closes session calling CloseSession.
            /// Client logs and/or display an error to the end user.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_023()
            {
                TestCase_Err_023(false);
            }

            /// <description>serverSoftwareCertificates Empty.
            ///</description>
            /// <expectedresult>The client should log (for debug purposes) that the certificates are empty.</expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_024()
            {
                TestCase_Err_024(false);
            }

            /// <description>serverSignature returns empty information.
            /// </description>
            /// <expectedresult>Client should close the session since the signature is used to validate the authenticity of 
            /// the connection. As a minimum, the Client should log (Debug purposes) why the session was dropped.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_025()
            {
                TestCase_Err_025(false);
            }

            /// <description>maxRequestMessageSize is decades larger than maxResponseMessageSize, e.g. maxResponseMessageSize * 10.
            /// If the requested MaxResponseMessageSize == 0 then set revise the MaxRequestMessageSize = 500;
            /// </description>
            /// <expectedresult>Client reports an informational message to the end-user to report the limitations on the channel. 
            /// The Client continues to communicate with the UA Server while abiding by the message size constraint.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_027()
            {
                TestCase_Err_027(false);
            }

            // DUP: Err-023. 20130401 NathanP: removed
            ///// <description>serverSoftwareCertificates contains a certificate where the productUri is different from that found 
            ///// in the EndpointDescription.
            ///// </description>
            ///// <expectedresult>The Client rejects that certificate and reports a message to the end-user.
            ///// </expectedresult>
            //[TestCategory("Keyword:CreateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_028()
            //{
            //    //  serverSoftwareCertificates contains a certificate where the productUri is different from that
            //    // found in the EndpointDescription..
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-028.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //            EasyUAClient client = MakeClient();
            //            client.IsolatedParameters.Session.RequireMatchingServerSoftwareCertificates = true;
            //            try
            //            {
            //                // Do any operation that requires a UA session, in order to cause the session be created.
            //                client.Read(new UAReadArguments(EndpointDescriptor,
            //                                                UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));
            //            }
            //            catch (UAException uaException)
            //            {
            //                AssertUtilities.WriteLine(uaException.Message);
            //                Exception innerException = uaException.InnerException;
            //                Assert.IsNotNull(innerException);
            //                AssertUtilities.WriteLine(innerException.Message);
            //                var clientEngineException = innerException as UAClientEngineException;
            //                Assert.IsNotNull(clientEngineException);
            //                Assert.AreEqual(UAErrorCodes.ServerSoftwareCertificateProductUriNotFound,
            //                                clientEngineException.ErrorCode);

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //                // ReSharper disable PossibleNullReferenceException
            //                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //                                                                                 (args.EventId ==
            //                                                                                  UAEngineEventIds
            //                                                                                      .ServerSoftwareCertificateProductUriNotFound));
            //                // ReSharper restore PossibleNullReferenceException
            //                Assert.IsTrue(logEntryEventArgs != null);
            //                AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            //                return;
            //            }
            //            Assert.Fail();
            //    }
            //}

            /// <description>Randomly inject one of the following codes into the Service Result.
            /// Repeat this test several times, each time varying the code: 
            ///
            /// 1. Bad_IdentityTokenInvalid
            /// 2. Bad_IdentityTokenRejected
            /// 3. Bad_UserAccessDenied
            /// 4. Bad_ApplicationSignatureInvalid
            /// 5. Bad_UserSignatureInvalid
            /// 6. Bad_NoValidCertificates
            /// 7. Bad_IdentityChangeNotSupported
            /// 8. Bad_CertificateTimeInvalid
            /// 9. Bad_CertificateIssuerTimeInvalid
            /// 10.Bad_CertificateHostNameInvalid
            /// 11.Bad_CertificateUriInvalid
            /// 12.Bad_CertificateUseNotAllowed
            /// 13.Bad_CertificateIssuerUseNotAllowed
            /// 14.Bad_CertificateUntrusted
            /// 15.Bad_CertificateRevocationUnknown
            /// 16.Bad_CertificateIssuerRevocationUnknown
            /// 17.Bad_CertificateRevoked
            /// 18.Bad_CertificateIssuerRevoked
            /// 19.Bad_SecurityChecksFailed. 
            /// </description>
            /// <expectedresult>Client handles the situation, and then logs and/or reports the error to the end-user.
            /// Client closes the session.
            /// Client does not reconnect to the server until the credentials have been fixed and the end-user initiates the 
            /// connection, or a reasonable delay (5 minutes?) occurs before the client retries. 
            /// </expectedresult>
            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_01()
            {
                // Sets the service result code to "Bad_IdentityTokenInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-01.js", UACodeBits.BadIdentityTokenInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_02()
            {
                // Sets the service result code to "Bad_IdentityTokenRejected" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-02.js", UACodeBits.BadIdentityTokenRejected);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_03()
            {
                // Sets the service result code to "Bad_UserAccessDenied" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-03.js", UACodeBits.BadUserAccessDenied);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_04()
            {
                // Sets the service result code to "Bad_ApplicationSignatureInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-04.js", UACodeBits.BadApplicationSignatureInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_05()
            {
                // Sets the service result code to "Bad_UserSignatureInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-05.js", UACodeBits.BadUserSignatureInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_06()
            {
                // Sets the service result code to "Bad_NoValidCertificates" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-06.js", UACodeBits.BadNoValidCertificates);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_07()
            {
                // Sets the service result code to "Bad_IdentityChangeNotSupported" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-07.js", UACodeBits.BadIdentityChangeNotSupported);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_08()
            {
                // Sets the service result code to "Bad_CertificateTimeInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-08.js", UACodeBits.BadCertificateTimeInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_09()
            {
                // Sets the service result code to "Bad_CertificateIssuerTimeInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-09.js", UACodeBits.BadCertificateIssuerTimeInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_10()
            {
                // Sets the service result code to "Bad_CertificateHostNameInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-10.js", UACodeBits.BadCertificateHostNameInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_11()
            {
                // Sets the service result code to "Bad_CertificateUriInvalid" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-11.js", UACodeBits.BadCertificateUriInvalid);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_12()
            {
                // Sets the service result code to "Bad_CertificateUseNotAllowed" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-12.js", UACodeBits.BadCertificateUseNotAllowed);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_13()
            {
                // Sets the service result code to "Bad_CertificateIssuerUseNotAllowed" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-13.js", UACodeBits.BadCertificateIssuerUseNotAllowed);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_14()
            {
                // Sets the service result code to "Bad_CertificateUntrusted" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-14.js", UACodeBits.BadCertificateUntrusted);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_15()
            {
                // Sets the service result code to "Bad_CertificateRevocationUnknown" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-15.js", UACodeBits.BadCertificateRevocationUnknown);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_16()
            {
                // Sets the service result code to "Bad_CertificateIssuerRevocationUnknown" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-16.js", UACodeBits.BadCertificateIssuerRevocationUnknown);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_17()
            {
                // Sets the service result code to "Bad_CertificateRevoked" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-17.js", UACodeBits.BadCertificateRevoked);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_18()
            {
                // Sets the service result code to "Bad_CertificateIssuerRevoked" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-18.js", UACodeBits.BadCertificateIssuerRevoked);
            }

            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]  // was Lab on the Web, but that doesn't seem right
            [TestMethod]
            public void TestCase_Err_029_19()
            {
                // Sets the service result code to "Bad_SecurityChecksFailed" when calling ActivateSessionHelper.
                TestCase_Err_029(@"Session Services\Session Client Base\Test Cases\Err-029-19.js", UACodeBits.BadSecurityChecksFailed);
            }

            // NOTE: Test case to be modified by OPC Foundation (20130528 DeepthiT: “If the connection is NOT secure, then 
            // the client may ignore the ServerNonce, or Client may close the session when security is used and then logs 
            // and/or displays an error to the end-user.”
            /// <description>serverNonce is null/empty.
            /// </description>
            /// <expectedresult>Client logs and/or reports an error to the end-user.</expectedresult>
            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_030_alpha()
            {
                // Return empty ServerNonce.
                using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-030.js"))
                {
                    var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
                    EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.AllowedMessageSecurityModes =
                        UAMessageSecurityModes.SecurityNone;

                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                    var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                }
            }

            // CTT cannot do secure connections needed for this test
            //[TestCategory("Keyword:ActivateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_030_beta()
            //{
            //    // Return empty ServerNonce.
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-030.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //        EasyUAClient client = MakeClient();
            //        client.IsolatedParameters.Session.EndpointSelectionPolicy.AllowedMessageSecurityModes =
            //            UAMessageSecurityModes.SecuritySign;
            //        try
            //        {
            //            // Do any operation that requires a UA session, in order to cause the session be created.
            //            client.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            //        }
            //        catch (UAException uaException)
            //        {
            //            AssertUtilities.WriteLine(uaException.Message);
            //            Exception innerException = uaException.InnerException;
            //            Assert.IsNotNull(innerException);
            //            AssertUtilities.WriteLine(innerException.Message);
            //            var clientEngineException = innerException as UAClientEngineException;
            //            Assert.IsNotNull(clientEngineException);
            //            Assert.AreEqual(UAErrorCodes.ServerNonceNull, clientEngineException.ErrorCode);

            //            // IMPROVE: verify that the session has been truly closed

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //            // ReSharper disable PossibleNullReferenceException
            //            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //                (args.EventId == UAEngineEventIds.ServerNonceNull));
            //            // ReSharper restore PossibleNullReferenceException
            //            Assert.IsTrue(logEntryEventArgs != null);
            //            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            //            return;
            //        }
            //        Assert.Fail();
            //    }
            //}

            /// <description>results[0] is Bad_CertificateUriInvalid.
            ///</description>
            /// <expectedresult>Client logs and/or reports the error to the end-user. The client should extract as much 
            /// information as possible from the diagnosticInfos parameter.</expectedresult>
            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_032()
            {
                TestCase_Err_032(false);
            }

            // cannot currently test this, the available servers return 0 validation results
            ///// <description>All results are Bad_.
            /////</description>
            ///// <expectedresult>Client logs and/or reports the error to the end-user. The client should extract as much 
            ///// information as possible from the diagnosticInfos parameter.</expectedresult>
            //[TestCategory("Keyword:ActivateSession")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_033()
            //{
            //    // All results are Bad_<some error>.
            //    using (Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-033.js"))
            //    {
            //        var certificate = new X509Certificate2();
            //        EasyUAClient.EngineParameters.SoftwareCertificates.Add(certificate);
            //        try
            //        {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //            EasyUAClient client = MakeClient();
            //            // Do any operation that requires a UA session, in order to cause the session be created.
            //            client.Read(new UAReadArguments(EndpointDescriptor, UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //            // ReSharper disable PossibleNullReferenceException
            //            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
            //                (args.EventId == UAEngineEventIds.ClientSoftwareCertificateRejectedByServer));
            //            // ReSharper restore PossibleNullReferenceException
            //            Assert.IsNotNull(logEntryEventArgs);
            //            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            //            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            //        }
            //        finally 
            //        {
            //            EasyUAClient.EngineParameters.SoftwareCertificates.Remove(certificate);
            //        }
            //    }
            //}

            /// <description>ServiceResult = Bad_SessionIdInvalid.
            ///</description>
            /// <expectedresult>Client does not crash, handles the situation gracefully, and then logs and/or displays this 
            /// information to the end-user.</expectedresult>
            [TestCategory("Keyword:CloseSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("TimeDomain")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_036()
            {
                TestCase_Err_036(false);
            }

            private void TestCase_Err_036(bool inverted)
            {
                var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
                EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

                // Sets the service result code to "Bad_SessionIdInvalid" when calling CloseSession.
                using (
                    inverted
                        ? null
                        : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-036.js"))
                {
                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.HoldPeriod = 0;
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));

                    int millisecondsTimeout =
                        EasyUAClient.SharedParameters.EngineParameters.GarbageCollectionPeriod +
                        client.IsolatedParameters.SessionParameters.HoldPeriod +
                        client.IsolatedParameters.SessionParameters.CloseTimeout +
                        5*1000;
                    AssertUtilities.WriteLine(millisecondsTimeout);
                    AssertUtilities.WaitingPhase(millisecondsTimeout); // allow time for session be closed
                }

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId ==
                                                                                  UAEngineEventIds.SessionCloseFailure));
                // ReSharper restore PossibleNullReferenceException

                if (inverted)
                    Assert.IsNull(logEntryEventArgs);
                else
                {
                    Assert.IsNotNull(logEntryEventArgs);
                    AssertUtilities.WriteLine(logEntryEventArgs.Message);
                    Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);
                    string message = logEntryEventArgs.Message;
                    StringAssert.Contains(message, "{BadSessionIdInvalid}");
                }
            }

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_001([NotNull] string scriptName, int codeBitsExpected)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    try
                    {
                        EasyUAClient client = MakeClient();
                        // Do any operation that requires a UA session, in order to cause the session be created.
                        client.Read(new UAReadArguments(EndpointDescriptor,
                                                        UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException.Message);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        AssertUtilities.WriteLine(innerException.Message);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);
                        Assert.AreEqual(codeBitsExpected, serviceException.ServiceResult.StatusCode.CodeBits);

                        // IMPROVE: verify that the session has been truly closed? Is the closing even proper?
                        return;
                    }
                    Assert.Fail();
                }
            }

            // ReSharper disable InconsistentNaming
            private void TestCase_Err_029([NotNull] string scriptName, int codeBitsExpected)
            // ReSharper restore InconsistentNaming
            {
                using (Ctt.RunScript(CttProjectName, scriptName))
                using (var recorder = new UAOperationRecorder())
                {
                    try
                    {
                        EasyUAClient client = MakeClient();
                        // Do any operation that requires a UA session, in order to cause the session be created.
                        client.Read(new UAReadArguments(EndpointDescriptor,
                                                        UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException.Message);
                        Exception innerException = uaException.InnerException;
                        Assert.IsNotNull(innerException);
                        AssertUtilities.WriteLine(innerException.Message);
                        var serviceException = innerException as UAServiceException;
                        Assert.IsNotNull(serviceException);
                        Assert.AreEqual(codeBitsExpected, serviceException.ServiceResult.StatusCode.CodeBits);

                        // Verify that the session has been truly closed
                        int closeSessionCount = recorder.CloseSessionRecordList.Count;
                        Assert.AreEqual(1, closeSessionCount);

                        // IMPROVE: verify the reconnection mechanism as stated by the test case
                        return;
                    }
                    Assert.Fail();
                }
            }

            #endregion
        }
    }
}
