﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Xml;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;
using UserIdentity = OpcLabs.BaseLib.IdentityModel.User.UserIdentity;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SessionServices
{
    // ReSharper disable InconsistentNaming
    public class BaseTest_NoCtt : BaseTest
    // ReSharper restore InconsistentNaming
    {
        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp : BaseTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public BaseTest_NoCtt_OpcFoundation_SampleServer_OpcTcp()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp);
            }



            #region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
            /// <description>Client can vary the userIdentityToken for login credentials:using: UserNameIdentityToken.
            /// Credentials must be encrypted.
            /// </description>
            /// <expectedresult>Verify the users/specified login credentials are used.
            /// The credentials are encrypted.
            /// Client should be prepared for the server to reject the credentials.</expectedresult>
            [TestCategory("Keyword:ActivateSession")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:Lab")]
            [TestMethod]
            public void TestCase_026()
            {
                TestCase_025_To_028(UserIdentity.CreateUserNameIdentity("Alice", "1234"), UserTokenType.UserName, null);
            }
#endif

            // ReSharper restore InconsistentNaming
#endregion
        }



        [TestClass]
        // ReSharper disable InconsistentNaming
        public class BaseTest_NoCtt_Softing_DemoServer_Http : BaseTest_NoCtt
        // ReSharper restore InconsistentNaming
        {
            public BaseTest_NoCtt_Softing_DemoServer_Http()
            {
                EndpointDescriptorDirect = LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http;
            }

            [ClassInitialize]
            static public void SocketPrecheck(TestContext context)
            {
                LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.Http);
            }
        }



#region Test Methods
// ReSharper disable InconsistentNaming

#if (NETFRAMEWORK)
        /// <description>Monitor service call and analyze passed parameter for correctness and adherence to the specifications.
        ///</description>
        /// <expectedresult>Check that the Client specifies a Session Name that uniquely identifies its Instance.
        /// </expectedresult>
        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_001()
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client1 = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int sessionOpenCount = recorder.SessionOpenRecordList.Count;
                Assert.AreEqual(1, sessionOpenCount);

                SessionOpenRecord sessionOpenRecord = recorder.SessionOpenRecordList[0];
                Debug.Assert(sessionOpenRecord != null);
                string sessionName = sessionOpenRecord.SessionName;
                Assert.IsNotNull(sessionName);
                Assert.IsTrue(sessionOpenRecord.SessionTimeout > 0);

                AssertUtilities.WriteLine(sessionName);
                StringAssert.Contains(sessionName, "EasyUASession;");
            }
        }
#endif

#if (NETFRAMEWORK)
        /// <description>Client opens multiple/concurrent sessions, i.e. 2 or more. Each session must be individually 
        /// controllable within the same UA Server.
        /// </description>
        /// <expectedresult>Each session must be equally maintainable.
        /// This can be tested by trying to Read, Write and/or Browse using each session.
        /// </expectedresult>
        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_002()
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client1 = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                EasyUAClient client2 = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client2.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int sessionOpenCount = recorder.SessionOpenRecordList.Count;
                Assert.AreEqual(2, sessionOpenCount);
                SessionOpenRecord sessionOpenRecord1 = recorder.SessionOpenRecordList[0];
                Debug.Assert(sessionOpenRecord1 != null);
                SessionOpenRecord sessionOpenRecord2 = recorder.SessionOpenRecordList[1];
                Debug.Assert(sessionOpenRecord2 != null);

                int readCount = recorder.ReadRecordList.Count;
                Assert.AreEqual(2, readCount);
                ReadRecord readRecord1 = recorder.ReadRecordList[0];
                Debug.Assert(readRecord1 != null);
                ReadRecord readRecord2 = recorder.ReadRecordList[1];
                Debug.Assert(readRecord2 != null);

                Assert.AreSame(readRecord1.SessionClient, sessionOpenRecord1.Session);
                Assert.AreSame(readRecord2.SessionClient, sessionOpenRecord2.Session);
            }
        }
#endif

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_005_Inverted()
        {
            TestCase_005(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_006_Inverted()
        {
            TestCase_006(true);
        }

#if (NETFRAMEWORK)
        /// <description>a.Client specifies a name for the session.
        ///</description>
        /// <expectedresult>The name should be unique for that Client. If the Client supports multiple Sessions to the same 
        /// Server then each session would have a unique name.
        /// </expectedresult>
        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_011()
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client1 = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                EasyUAClient client2 = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client2.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int sessionOpenCount = recorder.SessionOpenRecordList.Count;
                Assert.AreEqual(2, sessionOpenCount);

                SessionOpenRecord sessionOpenRecord1 = recorder.SessionOpenRecordList[0];
                Debug.Assert(sessionOpenRecord1 != null);
                string sessionName1 = sessionOpenRecord1.SessionName;
                Assert.IsNotNull(sessionName1);

                SessionOpenRecord sessionOpenRecord2 = recorder.SessionOpenRecordList[1];
                Debug.Assert(sessionOpenRecord2 != null);
                string sessionName2 = sessionOpenRecord2.SessionName;
                Assert.IsNotNull(sessionName2);

                AssertUtilities.WriteLine(sessionName1);
                AssertUtilities.WriteLine(sessionName2);
                Assert.AreNotEqual(sessionName1, sessionName2);
            }
        }
#endif

#if (NETFRAMEWORK)
        /// <description>Client can specify a primary LocaleId such as `EN-US` or `DE-DE`.
        /// </description>
        /// <expectedresult>The client should specify a primary locale by minimum.</expectedresult>
        [TestCategory("Keyword:ActivateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_023_alpha()
        {
            TestCase_023("EN-US");
        }

        [TestCategory("Keyword:ActivateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_023_beta()
        {
            TestCase_023("DE-DE");
        }
#endif

#if (NETFRAMEWORK)
        /// <description>Client can vary the userIdentityToken for login credentials:using: AnonymousIdentityToken.
        /// </description>
        /// <expectedresult>Verify the connection is anonymous.
        /// Client should be prepared for the server to reject the credentials.</expectedresult>
        [TestCategory("Keyword:ActivateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_025()
        {
            TestCase_025_To_028(UserIdentity.CreateAnonymousIdentity(), UserTokenType.Anonymous, null);
        }

        ///// <description>Client can vary the userIdentityToken for login credentials:using: IssuedIdentityToken
        ///// </description>
        ///// <expectedresult>Client should be prepared for the server to reject the credentials.
        ///// The credentials are encrypted.</expectedresult>
        //[TestCategory("Keyword:ActivateSession")]
        //[TestCategory("Live")]
        //[TestCategory("OpcCompliance")]
        //[TestCategory("Type:Lab")]
        //[TestMethod]
        //public void TestCase_028()
        //{
        //    Assert.Inconclusive("Need to find a server that supports this.");
        //    TestCase_025_To_028(UAUserIdentity.CreateKerberosIdentity(new NetworkCredential2("Alice", "1234", "OPC")), 
        //        UserTokenType.IssuedToken, null);
        //}
#endif

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_002_Inverted()
        {
            TestCase_Err_002(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_003_Inverted()
        {
            TestCase_Err_003(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_006_Inverted()
        {
            TestCase_Err_006(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_007_Inverted()
        {
            TestCase_Err_007(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_009_Inverted()
        {
            TestCase_Err_009(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_011_Inverted()
        {
            TestCase_Err_011(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_020_Inverted()
        {
            TestCase_Err_020(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_021_Inverted()
        {
            TestCase_Err_021(true);
        }

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_022_Inverted()
        {
            TestCase_Err_022(true);
        }

        // Cannot test until our test servers do it right
        //[TestCategory("Keyword:CreateSession")]
        //[TestCategory("Live")]
        //[TestCategory("OpcCompliance")]
        //[TestMethod]
        //public void TestCase_Err_023_Inverted()
        //{
        //    TestCase_Err_023(true);
        //}

        // Cannot test until our test servers do it right
        //[TestCategory("Keyword:CreateSession")]
        //[TestCategory("Live")]
        //[TestCategory("OpcCompliance")]
        //[TestMethod]
        //public void TestCase_Err_024_Inverted()
        //{
        //    TestCase_Err_024(true);
        //}

        // Cannot test until our test servers do it right
        //[TestCategory("Keyword:CreateSession")]
        //[TestCategory("Live")]
        //[TestCategory("OpcCompliance")]
        //[TestMethod]
        //public void TestCase_Err_025_Inverted()
        //{
        //    TestCase_Err_025(true);
        //}

        [TestCategory("Keyword:CreateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_027_Inverted()
        {
            TestCase_Err_027(true);
        }

        [TestCategory("Keyword:ActivateSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestMethod]
        public void TestCase_Err_032_Inverted()
        {
            TestCase_Err_032(true);
        }

#if (NETFRAMEWORK)
        /// <description>With one session open.
        /// Monitor service call (CloseSession) and analyze passed parameter (defaults) for correctness and adherence to the 
        /// specifications.
        ///</description>
        /// <expectedresult>ServiceResult = Good.</expectedresult>
        [TestCategory("Keyword:CloseSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("TimeDomain")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_Err_037()
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.HoldPeriod = 0;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int millisecondsTimeout =
                    EasyUAClient.SharedParameters.EngineParameters.GarbageCollectionPeriod +
                    client.IsolatedParameters.SessionParameters.HoldPeriod +
                    client.IsolatedParameters.SessionParameters.CloseTimeout +
                    5 * 1000;
                AssertUtilities.WaitingPhase(millisecondsTimeout); // allow time for session be closed

                int closeSessionCount = recorder.CloseSessionRecordList.Count;
                Assert.AreEqual(1, closeSessionCount);
                CloseSessionRecord closeSessionRecord = recorder.CloseSessionRecordList[0];
                Debug.Assert(closeSessionRecord != null);

                // IMPROVE: verify RequestHeader
                Assert.IsTrue(closeSessionRecord.DeleteSubscriptions);
            }
        }
#endif

#if (NETFRAMEWORK)
        /// <description>With multiple sessions open, close all of them.
        ///</description>
        /// <expectedresult>ServiceResult = Good.
        /// Multiple calls are made, one per CloseSession request.
        /// </expectedresult>
        [TestCategory("Keyword:CloseSession")]
        [TestCategory("Live")]
        [TestCategory("OpcCompliance")]
        [TestCategory("TimeDomain")]
        [TestCategory("Type:Lab")]
        [TestMethod]
        public void TestCase_Err_041()
        {
            const int sessionCount = 5;
            using (var recorder = new UAOperationRecorder())
            {
                const int holdPeriod = 0;
                int closeTimeout = MakeClient().IsolatedParameters.SessionParameters.CloseTimeout;
                for (int i = 0; i < sessionCount; i++)
                {
                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.HoldPeriod = holdPeriod;
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }

                int millisecondsTimeout =
                    EasyUAClient.SharedParameters.EngineParameters.GarbageCollectionPeriod +
                    holdPeriod +
                    closeTimeout +
                    5 * 1000;
                AssertUtilities.WaitingPhase(millisecondsTimeout); // allow time for session be closed

                int closeSessionCount = recorder.CloseSessionRecordList.Count;
                Assert.AreEqual(sessionCount, closeSessionCount);
            }
        }
#endif

// ReSharper restore InconsistentNaming
#endregion


#region Private Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_023([NotNull] string localeId)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.LocaleId = localeId;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int sessionOpenCount = recorder.SessionOpenRecordList.Count;
                Assert.AreEqual(1, sessionOpenCount);
                SessionOpenRecord sessionOpenRecord = recorder.SessionOpenRecordList[0];
                Debug.Assert(sessionOpenRecord != null);

                IList<string> preferredLocales = sessionOpenRecord.PreferredLocales;
                Assert.IsNotNull(preferredLocales);
                int preferredLocalesCount = preferredLocales.Count;
                Assert.AreEqual(1, preferredLocalesCount);
                string preferredLocale = preferredLocales[0];
                Assert.IsNotNull(preferredLocale);
                Assert.AreEqual(localeId, preferredLocale);
            }
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        private void TestCase_025_To_028([NotNull] UserIdentity userIdentity, UserTokenType userTokenType,
            [CanBeNull] XmlQualifiedName xmlQualifiedName)
        // ReSharper restore InconsistentNaming
        {
            using (var recorder = new UAOperationRecorder())
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.UserIdentity = userIdentity;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                int sessionOpenCount = recorder.SessionOpenRecordList.Count;
                Assert.AreEqual(1, sessionOpenCount);
                SessionOpenRecord sessionOpenRecord = recorder.SessionOpenRecordList[0];
                Debug.Assert(sessionOpenRecord != null);

                IUserIdentity sdkUserIdentity = sessionOpenRecord.Identity;
                Assert.IsNotNull(sdkUserIdentity);
                string displayName = sdkUserIdentity.DisplayName;
                Assert.IsNotNull(displayName);
                AssertUtilities.WriteLine(displayName);
                Assert.AreEqual(userTokenType, sdkUserIdentity.TokenType);
                if (userTokenType == UserTokenType.IssuedToken)
                    Assert.AreEqual(xmlQualifiedName, sdkUserIdentity.IssuedTokenType);

                // IMPROVE: verify more
            }
        }
#endif

#endregion
    }
}
