﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SessionServices
{
    public partial class BaseTest : EndpointTestBase
    {
        #region Protected Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_005(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Modifies the revisedSessionTimeout to 60-seconds.
            using (inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\005.js"))
            {
                EasyUAClient client = MakeClient();
                if (!inverted)
                    client.IsolatedParameters.SessionParameters.SessionTimeout = 1000;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs requestedSessionTimeoutUnusuallyShort = logEntryEventArgsList.Find(args =>
                                                                                                 (args.EventId ==
                                                                                                  UAEngineEventIds
                                                                                                      .RequestedSessionTimeoutUnusuallyShort));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((requestedSessionTimeoutUnusuallyShort == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", requestedSessionTimeoutUnusuallyShort));

            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs revisedSessionTimeoutDifferent = logEntryEventArgsList.Find(args =>
                                                                                          (args.EventId ==
                                                                                           UAEngineEventIds
                                                                                               .RevisedSessionTimeoutDifferent));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((revisedSessionTimeoutDifferent == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", revisedSessionTimeoutDifferent));

            if (inverted)
            {
                Assert.IsNull(requestedSessionTimeoutUnusuallyShort);
                Assert.IsNull(revisedSessionTimeoutDifferent);
                return;
            }

            Assert.IsNotNull(requestedSessionTimeoutUnusuallyShort);
            AssertUtilities.WriteLine(requestedSessionTimeoutUnusuallyShort.Message);
            Assert.AreEqual(LogEntryType.Information, requestedSessionTimeoutUnusuallyShort.EntryType);

            Assert.IsNotNull(revisedSessionTimeoutDifferent);
            AssertUtilities.WriteLine(revisedSessionTimeoutDifferent.Message);
            Assert.AreEqual(LogEntryType.Information, revisedSessionTimeoutDifferent.EntryType);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_006(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Modifies the requestedSessionTimeout to 60-seconds, if the initial
            // requested timeout is long (more than 60-seconds in this case).
            using (inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\006.js"))
            {
                EasyUAClient client = MakeClient();
                if (!inverted)
                    client.IsolatedParameters.SessionParameters.SessionTimeout = 7 * 24 * 60 * 60 * 1000;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs requestedSessionTimeoutUnusuallyLong = logEntryEventArgsList.Find(args =>
                                                                                                (args.EventId ==
                                                                                                 UAEngineEventIds
                                                                                                     .RequestedSessionTimeoutUnusuallyLong));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((requestedSessionTimeoutUnusuallyLong == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", requestedSessionTimeoutUnusuallyLong));

            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs revisedSessionTimeoutDifferent = logEntryEventArgsList.Find(args =>
                                                                                          (args.EventId ==
                                                                                           UAEngineEventIds
                                                                                               .RevisedSessionTimeoutDifferent));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((revisedSessionTimeoutDifferent == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", revisedSessionTimeoutDifferent));

            if (inverted)
            {
                Assert.IsNull(requestedSessionTimeoutUnusuallyLong);
                Assert.IsNull(revisedSessionTimeoutDifferent);
                return;
            }

            Assert.IsNotNull(requestedSessionTimeoutUnusuallyLong);
            AssertUtilities.WriteLine(requestedSessionTimeoutUnusuallyLong.Message);
            Assert.AreEqual(LogEntryType.Information, requestedSessionTimeoutUnusuallyLong.EntryType);

            Assert.IsNotNull(revisedSessionTimeoutDifferent);
            AssertUtilities.WriteLine(revisedSessionTimeoutDifferent.Message);
            Assert.AreEqual(LogEntryType.Information, revisedSessionTimeoutDifferent.EntryType);
        }
#endif

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_002(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // sessionId returned is the same as another already-open session.
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-002.js"))
            {
                EasyUAClient client1 = MakeClient();
                client1.IsolatedParameters.SessionParameters.HoldPeriod = UASmartSessionParameters.DefaultHoldPeriod;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                 UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));

                UAException exception = null;
                try
                {
                    EasyUAClient client2 = MakeClient();
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client2.Read(new UAReadArguments(EndpointDescriptor,
                                                     UATestUtilities.MakeNodeDescriptor(typeof(Int32),
                                                                                        dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId ==
                                                                                  UAEngineEventIds
                                                                                      .SessionIdDuplicate));
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                    String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

                if (inverted)
                {
                    Assert.IsNull(exception);
                    Assert.IsNull(logEntryEventArgs);
                }
                else
                {
                    Assert.IsNotNull(exception);
                    AssertUtilities.WriteLine(exception.Message);
                    Exception innerException = exception.InnerException;
                    Assert.IsNotNull(innerException);
                    AssertUtilities.WriteLine(innerException.Message);
                    var clientEngineException = innerException as UAClientEngineException;
                    Assert.IsNotNull(clientEngineException);
                    Assert.AreEqual(UAClientErrorCodes.SessionIdDuplicate, clientEngineException.ErrorCode);

                    // IMPROVE: verify that the session has been truly closed

                    Assert.IsNotNull(logEntryEventArgs);
                    AssertUtilities.WriteLine(logEntryEventArgs.Message);
                    Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                }

                // The call below uses the original session and must not throw.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                 UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }
        }
#endif

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_003(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            const bool workAroundCttScriptBug = false;
            // sessionId returned is empty
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-003.js"))
            {
                var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
                EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

                // ReSharper disable RedundantAssignment
                EasyUAClient client1 = MakeClient();
                // ReSharper restore RedundantAssignment
                // ReSharper disable ConditionIsAlwaysTrueOrFalse
                if (!workAroundCttScriptBug)
                // ReSharper restore ConditionIsAlwaysTrueOrFalse
#pragma warning disable 162
                // ReSharper disable HeuristicUnreachableCode
                {
                    AssertUtilities.Phase("First reading");
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client1.Read(new UAReadArguments(EndpointDescriptor,
                                                     UATestUtilities.MakeNodeDescriptor(typeof(Int32),
                                                                                        dynamic: true)));
                }
                // ReSharper restore HeuristicUnreachableCode
#pragma warning restore 162

                UAException exception = null;
                try
                {
                    EasyUAClient client2 = MakeClient();
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    AssertUtilities.Phase("Second reading");
                    client2.Read(new UAReadArguments(EndpointDescriptor,
                                                     UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId == UAEngineEventIds.SessionIdNull));
                // ReSharper restore PossibleNullReferenceException

                if (inverted)
                {
                    Assert.IsNull(exception);
                    Assert.IsNull(logEntryEventArgs);
                }
                else
                {
                    Assert.IsNotNull(exception);
                    AssertUtilities.WriteLine(exception.Message);
                    Exception innerException = exception.InnerException;
                    Assert.IsNotNull(innerException);
                    AssertUtilities.WriteLine(innerException.Message);
                    var clientEngineException = innerException as UAClientEngineException;
                    Assert.IsNotNull(clientEngineException);
                    Assert.AreEqual(UAClientErrorCodes.SessionIdNull, clientEngineException.ErrorCode);

                    // IMPROVE: verify that the session has been truly closed

                    Assert.IsNotNull(logEntryEventArgs);
                    AssertUtilities.WriteLine(logEntryEventArgs.Message);
                    Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                }

                // The part below is commented out because the current CTT script does not support it
                //                        // ReSharper disable ConditionIsAlwaysTrueOrFalse
                //                        if (!workAroundCttScriptBug)
                //                        // ReSharper restore ConditionIsAlwaysTrueOrFalse
                //#pragma warning disable 162
                //// ReSharper disable HeuristicUnreachableCode
                //                        {
                //                            // The call below uses the original session and must not throw.
                //                            client1.Read(new UAReadArguments(EndpointDescriptor,
                //                                UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));
                //                        }
                //// ReSharper restore HeuristicUnreachableCode
                //#pragma warning restore 162
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_006(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // sessionId returned is the same as another already-open session.
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-006.js"))
            {
                EasyUAClient client1 = MakeClient();
                client1.IsolatedParameters.SessionParameters.HoldPeriod = UASmartSessionParameters.DefaultHoldPeriod;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client1.Read(new UAReadArguments(EndpointDescriptor,
                                                 UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                AssertUtilities.WriteLine("First Read on client1 complete (no exception).");
                AssertUtilities.WriteSeparatorLine();

                UAException exception2 = null;
                try
                {
                    EasyUAClient client2 = MakeClient();
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client2.Read(new UAReadArguments(EndpointDescriptor,
                                                     UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException2)
                {
                    exception2 = uaException2;
                }
                AssertUtilities.WriteLine("Read on client2 complete.");
                AssertUtilities.WriteSeparatorLine();

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId ==
                                                                                  UAEngineEventIds.AuthenticationTokenDuplicate));
                // ReSharper restore PossibleNullReferenceException

                if (inverted)
                {
                    Assert.IsNull(exception2);
                    Assert.IsNull(logEntryEventArgs);
                }
                else
                {
                    Assert.IsNotNull(exception2);
                    AssertUtilities.WriteLine(exception2.Message);
                    Exception innerException = exception2.InnerException;
                    Assert.IsNotNull(innerException);
                    AssertUtilities.WriteLine(innerException.Message);
                    var clientEngineException = innerException as UAClientEngineException;
                    Assert.IsNotNull(clientEngineException);
                    Assert.AreEqual(UAClientErrorCodes.AuthenticationTokenDuplicate, clientEngineException.ErrorCode);

                    // IMPROVE: verify that the session has been truly closed

                    Assert.IsNotNull(logEntryEventArgs);
                    AssertUtilities.WriteLine(logEntryEventArgs.Message);
                    Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                }

                // The call below uses the original session and throws as well, because the sessions cannot be distinguished.
                AssertUtilities.WriteSeparatorLine();
                UAException exception1 = null;
                try
                {
                    client1.Read(new UAReadArguments(EndpointDescriptor,
                                                     UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException1)
                {
                    exception1 = uaException1;
                }
                AssertUtilities.WriteLine("Second Read on client1 complete.");
                AssertUtilities.WriteSeparatorLine();

                if (inverted)
                    Assert.IsNull(exception1);
                else
                {
                    Assert.IsNotNull(exception1);
                    AssertServiceException(exception1.InnerException, UACodeBits.BadSessionClosed);
                }
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_007(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Revises the session timeout by taking the clients request and * 10.
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-007.js"))
            {
                EasyUAClient client = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .RevisedSessionTimeoutUnusuallyExtended));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_009(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //  Revises the session timeout to an insanely long amount of time, i.e. 1-week.
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-009.js"))
            {
                EasyUAClient client = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .RevisedSessionTimeoutUnusuallyLong));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_011(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // ServerNonce is less than 32 bytes
            UAException exception = null;
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-011.js"))
            {
                EasyUAClient client = MakeClient();
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds.ServerNonceTooShort));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var clientEngineException = innerException as UAClientEngineException;
                Assert.IsNotNull(clientEngineException);
                Assert.AreEqual(UAClientErrorCodes.ServerNonceTooShort, clientEngineException.ErrorCode);

                // IMPROVE: verify that the session has been truly closed

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_020(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Clears all endpoints.
            UAException exception = null;
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-020.js"))
            {
                EasyUAClient client = MakeClient();
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ServerEndpointsDoNotMatchDiscovery));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var clientEngineException = innerException as UAClientEngineException;
                Assert.IsNotNull(clientEngineException);
                Assert.AreEqual(UAClientErrorCodes.ServerEndpointsDoNotMatchDiscovery, clientEngineException.ErrorCode);

                // IMPROVE: verify that the session has been truly closed

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_021(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Clears all endpoints.
            UAException exception = null;
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-021.js"))
            {
                EasyUAClient client = MakeClient();
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds.SessionConnectFailure));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var serviceException = innerException as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadSecurityChecksFailed, serviceException.ServiceResult.StatusCode.CodeBits);

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                string message = logEntryEventArgs.Message;
                StringAssert.Contains(message, "{BadSecurityChecksFailed}");
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_022(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Returns just 1 endpoint, which is empty.
            UAException exception = null;
            using (inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-022.js"))
            {
                EasyUAClient client = MakeClient();
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds.SessionConnectFailure));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var serviceException = innerException as UAServiceException;
                Assert.IsNotNull(serviceException);
                Assert.AreEqual(UACodeBits.BadSecurityChecksFailed, serviceException.ServiceResult.StatusCode.CodeBits);

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                string message = logEntryEventArgs.Message;
                StringAssert.Contains(message, "{BadSecurityChecksFailed}");
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_023(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // productUris contained in the returned SoftwareCertificates do not match the
            // productUris in the EndpointDescription used by the Client to connect to the
            // Server.
            UAException exception = null;
            using (inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-023.js"))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.RequireMatchingServerSoftwareCertificates = true; // not a default!
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ServerSoftwareCertificateProductUriNotFound));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var clientEngineException = innerException as UAClientEngineException;
                Assert.IsNotNull(clientEngineException);
                Assert.AreEqual(UAClientErrorCodes.ServerSoftwareCertificateProductUriNotFound,
                                clientEngineException.ErrorCode);

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_024(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // serverSoftwareCertificates Empty.
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-024.js"))
            {
                EasyUAClient client = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds.ServerSoftwareCertificatesEmpty));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_025(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //  Clears the serverSignature parameter when calling CreateSessionHelper.
            UAException exception = null;
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-025.js"))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.RequireNonEmptyServerSignature = true;
                try
                {
                    // Do any operation that requires a UA session, in order to cause the session be created.
                    client.Read(new UAReadArguments(EndpointDescriptor,
                                                    UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
                }
                catch (UAException uaException)
                {
                    exception = uaException;
                }
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ServerSignatureEmpty));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(exception);
                Assert.IsNull(logEntryEventArgs);
            }
            else
            {
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception.Message);
                Exception innerException = exception.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.Message);
                var clientEngineException = innerException as UAClientEngineException;
                Assert.IsNotNull(clientEngineException);
                Assert.AreEqual(UAClientErrorCodes.ServerSignatureEmpty, clientEngineException.ErrorCode);

                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_027(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // Modify the max response message size by taking the client request * 10, or if the client did not
            // specify a max, then specify one for it!
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-027.js"))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.SessionTimeout = 1000;
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .MaxRequestMessageSizeUnusuallyLargeRelative));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
                Assert.IsNull(logEntryEventArgs);
            else
            {
                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_032(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            // results[0] is Bad_CertificateUriInvalid.
            using (
                inverted ? null : Ctt.RunScript(CttProjectName, @"Session Services\Session Client Base\Test Cases\Err-032.js"))
            {
                EasyUAClient client = MakeClient();
                // Do any operation that requires a UA session, in order to cause the session be created.
                client.Read(new UAReadArguments(EndpointDescriptor,
                                                UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true)));
            }

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .ClientSoftwareCertificateRejectedByServer));
            // ReSharper restore PossibleNullReferenceException

            if (inverted)
                Assert.IsNull(logEntryEventArgs);
            else
            {
                Assert.IsNotNull(logEntryEventArgs);
                AssertUtilities.WriteLine(logEntryEventArgs.Message);
                Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
                StringAssert.Contains(logEntryEventArgs.Message, "CertificateUriInvalid");
            }
        }

#endregion
    }
}
