﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SubscriptionServices
{
    public abstract partial class BasicTest
    {
        // ReSharper disable InconsistentNaming
        public abstract class BasicTestCtt : BasicTest
        // ReSharper restore InconsistentNaming
        {
            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class BasicTest_Ctt_OpcFoundationSampleServer_OpcTcp : BasicTestCtt
            // ReSharper restore InconsistentNaming
            {
                public BasicTest_Ctt_OpcFoundationSampleServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
                }
            }



            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class BasicTest_Ctt_SoftingDemoServer_OpcTcp : BasicTestCtt
            // ReSharper restore InconsistentNaming
            {
                public BasicTest_Ctt_SoftingDemoServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
                }

                [ClassInitialize]
                static public void SocketPrecheck(TestContext context)
                {
                    LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
                }
            }




            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>Client creates two subscriptions where the 
            /// subscriptionId returned for the 2nd create request has the same id used previously.
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays an error to the end-user. Client cannot use the subscription because it conflicts 
            /// with the existing subscription of the same id.
            /// Client prompts the end-user to disconnect from the Server and deletes its subscriptions. If the end-user 
            /// rejects the disconnect request then the client warns of the potential danger of using the subscription in its conflicted state.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_002()
            {
                TestCase_Err_002(false);
            }

            /// <description>Inject the following RevisedPublishInterval error codes:
            /// 1.revisedPublishingInterval returned decades larger than the requestedPublishingInterval, e.g. requested*10.
            /// 2. revisedPublishingInterval returned decades smaller than the requestedPublishingInterval, e.g. requested / 10 
            /// (or a reasonable difference). 
            /// </description>
            /// <expectedresult>
            /// a.Client checks the revised value and if it is different to the requested then it logs a message.
            /// b.Client logs a message that the requested and revised value differs. 
            /// </expectedresult>
            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_01()
            {
                TestCase_003_01(false);
            }

            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_02()
            {
                TestCase_003_02(false);
            }

            /// <description>Inject the following RevisedLifeTimeCount error codes:
            /// 1. revisedLifetimeCount returned decades larger than the requestedLifetimeCount, e.g. requested*10.
            /// 2. revisedLifetimeCount returned decades smaller than requestedLifetimeCount, e.g. requested / 10 (or 
            /// a reasonable difference).
            /// 3. revisedLifetimeCount is less than 3 times the revisedMaxKeepAliveCount. 
            /// </description>
            /// <expectedresult>
            /// a.Client checks the revised value and if it is different to the requested then it logs a message.
            /// b.Client logs a message that the requested and revised value differs. 
            /// c.Client logs and/or displays a message to the end-user.
            /// Subscription is removed because it is not valid, i.e. revisedLifetimeCounter must be 3 times that of 
            /// revisedKeepAliveCount.
            /// </expectedresult>
            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_01()
            {
                TestCase_004_01(false);
            }

            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_02()
            {
                TestCase_Err_004_02(false);
            }

            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_03()
            {
                TestCase_Err_004_03(false);
            }

            /// <description>Inject the following RevisedMaxKeepAliveCount error codes:
            /// 1. revisedMaxKeepAliveCount returned decades larger than the requestedMaxKeepAliveCount*10.
            /// 2. revisedMaxKeepAliveCount returned decades smaller than requestedMaxKeepAliveCount, e.g. requested / 10 
            /// (or a reasonable difference). 
            /// </description>
            /// <expectedresult>
            /// a.Client checks the revised value and if it is different to the requested then it logs a message.
            /// b.Client logs a message that the requested and revised value differs. 
            /// </expectedresult>
            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_01()
            {
                TestCase_Err_005_01(false);
            }

            [TestCategory("Keyword:CreateSubscription")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_005_02()
            {
                TestCase_Err_005_02(false);
            }

            // LATER: Found that BadNoSubscription also appears legitimately, not closing for this reason now.
            //From: Nathan Pocock [mailto:nathan.pocock@opcfoundation.org] 
            //Sent: Monday, June 10, 2013 9:31 PM
            //...
            //I can see where this is confusing. I would agree with your points. I think it is safe to say that such behavior 
            //would indicate a really bad Server (or Client). I think in both cases it would be safe to restart the session, 
            //given the behavior you’ve described.
            //...
            //From: Zbynek Zahradnik [mailto:zbynekz@opclabs.com] 
            //Sent: Monday, June 10, 2013 11:35 AM
            //To: deepthi.teegala@opcfoundation.org; nathan.pocock@opcfoundation.org
            //Subject: Issues with Err-008 in Subscription Services\Subscription Client Basic\Test Cases\Err-008-01/02.js
            //...
            //Alpha) To “Client establishes a new Subscription, relinquishing any local resources”: Well but this failed 
            //Publish is not related to any particular subscription, is it? I do not quite get what new subscription should be 
            // established then.
            //If Bad_NoSubscription is meant to indicate that Publish was called “unnecessarily”, while no subscriptions 
            // existed, then unless the client is buggy, the server must be. It is difficult to say what the right reaction 
            // should be in such case, but I would handle it as a (presumably) server failure on the session level, close it 
            // and reestablish a new Session (not a subscription). Any objections?
            //Beta) “Client does not issue any more Publish requests until a Publish response is received (async API).” If we 
            // have max one Publish pending, stopping the Publishes and waiting for Publish response would mean that the 
            // session will be in bad state from which it will recover only by accident (when closed for some other reason). 
            // So, besides stopping new Publishes, if we want a resilient client, our reaction should be to close and 
            // reestablish the Session as well, shouldn’t it?
            /// <description>
            /// Inject the following ServiceResult error codes:
            /// 1. Bad_TooManyPublishRequests
            /// 2. Bad_NoSubscription 
            /// </description>
            /// <expectedresult>
            /// a.Client handles the error, logs and/or displays an error to the end-user.
            /// Client does not issue any more Publish requests until a Publish response is received (async API).
            /// b.Client handles the error, logs and/or displays an error to the end-user.
            /// Client establishes a new Subscription, relinquishing any local resources.
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_008_01()
            {
                TestCase_Err_008(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-008-01.js",
                    UAEngineEventIds.SessionPublishError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadTooManyPublishRequests, 0));
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_008_02()
            {
                TestCase_Err_008(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-008-02.js",
                    UAEngineEventIds.SessionPublishError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadNoSubscription, 0));
            }

            /// <description>
            /// ServiceResult = Good, but the SubscriptionId is unknown to the Client.
            /// </description>
            /// <expectedresult>
            /// Client handles the error, logs and/or displays an error to the end-user.
            /// Client discards the entire message.
            /// Optionally, the Client may choose to call DeleteSubscription passing in this unknown SubscriptionId.
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_009()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-009.js",
                    UAEngineEventIds.PublishResponseSubscriptionIdUnknown,
                    LogEntryType.Warning);
            }

            /// <description>
            /// Inject the following error codes
            /// 1.Inject previously acknowledged SequenceNumbers into this parameter upon every Publish response.
            /// 2.Inject some previously acknowledged SequenceNumbers into this parameter where valid sequenceNumbers are 
            /// already present (sent from the Server).
            /// 3.Inject invalid (most recent SequenceNumber+1) SequenceNumbers into this parameter.
            /// 4.Set AvailableSequenceNumbers to NULL/empty although there are NotificationMessages to acknowledge. 
            /// </description>
            /// <expectedresult>
            /// a,b.Client reports an error that previously acknowledged SequenceNumbers are showing as unacknowledged. 
            /// Client re-acknowledges them (see clause 5.1.5 Client/Server Negotiations on page 17).
            /// c.Client reports an error that it’s receiving bogus SequenceNumbers.
            /// Client does NOT acknowledge the SequenceNumber.
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010_01()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-010-01.js",
                    UAEngineEventIds.AvailableSequenceNumberAlreadyAcknowledged,
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010_02_2_MonitoredItems()
            {
                TestCase_Err_Subscription_Volume(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-010-02.js",
                    2,
                    20*1000,
                    UAEngineEventIds.AvailableSequenceNumberAlreadyAcknowledged,
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010_02_100_MonitoredItems()
            {
                TestCase_Err_Subscription_Volume(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-010-02.js",
                    100,
                    20 * 1000,
                    UAEngineEventIds.AvailableSequenceNumberAlreadyAcknowledged,
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010_03()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-010-03.js",
                    UAEngineEventIds.AvailableSequenceNumberUnknown,
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_010_04()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-010-04.js",
                    UAEngineEventIds.AvailableSequenceNumbersEmpty,
                    LogEntryType.Warning);
            }

            /// <description>
            /// Inject the following error codes:
            /// 1.NotificationMessage is empty. 
            /// 2.Re-use a SequenceNumber within a new DataChange notification.
            /// </description>
            /// <expectedresult>
            /// a.Client reports an error that it’s receiving invalid Publish responses.
            /// b.Client dismisses the event.
            /// Client reports an error that a previously acknowledged SequenceNumber has been received. Client 
            /// re-acknowledges it (see clause 5.1.5 Client/Server Negotiations on page 17).
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011_01()
            {
                TestCase_Err_011(
                    false, 
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-011-01.js", 
                    UAEngineEventIds.NotificationMessageEmpty, 
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_011_02()
            {
                Ctt.ClaimScriptBuggyVersion(0x19187fea);
                TestCase_Err_011(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-011-02.js",
                    UAEngineEventIds.SequenceNumberReused,
                    LogEntryType.Warning);
            }

            /// <description>
            /// Inject the following error codes:
            /// 1.Set the PublishTime (in the NotificationMessage) to a time in the future, e.g. now + 25 hours.
            /// 2.Set the PublishTime (in the NotificationMessage) to a time in the past, e.g. now – 25 hours. 
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays a warning message stating that it is receiving events from the future.
            /// Client by default should discard the data, unless the end-user selects an option to consume the data.
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_01()
            {
                TestCase_Err_012(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-012-01.js",
                    UAEngineEventIds.PublishTimeAhead,
                    LogEntryType.Information);
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_012_02()
            {
                TestCase_Err_012(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-012-02.js",
                    UAEngineEventIds.PublishTimeStale,
                    LogEntryType.Information);
            }

            /// <description>
            /// Inject one of the following Operation Level Results into the first and all returned item.
            /// 1.ServiceResult = Good.
            /// Results[0] = Bad_SubscriptionIdInvalid
            /// 2.Results[0] = Bad_SequenceNumberUnknown
            /// 3.ServiceResult = Good.
            /// Results[0] = Bad_ Bad_InvalidSequenceNumber
            /// 4.ServiceResult = Good.
            /// Results[0] = Bad_SubscriptionIdInvalid
            /// Results[1] = Bad_ SequenceNumberUnknown
            /// 5. Results[*] = Bad_SubscriptionIdInvalid 
            /// </description>
            /// <expectedresult>
            /// Step Number
            /// Action
            /// Expected Result(s)
            /// 1 
            /// TestCase 1:
            /// Client logs and/or displays an error to the end-user specifying the subscription/sequenceNumber(s) that were 
            /// not successfully acknowledged. 
            /// 2 
            /// TestCase 2:
            /// Client handles the error, logs and/or displays an error to the end-user. The Client may choose to handle this 
            /// in a number of ways, such as establishes a new Subscription, or disconnecting etc. 
            /// 3 
            /// TestCase 3:
            /// Client reports an error to the end-user.
            /// 4 
            /// TestCase 4:
            /// For the result in error Bad_InvalidSequenceNumber the Client reports an error to the end-user. For the result 
            /// in error Bad_SubscriptionIdInvalid the Client may choose to handle this in a number of ways, such as 
            /// establishes new Subscriptions, or disconnecting etc.  
            /// </expectedresult>
            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_018_01()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-018-01.js",
                    UAEngineEventIds.NotificationAcknowledgeError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadSubscriptionIdInvalid, 0));
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_018_02()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-018-02.js",
                    UAEngineEventIds.NotificationAcknowledgeError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadSequenceNumberUnknown, 0));
            }

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_018_03()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-018-03.js",
                    UAEngineEventIds.NotificationAcknowledgeError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadSequenceNumberInvalid, 0));
            }

            // IMPROVE: make TestCase_Err_018_04 work (needs more to be acknowledged - how to enforce it?)
            //[TestCategory("Keyword:Publish")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_018_04()
            //{
            //    using (Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-018-04.js"))
            //    {
            //var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            //EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            //        var monitoredItemArgumentsArray = MakeMultipleMonitoredItemArguments(
            //            new[] { typeof(Single), typeof(Int32), typeof(Int16) }, dynamic: true);
            //        EasyUAClient client = MakeClient();
            //        client.IsolatedParameters.Session.StatusSubscriptionSamplingInterval = Timeout.Infinite;
            //        CollectMultipleMonitoredItemChangedEventArgs(
            //            client,
            //            monitoredItemArgumentsArray,
            //            MillisecondsToCollectSingleNotification);

            //        AssertUtilities.WriteSeparatorLine();

            //var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            //        // ReSharper disable PossibleNullReferenceException
            //        var statusCodeExpected1 = new UAStatusCode(UACodeBits.BadSubscriptionIdInvalid, 0);
            //        LogEntryEventArgs logEntryEventArgs1 = logEntryEventArgsList.Find(args =>
            //            (args.EventId == UAEngineEventIds.NotificationAcknowledgeError) &&
            //            (args.Message.Contains(statusCodeExpected1.ToString())));
            //        // ReSharper restore PossibleNullReferenceException
            //        Assert.IsTrue(logEntryEventArgs1 != null);
            //        AssertUtilities.WriteLine(logEntryEventArgs1.Message);

            //        // ReSharper disable PossibleNullReferenceException
            //        var statusCodeExpected2 = new UAStatusCode(UACodeBits.BadSequenceNumberUnknown, 0);
            //        LogEntryEventArgs logEntryEventArgs2 = logEntryEventArgsList.Find(args =>
            //            (args.EventId == UAEngineEventIds.NotificationAcknowledgeError) &&
            //            (args.Message.Contains(statusCodeExpected2.ToString())));
            //        // ReSharper restore PossibleNullReferenceException
            //        Assert.IsTrue(logEntryEventArgs2 != null);
            //        AssertUtilities.WriteLine(logEntryEventArgs2.Message);
            //    }
            //}

            [TestCategory("Keyword:Publish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_018_05()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-018-05.js",
                    UAEngineEventIds.NotificationAcknowledgeError,
                    LogEntryType.Warning,
                    /*((Ctt.ScriptCrc == 0x7c743fe3) || (Ctt.ScriptCrc == 0xaa4eacb2))
                        ?*/ new UAStatusCode(UACodeBits.BadSubscriptionIdInvalid, 0)/*
                        : new UAStatusCode(UACodeBits.BadSequenceNumberInvalid, 0)*/);
            }

            /// <description>
            /// Inject the following error codes
            /// 1.ServiceResult = Good, MonitoredItems parameter is empty but the SequenceNumber has incremented.
            /// </description>
            /// <expectedresult>
            /// Client logs and/or reports the error to the end-user.
            /// Client calls Republish to retrieve the missing SequenceNumber.
            /// </expectedresult>
            [TestCategory("Keyword:DataChangeNotification")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_020()
            {
                TestCase_Err_020(false);
            }

            ///// <description>
            ///// Inject the following error codes:
            ///// 1.Where ‘x’ is the MaxNotificationsPerPublish from CreateSubscription 
            ///// (or ModifySubscription) inject x + 1 notifications into the DataChange.
            ///// Note: MaxNotificationsPerPublish must be specified as a non-zero value for this test!
            ///// 2.Where ‘x’ is the MaxNotificationsPerPublish from CreateSubscription 
            ///// (or ModifySubscription) inject 10*x notifications into the DataChange.
            ///// 3.Where ‘x’ is the MaxNotificationsPerPublish from CreateSubscription 
            ///// (or ModifySubscription) inject 2*x notifications into the DataChange.
            ///// 1,2,3 
            ///// </description>
            ///// <expectedresult>
            ///// Client reports that more notifications were received than requested (server at fault!). If the Client can 
            ///// handle the response, then it processes the response normally. If the Client cannot handle the response, then 
            ///// the SecureChannel will be terminated and restarted. All resources will be cleanly relinquished before being 
            ///// recreated. 
            ///// </expectedresult>
            //[TestCategory("Keyword:Publish")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_019_01()
            //{
            //    TestCase_Err_Subscription(
            //        @"Subscription Services\Subscription Client Basic\Test Cases\Err-019-01.js",
            //        0/*fill in...*/,
            //        LogEntryType.Information);
            //}

            /// <description>Inject the following error codes:
            /// 1. ServiceResult = Good, the first items Timestamps are set to a date in the future, e.g. now + 1 day.
            /// 2. ServiceResult = Good, the first items StatusCode is modified to show BAD quality. 
            /// </description>
            /// <expectedresult>
            /// Step Number Action Expected Result(s) 
            /// 1 1. Client logs and/or reports the error to the end-user.Client changes the StatusCode to Bad for the 
            /// affected item(s).  
            /// 2 2. Client handles the bad data accordingly, and continues processing the other values. 
            /// </expectedresult>
            [TestCategory("Keyword:DataChangeNotification")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_021_01()
            {
                TestCase_Err_021_01(false);
            }

            /// <description>Inject the following error codes
            /// 1.ServiceResult = Good, all StatusCode values are made BAD.
            /// 2.ServiceResult = Bad, all StatusCode values are made GOOD. 
            /// </description>
            /// <expectedresult>
            /// Step Number Action Expected Result(s) 
            /// 1 1 Client handles the bad data accordingly, and continues processing the other values.  
            /// 2 2 Client discards the message. Client logs and/or reports the error to the end-user. 
            /// </expectedresult>
            [TestCategory("Keyword:DataChangeNotification")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_022_02()
            {
                TestCase_Err_022_02(false);
            }

            /// <description>ServiceResult = Good, MonitoredItems parameter contains items that were not
            /// requested or not valid, i.e. ClientHandle = ClientHandle*10. 
            /// </description>
            /// <expectedresult>
            /// Client logs and/or reports the error to the end-user.
            /// </expectedresult>
            [TestCategory("Keyword:DataChangeNotification")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_023()
            {
                TestCase_Err_023(false);
            }

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            #endregion
        }
    }
}
