﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SubscriptionServices
{
    public abstract partial class BasicTest : EndpointTestBase
    {
        #region Protected Methods

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_002(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            var monitoredItemChangedEventArgsList1 = new List<EasyUAMonitoredItemChangedEventArgs>();
            var monitoredItemChangedEventArgsList2 = new List<EasyUAMonitoredItemChangedEventArgs>();
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName,
                                    @"Subscription Services\Subscription Client Basic\Test Cases\Err-002.js"))
            {

                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;

                    client.SubscribeDataChange(
                        EndpointDescriptor,
                        UATestUtilities.MakeNodeDescriptor(typeof (Int32), dynamic: true),
                        3*1000,
                        (_, args) => monitoredItemChangedEventArgsList1.Add(args));

                    client.SubscribeDataChange(
                        EndpointDescriptor,
                        UATestUtilities.MakeNodeDescriptor(typeof (Int16), dynamic: true),
                        2*1000, // must be different from first one, to enforce a separate subscription
                        (_, args) => monitoredItemChangedEventArgsList2.Add(args));

                    AssertUtilities.WaitingPhase(10*1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            int count1 = monitoredItemChangedEventArgsList1.Count;
            AssertUtilities.WriteLine(count1);

            int count2 = monitoredItemChangedEventArgsList2.Count;
            AssertUtilities.WriteLine(count2);

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                args.EventId == UAEngineEventIds.SubscriptionIdDuplicate);
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            Assert.AreNotEqual(0, count1);
            Assert.AreNotEqual(0, count2);

            EasyUAMonitoredItemChangedEventArgs eventArgs1 = monitoredItemChangedEventArgsList1[0];
            Assert.IsNotNull(eventArgs1);
            AssertUtilities.WriteLine(eventArgs1);
            EasyUAMonitoredItemChangedEventArgs eventArgs2 = monitoredItemChangedEventArgsList2[0];
            Assert.IsNotNull(eventArgs2);
            AssertUtilities.WriteLine(eventArgs2);

            // The component gives no guarantee about the order in which it will call the CreateSubscription service.
            // We therefore cannot realy on monitoredItemChangedEventArgsList1 containing the successful results,
            // and monitoredItemChangedEventArgsList2 the failure. Need to recognize it dynamically, with some heuristics.

            EasyUAMonitoredItemChangedEventArgs failureEventArgs = (eventArgs1.Exception != null) ? eventArgs1 : eventArgs2;
            Exception exception = failureEventArgs.Exception;
            Assert.IsNotNull(exception);
            AssertUtilities.WriteLine(exception);
            var clientEngineException = exception as UAClientEngineException;
            Assert.IsNotNull(clientEngineException);
            Assert.AreEqual(UAClientErrorCodes.SubscriptionIdDuplicate, clientEngineException.ErrorCode);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_003_01(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-003-01.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedPublishingIntervalAboveToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_003_02(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-003-02.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedPublishingIntervalBelowToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_004_01(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-004-01.js"))
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedLifetimeCountAboveToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_004_02(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-004-02.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedLifetimeCountBelowToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_004_03(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            var monitoredItemChangedEventArgsList = new List<EasyUAMonitoredItemChangedEventArgs>();
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-004-03.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000,
                                                  (_, args) => monitoredItemChangedEventArgsList.Add(args));
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            int count = monitoredItemChangedEventArgsList.Count;
            AssertUtilities.WriteLine(count);

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedLifetimeCountBelowMinimum));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
            Assert.AreEqual(1, count);

            EasyUAMonitoredItemChangedEventArgs eventArgs = monitoredItemChangedEventArgsList[0];
            Assert.IsNotNull(eventArgs);
            Exception exception = eventArgs.Exception;
            Assert.IsNotNull(exception);
            AssertUtilities.WriteLine(exception);
            var clientEngineException = exception as UAClientEngineException;
            Assert.IsNotNull(clientEngineException);
            Assert.AreEqual(UAClientErrorCodes.RevisedLifetimeCountBelowMinimum, clientEngineException.ErrorCode);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_005_01(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-005-01.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedMaxKeepAliveCountAboveToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_005_02(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-005-02.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId ==
                                                                                UAEngineEventIds
                                                                                    .RevisedMaxKeepAliveCountBelowToleranceRange));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Information, logEntryEventArgs.EntryType);

            // IMPROVE: check that the subscription normally works
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_008(
            // ReSharper restore InconsistentNaming
            bool inverted,
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected,
            [NotNull] UAStatusCode statusCodeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);

                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;

                client.IsolatedParameters.SessionParameters.RetrialDelay = 10 * 1000;
                client.IsolatedParameters.SubscriptionParameters.RetrialDelay = 20 * 1000;
                client.IsolatedParameters.MonitoredItemParameters.RetrialDelay = 15 * 1000;

                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    60 * 1000);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == eventIdExpected) && args.Message.Contains(statusCodeExpected.ToString()));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);

            // IMPROVE: verify that the reconnections are taking place
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_011(
        // ReSharper restore InconsistentNaming
            bool inverted, 
            [NotNull] string scriptName, 
            int eventIdExpected, 
            LogEntryType entryTypeExpected)
        {
            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
                EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: true);
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    MillisecondsToCollectSingleNotification);

                AssertUtilities.WriteSeparatorLine();

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                 (args.EventId == eventIdExpected));
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                    String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

                if (inverted)
                {
                    Assert.IsNull(logEntryEventArgs);
                    return;
                }

                Assert.IsNotNull(logEntryEventArgs);
                Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);

                // IMPROVE: test what happens to the server (re-ack etc.)
            }
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_012(
            bool inverted,
            // ReSharper restore InconsistentNaming
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            TestCase_Err_Subscription(inverted, scriptName, eventIdExpected, entryTypeExpected);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_020(
            bool inverted)
            // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName,
                @"Subscription Services\Subscription Client Basic\Test Cases\Err-020.js"))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    10*1000);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == UAEngineEventIds.DataChangeNotificationMonitoredItemsEmpty));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs republishMissingMessagesEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == UAEngineEventIds.RepublishMissingMessages));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((republishMissingMessagesEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", republishMissingMessagesEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                Assert.IsNull(republishMissingMessagesEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);

            Assert.IsNotNull(republishMissingMessagesEventArgs);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_021_01(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            var dataChangeNotificationEventArgsList = new List<EasyUADataChangeNotificationEventArgs>();
            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-021-01.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(
                        EndpointDescriptor,
                        UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                        3 * 1000,
                        (_, args) => dataChangeNotificationEventArgsList.Add(args));
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgsServer = logEntryEventArgsList.Find(args =>
                                                                                   (args.EventId ==
                                                                                    UAEngineEventIds
                                                                                        .MonitoredItemNotificationServerTimestampAhead));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgsServer == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgsServer));

            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgsSource = logEntryEventArgsList.Find(args =>
                                                                                   (args.EventId ==
                                                                                    UAEngineEventIds
                                                                                        .MonitoredItemNotificationSourceTimestampAhead));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgsSource == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgsSource));

            int count = dataChangeNotificationEventArgsList.Count;
            AssertUtilities.WriteLine(count);
            Assert.AreNotEqual(0, count);

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgsServer);
                Assert.IsNull(logEntryEventArgsSource);

                Assert.IsTrue(dataChangeNotificationEventArgsList.All(args =>
                    // ReSharper disable once MergeSequentialChecks
                                                                    (args != null) &&
                                                                    (args.AttributeData != null) &&
                                                                    (args.AttributeData.StatusCode.CodeBits == UACodeBits.Good)));

                return;
            }

            Assert.IsNotNull(logEntryEventArgsServer);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgsServer.EntryType);

            Assert.IsNotNull(logEntryEventArgsSource);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgsSource.EntryType);

            Assert.IsTrue(dataChangeNotificationEventArgsList.All(args =>
                // ReSharper disable once MergeSequentialChecks
                                                                (args != null) &&
                                                                (args.AttributeData != null) &&
                                                                (args.AttributeData.StatusCode.CodeBits ==
                                                                 UACodeBits.BadInvalidTimestamp)));
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_022_02(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-022-02.js")
                )
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds.SessionPublishError));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_023(bool inverted)
        // ReSharper restore InconsistentNaming
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (
                inverted
                    ? null
                    : Ctt.RunScript(CttProjectName, @"Subscription Services\Subscription Client Basic\Test Cases\Err-023.js"))
            {
                using (EasyUAClient client = MakeClient())
                {
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                    client.SubscribeDataChange(EndpointDescriptor,
                                                  UATestUtilities.MakeNodeDescriptor(typeof(Int32), dynamic: true),
                                                  3 * 1000);
                    AssertUtilities.WaitingPhase(10 * 1000);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                             (args.EventId ==
                                                                              UAEngineEventIds
                                                                                  .MonitoredItemNotificationClientHandleUnknown));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(LogEntryType.Warning, logEntryEventArgs.EntryType);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_Subscription(
            bool inverted,
            // ReSharper restore InconsistentNaming
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                EasyUAMonitoredItemArguments monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    MillisecondsToCollectSingleNotification);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_Subscription(
            // ReSharper restore InconsistentNaming
            bool inverted,
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected,
            [NotNull] UAStatusCode statusCodeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                EasyUAMonitoredItemArguments monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    MillisecondsToCollectSingleNotification);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);

            StringAssert.Contains(logEntryEventArgs.Message, statusCodeExpected.ToString());
        }

        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_Subscription_Volume(
            bool inverted,
            // ReSharper restore InconsistentNaming
            [NotNull] string scriptName,
            int count,
            int millisecondsToCollect,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                EasyUAMonitoredItemArguments[] monitoredItemArgumentsArray = 
                    MakeVolumeMonitoredItemArguments(count, dynamic: true);
                CollectMultipleDataChangeNotificationEventArgs(
                    client,
                    monitoredItemArgumentsArray,
                    millisecondsToCollect);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
        }

        #endregion
    }
}
