﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SubscriptionServices
{
    public abstract partial class ModifyTest : EndpointTestBase
    {
        #region Protected Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_Subscription(
            bool inverted,
            // ReSharper restore InconsistentNaming
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected,
            [NotNull] UAStatusCode statusCodeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (inverted ? null : Ctt.RunScript(CttProjectName, scriptName))
            {
                var monitoredItemArguments1 = MakeMonitoredItemArguments(typeof(Int32), dynamic: false);
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;

                // If we had just one subscription, the DeleteSubscription service may not be called, because of 
                // optimization. We therefore neeed at least one more subscription.
                EasyUAMonitoredItemArguments monitoredItemArguments2 = MakeMonitoredItemArguments(typeof(Int16), dynamic: false);
                int handle2 = client.SubscribeMonitoredItem(monitoredItemArguments2);

                try
                {
                    CollectMonitoredItemChangedEventArgs(
                        client,
                        monitoredItemArguments1,
                        MillisecondsToCollectSingleNotification);
                }
                finally
                {
                    client.UnsubscribeMonitoredItem(handle2);
                }
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            AssertUtilities.WriteLine(logEntryEventArgs.Message);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);

            StringAssert.Contains(logEntryEventArgs.Message, statusCodeExpected.ToString());
        }
#endif

#endregion
    }
}
