﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SubscriptionServices
{
    public abstract partial class RepublishTest
    {
        // ReSharper disable InconsistentNaming
        public abstract class RepublishTestCtt : RepublishTest
        // ReSharper restore InconsistentNaming
        {
            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class RepublishTest_Ctt_OpcFoundationSampleServer_OpcTcp : RepublishTestCtt
            // ReSharper restore InconsistentNaming
            {
                public RepublishTest_Ctt_OpcFoundationSampleServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\OpcFoundation_SampleServer_OpcTcp.ctt.xml";
                }
            }



            [DeploymentItem("UAClientCttProjects", "UAClientCttProjects")]
            [TestClass]
            // ReSharper disable InconsistentNaming
            public class RepublishTest_Ctt_SoftingDemoServer_OpcTcp : RepublishTestCtt
            // ReSharper restore InconsistentNaming
            {
                public RepublishTest_Ctt_SoftingDemoServer_OpcTcp()
                {
                    CttProjectName = @"UAClientCttProjects\Softing_DemoServer_OpcTcp.ctt.xml";
                }

                [ClassInitialize]
                static public void SocketPrecheck(TestContext context)
                {
                    LegacyUATestEnvironment.Current.PrecheckEndpoint(LegacyUATestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);
                }
            }




            #region Test Methods
            // ReSharper disable InconsistentNaming

            /// <description>
            /// Randomly inject one of the following ServiceResult codes. Repeat this test several times each tmie varying 
            /// the result
            /// 1. Bad_SubscriptionIdInvalid
            /// 2. Bad_MessageNotAvailable. 
            /// </description>
            /// <expectedresult>
            /// Client reports the error.
            /// </expectedresult>
            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_001_01()
            {
                TestCase_Err_001(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-001-01-WithInvoke.js",
                    UAEngineEventIds.SessionPublishError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadSubscriptionIdInvalid, 0));
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_001_01_Inverted()
            {
                TestCase_Err_001(
                    true,
                    @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
                    UAEngineEventIds.SessionPublishError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadSubscriptionIdInvalid, 0));
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_001_02()
            {
                TestCase_Err_001(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-001-02-WithInvoke.js",
                    UAEngineEventIds.SessionPublishError,
                    LogEntryType.Warning,
                    new UAStatusCode(UACodeBits.BadMessageNotAvailable, 0));
            }

            //[TestCategory("Keyword:Republish")]
            //[TestCategory("Live")]
            //[TestCategory("OpcCompliance")]
            //[TestCategory("Type:CTT")]
            //[TestMethod]
            //public void TestCase_Err_001_02_Inverted()
            //{
            //    TestCase_Err_001(
            //        true,
            //        @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
            //        UAEngineEventIds.SessionPublishError,
            //        LogEntryType.Warning,
            //        new UAStatusCode(UACodeBits.BadMessageNotAvailable, 0));
            //}

            /// <description>
            /// ServiceResult = Good, but clear the NotificationMessage collection.
            /// </description>
            /// <expectedresult>
            /// Client handles the situation and then reports an error to the end-user.
            /// </expectedresult>
            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_002()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-002-WithInvoke.js",
                    UAEngineEventIds.NotificationMessageEmpty,
                    LogEntryType.Warning);
            }

            // Event the inverted test runs with CTT, for invoking the Republish.
            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_002_Inverted()
            {
                TestCase_Err_Subscription(
                    true,
                    @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
                    UAEngineEventIds.NotificationMessageEmpty,
                    LogEntryType.Warning);
            }

            /// <description>
            /// ServiceResult = Good, but modify the sequenceNumber to use a previously acknowledged number.
            /// </description>
            /// <expectedresult>
            /// Client handles the situation and then reports an error to the end-user.
            /// </expectedresult>
            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-003-WithInvoke.js",
                    UAEngineEventIds.SequenceNumberNotAsRequested,
                    LogEntryType.Warning);
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_003_Inverted()
            {
                TestCase_Err_Subscription(
                    true,
                    @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
                    UAEngineEventIds.SequenceNumberNotAsRequested,
                    LogEntryType.Warning);
            }

            // We are intentionally expecting a different result: the one borrowed from Publish, Err-012.
            //From: Nathan Pocock [mailto:nathan.pocock@opcfoundation.org] 
            //Sent: Thursday, June 13, 2013 3:27 PM
            //To: Zbynek Zahradnik; deepthi.teegala@opcfoundation.org
            //Subject: RE: Inconsistencies between Publish and Republish
            //Hello Zbynek,
            //I would agree with you. The Republish expectation is too much. I will revise the test-case expectation to match 
            //that of the Publish test-case.
            /// <description>
            /// Randomly inject one of the following code. Repeat this test several times each tmie varying the result
            /// 1. ServiceResult = Good, but the publishTime parameter uses a time in the future, i.e. Now + 1 day.
            /// 2.ServiceResult = Good, but the publishTime parameter uses a time in the
            /// future, i.e. Now + 1 year. 
            /// </description>
            /// <expectedresult>
            /// Client logs and/or displays a Warning to the end-user.
            /// Client adjusts the StatusCode for each monitoredItem to Bad_UnexpectedError.
            /// </expectedresult>
            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_01()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-004-01-WithInvoke.js",
                    UAEngineEventIds.PublishTimeAhead,
                    LogEntryType.Information);
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_01_Inverted()
            {
                TestCase_Err_Subscription(
                    true,
                    @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
                    UAEngineEventIds.PublishTimeAhead,
                    LogEntryType.Information);
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_02()
            {
                TestCase_Err_Subscription(
                    false,
                    @"Subscription Services\Subscription Client Republish\Test Cases\Err-004-02-WithInvoke.js",
                    UAEngineEventIds.PublishTimeAhead,
                    LogEntryType.Information);
            }

            [TestCategory("Keyword:Republish")]
            [TestCategory("Live")]
            [TestCategory("OpcCompliance")]
            [TestCategory("Type:CTT")]
            [TestMethod]
            public void TestCase_Err_004_02_Inverted()
            {
                TestCase_Err_Subscription(
                    true,
                    @"Subscription Services\Subscription Client Republish\Test Cases\InvokeRepublish.js",
                    UAEngineEventIds.PublishTimeAhead,
                    LogEntryType.Information);
            }

            // ReSharper restore InconsistentNaming
            #endregion



            #region Private Methods

            // ReSharper disable InconsistentNaming
            protected void TestCase_Err_001(
            // ReSharper restore InconsistentNaming
                bool inverted,
                [NotNull] string scriptName,
                int eventIdExpected,
                LogEntryType entryTypeExpected,
                [NotNull] UAStatusCode statusCodeExpected)
            {
                var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
                EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

                using (Ctt.RunScript(CttProjectName, scriptName))
                {
                    var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: true);

                    EasyUAClient client = MakeClient();
                    client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;

                    CollectMonitoredItemChangedEventArgs(
                        client,
                        monitoredItemArguments,
                        10 * 1000);
                }
                AssertUtilities.WriteSeparatorLine();

                var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs republishMissingMessagesEventArgs = logEntryEventArgsList.Find(args =>
                    (args.EventId == UAEngineEventIds.RepublishMissingMessages));
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WriteLine((republishMissingMessagesEventArgs == null) ? "Not found" :
                    String.Format(CultureInfo.InvariantCulture, "Found {0}", republishMissingMessagesEventArgs));

                // ReSharper disable PossibleNullReferenceException
                LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                    (args.EventId == eventIdExpected) && args.Message.Contains(statusCodeExpected.ToString()));
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                    String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

                Assert.IsNotNull(republishMissingMessagesEventArgs);

                if (inverted)
                {
                    Assert.IsNull(logEntryEventArgs);
                    return;
                }

                Assert.IsNotNull(logEntryEventArgs);
                Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
            }

            #endregion
        }
    }
}
