﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Globalization;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Instrumentation;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Toolkit.Diagnostics;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.SubscriptionServices
{
    public abstract partial class RepublishTest : EndpointTestBase
    {
        #region Protected Methods

#if (NETFRAMEWORK)
        // ReSharper disable InconsistentNaming
        protected void TestCase_Err_Subscription(
            bool inverted,
            // ReSharper restore InconsistentNaming
            [NotNull] string scriptName,
            int eventIdExpected,
            LogEntryType entryTypeExpected)
        {
            var logEntryEventArgsListQueue = new ConcurrentQueue<LogEntryEventArgs>();
            EasyUAClient.LogEntry += (sender, args) => logEntryEventArgsListQueue.Enqueue(args);

            using (Ctt.RunScript(CttProjectName, scriptName))
            {
                EasyUAClient client = MakeClient();
                client.IsolatedParameters.SessionParameters.StatusSubscriptionSamplingInterval = Timeout.Infinite;
                var monitoredItemArguments = MakeMonitoredItemArguments(typeof(Int32), dynamic: true);
                CollectMonitoredItemChangedEventArgs(
                    client,
                    monitoredItemArguments,
                    10*1000);
            }
            AssertUtilities.WriteSeparatorLine();

            var logEntryEventArgsList = new List<LogEntryEventArgs>(logEntryEventArgsListQueue);
            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs republishMissingMessagesEventArgs = logEntryEventArgsList.Find(args =>
                (args.EventId == UAEngineEventIds.RepublishMissingMessages));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((republishMissingMessagesEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", republishMissingMessagesEventArgs));

            // ReSharper disable PossibleNullReferenceException
            LogEntryEventArgs logEntryEventArgs = logEntryEventArgsList.Find(args =>
                                                                                (args.EventId == eventIdExpected));
            // ReSharper restore PossibleNullReferenceException
            AssertUtilities.WriteLine((logEntryEventArgs == null) ? "Not found" :
                String.Format(CultureInfo.InvariantCulture, "Found {0}", logEntryEventArgs));

            Assert.IsNotNull(republishMissingMessagesEventArgs);

            if (inverted)
            {
                Assert.IsNull(logEntryEventArgs);
                return;
            }

            Assert.IsNotNull(logEntryEventArgs);
            Assert.AreEqual(entryTypeExpected, logEntryEventArgs.EntryType);
        }
#endif

#endregion
    }
}
