﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.QualityTools.Testing.Fakes;
using Opc.Ua;
using Opc.Ua.Client;
using Opc.Ua.Client.Fakes;
using Opc.Ua.Fakes;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    class UAOperationRecorder : IDisposable
    {
        #region Public Constructors

        public UAOperationRecorder()
        {
            _context = ShimsContext.Create();

            CloseSessionRecordList = new List<CloseSessionRecord>();
            CreateMonitoredItemsRecordList = new List<CreateMonitoredItemsRecord>();
            ReadRecordList = new List<ReadRecord>();
            SessionOpenRecordList = new List<SessionOpenRecord>();
            WriteRecordList = new List<WriteRecord>();

            ShimSession.AllInstances.OpenStringUInt32IUserIdentityIListOfString 
                = FakeSessionOpen;
            ShimSessionClient.AllInstances.CloseSessionRequestHeaderBoolean 
                = FakeCloseSession;
            ShimSessionClient.AllInstances.CreateMonitoredItemsRequestHeaderUInt32TimestampsToReturnMonitoredItemCreateRequestCollectionMonitoredItemCreateResultCollectionOutDiagnosticInfoCollectionOut
                = FakeCreateMonitoredItems;
            ShimSessionClient.AllInstances.ReadRequestHeaderDoubleTimestampsToReturnReadValueIdCollectionDataValueCollectionOutDiagnosticInfoCollectionOut
                = FakeRead;
            ShimSessionClient.AllInstances.WriteRequestHeaderWriteValueCollectionStatusCodeCollectionOutDiagnosticInfoCollectionOut
                = FakeWrite;
        }

        #endregion



        ~UAOperationRecorder()
        {
            Dispose(false);
        }



        #region Public Properties

        //[NotNull]
        //private List<ActivateSessionRecord> ActivateSessionRecordList { get; set; }

        [NotNull]
        public List<CloseSessionRecord> CloseSessionRecordList { get; }

        [NotNull]
        public List<CreateMonitoredItemsRecord> CreateMonitoredItemsRecordList { get; }

        [NotNull]
        public List<ReadRecord> ReadRecordList { get; }

        [NotNull]
        public List<SessionOpenRecord> SessionOpenRecordList { get; }

        [NotNull]
        public List<WriteRecord> WriteRecordList { get; }

        #endregion



        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion



        #region Private Methods

        private void Dispose(bool disposing)
        {
            //Check.Ensures(_disposed);

            if (!_disposed)
            {
                if (disposing)
                {
                    // Free other state (managed objects).
                }
                // Free your own state (unmanaged objects).
                // Set large fields to null.
                if (_context != null)
                {
                    _context.Dispose();
                    _context = null;
                }

                _disposed = true;
            }
        }

        //private ResponseHeader FakeActivateSession([NotNull] SessionClient sessionClient,
        //    RequestHeader requestHeader,
        //    SignatureData clientSignature,
        //    SignedSoftwareCertificateCollection clientSoftwareCertificates,
        //    StringCollection localeIds,
        //    ExtensionObject userIdentityToken,
        //    SignatureData userTokenSignature,
        //    out byte[] serverNonce,
        //    out StatusCodeCollection results,
        //    out DiagnosticInfoCollection diagnosticInfos)
        //{
        //    var activateSessionRecord = new ActivateSessionRecord
        //    {
        //        SessionClient = sessionClient,
        //        RequestHeader = requestHeader,
        //        ClientSignature = clientSignature,
        //        ClientSoftwareCertificates = clientSoftwareCertificates,
        //        LocaleIds = localeIds,
        //        UserIdentityToken = userIdentityToken,
        //        UserTokenSignature = userTokenSignature,
        //    };
        //    try
        //    {
        //        ShimsContext.ExecuteWithoutShims(() => OriginalActivateSession(activateSessionRecord));
        //        serverNonce = activateSessionRecord.ServerNonce;
        //        results = activateSessionRecord.Results;
        //        diagnosticInfos = activateSessionRecord.DiagnosticInfos;
        //    }
        //    catch (Exception exception)
        //    {
        //        activateSessionRecord.Exception = exception;
        //        throw;
        //    }
        //    finally
        //    {
        //        ActivateSessionRecordList.Add(activateSessionRecord);
        //    }
        //    return activateSessionRecord.Result;
        //}

        private ResponseHeader FakeCloseSession([NotNull] SessionClient sessionClient,
            RequestHeader requestHeader, bool deleteSubscriptions)
        {
            var closeSessionRecord = new CloseSessionRecord(sessionClient)
            {
                RequestHeader = requestHeader,
                DeleteSubscriptions = deleteSubscriptions,
            };
            try
            {
                ShimsContext.ExecuteWithoutShims(() => OriginalCloseSession(closeSessionRecord));
            }
            catch (Exception exception)
            {
                closeSessionRecord.Exception = exception;
                throw;
            }
            finally
            {
                CloseSessionRecordList.Add(closeSessionRecord);
            }
            return closeSessionRecord.Result;
        }

        private ResponseHeader FakeCreateMonitoredItems([NotNull] SessionClient sessionClient, RequestHeader requestHeader,
            uint subscriptionId, TimestampsToReturn timestampsToReturn, MonitoredItemCreateRequestCollection itemsToCreate,
            out MonitoredItemCreateResultCollection results, out DiagnosticInfoCollection diagnosticInfos)
        {
            var createMonitoredItemsRecord = new CreateMonitoredItemsRecord(sessionClient)
            {
                RequestHeader = requestHeader,
                SubscriptionId = subscriptionId,
                TimestampsToReturn = timestampsToReturn,
                ItemsToCreate = itemsToCreate,
            };
            try
            {
                ShimsContext.ExecuteWithoutShims(() => OriginalCreateMonitoredItems(createMonitoredItemsRecord));
                results = createMonitoredItemsRecord.Results;
                diagnosticInfos = createMonitoredItemsRecord.DiagnosticInfos;
            }
            catch (Exception exception)
            {
                createMonitoredItemsRecord.Exception = exception;
                throw;
            }
            finally
            {
                CreateMonitoredItemsRecordList.Add(createMonitoredItemsRecord);
            }
            return createMonitoredItemsRecord.Result;
        }

        private ResponseHeader FakeRead([NotNull] SessionClient sessionClient, RequestHeader requestHeader,
            double maxAge, TimestampsToReturn timestampsToReturn, ReadValueIdCollection nodesToRead,
            out DataValueCollection results, out DiagnosticInfoCollection diagnosticInfos)
        {
            var readRecord = new ReadRecord(sessionClient)
            {
                RequestHeader = requestHeader,
                MaxAge = maxAge,
                TimestampsToReturn = timestampsToReturn,
                NodesToRead = nodesToRead,
            };
            try
            {
                ShimsContext.ExecuteWithoutShims(() => OriginalRead(readRecord));
                results = readRecord.Results;
                diagnosticInfos = readRecord.DiagnosticInfos;
            }
            catch (Exception exception)
            {
                readRecord.Exception = exception;
                throw;
            }
            finally
            {
                ReadRecordList.Add(readRecord);
            }
            return readRecord.Result;
        }

        private void FakeSessionOpen([NotNull] Session session, string sessionName, uint sessionTimeout,
            IUserIdentity identity, IList<string> preferredLocales)
        {
            var sessionOpenRecord = new SessionOpenRecord(session)
            {
                SessionName = sessionName,
                SessionTimeout = sessionTimeout,
                Identity = identity,
                PreferredLocales = preferredLocales,
            };
            try
            {
                ShimsContext.ExecuteWithoutShims(() => OriginalSessionOpen(sessionOpenRecord));
            }
            catch (Exception exception)
            {
                sessionOpenRecord.Exception = exception;
                throw;
            }
            finally
            {
                SessionOpenRecordList.Add(sessionOpenRecord);
            }
        }

        private ResponseHeader FakeWrite([NotNull] SessionClient sessionClient, RequestHeader requestHeader,
            WriteValueCollection nodesToWrite,
            out StatusCodeCollection results, out DiagnosticInfoCollection diagnosticInfos)
        {
            var writeRecord = new WriteRecord(sessionClient)
            {
                RequestHeader = requestHeader,
                NodesToWrite = nodesToWrite,
            };
            try
            {
                ShimsContext.ExecuteWithoutShims(() => OriginalWrite(writeRecord));
                results = writeRecord.Results;
                diagnosticInfos = writeRecord.DiagnosticInfos;
            }
            catch (Exception exception)
            {
                writeRecord.Exception = exception;
                throw;
            }
            finally
            {
                WriteRecordList.Add(writeRecord);
            }
            return writeRecord.Result;
        }

        //private void OriginalActivateSession([NotNull] ActivateSessionRecord record)
        //{
        //    record.Result = record.SessionClient.ActivateSession(
        //        record.RequestHeader, record.ClientSignature, record.ClientSoftwareCertificates, record.LocaleIds,
        //        record.UserIdentityToken, record.UserTokenSignature,
        //        out record.ServerNonce, out record.Results, out record.DiagnosticInfos);
        //}

        private void OriginalCloseSession([NotNull] CloseSessionRecord record)
        {
            record.Result = record.SessionClient.CloseSession(record.RequestHeader, record.DeleteSubscriptions);
        }

        private void OriginalCreateMonitoredItems([NotNull] CreateMonitoredItemsRecord record)
        {
            record.Result = record.SessionClient.CreateMonitoredItems(record.RequestHeader, record.SubscriptionId,
                record.TimestampsToReturn, record.ItemsToCreate, 
                out MonitoredItemCreateResultCollection results, out DiagnosticInfoCollection diagnosticInfos);
            record.Results = results;
            record.DiagnosticInfos = diagnosticInfos;
        }

        private void OriginalRead([NotNull] ReadRecord record)
        {
            record.Result = record.SessionClient.Read(record.RequestHeader, record.MaxAge,
                record.TimestampsToReturn, record.NodesToRead, 
                out DataValueCollection results, out DiagnosticInfoCollection diagnosticInfos);
            record.Results = results;
            record.DiagnosticInfos = diagnosticInfos;
        }

        private void OriginalSessionOpen([NotNull] SessionOpenRecord record)
        {
            record.Session.Open(record.SessionName, record.SessionTimeout, record.Identity, record.PreferredLocales);
        }

        private void OriginalWrite([NotNull] WriteRecord record)
        {
            record.Result = record.SessionClient.Write(record.RequestHeader,
                record.NodesToWrite,
                out StatusCodeCollection results, out DiagnosticInfoCollection diagnosticInfos);
            record.Results = results;
            record.DiagnosticInfos = diagnosticInfos;
        }

        #endregion



        #region Private Fields

        private IDisposable _context;

        private bool _disposed;

        #endregion
    }
}
