﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using JetBrains.Annotations;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Navigation;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    static class UATestUtilities
    {
        public const int BuiltInCreateMonitoredItemsCount = 1;



        [NotNull]
        static public UANodeDescriptor MakeNodeDescriptor([NotNull] Type type, bool dynamic, bool preferArray = false,
            bool shared = false)
        {
            string pathPart = null;
            if (type == typeof(Boolean)) pathPart = "Scalar/BooleanValue";
            else if (type == typeof(Boolean[])) pathPart = "Array/BooleanValue";
            else if (type == typeof(Byte)) pathPart = "Scalar/ByteValue";
            else if (type == typeof(Byte[])) pathPart = preferArray ? "Array/ByteValue"  : "Scalar/ByteStringValue";
            else if (type == typeof(Byte[][])) pathPart = "Array/ByteStringValue";
            //else if (type == typeof (Char)) name = "Scalar/CharValue";
            //else if (type == typeof (Char[])) name = "Array/CharValue";
            else if (type == typeof(DateTime)) pathPart = "Scalar/DateTimeValue";
            else if (type == typeof(DateTime[])) pathPart = "Array/DateTimeValue";
            else if (type == typeof(Double)) pathPart = "Scalar/DoubleValue";
            else if (type == typeof(Double[])) pathPart = "Array/DoubleValue";
            else if (type == typeof(Guid)) pathPart = "Scalar/GuidValue";
            else if (type == typeof(Guid[])) pathPart = "Array/GuidValue";
            else if (type == typeof(Int16)) pathPart = "Scalar/Int16Value";
            else if (type == typeof(Int16[])) pathPart = "Array/Int16Value";
            else if (type == typeof(Int32)) pathPart = "Scalar/Int32Value";
            else if (type == typeof(Int32[])) pathPart = "Array/Int32Value";
            else if (type == typeof(Int64)) pathPart = "Scalar/Int64Value";
            else if (type == typeof(Int64[])) pathPart = "Array/Int64Value";
            else if (type == typeof(SByte)) pathPart = "Scalar/SByteValue";
            else if (type == typeof(SByte[])) pathPart = "Array/SByteValue";
            else if (type == typeof(Single)) pathPart = "Scalar/FloatValue";
            else if (type == typeof(Single[])) pathPart = "Array/FloatValue";
            else if (type == typeof(String)) pathPart = "Scalar/StringValue";
            else if (type == typeof(String[])) pathPart = "Array/StringValue";
            else if (type == typeof(UInt16)) pathPart = "Scalar/UInt16Value";
            else if (type == typeof(UInt16[])) pathPart = "Array/UInt16Value";
            else if (type == typeof(UInt32)) pathPart = "Scalar/UInt32Value";
            else if (type == typeof(UInt32[])) pathPart = "Array/UInt32Value";
            else if (type == typeof(UInt64)) pathPart = "Scalar/UInt64Value";
            else if (type == typeof(UInt64[])) pathPart = "Array/UInt64Value";
            Debug.Assert(pathPart != null);

            if (!shared)
                pathPart = "User" + pathPart;

            string browsePathString = String.Format(CultureInfo.InvariantCulture,
                dynamic ? "[ObjectsFolder]/Data/Dynamic/{0}" : "[ObjectsFolder]/Data/Static/{0}",
                pathPart);
            return new UANodeDescriptor(UABrowsePath.Parse(browsePathString, "http://test.org/UA/Data/"));
        }

        [NotNull]
        static public UANodeDescriptor MakeVolumeNodeDescriptor(int index, bool dynamic)
        {
            if (dynamic)
            {
                const string memoryBuffersUInt32Format = "nsu=http://samples.org/UA/memorybuffer/Instance;s=UInt32[{0}]";
                string expandedNodeIdText = String.Format(CultureInfo.InvariantCulture, memoryBuffersUInt32Format, index);
                return new UANodeDescriptor(expandedNodeIdText);
            }
            //else
            {
                const string pathFormat = "[ObjectsFolder]/CTT/Static/All Profiles/Scalar Mass/Int32/Int32{0:D3}";
                string browsePathString = String.Format(CultureInfo.InvariantCulture, pathFormat, index);
                return new UANodeDescriptor(UABrowsePath.Parse(browsePathString, "http://opcfoundation.org/Quickstarts/ReferenceApplications"));
            }
        }

        [NotNull]
        static public UANodeDescriptor[] MakeVolumeNodeDescriptors(int count, bool dynamic)
        {
            var list = new List<UANodeDescriptor>();
            for (int i = 0; i < count; i++)
                list.Add(MakeVolumeNodeDescriptor(i, dynamic));
            return list.ToArray();
        }
    }
}
