﻿using System;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.Timing;
using OpcLabs.EasyOpc.UA.PubSub.Engine;
using OpcLabs.EasyOpc.UA.Sdk.PubSub.Configuration;
using OpcLabs.EasyOpc.UA.Toolkit.PubSub;
using OpcLabs.EasyOpc.UA.Toolkit.PubSub.Sdk;

namespace OpcLabs.EasyOpcLiveTest.UA.Toolkit.PubSub.Sdk
{
    [TestClass]
    public class PublisherSubscriberTest
    {
        class DataSink : IDataSink
        {
            public void OnDataSet(PubSubConnectionDataType pubSubConnectionData, ReaderGroupDataType readerGroupData,
                DataSetReaderDataType dataSetReaderData, DataSetData dataSetData, bool canBeDataMessage, bool canBeEventMessage, bool isDeltaFrame)
            {
                throw new NotImplementedException();
            }

            public void OnError(OperationResult operationResult, PubSubConnectionDataType pubSubConnectionData,
                ReaderGroupDataType readerGroupData, DataSetReaderDataType dataSetReaderData)
            {
                throw new NotImplementedException();
            }

            public void OnEvent(PubSubConnectionDataType pubSubConnectionData, ReaderGroupDataType readerGroupData,
                DataSetReaderDataType dataSetReaderData, DataSetData dataSetData)
            {
                throw new NotImplementedException();
            }
        }



        [TestMethod]
        public void Test_Open62541_UdpUadp()
        {
            // DataSetReaderData
            var dataSetReaderData = new DataSetReaderDataType
            {
                MessageSettings = new ExtensionObject(new UadpDataSetReaderMessageDataType()),
                SecurityMode = MessageSecurityMode.None
            };

            // ReaderGroupData
            var readerGroupData = new ReaderGroupDataType
            {
                DataSetReaders = { dataSetReaderData }
            };

            // PubSubConnectionData
            var networkAddressUrlData = new NetworkAddressUrlDataType
            {
                Url = "opc.udp://224.0.0.22"
            };
            var datagramConnectionTransportData = new DatagramConnectionTransportDataType();
            var pubSubConnectionData = new PubSubConnectionDataType
            {
                TransportProfileUri = UAPubSubTransportProfileUriStrings.UdpUadp,
                Address = new ExtensionObject(networkAddressUrlData),
                TransportSettings = new ExtensionObject(datagramConnectionTransportData),
                ReaderGroups = { readerGroupData }
            };

            // PubSubConfigurationData
            var pubSubConfigurationData = new PubSubConfigurationDataType
            {
                Connections = { pubSubConnectionData }
            };

            // 
            var publishSubscribe = new PublishSubscribe();
            var pubSubConfiguration = new PubSubConfiguration(publishSubscribe, pubSubConfigurationData);
            publishSubscribe.Children.Add(pubSubConfiguration);

            var readerSink = new DataSink();
            var subscriber = new PublisherSubscriber(new SystemTiming(), pubSubConfiguration, readerSink);

            subscriber.Attach();

            bool b = publishSubscribe.Status.Enable();
            Assert.IsTrue(b);

            Thread.Sleep(5*60*1000);  // TODO: 2019.2





        }
    }
}
