﻿
using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Graphs;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    partial class EasyUAClientTest
    {
        #region Static Private Methods

        static private void Asserts_OpcFoundation_AlarmConditionServer_EventTypes([CanBeNull] UANodeElementCollection nodeElements)
        {
            Assert.IsNotNull(nodeElements);
            int count = nodeElements.Count;
            AssertUtilities.WriteLine(count);
            foreach (UANodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine(nodeElement);
            Assert.AreEqual(7, count);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            // IMPROVE: assert more
        }

        static private void Asserts_OpcFoundation_AlarmConditionServer_ObjectTypes([CanBeNull] UANodeElementCollection nodeElements)
        {
            Assert.IsNotNull(nodeElements);
            int count = nodeElements.Count;
            AssertUtilities.WriteLine(count);
            foreach (UANodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine(nodeElement);
            Assert.AreEqual(25, count);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            // IMPROVE: assert more
        }

        #endregion



        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseDataNodesTest_AllReferenceTypesAndNodeClasses()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();

            TestSection.Act();
            UANodeElementCollection nodeElementCollection = target.Browse(
                endpointDescriptor,
                UAObjectIds.Server,
                new UABrowseParameters(UANodeClass.All, new[] { UAReferenceTypeIds.References }));

            TestSection.Assert();

            Assert.IsNotNull(nodeElementCollection);
            int count = nodeElementCollection.Count;
            AssertUtilities.WriteLine(count);
            foreach (UANodeElement nodeElement in nodeElementCollection)
                AssertUtilities.WriteLine(nodeElement);
            Assert.IsTrue(count == 14);
            CollectionAssert.AllItemsAreNotNull(nodeElementCollection);
            CollectionAssert.AllItemsAreUnique(nodeElementCollection);

            // TODO: assert more
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseDataNodesTest_OpcFoundation_AlarmConditionServer()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            var target = new EasyUAClient();

            TestSection.Act();
            UANodeElementCollection nodeElements = target.BrowseDataNodes(endpointDescriptor);

            TestSection.Assert();
            Assert.IsNotNull(nodeElements);
            int count = nodeElements.Count;
            AssertUtilities.WriteLine(count);
            foreach (UANodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine(nodeElement);
            Assert.IsTrue((count == 1 /*newer server*/) || (count == 2/*older server*/));
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            // IMPROVE: assert more
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("UAAlarmsAndConditions")]
        [TestMethod]
        public void BrowseEventSources_RecursiveTest_OpcFoundation_AlarmConditionServer()
        {
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            BrowseEventSources_RecursiveTest(endpointDescriptor, UAObjectIds.ObjectsFolder, depth: 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseMultipleTest_OpcFoundation_AlarmConditionServer()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            var target = new EasyUAClient();
            UABrowseArguments[] browseArgumentsArray = new[]
            {
                new UABrowseArguments(endpointDescriptor, UAObjectTypeIds.BaseObjectType, UABrowseParameters.ObjectTypes),
                new UABrowseArguments(endpointDescriptor, UAObjectTypeIds.BaseEventType, UABrowseParameters.ObjectTypes)
            };

            TestSection.Act();
            UANodeElementCollectionResult[] nodeElementCollectionResultArray = target.BrowseMultiple(browseArgumentsArray);

            TestSection.Assert();
            Assert.IsNotNull(nodeElementCollectionResultArray);
            Assert.AreEqual(2, nodeElementCollectionResultArray.Length);
            UANodeElementCollectionResult result0 = nodeElementCollectionResultArray[0];
            Assert.IsNotNull(result0);
            Assert.IsTrue(result0.Succeeded);
            Asserts_OpcFoundation_AlarmConditionServer_ObjectTypes(result0.Value);
            UANodeElementCollectionResult result1 = nodeElementCollectionResultArray[1];
            Assert.IsNotNull(result1);
            Assert.IsTrue(result1.Succeeded);
            Asserts_OpcFoundation_AlarmConditionServer_EventTypes(result1.Value);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("UAAlarmsAndConditions")]
        [TestMethod]
        public void BrowseTest_EventTypes_OpcFoundation_AlarmConditionServer()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UABrowseParameters browseParameters = UABrowseParameters.ObjectTypes;

            TestSection.Act();
            UANodeElementCollection nodeElements = target.Browse(endpointDescriptor,
                UAObjectTypeIds.BaseEventType, browseParameters);

            TestSection.Assert();
            Asserts_OpcFoundation_AlarmConditionServer_EventTypes(nodeElements);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseTest_ObjectTypes_OpcFoundation_AlarmConditionServer()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            var target = new EasyUAClient();
            UABrowseParameters browseParameters = UABrowseParameters.ObjectTypes;

            TestSection.Act();
            UANodeElementCollection nodeElements = target.Browse(endpointDescriptor,
                UAObjectTypeIds.BaseObjectType, browseParameters);

            TestSection.Assert();
            Asserts_OpcFoundation_AlarmConditionServer_ObjectTypes(nodeElements);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void TraverseNodeTreeTest_EventTypes_OpcFoundation_AlarmConditionServer()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UAAlarmConditionServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            var target = new EasyUAClient();
            UABrowseParameters browseParameters = UABrowseParameters.ObjectTypes;

            TestSection.Act();
            //IEnumerable<UATreePosition> positions = target.TraverseNodeTree(endpointDescriptor, UAObjectIds.EventTypesFolder, browseParameters);
            IEnumerable<UATreePosition> positions = target.TraverseNodeTree(endpointDescriptor, UAObjectTypeIds.BaseEventType, browseParameters);

            TestSection.Assert();
            int count = 0;
            foreach (UATreePosition treePosition in positions)
            {
                AssertUtilities.WriteLine(treePosition);
                count++;
            }
            AssertUtilities.WriteLine(count);
            Assert.AreEqual(71, count);
            // IMPROVE: assert more
        }

        [TestCategory("Live")]  // IMPROVE: is not really live; move elsewhere
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseDataNodesTest_NullEndpointDescriptor()
        {
            TestSection.Arrange();
            var endpointDescriptor = new UAEndpointDescriptor();
            var target = new EasyUAClient();

            TestSection.Act();
            try
            {
                target.BrowseDataNodes(endpointDescriptor);
            }
            catch (UAException uaException)
            {
                TestSection.Assert();
                AssertUtilities.WriteLine(uaException);
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException);
                var clientEngineException = innerException as UAClientEngineException;
                Assert.IsNotNull(clientEngineException);
                return;
            }
            TestSection.Assert();
            Assert.Fail();
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BrowseDataNodesTest_Objects()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();

            TestSection.Act();
            UANodeElementCollection nodeElementCollection = target.BrowseDataNodes(endpointDescriptor);

            TestSection.Assert();

            Assert.IsNotNull(nodeElementCollection);
            int count = nodeElementCollection.Count;
            AssertUtilities.WriteLine(count);
            foreach (UANodeElement nodeElement in nodeElementCollection)
                AssertUtilities.WriteLine(nodeElement);
            Assert.IsTrue((count == 4) || (count == 5));
            CollectionAssert.AllItemsAreNotNull(nodeElementCollection);
            CollectionAssert.AllItemsAreUnique(nodeElementCollection);

            // TODO: assert more
        }

        #endregion



        #region Private Methods

        private void BrowseEventSources_RecursiveTest(
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] UANodeDescriptor nodeDescriptor,
            int depth)
        {
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Parent node: {nodeDescriptor}");
            AssertUtilities.WriteLine($"Depth: {depth}");

            // Instantiate the client object
            var target = new EasyUAClient();

            // Obtain notifiers
            UANodeElementCollection nodeElements = target.BrowseEventSources(endpointDescriptor, nodeDescriptor);

            Assert.IsNotNull(nodeElements);
            int count = nodeElements.Count;
            AssertUtilities.WriteLine($"Count: {count}");
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Event sources:");
            foreach (UANodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine(nodeElement);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            // IMPROVE: assert more

            // Obtain objects
            UANodeElementCollection objectNodeElementCollection = target.BrowseObjects(endpointDescriptor, nodeDescriptor);

            // Recurse
            foreach (UANodeElement objectNodeElement in objectNodeElementCollection)
            {
                Assert.IsNotNull(objectNodeElement);
                BrowseEventSources_RecursiveTest(endpointDescriptor, objectNodeElement, depth + 1);
            }
        }

        #endregion
    }
}
