﻿
using System;
using System.Collections.Generic;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    public partial class EasyUAClientTest
    {
        #region Test Methods

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ServerConditionChangedTest_ClosedPort()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = ClosedPortEndpointDescriptor;
            PrecheckClosedPort(endpointDescriptor);

            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            ServerConditionChangedTest(
                endpointDescriptor,
                // ReSharper disable AssignNullToNotNullAttribute
                (client, descriptor) =>
                {
                    try
                    {
                        client.Read(descriptor, nodeId);
                    }
                    catch (UAException uaException)
                    {
                        AssertUtilities.WriteLine(uaException);
                        return;
                    }
                    Assert.Fail("UAException not thrown.");
                },
                // ReSharper restore AssignNullToNotNullAttribute
                out List<EasyUAServerConditionChangedEventArgs> readList,
                out List<EasyUAServerConditionChangedEventArgs> afterReadList);

            Assert.IsTrue(readList.Count > 0);
            {
                EasyUAServerConditionChangedEventArgs readEventArgs1 = readList[0];
                Assert.IsNotNull(readEventArgs1);
                Assert.AreEqual(endpointDescriptor, readEventArgs1.EndpointDescriptor);
                Assert.AreEqual(0, readEventArgs1.ErrorCode);
                Assert.IsNull(readEventArgs1.Exception);
                Assert.IsFalse(readEventArgs1.Connected);
                Assert.AreEqual(ConnectionState.Connecting, readEventArgs1.ConnectionState);
                Assert.AreEqual(0, readEventArgs1.RetrialDelay);
            }
            Assert.IsTrue(readList.Count > 1);
            {
                EasyUAServerConditionChangedEventArgs readEventArgs2 = readList[1];
                Assert.IsNotNull(readEventArgs2);
                Assert.AreEqual(endpointDescriptor, readEventArgs2.EndpointDescriptor);
                Assert.IsTrue(readEventArgs2.ErrorCode < 0);
                Assert.IsNotNull(readEventArgs2.Exception);
                Assert.IsFalse(readEventArgs2.Connected);
                Assert.AreEqual(ConnectionState.Connecting, readEventArgs2.ConnectionState);
                Assert.AreEqual(0, readEventArgs2.RetrialDelay);
            }
            Assert.IsFalse(readList.Count > 2);

            Assert.IsTrue(afterReadList.Count > 0);
            {
                EasyUAServerConditionChangedEventArgs afterReadEventArgs1 = afterReadList[0];
                Assert.IsNotNull(afterReadEventArgs1);
                Assert.AreEqual(endpointDescriptor, afterReadEventArgs1.EndpointDescriptor);
                Assert.IsTrue(afterReadEventArgs1.ErrorCode < 0);
                Assert.IsNotNull(afterReadEventArgs1.Exception);
                Assert.IsFalse(afterReadEventArgs1.Connected);
                Assert.AreEqual(ConnectionState.Disconnected, afterReadEventArgs1.ConnectionState);
                Assert.AreEqual(EasyUAClient.AdaptableParameters.SessionParameters.RetrialDelay, afterReadEventArgs1.RetrialDelay);
            }
            Assert.IsTrue(afterReadList.Count > 1);
            {
                EasyUAServerConditionChangedEventArgs afterReadEventArgs2 = afterReadList[1];
                Assert.IsNotNull(afterReadEventArgs2);
                Assert.AreEqual(endpointDescriptor, afterReadEventArgs2.EndpointDescriptor);
                Assert.AreEqual(0, afterReadEventArgs2.ErrorCode);
                Assert.IsNull(afterReadEventArgs2.Exception);
                Assert.IsFalse(afterReadEventArgs2.Connected);
                Assert.AreEqual(ConnectionState.Disconnected, afterReadEventArgs2.ConnectionState);
                Assert.AreEqual(Timeout.Infinite, afterReadEventArgs2.RetrialDelay);
            }
            Assert.IsFalse(afterReadList.Count > 2);
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ServerConditionChangedTest_Normal()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            ServerConditionChangedTest(
                endpointDescriptor,
                // ReSharper disable AssignNullToNotNullAttribute
                (client, descriptor) => { client.Read(descriptor, nodeId); },
                // ReSharper restore AssignNullToNotNullAttribute
                out List<EasyUAServerConditionChangedEventArgs> readList,
                out List<EasyUAServerConditionChangedEventArgs> afterReadList);

            Assert.IsTrue(readList.Count > 0);
            {
                EasyUAServerConditionChangedEventArgs readEventArgs1 = readList[0];
                Assert.IsNotNull(readEventArgs1);
                Assert.AreEqual(endpointDescriptor, readEventArgs1.EndpointDescriptor);
                Assert.AreEqual(0, readEventArgs1.ErrorCode);
                Assert.IsNull(readEventArgs1.Exception);
                Assert.IsFalse(readEventArgs1.Connected);
                Assert.AreEqual(ConnectionState.Connecting, readEventArgs1.ConnectionState);
                Assert.AreEqual(0, readEventArgs1.RetrialDelay);
            }
            Assert.IsTrue(readList.Count > 1);
            {
                EasyUAServerConditionChangedEventArgs readEventArgs2 = readList[1];
                Assert.IsNotNull(readEventArgs2);
                Assert.AreEqual(endpointDescriptor, readEventArgs2.EndpointDescriptor);
                Assert.AreEqual(0, readEventArgs2.ErrorCode);
                Assert.IsNull(readEventArgs2.Exception);
                Assert.IsTrue(readEventArgs2.Connected);
                Assert.AreEqual(ConnectionState.Connected, readEventArgs2.ConnectionState);
                Assert.AreEqual(0, readEventArgs2.RetrialDelay);
            }
            Assert.IsFalse(readList.Count > 2);

            Assert.IsTrue(afterReadList.Count > 0);
            {
                EasyUAServerConditionChangedEventArgs afterReadEventArgs1 = afterReadList[0];
                Assert.IsNotNull(afterReadEventArgs1);
                Assert.AreEqual(endpointDescriptor, afterReadEventArgs1.EndpointDescriptor);
                Assert.AreEqual(0, afterReadEventArgs1.ErrorCode);
                Assert.IsNull(afterReadEventArgs1.Exception);
                Assert.IsFalse(afterReadEventArgs1.Connected);
                Assert.AreEqual(ConnectionState.Disconnecting, afterReadEventArgs1.ConnectionState);
                Assert.AreEqual(0, afterReadEventArgs1.RetrialDelay);
            }
            Assert.IsTrue(afterReadList.Count > 1);
            {
                EasyUAServerConditionChangedEventArgs afterReadEventArgs2 = afterReadList[1];
                Assert.IsNotNull(afterReadEventArgs2);
                Assert.AreEqual(endpointDescriptor, afterReadEventArgs2.EndpointDescriptor);
                Assert.AreEqual(0, afterReadEventArgs2.ErrorCode);
                Assert.IsNull(afterReadEventArgs2.Exception);
                Assert.IsFalse(afterReadEventArgs2.Connected);
                Assert.AreEqual(ConnectionState.Disconnected, afterReadEventArgs2.ConnectionState);
                Assert.AreEqual(Timeout.Infinite, afterReadEventArgs2.RetrialDelay);
            }
            Assert.IsFalse(afterReadList.Count > 2);
        }

        #endregion



        #region Private Methods

        private void ServerConditionChangedTest(
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] Action<EasyUAClient, UAEndpointDescriptor> operation,
            [NotNull] out List<EasyUAServerConditionChangedEventArgs> operationList,
            [NotNull] out List<EasyUAServerConditionChangedEventArgs> afterOperationList)
        {
            TestSection.Arrange();

            var preConnectionList = new List<EasyUAServerConditionChangedEventArgs>();
            operationList = new List<EasyUAServerConditionChangedEventArgs>();
            afterOperationList = new List<EasyUAServerConditionChangedEventArgs>();
            var afterWaitList = new List<EasyUAServerConditionChangedEventArgs>();

            List<EasyUAServerConditionChangedEventArgs> currentList = preConnectionList;

            var client = new EasyUAClient { Isolated = true };

            void Handler(object sender, EasyUAServerConditionChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                Assert.AreSame(client, sender);
                // ReSharper disable once AccessToModifiedClosure
                currentList.Add(args);
            }

            client.ServerConditionChanged += Handler;

            TestSection.Act();

            AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.Phase("Operating");

            currentList = operationList;
            operation(client, endpointDescriptor);
            AssertUtilities.WaitingPhase(1 * 1000);

            currentList = afterOperationList;
            AssertUtilities.WaitingPhase(10 * 1000);

            currentList = afterWaitList;
            AssertUtilities.WaitingPhase(10 * 1000);

            TestSection.Assert();

            AssertUtilities.WriteLine("Pre-connection ServerConditionChanged notifications:");
            foreach (EasyUAServerConditionChangedEventArgs eventArgs in preConnectionList)
                AssertUtilities.WriteLine(eventArgs);

            AssertUtilities.WriteLine("Operation ServerConditionChanged notifications:");
            foreach (EasyUAServerConditionChangedEventArgs eventArgs in operationList)
                AssertUtilities.WriteLine(eventArgs);

            AssertUtilities.WriteLine("After-operation ServerConditionChanged notifications:");
            foreach (EasyUAServerConditionChangedEventArgs eventArgs in afterOperationList)
                AssertUtilities.WriteLine(eventArgs);

            AssertUtilities.WriteLine("After-wait ServerConditionChanged notifications:");
            foreach (EasyUAServerConditionChangedEventArgs eventArgs in afterWaitList)
                AssertUtilities.WriteLine(eventArgs);

            Assert.AreEqual(0, preConnectionList.Count);
            Assert.AreEqual(0, afterWaitList.Count);
        }

        #endregion
    }
}
