﻿using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.OperationModel
{
    public partial class DAReadItemArgumentsTest
    {
        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTestArray()
        {
            Tests.BinarySerializationTest(CreateSerializationTargetArray(), SerializationAssertionArray);
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestArray()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetArray(), SerializationAssertionArray);
        }

        [NotNull]
        private DAReadItemArguments CreateSerializationTarget0()
        {
            return new DAReadItemArguments();
        }

        [NotNull]
        private DAReadItemArguments CreateSerializationTarget1()
        {
            return new DAReadItemArguments(
                new ServerDescriptor("machine1", "company1.class1"),
                new DAItemDescriptor("branch1.item1", VarTypes.I4),
                new DAReadParameters(23456), state: 888888881);
        }

        [NotNull]
        private DAReadItemArguments CreateSerializationTarget2()
        {
            return new DAReadItemArguments("machine1", "company1.class1",
                "branch1.item2", VarTypes.R8, 12345, state: 888888882);
        }

        [NotNull]
        private DAReadItemArguments[] CreateSerializationTargetArray()
        {
            return new[] { CreateSerializationTarget0(), CreateSerializationTarget1(), CreateSerializationTarget2() };
        }

        public void SerializationAssertionArray(DAReadItemArguments[] target, DAReadItemArguments[] actual)
        {
            Check.Assert(target != null);

            Assert.IsNotNull(actual);
            Assert.IsTrue(target.SequenceEqual(actual));
        }
    }
}
