﻿using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.OperationModel
{
    public partial class EasyDAItemSubscriptionArgumentsTest
    {
        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTest2()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget2(), SerializationAssertion2);
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest2()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget2(), SerializationAssertion2);
        }

        [NotNull]
        private EasyDAItemSubscriptionArguments CreateSerializationTarget0()
        {
            return new EasyDAItemSubscriptionArguments();
        }

        [NotNull]
        private EasyDAItemSubscriptionArguments CreateSerializationTarget1()
        {
            return new EasyDAItemSubscriptionArguments(
                new ServerDescriptor("TestServerClass"),
                new DAItemDescriptor("TestItemId"),
                new DAGroupParameters(3333),
                null);
        }

        [NotNull]
        private EasyDAItemSubscriptionArguments CreateSerializationTarget2()
        {
            var arguments = CreateSerializationTarget1();
            arguments.State = "TestState";
            return arguments;
        }

        private void SerializationAssertion2(EasyDAItemSubscriptionArguments target, EasyDAItemSubscriptionArguments actual)
        {
            Check.Assert(target != null);

            Assert.IsNotNull(actual);
            Assert.AreEqual(target.State, actual.State);
        }
    }
}
