﻿using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess
{
    public partial class DAQualityTest
    {
        #region Static Private Methods

        // Helper method to obtain the dictionaries needed in test methods.
        static private void TraceFormatResults(string format)
        {
            for (int numericalValue = 0; numericalValue < 259; numericalValue++)
            {
                var quality = DAQuality.FromInt32(numericalValue);
                string result = quality.ToString(format, formatProvider: null);
                AssertUtilities.WriteLine($"{{ {numericalValue}, \"{result}\" }},");
            }
        }

        #endregion



        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_CustomFormatStrings_Example1a_Test()
        {
            Assert.AreEqual("GoodNonspecific, 0xC0", new DAQuality(DAQualities.GoodNonspecific).ToString("S, 0xN(X)"));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_CustomFormatStrings_Example1b_Test()
        {
            Assert.AreEqual("GoodNonspecific, 0xC0", $"{new DAQuality(DAQualities.GoodNonspecific):S, 0xN(X)}");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_LiteralString_Test()
        {
            Assert.AreEqual("Quality is Good", $"{new DAQuality(DAQualities.GoodNonspecific):'Quality is 'C}");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_SectionSeparator_Example1a_Test1()
        {
            Assert.AreEqual("ERROR 24", new DAQuality(DAQualities.CommFailure).ToString("OK;ERROR N"));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_SectionSeparator_Example1a_Test2()
        {
            Assert.AreEqual("OK", new DAQuality(DAQualities.GoodNonspecific).ToString("OK;ERROR N"));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_SectionSeparator_Example1b_Test1()
        {
            Assert.AreEqual("ERROR 24", $"{new DAQuality(DAQualities.CommFailure):OK;ERROR N}");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void Formatting_SectionSeparator_Example1b_Test2()
        {
            Assert.AreEqual("OK", $"{new DAQuality(DAQualities.GoodNonspecific):OK;ERROR N}");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTests_F()
        {
            Dictionary<int, string> expected = new Dictionary<int, string>
            {
                { 0, "Bad BadNonspecific LimitOk (0)" },
                { 1, "Bad BadNonspecific LimitLow (1)" },
                { 2, "Bad BadNonspecific LimitHigh (2)" },
                { 3, "Bad BadNonspecific LimitConst (3)" },
                { 4, "Bad ConfigError LimitOk (4)" },
                { 5, "Bad ConfigError LimitLow (5)" },
                { 6, "Bad ConfigError LimitHigh (6)" },
                { 7, "Bad ConfigError LimitConst (7)" },
                { 8, "Bad NotConnected LimitOk (8)" },
                { 9, "Bad NotConnected LimitLow (9)" },
                { 10, "Bad NotConnected LimitHigh (10)" },
                { 11, "Bad NotConnected LimitConst (11)" },
                { 12, "Bad DeviceFailure LimitOk (12)" },
                { 13, "Bad DeviceFailure LimitLow (13)" },
                { 14, "Bad DeviceFailure LimitHigh (14)" },
                { 15, "Bad DeviceFailure LimitConst (15)" },
                { 16, "Bad SensorFailure LimitOk (16)" },
                { 17, "Bad SensorFailure LimitLow (17)" },
                { 18, "Bad SensorFailure LimitHigh (18)" },
                { 19, "Bad SensorFailure LimitConst (19)" },
                { 20, "Bad LastKnown LimitOk (20)" },
                { 21, "Bad LastKnown LimitLow (21)" },
                { 22, "Bad LastKnown LimitHigh (22)" },
                { 23, "Bad LastKnown LimitConst (23)" },
                { 24, "Bad CommFailure LimitOk (24)" },
                { 25, "Bad CommFailure LimitLow (25)" },
                { 26, "Bad CommFailure LimitHigh (26)" },
                { 27, "Bad CommFailure LimitConst (27)" },
                { 28, "Bad OutOfService LimitOk (28)" },
                { 29, "Bad OutOfService LimitLow (29)" },
                { 30, "Bad OutOfService LimitHigh (30)" },
                { 31, "Bad OutOfService LimitConst (31)" },
                { 32, "Bad WaitingForInitialData LimitOk (32)" },
                { 33, "Bad WaitingForInitialData LimitLow (33)" },
                { 34, "Bad WaitingForInitialData LimitHigh (34)" },
                { 35, "Bad WaitingForInitialData LimitConst (35)" },
                { 36, "Bad 9 LimitOk (36)" },
                { 37, "Bad 9 LimitLow (37)" },
                { 38, "Bad 9 LimitHigh (38)" },
                { 39, "Bad 9 LimitConst (39)" },
                { 40, "Bad 10 LimitOk (40)" },
                { 41, "Bad 10 LimitLow (41)" },
                { 42, "Bad 10 LimitHigh (42)" },
                { 43, "Bad 10 LimitConst (43)" },
                { 44, "Bad 11 LimitOk (44)" },
                { 45, "Bad 11 LimitLow (45)" },
                { 46, "Bad 11 LimitHigh (46)" },
                { 47, "Bad 11 LimitConst (47)" },
                { 48, "Bad 12 LimitOk (48)" },
                { 49, "Bad 12 LimitLow (49)" },
                { 50, "Bad 12 LimitHigh (50)" },
                { 51, "Bad 12 LimitConst (51)" },
                { 52, "Bad 13 LimitOk (52)" },
                { 53, "Bad 13 LimitLow (53)" },
                { 54, "Bad 13 LimitHigh (54)" },
                { 55, "Bad 13 LimitConst (55)" },
                { 56, "Bad 14 LimitOk (56)" },
                { 57, "Bad 14 LimitLow (57)" },
                { 58, "Bad 14 LimitHigh (58)" },
                { 59, "Bad 14 LimitConst (59)" },
                { 60, "Bad 15 LimitOk (60)" },
                { 61, "Bad 15 LimitLow (61)" },
                { 62, "Bad 15 LimitHigh (62)" },
                { 63, "Bad 15 LimitConst (63)" },
                { 64, "Uncertain UncertainNonspecific LimitOk (64)" },
                { 65, "Uncertain UncertainNonspecific LimitLow (65)" },
                { 66, "Uncertain UncertainNonspecific LimitHigh (66)" },
                { 67, "Uncertain UncertainNonspecific LimitConst (67)" },
                { 68, "Uncertain LastUsable LimitOk (68)" },
                { 69, "Uncertain LastUsable LimitLow (69)" },
                { 70, "Uncertain LastUsable LimitHigh (70)" },
                { 71, "Uncertain LastUsable LimitConst (71)" },
                { 72, "Uncertain 18 LimitOk (72)" },
                { 73, "Uncertain 18 LimitLow (73)" },
                { 74, "Uncertain 18 LimitHigh (74)" },
                { 75, "Uncertain 18 LimitConst (75)" },
                { 76, "Uncertain 19 LimitOk (76)" },
                { 77, "Uncertain 19 LimitLow (77)" },
                { 78, "Uncertain 19 LimitHigh (78)" },
                { 79, "Uncertain 19 LimitConst (79)" },
                { 80, "Uncertain SensorCal LimitOk (80)" },
                { 81, "Uncertain SensorCal LimitLow (81)" },
                { 82, "Uncertain SensorCal LimitHigh (82)" },
                { 83, "Uncertain SensorCal LimitConst (83)" },
                { 84, "Uncertain EguExceeded LimitOk (84)" },
                { 85, "Uncertain EguExceeded LimitLow (85)" },
                { 86, "Uncertain EguExceeded LimitHigh (86)" },
                { 87, "Uncertain EguExceeded LimitConst (87)" },
                { 88, "Uncertain Subnormal LimitOk (88)" },
                { 89, "Uncertain Subnormal LimitLow (89)" },
                { 90, "Uncertain Subnormal LimitHigh (90)" },
                { 91, "Uncertain Subnormal LimitConst (91)" },
                { 92, "Uncertain 23 LimitOk (92)" },
                { 93, "Uncertain 23 LimitLow (93)" },
                { 94, "Uncertain 23 LimitHigh (94)" },
                { 95, "Uncertain 23 LimitConst (95)" },
                { 96, "Uncertain 24 LimitOk (96)" },
                { 97, "Uncertain 24 LimitLow (97)" },
                { 98, "Uncertain 24 LimitHigh (98)" },
                { 99, "Uncertain 24 LimitConst (99)" },
                { 100, "Uncertain 25 LimitOk (100)" },
                { 101, "Uncertain 25 LimitLow (101)" },
                { 102, "Uncertain 25 LimitHigh (102)" },
                { 103, "Uncertain 25 LimitConst (103)" },
                { 104, "Uncertain 26 LimitOk (104)" },
                { 105, "Uncertain 26 LimitLow (105)" },
                { 106, "Uncertain 26 LimitHigh (106)" },
                { 107, "Uncertain 26 LimitConst (107)" },
                { 108, "Uncertain 27 LimitOk (108)" },
                { 109, "Uncertain 27 LimitLow (109)" },
                { 110, "Uncertain 27 LimitHigh (110)" },
                { 111, "Uncertain 27 LimitConst (111)" },
                { 112, "Uncertain 28 LimitOk (112)" },
                { 113, "Uncertain 28 LimitLow (113)" },
                { 114, "Uncertain 28 LimitHigh (114)" },
                { 115, "Uncertain 28 LimitConst (115)" },
                { 116, "Uncertain 29 LimitOk (116)" },
                { 117, "Uncertain 29 LimitLow (117)" },
                { 118, "Uncertain 29 LimitHigh (118)" },
                { 119, "Uncertain 29 LimitConst (119)" },
                { 120, "Uncertain 30 LimitOk (120)" },
                { 121, "Uncertain 30 LimitLow (121)" },
                { 122, "Uncertain 30 LimitHigh (122)" },
                { 123, "Uncertain 30 LimitConst (123)" },
                { 124, "Uncertain 31 LimitOk (124)" },
                { 125, "Uncertain 31 LimitLow (125)" },
                { 126, "Uncertain 31 LimitHigh (126)" },
                { 127, "Uncertain 31 LimitConst (127)" },
                { 128, "2 32 LimitOk (128)" },
                { 129, "2 32 LimitLow (129)" },
                { 130, "2 32 LimitHigh (130)" },
                { 131, "2 32 LimitConst (131)" },
                { 132, "2 33 LimitOk (132)" },
                { 133, "2 33 LimitLow (133)" },
                { 134, "2 33 LimitHigh (134)" },
                { 135, "2 33 LimitConst (135)" },
                { 136, "2 34 LimitOk (136)" },
                { 137, "2 34 LimitLow (137)" },
                { 138, "2 34 LimitHigh (138)" },
                { 139, "2 34 LimitConst (139)" },
                { 140, "2 35 LimitOk (140)" },
                { 141, "2 35 LimitLow (141)" },
                { 142, "2 35 LimitHigh (142)" },
                { 143, "2 35 LimitConst (143)" },
                { 144, "2 36 LimitOk (144)" },
                { 145, "2 36 LimitLow (145)" },
                { 146, "2 36 LimitHigh (146)" },
                { 147, "2 36 LimitConst (147)" },
                { 148, "2 37 LimitOk (148)" },
                { 149, "2 37 LimitLow (149)" },
                { 150, "2 37 LimitHigh (150)" },
                { 151, "2 37 LimitConst (151)" },
                { 152, "2 38 LimitOk (152)" },
                { 153, "2 38 LimitLow (153)" },
                { 154, "2 38 LimitHigh (154)" },
                { 155, "2 38 LimitConst (155)" },
                { 156, "2 39 LimitOk (156)" },
                { 157, "2 39 LimitLow (157)" },
                { 158, "2 39 LimitHigh (158)" },
                { 159, "2 39 LimitConst (159)" },
                { 160, "2 40 LimitOk (160)" },
                { 161, "2 40 LimitLow (161)" },
                { 162, "2 40 LimitHigh (162)" },
                { 163, "2 40 LimitConst (163)" },
                { 164, "2 41 LimitOk (164)" },
                { 165, "2 41 LimitLow (165)" },
                { 166, "2 41 LimitHigh (166)" },
                { 167, "2 41 LimitConst (167)" },
                { 168, "2 42 LimitOk (168)" },
                { 169, "2 42 LimitLow (169)" },
                { 170, "2 42 LimitHigh (170)" },
                { 171, "2 42 LimitConst (171)" },
                { 172, "2 43 LimitOk (172)" },
                { 173, "2 43 LimitLow (173)" },
                { 174, "2 43 LimitHigh (174)" },
                { 175, "2 43 LimitConst (175)" },
                { 176, "2 44 LimitOk (176)" },
                { 177, "2 44 LimitLow (177)" },
                { 178, "2 44 LimitHigh (178)" },
                { 179, "2 44 LimitConst (179)" },
                { 180, "2 45 LimitOk (180)" },
                { 181, "2 45 LimitLow (181)" },
                { 182, "2 45 LimitHigh (182)" },
                { 183, "2 45 LimitConst (183)" },
                { 184, "2 46 LimitOk (184)" },
                { 185, "2 46 LimitLow (185)" },
                { 186, "2 46 LimitHigh (186)" },
                { 187, "2 46 LimitConst (187)" },
                { 188, "2 47 LimitOk (188)" },
                { 189, "2 47 LimitLow (189)" },
                { 190, "2 47 LimitHigh (190)" },
                { 191, "2 47 LimitConst (191)" },
                { 192, "Good GoodNonspecific LimitOk (192)" },
                { 193, "Good GoodNonspecific LimitLow (193)" },
                { 194, "Good GoodNonspecific LimitHigh (194)" },
                { 195, "Good GoodNonspecific LimitConst (195)" },
                { 196, "Good 49 LimitOk (196)" },
                { 197, "Good 49 LimitLow (197)" },
                { 198, "Good 49 LimitHigh (198)" },
                { 199, "Good 49 LimitConst (199)" },
                { 200, "Good 50 LimitOk (200)" },
                { 201, "Good 50 LimitLow (201)" },
                { 202, "Good 50 LimitHigh (202)" },
                { 203, "Good 50 LimitConst (203)" },
                { 204, "Good 51 LimitOk (204)" },
                { 205, "Good 51 LimitLow (205)" },
                { 206, "Good 51 LimitHigh (206)" },
                { 207, "Good 51 LimitConst (207)" },
                { 208, "Good 52 LimitOk (208)" },
                { 209, "Good 52 LimitLow (209)" },
                { 210, "Good 52 LimitHigh (210)" },
                { 211, "Good 52 LimitConst (211)" },
                { 212, "Good 53 LimitOk (212)" },
                { 213, "Good 53 LimitLow (213)" },
                { 214, "Good 53 LimitHigh (214)" },
                { 215, "Good 53 LimitConst (215)" },
                { 216, "Good LocalOverride LimitOk (216)" },
                { 217, "Good LocalOverride LimitLow (217)" },
                { 218, "Good LocalOverride LimitHigh (218)" },
                { 219, "Good LocalOverride LimitConst (219)" },
                { 220, "Good 55 LimitOk (220)" },
                { 221, "Good 55 LimitLow (221)" },
                { 222, "Good 55 LimitHigh (222)" },
                { 223, "Good 55 LimitConst (223)" },
                { 224, "Good 56 LimitOk (224)" },
                { 225, "Good 56 LimitLow (225)" },
                { 226, "Good 56 LimitHigh (226)" },
                { 227, "Good 56 LimitConst (227)" },
                { 228, "Good 57 LimitOk (228)" },
                { 229, "Good 57 LimitLow (229)" },
                { 230, "Good 57 LimitHigh (230)" },
                { 231, "Good 57 LimitConst (231)" },
                { 232, "Good 58 LimitOk (232)" },
                { 233, "Good 58 LimitLow (233)" },
                { 234, "Good 58 LimitHigh (234)" },
                { 235, "Good 58 LimitConst (235)" },
                { 236, "Good 59 LimitOk (236)" },
                { 237, "Good 59 LimitLow (237)" },
                { 238, "Good 59 LimitHigh (238)" },
                { 239, "Good 59 LimitConst (239)" },
                { 240, "Good 60 LimitOk (240)" },
                { 241, "Good 60 LimitLow (241)" },
                { 242, "Good 60 LimitHigh (242)" },
                { 243, "Good 60 LimitConst (243)" },
                { 244, "Good 61 LimitOk (244)" },
                { 245, "Good 61 LimitLow (245)" },
                { 246, "Good 61 LimitHigh (246)" },
                { 247, "Good 61 LimitConst (247)" },
                { 248, "Good 62 LimitOk (248)" },
                { 249, "Good 62 LimitLow (249)" },
                { 250, "Good 62 LimitHigh (250)" },
                { 251, "Good 62 LimitConst (251)" },
                { 252, "Good 63 LimitOk (252)" },
                { 253, "Good 63 LimitLow (253)" },
                { 254, "Good 63 LimitHigh (254)" },
                { 255, "Good 63 LimitConst (255)" },
                { 256, "Bad BadNonspecific LimitOk (256)" },
                { 257, "Bad BadNonspecific LimitLow (257)" },
                { 258, "Bad BadNonspecific LimitHigh (258)" },
            };

            foreach (KeyValuePair<int, string> pair in expected)
            {
                int numericalValue = pair.Key;
                var quality = DAQuality.FromInt32(numericalValue);
                string result = quality.ToString("F", formatProvider: null);
                Assert.AreEqual(pair.Value, result);
            }
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTests_f()
        {
            Dictionary<int, string> expected = new Dictionary<int, string>
            {
                { 0, "Bad BadNonspecific LimitOk" },
                { 1, "Bad BadNonspecific LimitLow" },
                { 2, "Bad BadNonspecific LimitHigh" },
                { 3, "Bad BadNonspecific LimitConst" },
                { 4, "Bad ConfigError LimitOk" },
                { 5, "Bad ConfigError LimitLow" },
                { 6, "Bad ConfigError LimitHigh" },
                { 7, "Bad ConfigError LimitConst" },
                { 8, "Bad NotConnected LimitOk" },
                { 9, "Bad NotConnected LimitLow" },
                { 10, "Bad NotConnected LimitHigh" },
                { 11, "Bad NotConnected LimitConst" },
                { 12, "Bad DeviceFailure LimitOk" },
                { 13, "Bad DeviceFailure LimitLow" },
                { 14, "Bad DeviceFailure LimitHigh" },
                { 15, "Bad DeviceFailure LimitConst" },
                { 16, "Bad SensorFailure LimitOk" },
                { 17, "Bad SensorFailure LimitLow" },
                { 18, "Bad SensorFailure LimitHigh" },
                { 19, "Bad SensorFailure LimitConst" },
                { 20, "Bad LastKnown LimitOk" },
                { 21, "Bad LastKnown LimitLow" },
                { 22, "Bad LastKnown LimitHigh" },
                { 23, "Bad LastKnown LimitConst" },
                { 24, "Bad CommFailure LimitOk" },
                { 25, "Bad CommFailure LimitLow" },
                { 26, "Bad CommFailure LimitHigh" },
                { 27, "Bad CommFailure LimitConst" },
                { 28, "Bad OutOfService LimitOk" },
                { 29, "Bad OutOfService LimitLow" },
                { 30, "Bad OutOfService LimitHigh" },
                { 31, "Bad OutOfService LimitConst" },
                { 32, "Bad WaitingForInitialData LimitOk" },
                { 33, "Bad WaitingForInitialData LimitLow" },
                { 34, "Bad WaitingForInitialData LimitHigh" },
                { 35, "Bad WaitingForInitialData LimitConst" },
                { 36, "Bad 9 LimitOk" },
                { 37, "Bad 9 LimitLow" },
                { 38, "Bad 9 LimitHigh" },
                { 39, "Bad 9 LimitConst" },
                { 40, "Bad 10 LimitOk" },
                { 41, "Bad 10 LimitLow" },
                { 42, "Bad 10 LimitHigh" },
                { 43, "Bad 10 LimitConst" },
                { 44, "Bad 11 LimitOk" },
                { 45, "Bad 11 LimitLow" },
                { 46, "Bad 11 LimitHigh" },
                { 47, "Bad 11 LimitConst" },
                { 48, "Bad 12 LimitOk" },
                { 49, "Bad 12 LimitLow" },
                { 50, "Bad 12 LimitHigh" },
                { 51, "Bad 12 LimitConst" },
                { 52, "Bad 13 LimitOk" },
                { 53, "Bad 13 LimitLow" },
                { 54, "Bad 13 LimitHigh" },
                { 55, "Bad 13 LimitConst" },
                { 56, "Bad 14 LimitOk" },
                { 57, "Bad 14 LimitLow" },
                { 58, "Bad 14 LimitHigh" },
                { 59, "Bad 14 LimitConst" },
                { 60, "Bad 15 LimitOk" },
                { 61, "Bad 15 LimitLow" },
                { 62, "Bad 15 LimitHigh" },
                { 63, "Bad 15 LimitConst" },
                { 64, "Uncertain UncertainNonspecific LimitOk" },
                { 65, "Uncertain UncertainNonspecific LimitLow" },
                { 66, "Uncertain UncertainNonspecific LimitHigh" },
                { 67, "Uncertain UncertainNonspecific LimitConst" },
                { 68, "Uncertain LastUsable LimitOk" },
                { 69, "Uncertain LastUsable LimitLow" },
                { 70, "Uncertain LastUsable LimitHigh" },
                { 71, "Uncertain LastUsable LimitConst" },
                { 72, "Uncertain 18 LimitOk" },
                { 73, "Uncertain 18 LimitLow" },
                { 74, "Uncertain 18 LimitHigh" },
                { 75, "Uncertain 18 LimitConst" },
                { 76, "Uncertain 19 LimitOk" },
                { 77, "Uncertain 19 LimitLow" },
                { 78, "Uncertain 19 LimitHigh" },
                { 79, "Uncertain 19 LimitConst" },
                { 80, "Uncertain SensorCal LimitOk" },
                { 81, "Uncertain SensorCal LimitLow" },
                { 82, "Uncertain SensorCal LimitHigh" },
                { 83, "Uncertain SensorCal LimitConst" },
                { 84, "Uncertain EguExceeded LimitOk" },
                { 85, "Uncertain EguExceeded LimitLow" },
                { 86, "Uncertain EguExceeded LimitHigh" },
                { 87, "Uncertain EguExceeded LimitConst" },
                { 88, "Uncertain Subnormal LimitOk" },
                { 89, "Uncertain Subnormal LimitLow" },
                { 90, "Uncertain Subnormal LimitHigh" },
                { 91, "Uncertain Subnormal LimitConst" },
                { 92, "Uncertain 23 LimitOk" },
                { 93, "Uncertain 23 LimitLow" },
                { 94, "Uncertain 23 LimitHigh" },
                { 95, "Uncertain 23 LimitConst" },
                { 96, "Uncertain 24 LimitOk" },
                { 97, "Uncertain 24 LimitLow" },
                { 98, "Uncertain 24 LimitHigh" },
                { 99, "Uncertain 24 LimitConst" },
                { 100, "Uncertain 25 LimitOk" },
                { 101, "Uncertain 25 LimitLow" },
                { 102, "Uncertain 25 LimitHigh" },
                { 103, "Uncertain 25 LimitConst" },
                { 104, "Uncertain 26 LimitOk" },
                { 105, "Uncertain 26 LimitLow" },
                { 106, "Uncertain 26 LimitHigh" },
                { 107, "Uncertain 26 LimitConst" },
                { 108, "Uncertain 27 LimitOk" },
                { 109, "Uncertain 27 LimitLow" },
                { 110, "Uncertain 27 LimitHigh" },
                { 111, "Uncertain 27 LimitConst" },
                { 112, "Uncertain 28 LimitOk" },
                { 113, "Uncertain 28 LimitLow" },
                { 114, "Uncertain 28 LimitHigh" },
                { 115, "Uncertain 28 LimitConst" },
                { 116, "Uncertain 29 LimitOk" },
                { 117, "Uncertain 29 LimitLow" },
                { 118, "Uncertain 29 LimitHigh" },
                { 119, "Uncertain 29 LimitConst" },
                { 120, "Uncertain 30 LimitOk" },
                { 121, "Uncertain 30 LimitLow" },
                { 122, "Uncertain 30 LimitHigh" },
                { 123, "Uncertain 30 LimitConst" },
                { 124, "Uncertain 31 LimitOk" },
                { 125, "Uncertain 31 LimitLow" },
                { 126, "Uncertain 31 LimitHigh" },
                { 127, "Uncertain 31 LimitConst" },
                { 128, "2 32 LimitOk" },
                { 129, "2 32 LimitLow" },
                { 130, "2 32 LimitHigh" },
                { 131, "2 32 LimitConst" },
                { 132, "2 33 LimitOk" },
                { 133, "2 33 LimitLow" },
                { 134, "2 33 LimitHigh" },
                { 135, "2 33 LimitConst" },
                { 136, "2 34 LimitOk" },
                { 137, "2 34 LimitLow" },
                { 138, "2 34 LimitHigh" },
                { 139, "2 34 LimitConst" },
                { 140, "2 35 LimitOk" },
                { 141, "2 35 LimitLow" },
                { 142, "2 35 LimitHigh" },
                { 143, "2 35 LimitConst" },
                { 144, "2 36 LimitOk" },
                { 145, "2 36 LimitLow" },
                { 146, "2 36 LimitHigh" },
                { 147, "2 36 LimitConst" },
                { 148, "2 37 LimitOk" },
                { 149, "2 37 LimitLow" },
                { 150, "2 37 LimitHigh" },
                { 151, "2 37 LimitConst" },
                { 152, "2 38 LimitOk" },
                { 153, "2 38 LimitLow" },
                { 154, "2 38 LimitHigh" },
                { 155, "2 38 LimitConst" },
                { 156, "2 39 LimitOk" },
                { 157, "2 39 LimitLow" },
                { 158, "2 39 LimitHigh" },
                { 159, "2 39 LimitConst" },
                { 160, "2 40 LimitOk" },
                { 161, "2 40 LimitLow" },
                { 162, "2 40 LimitHigh" },
                { 163, "2 40 LimitConst" },
                { 164, "2 41 LimitOk" },
                { 165, "2 41 LimitLow" },
                { 166, "2 41 LimitHigh" },
                { 167, "2 41 LimitConst" },
                { 168, "2 42 LimitOk" },
                { 169, "2 42 LimitLow" },
                { 170, "2 42 LimitHigh" },
                { 171, "2 42 LimitConst" },
                { 172, "2 43 LimitOk" },
                { 173, "2 43 LimitLow" },
                { 174, "2 43 LimitHigh" },
                { 175, "2 43 LimitConst" },
                { 176, "2 44 LimitOk" },
                { 177, "2 44 LimitLow" },
                { 178, "2 44 LimitHigh" },
                { 179, "2 44 LimitConst" },
                { 180, "2 45 LimitOk" },
                { 181, "2 45 LimitLow" },
                { 182, "2 45 LimitHigh" },
                { 183, "2 45 LimitConst" },
                { 184, "2 46 LimitOk" },
                { 185, "2 46 LimitLow" },
                { 186, "2 46 LimitHigh" },
                { 187, "2 46 LimitConst" },
                { 188, "2 47 LimitOk" },
                { 189, "2 47 LimitLow" },
                { 190, "2 47 LimitHigh" },
                { 191, "2 47 LimitConst" },
                { 192, "Good GoodNonspecific LimitOk" },
                { 193, "Good GoodNonspecific LimitLow" },
                { 194, "Good GoodNonspecific LimitHigh" },
                { 195, "Good GoodNonspecific LimitConst" },
                { 196, "Good 49 LimitOk" },
                { 197, "Good 49 LimitLow" },
                { 198, "Good 49 LimitHigh" },
                { 199, "Good 49 LimitConst" },
                { 200, "Good 50 LimitOk" },
                { 201, "Good 50 LimitLow" },
                { 202, "Good 50 LimitHigh" },
                { 203, "Good 50 LimitConst" },
                { 204, "Good 51 LimitOk" },
                { 205, "Good 51 LimitLow" },
                { 206, "Good 51 LimitHigh" },
                { 207, "Good 51 LimitConst" },
                { 208, "Good 52 LimitOk" },
                { 209, "Good 52 LimitLow" },
                { 210, "Good 52 LimitHigh" },
                { 211, "Good 52 LimitConst" },
                { 212, "Good 53 LimitOk" },
                { 213, "Good 53 LimitLow" },
                { 214, "Good 53 LimitHigh" },
                { 215, "Good 53 LimitConst" },
                { 216, "Good LocalOverride LimitOk" },
                { 217, "Good LocalOverride LimitLow" },
                { 218, "Good LocalOverride LimitHigh" },
                { 219, "Good LocalOverride LimitConst" },
                { 220, "Good 55 LimitOk" },
                { 221, "Good 55 LimitLow" },
                { 222, "Good 55 LimitHigh" },
                { 223, "Good 55 LimitConst" },
                { 224, "Good 56 LimitOk" },
                { 225, "Good 56 LimitLow" },
                { 226, "Good 56 LimitHigh" },
                { 227, "Good 56 LimitConst" },
                { 228, "Good 57 LimitOk" },
                { 229, "Good 57 LimitLow" },
                { 230, "Good 57 LimitHigh" },
                { 231, "Good 57 LimitConst" },
                { 232, "Good 58 LimitOk" },
                { 233, "Good 58 LimitLow" },
                { 234, "Good 58 LimitHigh" },
                { 235, "Good 58 LimitConst" },
                { 236, "Good 59 LimitOk" },
                { 237, "Good 59 LimitLow" },
                { 238, "Good 59 LimitHigh" },
                { 239, "Good 59 LimitConst" },
                { 240, "Good 60 LimitOk" },
                { 241, "Good 60 LimitLow" },
                { 242, "Good 60 LimitHigh" },
                { 243, "Good 60 LimitConst" },
                { 244, "Good 61 LimitOk" },
                { 245, "Good 61 LimitLow" },
                { 246, "Good 61 LimitHigh" },
                { 247, "Good 61 LimitConst" },
                { 248, "Good 62 LimitOk" },
                { 249, "Good 62 LimitLow" },
                { 250, "Good 62 LimitHigh" },
                { 251, "Good 62 LimitConst" },
                { 252, "Good 63 LimitOk" },
                { 253, "Good 63 LimitLow" },
                { 254, "Good 63 LimitHigh" },
                { 255, "Good 63 LimitConst" },
                { 256, "Bad BadNonspecific LimitOk" },
                { 257, "Bad BadNonspecific LimitLow" },
                { 258, "Bad BadNonspecific LimitHigh" },            
            };

            foreach (KeyValuePair<int, string> pair in expected)
            {
                int numericalValue = pair.Key;
                var quality = DAQuality.FromInt32(numericalValue);
                string result = quality.ToString("f", formatProvider: null);
                Assert.AreEqual(pair.Value, result);
            }
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTests_G()
        {
            Dictionary<int, string> expected = new Dictionary<int, string>
            {
                { 0, "BadNonspecific LimitOk (0)" },
                { 1, "BadNonspecific LimitLow (1)" },
                { 2, "BadNonspecific LimitHigh (2)" },
                { 3, "BadNonspecific LimitConst (3)" },
                { 4, "ConfigError LimitOk (4)" },
                { 5, "ConfigError LimitLow (5)" },
                { 6, "ConfigError LimitHigh (6)" },
                { 7, "ConfigError LimitConst (7)" },
                { 8, "NotConnected LimitOk (8)" },
                { 9, "NotConnected LimitLow (9)" },
                { 10, "NotConnected LimitHigh (10)" },
                { 11, "NotConnected LimitConst (11)" },
                { 12, "DeviceFailure LimitOk (12)" },
                { 13, "DeviceFailure LimitLow (13)" },
                { 14, "DeviceFailure LimitHigh (14)" },
                { 15, "DeviceFailure LimitConst (15)" },
                { 16, "SensorFailure LimitOk (16)" },
                { 17, "SensorFailure LimitLow (17)" },
                { 18, "SensorFailure LimitHigh (18)" },
                { 19, "SensorFailure LimitConst (19)" },
                { 20, "LastKnown LimitOk (20)" },
                { 21, "LastKnown LimitLow (21)" },
                { 22, "LastKnown LimitHigh (22)" },
                { 23, "LastKnown LimitConst (23)" },
                { 24, "CommFailure LimitOk (24)" },
                { 25, "CommFailure LimitLow (25)" },
                { 26, "CommFailure LimitHigh (26)" },
                { 27, "CommFailure LimitConst (27)" },
                { 28, "OutOfService LimitOk (28)" },
                { 29, "OutOfService LimitLow (29)" },
                { 30, "OutOfService LimitHigh (30)" },
                { 31, "OutOfService LimitConst (31)" },
                { 32, "WaitingForInitialData LimitOk (32)" },
                { 33, "WaitingForInitialData LimitLow (33)" },
                { 34, "WaitingForInitialData LimitHigh (34)" },
                { 35, "WaitingForInitialData LimitConst (35)" },
                { 36, "9 LimitOk (36)" },
                { 37, "9 LimitLow (37)" },
                { 38, "9 LimitHigh (38)" },
                { 39, "9 LimitConst (39)" },
                { 40, "10 LimitOk (40)" },
                { 41, "10 LimitLow (41)" },
                { 42, "10 LimitHigh (42)" },
                { 43, "10 LimitConst (43)" },
                { 44, "11 LimitOk (44)" },
                { 45, "11 LimitLow (45)" },
                { 46, "11 LimitHigh (46)" },
                { 47, "11 LimitConst (47)" },
                { 48, "12 LimitOk (48)" },
                { 49, "12 LimitLow (49)" },
                { 50, "12 LimitHigh (50)" },
                { 51, "12 LimitConst (51)" },
                { 52, "13 LimitOk (52)" },
                { 53, "13 LimitLow (53)" },
                { 54, "13 LimitHigh (54)" },
                { 55, "13 LimitConst (55)" },
                { 56, "14 LimitOk (56)" },
                { 57, "14 LimitLow (57)" },
                { 58, "14 LimitHigh (58)" },
                { 59, "14 LimitConst (59)" },
                { 60, "15 LimitOk (60)" },
                { 61, "15 LimitLow (61)" },
                { 62, "15 LimitHigh (62)" },
                { 63, "15 LimitConst (63)" },
                { 64, "UncertainNonspecific LimitOk (64)" },
                { 65, "UncertainNonspecific LimitLow (65)" },
                { 66, "UncertainNonspecific LimitHigh (66)" },
                { 67, "UncertainNonspecific LimitConst (67)" },
                { 68, "LastUsable LimitOk (68)" },
                { 69, "LastUsable LimitLow (69)" },
                { 70, "LastUsable LimitHigh (70)" },
                { 71, "LastUsable LimitConst (71)" },
                { 72, "18 LimitOk (72)" },
                { 73, "18 LimitLow (73)" },
                { 74, "18 LimitHigh (74)" },
                { 75, "18 LimitConst (75)" },
                { 76, "19 LimitOk (76)" },
                { 77, "19 LimitLow (77)" },
                { 78, "19 LimitHigh (78)" },
                { 79, "19 LimitConst (79)" },
                { 80, "SensorCal LimitOk (80)" },
                { 81, "SensorCal LimitLow (81)" },
                { 82, "SensorCal LimitHigh (82)" },
                { 83, "SensorCal LimitConst (83)" },
                { 84, "EguExceeded LimitOk (84)" },
                { 85, "EguExceeded LimitLow (85)" },
                { 86, "EguExceeded LimitHigh (86)" },
                { 87, "EguExceeded LimitConst (87)" },
                { 88, "Subnormal LimitOk (88)" },
                { 89, "Subnormal LimitLow (89)" },
                { 90, "Subnormal LimitHigh (90)" },
                { 91, "Subnormal LimitConst (91)" },
                { 92, "23 LimitOk (92)" },
                { 93, "23 LimitLow (93)" },
                { 94, "23 LimitHigh (94)" },
                { 95, "23 LimitConst (95)" },
                { 96, "24 LimitOk (96)" },
                { 97, "24 LimitLow (97)" },
                { 98, "24 LimitHigh (98)" },
                { 99, "24 LimitConst (99)" },
                { 100, "25 LimitOk (100)" },
                { 101, "25 LimitLow (101)" },
                { 102, "25 LimitHigh (102)" },
                { 103, "25 LimitConst (103)" },
                { 104, "26 LimitOk (104)" },
                { 105, "26 LimitLow (105)" },
                { 106, "26 LimitHigh (106)" },
                { 107, "26 LimitConst (107)" },
                { 108, "27 LimitOk (108)" },
                { 109, "27 LimitLow (109)" },
                { 110, "27 LimitHigh (110)" },
                { 111, "27 LimitConst (111)" },
                { 112, "28 LimitOk (112)" },
                { 113, "28 LimitLow (113)" },
                { 114, "28 LimitHigh (114)" },
                { 115, "28 LimitConst (115)" },
                { 116, "29 LimitOk (116)" },
                { 117, "29 LimitLow (117)" },
                { 118, "29 LimitHigh (118)" },
                { 119, "29 LimitConst (119)" },
                { 120, "30 LimitOk (120)" },
                { 121, "30 LimitLow (121)" },
                { 122, "30 LimitHigh (122)" },
                { 123, "30 LimitConst (123)" },
                { 124, "31 LimitOk (124)" },
                { 125, "31 LimitLow (125)" },
                { 126, "31 LimitHigh (126)" },
                { 127, "31 LimitConst (127)" },
                { 128, "32 LimitOk (128)" },
                { 129, "32 LimitLow (129)" },
                { 130, "32 LimitHigh (130)" },
                { 131, "32 LimitConst (131)" },
                { 132, "33 LimitOk (132)" },
                { 133, "33 LimitLow (133)" },
                { 134, "33 LimitHigh (134)" },
                { 135, "33 LimitConst (135)" },
                { 136, "34 LimitOk (136)" },
                { 137, "34 LimitLow (137)" },
                { 138, "34 LimitHigh (138)" },
                { 139, "34 LimitConst (139)" },
                { 140, "35 LimitOk (140)" },
                { 141, "35 LimitLow (141)" },
                { 142, "35 LimitHigh (142)" },
                { 143, "35 LimitConst (143)" },
                { 144, "36 LimitOk (144)" },
                { 145, "36 LimitLow (145)" },
                { 146, "36 LimitHigh (146)" },
                { 147, "36 LimitConst (147)" },
                { 148, "37 LimitOk (148)" },
                { 149, "37 LimitLow (149)" },
                { 150, "37 LimitHigh (150)" },
                { 151, "37 LimitConst (151)" },
                { 152, "38 LimitOk (152)" },
                { 153, "38 LimitLow (153)" },
                { 154, "38 LimitHigh (154)" },
                { 155, "38 LimitConst (155)" },
                { 156, "39 LimitOk (156)" },
                { 157, "39 LimitLow (157)" },
                { 158, "39 LimitHigh (158)" },
                { 159, "39 LimitConst (159)" },
                { 160, "40 LimitOk (160)" },
                { 161, "40 LimitLow (161)" },
                { 162, "40 LimitHigh (162)" },
                { 163, "40 LimitConst (163)" },
                { 164, "41 LimitOk (164)" },
                { 165, "41 LimitLow (165)" },
                { 166, "41 LimitHigh (166)" },
                { 167, "41 LimitConst (167)" },
                { 168, "42 LimitOk (168)" },
                { 169, "42 LimitLow (169)" },
                { 170, "42 LimitHigh (170)" },
                { 171, "42 LimitConst (171)" },
                { 172, "43 LimitOk (172)" },
                { 173, "43 LimitLow (173)" },
                { 174, "43 LimitHigh (174)" },
                { 175, "43 LimitConst (175)" },
                { 176, "44 LimitOk (176)" },
                { 177, "44 LimitLow (177)" },
                { 178, "44 LimitHigh (178)" },
                { 179, "44 LimitConst (179)" },
                { 180, "45 LimitOk (180)" },
                { 181, "45 LimitLow (181)" },
                { 182, "45 LimitHigh (182)" },
                { 183, "45 LimitConst (183)" },
                { 184, "46 LimitOk (184)" },
                { 185, "46 LimitLow (185)" },
                { 186, "46 LimitHigh (186)" },
                { 187, "46 LimitConst (187)" },
                { 188, "47 LimitOk (188)" },
                { 189, "47 LimitLow (189)" },
                { 190, "47 LimitHigh (190)" },
                { 191, "47 LimitConst (191)" },
                { 192, "GoodNonspecific (192)" },
                { 193, "GoodNonspecific LimitLow (193)" },
                { 194, "GoodNonspecific LimitHigh (194)" },
                { 195, "GoodNonspecific LimitConst (195)" },
                { 196, "49 LimitOk (196)" },
                { 197, "49 LimitLow (197)" },
                { 198, "49 LimitHigh (198)" },
                { 199, "49 LimitConst (199)" },
                { 200, "50 LimitOk (200)" },
                { 201, "50 LimitLow (201)" },
                { 202, "50 LimitHigh (202)" },
                { 203, "50 LimitConst (203)" },
                { 204, "51 LimitOk (204)" },
                { 205, "51 LimitLow (205)" },
                { 206, "51 LimitHigh (206)" },
                { 207, "51 LimitConst (207)" },
                { 208, "52 LimitOk (208)" },
                { 209, "52 LimitLow (209)" },
                { 210, "52 LimitHigh (210)" },
                { 211, "52 LimitConst (211)" },
                { 212, "53 LimitOk (212)" },
                { 213, "53 LimitLow (213)" },
                { 214, "53 LimitHigh (214)" },
                { 215, "53 LimitConst (215)" },
                { 216, "LocalOverride LimitOk (216)" },
                { 217, "LocalOverride LimitLow (217)" },
                { 218, "LocalOverride LimitHigh (218)" },
                { 219, "LocalOverride LimitConst (219)" },
                { 220, "55 LimitOk (220)" },
                { 221, "55 LimitLow (221)" },
                { 222, "55 LimitHigh (222)" },
                { 223, "55 LimitConst (223)" },
                { 224, "56 LimitOk (224)" },
                { 225, "56 LimitLow (225)" },
                { 226, "56 LimitHigh (226)" },
                { 227, "56 LimitConst (227)" },
                { 228, "57 LimitOk (228)" },
                { 229, "57 LimitLow (229)" },
                { 230, "57 LimitHigh (230)" },
                { 231, "57 LimitConst (231)" },
                { 232, "58 LimitOk (232)" },
                { 233, "58 LimitLow (233)" },
                { 234, "58 LimitHigh (234)" },
                { 235, "58 LimitConst (235)" },
                { 236, "59 LimitOk (236)" },
                { 237, "59 LimitLow (237)" },
                { 238, "59 LimitHigh (238)" },
                { 239, "59 LimitConst (239)" },
                { 240, "60 LimitOk (240)" },
                { 241, "60 LimitLow (241)" },
                { 242, "60 LimitHigh (242)" },
                { 243, "60 LimitConst (243)" },
                { 244, "61 LimitOk (244)" },
                { 245, "61 LimitLow (245)" },
                { 246, "61 LimitHigh (246)" },
                { 247, "61 LimitConst (247)" },
                { 248, "62 LimitOk (248)" },
                { 249, "62 LimitLow (249)" },
                { 250, "62 LimitHigh (250)" },
                { 251, "62 LimitConst (251)" },
                { 252, "63 LimitOk (252)" },
                { 253, "63 LimitLow (253)" },
                { 254, "63 LimitHigh (254)" },
                { 255, "63 LimitConst (255)" },
                { 256, "BadNonspecific LimitOk (256)" },
                { 257, "BadNonspecific LimitLow (257)" },
                { 258, "BadNonspecific LimitHigh (258)" },
            };

            foreach (KeyValuePair<int, string> pair in expected)
            {
                int numericalValue = pair.Key;
                var quality = DAQuality.FromInt32(numericalValue);
                string result = quality.ToString("G", formatProvider: null);
                Assert.AreEqual(pair.Value, result);
            }
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTests_g()
        {
            Dictionary<int, string> expected = new Dictionary<int, string>
            {
                {0, "BadNonspecific LimitOk"},
                {1, "BadNonspecific LimitLow"},
                {2, "BadNonspecific LimitHigh"},
                {3, "BadNonspecific LimitConst"},
                {4, "ConfigError LimitOk"},
                {5, "ConfigError LimitLow"},
                {6, "ConfigError LimitHigh"},
                {7, "ConfigError LimitConst"},
                {8, "NotConnected LimitOk"},
                {9, "NotConnected LimitLow"},
                {10, "NotConnected LimitHigh"},
                {11, "NotConnected LimitConst"},
                {12, "DeviceFailure LimitOk"},
                {13, "DeviceFailure LimitLow"},
                {14, "DeviceFailure LimitHigh"},
                {15, "DeviceFailure LimitConst"},
                {16, "SensorFailure LimitOk"},
                {17, "SensorFailure LimitLow"},
                {18, "SensorFailure LimitHigh"},
                {19, "SensorFailure LimitConst"},
                {20, "LastKnown LimitOk"},
                {21, "LastKnown LimitLow"},
                {22, "LastKnown LimitHigh"},
                {23, "LastKnown LimitConst"},
                {24, "CommFailure LimitOk"},
                {25, "CommFailure LimitLow"},
                {26, "CommFailure LimitHigh"},
                {27, "CommFailure LimitConst"},
                {28, "OutOfService LimitOk"},
                {29, "OutOfService LimitLow"},
                {30, "OutOfService LimitHigh"},
                {31, "OutOfService LimitConst"},
                {32, "WaitingForInitialData LimitOk"},
                {33, "WaitingForInitialData LimitLow"},
                {34, "WaitingForInitialData LimitHigh"},
                {35, "WaitingForInitialData LimitConst"},
                {36, "9 LimitOk"},
                {37, "9 LimitLow"},
                {38, "9 LimitHigh"},
                {39, "9 LimitConst"},
                {40, "10 LimitOk"},
                {41, "10 LimitLow"},
                {42, "10 LimitHigh"},
                {43, "10 LimitConst"},
                {44, "11 LimitOk"},
                {45, "11 LimitLow"},
                {46, "11 LimitHigh"},
                {47, "11 LimitConst"},
                {48, "12 LimitOk"},
                {49, "12 LimitLow"},
                {50, "12 LimitHigh"},
                {51, "12 LimitConst"},
                {52, "13 LimitOk"},
                {53, "13 LimitLow"},
                {54, "13 LimitHigh"},
                {55, "13 LimitConst"},
                {56, "14 LimitOk"},
                {57, "14 LimitLow"},
                {58, "14 LimitHigh"},
                {59, "14 LimitConst"},
                {60, "15 LimitOk"},
                {61, "15 LimitLow"},
                {62, "15 LimitHigh"},
                {63, "15 LimitConst"},
                {64, "UncertainNonspecific LimitOk"},
                {65, "UncertainNonspecific LimitLow"},
                {66, "UncertainNonspecific LimitHigh"},
                {67, "UncertainNonspecific LimitConst"},
                {68, "LastUsable LimitOk"},
                {69, "LastUsable LimitLow"},
                {70, "LastUsable LimitHigh"},
                {71, "LastUsable LimitConst"},
                {72, "18 LimitOk"},
                {73, "18 LimitLow"},
                {74, "18 LimitHigh"},
                {75, "18 LimitConst"},
                {76, "19 LimitOk"},
                {77, "19 LimitLow"},
                {78, "19 LimitHigh"},
                {79, "19 LimitConst"},
                {80, "SensorCal LimitOk"},
                {81, "SensorCal LimitLow"},
                {82, "SensorCal LimitHigh"},
                {83, "SensorCal LimitConst"},
                {84, "EguExceeded LimitOk"},
                {85, "EguExceeded LimitLow"},
                {86, "EguExceeded LimitHigh"},
                {87, "EguExceeded LimitConst"},
                {88, "Subnormal LimitOk"},
                {89, "Subnormal LimitLow"},
                {90, "Subnormal LimitHigh"},
                {91, "Subnormal LimitConst"},
                {92, "23 LimitOk"},
                {93, "23 LimitLow"},
                {94, "23 LimitHigh"},
                {95, "23 LimitConst"},
                {96, "24 LimitOk"},
                {97, "24 LimitLow"},
                {98, "24 LimitHigh"},
                {99, "24 LimitConst"},
                {100, "25 LimitOk"},
                {101, "25 LimitLow"},
                {102, "25 LimitHigh"},
                {103, "25 LimitConst"},
                {104, "26 LimitOk"},
                {105, "26 LimitLow"},
                {106, "26 LimitHigh"},
                {107, "26 LimitConst"},
                {108, "27 LimitOk"},
                {109, "27 LimitLow"},
                {110, "27 LimitHigh"},
                {111, "27 LimitConst"},
                {112, "28 LimitOk"},
                {113, "28 LimitLow"},
                {114, "28 LimitHigh"},
                {115, "28 LimitConst"},
                {116, "29 LimitOk"},
                {117, "29 LimitLow"},
                {118, "29 LimitHigh"},
                {119, "29 LimitConst"},
                {120, "30 LimitOk"},
                {121, "30 LimitLow"},
                {122, "30 LimitHigh"},
                {123, "30 LimitConst"},
                {124, "31 LimitOk"},
                {125, "31 LimitLow"},
                {126, "31 LimitHigh"},
                {127, "31 LimitConst"},
                {128, "32 LimitOk"},
                {129, "32 LimitLow"},
                {130, "32 LimitHigh"},
                {131, "32 LimitConst"},
                {132, "33 LimitOk"},
                {133, "33 LimitLow"},
                {134, "33 LimitHigh"},
                {135, "33 LimitConst"},
                {136, "34 LimitOk"},
                {137, "34 LimitLow"},
                {138, "34 LimitHigh"},
                {139, "34 LimitConst"},
                {140, "35 LimitOk"},
                {141, "35 LimitLow"},
                {142, "35 LimitHigh"},
                {143, "35 LimitConst"},
                {144, "36 LimitOk"},
                {145, "36 LimitLow"},
                {146, "36 LimitHigh"},
                {147, "36 LimitConst"},
                {148, "37 LimitOk"},
                {149, "37 LimitLow"},
                {150, "37 LimitHigh"},
                {151, "37 LimitConst"},
                {152, "38 LimitOk"},
                {153, "38 LimitLow"},
                {154, "38 LimitHigh"},
                {155, "38 LimitConst"},
                {156, "39 LimitOk"},
                {157, "39 LimitLow"},
                {158, "39 LimitHigh"},
                {159, "39 LimitConst"},
                {160, "40 LimitOk"},
                {161, "40 LimitLow"},
                {162, "40 LimitHigh"},
                {163, "40 LimitConst"},
                {164, "41 LimitOk"},
                {165, "41 LimitLow"},
                {166, "41 LimitHigh"},
                {167, "41 LimitConst"},
                {168, "42 LimitOk"},
                {169, "42 LimitLow"},
                {170, "42 LimitHigh"},
                {171, "42 LimitConst"},
                {172, "43 LimitOk"},
                {173, "43 LimitLow"},
                {174, "43 LimitHigh"},
                {175, "43 LimitConst"},
                {176, "44 LimitOk"},
                {177, "44 LimitLow"},
                {178, "44 LimitHigh"},
                {179, "44 LimitConst"},
                {180, "45 LimitOk"},
                {181, "45 LimitLow"},
                {182, "45 LimitHigh"},
                {183, "45 LimitConst"},
                {184, "46 LimitOk"},
                {185, "46 LimitLow"},
                {186, "46 LimitHigh"},
                {187, "46 LimitConst"},
                {188, "47 LimitOk"},
                {189, "47 LimitLow"},
                {190, "47 LimitHigh"},
                {191, "47 LimitConst"},
                {192, "GoodNonspecific"},
                {193, "GoodNonspecific LimitLow"},
                {194, "GoodNonspecific LimitHigh"},
                {195, "GoodNonspecific LimitConst"},
                {196, "49 LimitOk"},
                {197, "49 LimitLow"},
                {198, "49 LimitHigh"},
                {199, "49 LimitConst"},
                {200, "50 LimitOk"},
                {201, "50 LimitLow"},
                {202, "50 LimitHigh"},
                {203, "50 LimitConst"},
                {204, "51 LimitOk"},
                {205, "51 LimitLow"},
                {206, "51 LimitHigh"},
                {207, "51 LimitConst"},
                {208, "52 LimitOk"},
                {209, "52 LimitLow"},
                {210, "52 LimitHigh"},
                {211, "52 LimitConst"},
                {212, "53 LimitOk"},
                {213, "53 LimitLow"},
                {214, "53 LimitHigh"},
                {215, "53 LimitConst"},
                {216, "LocalOverride LimitOk"},
                {217, "LocalOverride LimitLow"},
                {218, "LocalOverride LimitHigh"},
                {219, "LocalOverride LimitConst"},
                {220, "55 LimitOk"},
                {221, "55 LimitLow"},
                {222, "55 LimitHigh"},
                {223, "55 LimitConst"},
                {224, "56 LimitOk"},
                {225, "56 LimitLow"},
                {226, "56 LimitHigh"},
                {227, "56 LimitConst"},
                {228, "57 LimitOk"},
                {229, "57 LimitLow"},
                {230, "57 LimitHigh"},
                {231, "57 LimitConst"},
                {232, "58 LimitOk"},
                {233, "58 LimitLow"},
                {234, "58 LimitHigh"},
                {235, "58 LimitConst"},
                {236, "59 LimitOk"},
                {237, "59 LimitLow"},
                {238, "59 LimitHigh"},
                {239, "59 LimitConst"},
                {240, "60 LimitOk"},
                {241, "60 LimitLow"},
                {242, "60 LimitHigh"},
                {243, "60 LimitConst"},
                {244, "61 LimitOk"},
                {245, "61 LimitLow"},
                {246, "61 LimitHigh"},
                {247, "61 LimitConst"},
                {248, "62 LimitOk"},
                {249, "62 LimitLow"},
                {250, "62 LimitHigh"},
                {251, "62 LimitConst"},
                {252, "63 LimitOk"},
                {253, "63 LimitLow"},
                {254, "63 LimitHigh"},
                {255, "63 LimitConst"},
                {256, "BadNonspecific LimitOk"},
                {257, "BadNonspecific LimitLow"},
                {258, "BadNonspecific LimitHigh"},
            };

            foreach (KeyValuePair<int, string> pair in expected)
            {
                int numericalValue = pair.Key;
                var quality = DAQuality.FromInt32(numericalValue);
                string result = quality.ToString("g", formatProvider: null);
                Assert.AreEqual(pair.Value, result);
            }
        }

        /// <summary>
        /// <see cref="DAQuality.IsBad"/>
        /// <see cref="DAQualities.BadNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ib_BadNonspecific()
        {
            DAQuality target = DAQualities.BadNonspecific;
            Assert.AreEqual("True", target.ToString("Ib", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsBad"/>
        /// <see cref="DAQualities.GoodNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ib_GoodNonspecific()
        {
            DAQuality target = DAQualities.GoodNonspecific;
            Assert.AreEqual("False", target.ToString("Ib", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsBad"/>
        /// <see cref="DAQualities.UncertainNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ib_UncertainNonspecific()
        {
            DAQuality target = DAQualities.UncertainNonspecific;
            Assert.AreEqual("False", target.ToString("Ib", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsGood"/>
        /// <see cref="DAQualities.BadNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ig_BadNonspecific()
        {
            DAQuality target = DAQualities.BadNonspecific;
            Assert.AreEqual("False", target.ToString("Ig", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsGood"/>
        /// <see cref="DAQualities.GoodNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ig_GoodNonspecific()
        {
            DAQuality target = DAQualities.GoodNonspecific;
            Assert.AreEqual("True", target.ToString("Ig", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsGood"/>
        /// <see cref="DAQualities.UncertainNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Ig_UncertainNonspecific()
        {
            DAQuality target = DAQualities.UncertainNonspecific;
            Assert.AreEqual("False", target.ToString("Ig", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsUncertain"/>
        /// <see cref="DAQualities.BadNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Iu_BadNonspecific()
        {
            DAQuality target = DAQualities.BadNonspecific;
            Assert.AreEqual("False", target.ToString("Iu", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsUncertain"/>
        /// <see cref="DAQualities.GoodNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Iu_GoodNonspecific()
        {
            DAQuality target = DAQualities.GoodNonspecific;
            Assert.AreEqual("False", target.ToString("Iu", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.IsUncertain"/>
        /// <see cref="DAQualities.UncertainNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Iu_UncertainNonspecific()
        {
            DAQuality target = DAQualities.UncertainNonspecific;
            Assert.AreEqual("True", target.ToString("Iu", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.StatusInfo"/>
        /// <see cref="DAQualities.BadNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Z_BadNonspecific()
        {
            DAQuality target = DAQualities.BadNonspecific;
            Assert.AreEqual("Error", target.ToString("Z", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.StatusInfo"/>
        /// <see cref="DAQualities.GoodNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Z_GoodNonspecific()
        {
            DAQuality target = DAQualities.GoodNonspecific;
            Assert.AreEqual("Normal", target.ToString("Z", formatProvider: null));
        }

        /// <summary>
        /// <see cref="DAQuality.StatusInfo"/>
        /// <see cref="DAQualities.UncertainNonspecific"/>
        /// </summary>
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ToStringTest_Z_UncertainNonspecific()
        {
            DAQuality target = DAQualities.UncertainNonspecific;
            Assert.AreEqual("Warning", target.ToString("Z", formatProvider: null));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void TraceResultsTest_ToString_F()
        {
            TraceFormatResults("F");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void TraceResultsTest_ToString_f()
        {
            TraceFormatResults("f");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void TraceResultsTest_ToString_G()
        {
            TraceFormatResults("G");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void TraceResultsTest_ToString_g()
        {
            TraceFormatResults("g");
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
