﻿
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    partial class DeveloperTest
    {
        #region Static Private Properties

        [NotNull]
        static private EnumerationDataType NodeIdTypeDataType
        {
            get
            {
                var nodeIdTypeDataType = new EnumerationDataType("NodeIdType", "http://opcfoundation.org/UA/:NodeIdType", 6)
                {
                    new EnumerationMember("TwoByte", 0),
                    new EnumerationMember("FourByte", 1),
                    new EnumerationMember("Numeric", 2),
                    new EnumerationMember("String", 3),
                    new EnumerationMember("Guid", 4),
                    new EnumerationMember("ByteString", 5)
                };
                nodeIdTypeDataType.Description = "The possible encodings for a NodeId value.";
                return nodeIdTypeDataType;
            }
        }

        [NotNull]
        static private StructuredData SequenceOfEnumeratedTypeData => new StructuredData(SequenceOfEnumeratedTypeDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, NodeIdTypeDataType))
            {
                new EnumerationData(NodeIdTypeDataType, 1),
                new EnumerationData(NodeIdTypeDataType, 2)
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfEnumeratedTypeDataType => new StructuredDataType("SequenceOfEnumeratedType", "DeveloperTest:SequenceOfEnumeratedType")
        {
            new DataField("Sequence", new SequenceDataType(2, NodeIdTypeDataType))
        };

        [NotNull]
        static private StructuredData SequenceOfOpaqueTypeData => new StructuredData(SequenceOfOpaqueTypeDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, StatusCodeDataType))
            {
                new OpaqueData(StatusCodeDataType, new byte[] { 1, 0, 0, 0 }),
                new OpaqueData(StatusCodeDataType, new byte[] { 2, 0, 0, 0 })
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfOpaqueTypeDataType => new StructuredDataType("SequenceOfOpaqueType", "DeveloperTest:SequenceOfOpaqueType")
        {
            new DataField("Sequence", new SequenceDataType(2, StatusCodeDataType))
        };

        [NotNull]
        static private StructuredData SequenceOfSequenceWithLengthData => new StructuredData(SequenceOfSequenceWithLengthDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, UAOpcBinaryStandardDataTypes.CharArray))
            {
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, "A"),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, "BC")
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfSequenceWithLengthDataType => new StructuredDataType("SequenceOfSequenceWithLength", "DeveloperTest:SequenceOfSequenceWithLength")
        {
            new DataField("Sequence", new SequenceDataType(2, UAOpcBinaryStandardDataTypes.CharArray))
        };

        [NotNull]
        static private StructuredData SequenceOfSequenceWithLengthInBytesData => new StructuredData(SequenceOfSequenceWithLengthInBytesDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, IntegerArrayWithFixedLengthInBytesDataType))
            {
                IntegerArrayWithFixedLengthInBytes_3Data,
                IntegerArrayWithFixedLengthInBytes_3Data
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfSequenceWithLengthInBytesDataType => new StructuredDataType("SequenceOfSequenceWithLengthInBytes", "DeveloperTest:SequenceOfSequenceWithLengthInBytes")
        {
            new DataField("Sequence", new SequenceDataType(2, IntegerArrayWithFixedLengthInBytesDataType))
        };

        [NotNull]
        static private StructuredData SequenceOfSequenceWithLengthInBytesUnalignedData => new StructuredData(SequenceOfSequenceWithLengthInBytesUnalignedDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, IntegerArrayWithFixedLengthInBytesUnalignedDataType))
            {
                IntegerArrayWithFixedLengthInBytesUnaligned_3Data,
                IntegerArrayWithFixedLengthInBytesUnaligned_3Data
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfSequenceWithLengthInBytesUnalignedDataType => new StructuredDataType("SequenceOfSequenceWithLengthInBytesUnaligned", "DeveloperTest:SequenceOfSequenceWithLengthInBytesUnaligned")
        {
            new DataField("Sequence", new SequenceDataType(2, IntegerArrayWithFixedLengthInBytesUnalignedDataType))
        };

        [NotNull]
        static private StructuredData SequenceOfSequenceWithTerminatorData => new StructuredData(SequenceOfSequenceWithTerminatorDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, UAOpcBinaryStandardDataTypes.WideString))
            {
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "A"),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "BC")
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfSequenceWithTerminatorDataType => new StructuredDataType("SequenceOfSequenceWithTerminator", "DeveloperTest:SequenceOfSequenceWithTerminator")
        {
            new DataField("Sequence", new SequenceDataType(2, UAOpcBinaryStandardDataTypes.WideString))
        };

        [NotNull]
        static private StructuredData SequenceOfStructuredTypeData => new StructuredData(SequenceOfStructuredTypeDataType)
        {
            {"Sequence", new SequenceData(new SequenceDataType(2, XmlElementDataType))
            {
                new StructuredData(XmlElementDataType)
                {
                    {"Length", new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 1) },
                    {"Value", new SequenceData(new SequenceDataType(UAOpcBinaryStandardDataTypes.Char))
                    {
                        new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, 'A')
                    }}
                },
                new StructuredData(XmlElementDataType)
                {
                    {"Length", new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 2) },
                    {"Value", new SequenceData(new SequenceDataType(UAOpcBinaryStandardDataTypes.Char))
                    {
                        new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, 'B'),
                        new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, 'C')
                    }}
                }
            }}
        };

        [NotNull]
        static private StructuredDataType SequenceOfStructuredTypeDataType => new StructuredDataType("SequenceOfStructuredType", "DeveloperTest:SequenceOfStructuredType")
        {
            new DataField("Sequence", new SequenceDataType(2, XmlElementDataType))
        };

        [NotNull]
        static private byte[] SequenceOfEnumeratedTypeExtensionObjectBody => new byte[] {129, 0};

        [NotNull]
        static private byte[] SequenceOfOpaqueTypeExtensionObjectBody => new byte[] { 1, 0, 0, 0, 2, 0, 0, 0 };

        [NotNull]
        static private byte[] SequenceOfSequenceWithLengthExtensionObjectBody => new byte[] { 1, 0, 0, 0, 65, 2, 0, 0, 0, 66, 67 };

        [NotNull]
        static private byte[] SequenceOfSequenceWithLengthInBytesExtensionObjectBody => 
            IntegerArrayWithFixedLengthInBytes_3ExtensionObjectBody.Concat(IntegerArrayWithFixedLengthInBytes_3ExtensionObjectBody).ToArray();

        [NotNull]
        static private byte[] SequenceOfSequenceWithLengthInBytesUnalignedExtensionObjectBody =>
            IntegerArrayWithFixedLengthInBytesUnaligned_3ExtensionObjectBody.Concat(IntegerArrayWithFixedLengthInBytesUnaligned_3ExtensionObjectBody).ToArray();

        [NotNull]
        static private byte[] SequenceOfSequenceWithTerminatorExtensionObjectBody => new byte[] { 65, 0, 0, 0, 66, 0, 67, 0, 0, 0 };

        [NotNull]
        static private byte[] SequenceOfStructuredTypeExtensionObjectBody => new byte[] { 1, 0, 0, 0, 65, 2, 0, 0, 0, 66, 67 };

        [NotNull]
        static private OpaqueDataType StatusCodeDataType => new OpaqueDataType("StatusCode", "http://opcfoundation.org/UA/:StatusCode", 32) { Description = "A 32-bit status code value." };

        [NotNull]
        static private StructuredDataType XmlElementDataType
        {
            get
            {
                var xmlElementDataType = new StructuredDataType("XmlElement", "http://opcfoundation.org/UA/:XmlElement")
                {
                    new DataField("Length", UAOpcBinaryStandardDataTypes.Int32) { IsLength = true },
                    new DataField("Value", new SequenceDataType(UAOpcBinaryStandardDataTypes.Char)) { IsOptional = true }
                };
                xmlElementDataType.Description = "An XML element encoded as a UTF-8 string.";
                return xmlElementDataType;
            }
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeTest_SequenceOfEnumeratedType() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfEnumeratedType",
                SequenceOfEnumeratedTypeExtensionObjectBody,
                SequenceOfEnumeratedTypeData);

        [TestMethod]
        public void DecodeTest_SequenceOfOpaqueType() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfOpaqueType",
                SequenceOfOpaqueTypeExtensionObjectBody,
                SequenceOfOpaqueTypeData);

        [TestMethod]
        public void DecodeTest_SequenceOfSequenceWithLength() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfSequenceWithLength",
                SequenceOfSequenceWithLengthExtensionObjectBody,
                SequenceOfSequenceWithLengthData);

        [TestMethod]
        public void DecodeTest_SequenceOfSequenceWithLengthInBytes() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfSequenceWithLengthInBytes",
                SequenceOfSequenceWithLengthInBytesExtensionObjectBody,
                SequenceOfSequenceWithLengthInBytesData);

        [TestMethod]
        public void DecodeTest_SequenceOfSequenceWithLengthInBytesUnaligned() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfSequenceWithLengthInBytesUnaligned",
                SequenceOfSequenceWithLengthInBytesUnalignedExtensionObjectBody,
                SequenceOfSequenceWithLengthInBytesUnalignedData);

        [TestMethod]
        public void DecodeTest_SequenceOfSequenceWithTerminator() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfSequenceWithTerminator",
                SequenceOfSequenceWithTerminatorExtensionObjectBody,
                SequenceOfSequenceWithTerminatorData);

        [TestMethod]
        public void DecodeTest_SequenceOfStructuredType() =>
            DecodeTest(null,
                TypeDictionaryByteArray,
                "SequenceOfStructuredType",
                SequenceOfStructuredTypeExtensionObjectBody,
                SequenceOfStructuredTypeData);

        #endregion
    }
}
