﻿
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLibFactory;
using OpcLabs.EasyOpc.UA.DataTypeModel;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    partial class TypeDescriptionExamplesTest
    {
        #region Static Private Properties

        [NotNull]
        static private StructuredDataType DataValueDataType
        {
            get
            {
                var structuredDataType = new StructuredDataType("DataValue", "TypeDescriptionExamples:DataValue")
                {
                    new DataField("ValueSpecified", UAOpcBinaryStandardDataTypes.Bit) { IsSwitch = true },
                    new DataField("StatusCodeSpecified", UAOpcBinaryStandardDataTypes.Bit) { IsSwitch = true },
                    new DataField("TimestampSpecified", UAOpcBinaryStandardDataTypes.Bit) { IsSwitch = true },
                    new DataField("Reserved1", new SequenceDataType(5, UAOpcBinaryStandardDataTypes.Bit)),
                    new DataField("Value", VariantDataType) { IsOptional = true },
                    new DataField("Quality", QualityDataType) { IsOptional = true },
                    new DataField("Timestamp", UAOpcBinaryStandardDataTypes.DateTime) { IsOptional = true }
                };
                structuredDataType.Description = "A value with an associated timestamp, and quality.";
                return structuredDataType;
            }
        }

        [NotNull]
        static private StructuredData DataValue_0Data => new StructuredData(DataValueDataType)
        {
            {"ValueSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false)},
            {"StatusCodeSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false)},
            {"TimestampSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false)},
            {
                "Reserved1", new SequenceData(new SequenceDataType(5, UAOpcBinaryStandardDataTypes.Bit))
                {
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false)
                }
            }
        };

        [NotNull]
        static private Byte[] DataValue_0ExtensionObjectBody => new byte[1];

        [NotNull]
        static private StructuredData DataValue_7Data => new StructuredData(DataValueDataType)
        {
            {"ValueSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, true)},
            {"StatusCodeSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, true)},
            {"TimestampSpecified", new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, true)},
            {
                "Reserved1", new SequenceData(new SequenceDataType(5, UAOpcBinaryStandardDataTypes.Bit))
                {
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                    new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false)
                }
            },
            {"Value", Variant_Int32Data },
            {"Quality", Quality_192Data },
            {"Timestamp", new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTimeInstances.Kind_Utc) }
        };

        [NotNull]
        static private Byte[] DataValue_7ExtensionObjectBody => new byte[] { 0x07, 0x02, 0x21, 0x43, 0x65, 0x87, 0xC0, 0x00, 0x80, 0x22, 0xF1, 0xA8, 0x6B, 0x10, 0x9B, 0x01 };

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeTest_DataValue_0()
        {
            // ReSharper disable once PossibleNullReferenceException
            DecodeTest(dts => dts.Parameters.TypeNameRemappingDictionary.Clear(), 
                TypeDictionaryByteArray, "DataValue", DataValue_0ExtensionObjectBody, DataValue_0Data);
        }

        [TestMethod]
        public void DecodeTest_DataValue_7()
        {
            // ReSharper disable once PossibleNullReferenceException
            DecodeTest(dts => dts.Parameters.TypeNameRemappingDictionary.Clear(), 
                TypeDictionaryByteArray, "DataValue", DataValue_7ExtensionObjectBody, DataValue_7Data);
        }

        [TestMethod]
        public void EncodeTest_DataValue_0()
        {
            EncodeTest(null, TypeDictionaryByteArray, "DataValue", DataValue_0Data, DataValue_0ExtensionObjectBody);
        }

        [TestMethod]
        public void EncodeTest_DataValue_7()
        {
            EncodeTest(null, TypeDictionaryByteArray, "DataValue", DataValue_7Data, DataValue_7ExtensionObjectBody);
        }

        #endregion
    }
}
