﻿using System.Diagnostics;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Plugins.ComplexData;
using OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.UA.Plugins.ComplexData
{
    [TestClass]
    public class UAComplexDataPluginTest
    {
        #region Static Private Methods

        [NotNull]
        static private EasyUAClient CreateClientWithComplexDataPlugin()
        {
            var client = new EasyUAClient();
            // ReSharper disable once PossibleNullReferenceException
            client.InstanceParameters.PluginSetups[EasyUAClientUriStrings.ComplexDataPlugin].Enabled = true;
            return client;
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void WriteInvalidDataTest()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = "anEndpointDescriptor";
            UANodeDescriptor nodeDescriptor = "aNodeDescriptor";

            EasyUAClient client = CreateClientWithComplexDataPlugin();

            var invalidData = new OpaqueData(TypeDescriptionExamplesTest.Int128DataType, new byte[0]);
            Debug.Assert(!invalidData.IsValid);

            TestSection.Act();

            const int state = 555;
            UAWriteResult[] resultArray = client.WriteMultipleValues(new[]
            {
                new UAWriteValueArguments(endpointDescriptor, nodeDescriptor, new UAGenericObject(invalidData)) { State = state }
            });

            TestSection.Assert();
            Assert.IsNotNull(resultArray);
            Assert.AreEqual(1, resultArray.Length);
            UAWriteResult result = resultArray[0];
            Assert.IsNotNull(result);
            AssertUtilities.WriteLine($"result: {result}");
            Assert.AreEqual(state, result.State);
            Assert.IsFalse(result.Succeeded);
            AssertUtilities.WriteLine($"result.ErrorMessage: {result.ErrorMessage}");
            Assert.AreEqual(-1, result.ErrorCode);
        }

        #endregion
    }
}
