﻿
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.LiveMapping;
using OpcLabs.EasyOpc.UA.LiveMapping.Extensions;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA.Servers
{
    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class _3SSmartSoftwareSolution_Codesys : UAInteroperabilityWorkshopTest
    {
        public override IEnumerable<UANodeDescriptor> ArrayReadableNodes => new UANodeDescriptor[] { "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesArrayStatic.a_OpcUaType_Int32_static" };

        public override IEnumerable<UANodeDescriptor> ArrayWriteableNodes => new UANodeDescriptor[] { "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesArrayStatic.a_OpcUaType_Int32_static" };

        public override IEnumerable<UANodeDescriptor> DynamicNodes => new UANodeDescriptor[]
        {
            "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesDynamic.OpcUaType_Double_dynamic",
            "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesDynamic.OpcUaType_Float_dynamic"
        };

        public override IEnumerable<UAEndpointDescriptor> EndpointDescriptors => new UAEndpointDescriptor[] { "opc.tcp://scholladnb:4841" };

        public override IEnumerable<string> ServerNotSupportedTests => base.ServerNotSupportedTests.Concat(new[]
        {
            nameof(Group001_Session_Test02_CreateSession_UaTcp_Basic128Rsa15),
            //nameof(Group001_Session_Test03_ActivateSession_UserNameAndPassword),
            nameof(Group001_Session_Test09_CreateSession_Https_None),
            nameof(Group001_Session_Test10_ActivateSession_X509Certificate),
            nameof(Group005_InformationModel_Test04_Browse_Read_SessionDiagnosticsArray),
            nameof(Group009_DataChange_Test05_AbsoluteDeadbandFilter),
            nameof(Group013_Discovery_Test03_FindServers_LDS_ServerPresence),
            nameof(Group102_Session2_Test01_CreateSession_UaTcp_Basic256),
            nameof(Group102_Session2_Test03_CreateSession_Https_Basic256Sha256)
        });

        public override string Password => "user1";

        public override IEnumerable<UANodeDescriptor> StaticNodes => new UANodeDescriptor[] { "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_Int32_static" };

        public override string UserName => "opcua_user1";

        public override IEnumerable<UANodeDescriptor> WriteableNodes => new UANodeDescriptor[] { "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_Int32_static" };

        [UAType]
        private class MappedType
        {
            public MappedType()
            {
                var buffer = new byte[4];
                new Random(Environment.TickCount).NextBytes(buffer);
                _value = BitConverter.ToUInt32(buffer, 0);
            }

            // Currently does not work
            //[UANode(NodeId = "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_UInt32_static")]
            //[UAData(Operations = UADataMappingOperations.All, Kind = UADataMappingKind.AttributeData)]
            //[UAMonitoring(SamplingInterval = 250)]
            //[UASubscription(PublishingInterval = 500)]
            //public UAAttributeData<long> Variable
            //{
            //    get
            //    {
            //        var attributeData = new UAAttributeData<long>(_value);
            //        AssertUtilities.WriteLine($"Attribute data got for writing, {attributeData.GetType()}: {attributeData}");
            //        return attributeData;
            //    }
            //    set => AssertUtilities.WriteLine($"Attribute data set by reading, {value?.GetType()}: {value}");
            //}

            //[UANode(NodeId = "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_UInt32_static")]
            //[UAData(Operations = UADataMappingOperations.All, Kind = UADataMappingKind.Exception)]
            //[UAMonitoring(SamplingInterval = 250)]
            //[UASubscription(PublishingInterval = 500)]
            //public Exception Exception
            //{
            //    set => AssertUtilities.WriteLine($"Exception: {value}");
            //}

            [UANode(NodeId = "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_UInt32_static")]
            [UAData(Operations = UADataMappingOperations.All, Kind = UADataMappingKind.AttributeData)]
            [UAMonitoring(SamplingInterval = 250)]
            [UASubscription(PublishingInterval = 500)]
            public UAAttributeData Variable
            {
                get
                {
                    var attributeData = new UAAttributeData((object)_value);
                    AssertUtilities.WriteLine($"Attribute data got for writing, {attributeData.GetType()}: {attributeData}");
                    return attributeData;
                }
                set => AssertUtilities.WriteLine($"Attribute data set by reading, {value?.GetType()}: {value}");
            }

            [UANode(NodeId = "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.OpcUaTypesStatic.OpcUaType_UInt32_static")]
            [UAData(Operations = UADataMappingOperations.All, Kind = UADataMappingKind.Exception)]
            [UAMonitoring(SamplingInterval = 250)]
            [UASubscription(PublishingInterval = 500)]
            public Exception Exception
            {
                set => AssertUtilities.WriteLine($"Exception: {value}");
            }

            private readonly UInt32 _value;
        }

        [TestMethod]
        public void ComplexValueEnumerationTest()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorUATcp;
            var client = new EasyUAClient();

            TestSection.Act();
            object value = client.ReadValue(endpointDescriptor, "nsu=CODESYSSPV3/3S/IecVarAccess;ns=4;s=|var|CODESYS Control Win V3.Application.ComplexValuesStatic.enumeration");
            
            TestSection.Assert();



        }

        [TestMethod]
        public void UInt64MappingTest()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorUATcp;

            var mapper = new UAClientMapper();
            var mappedObject = new MappedType();
            mapper.Map(mappedObject, new UAMappingContext
            {
                EndpointDescriptor = endpointDescriptor,
                MonitoringParameters = 1000,
            });

            AssertUtilities.Phase("Reading 1");
            mapper.Read();

            AssertUtilities.Phase("Writing");
            mapper.WriteTarget(mappedObject, /*recurse:*/false);

            AssertUtilities.Phase("Reading 2");
            mapper.Read();
        }
    }
}
